;;
;;                    A N A L Y Z E
;;
;; Where /eval/ evaluates /analyze/ might do other things.
;; Let's start with the call graph. What's the call graph
;; of the following expression:
;;
;;   (define (make-entry key definition successors)
;;     (cons key (cons definition successors)))
;;
;; Maybe the second line already represents the call graph
;; adequately. Let's canonicalize /define/ first:
;;
;;   (define make-entry (lambda (key definition successors)
;;     (cons key (cons definition successors)) ))
;;
;; Call graph:
;;   (define (lambda))
;;
;; Maybe we shouldn't have asked for the call graph of the
;; expression but for the call graph of the resulting
;; function:
;;
;;   (call-graph make-entry) => (cons (cons))
;;
;; Another example:
;;   
;;   (define (get-entry branch keys)
;;    (cond
;;     ((not branch)     #f)
;;     ((null? keys) branch)
;;     (else
;;      (get-entry (assq (car keys) (get-successors branch))
;;                 (cdr keys)) )))
;;
;;   (define (lambda))
;;
;;   (call-graph get-entry) => (cond (branch#0 (not))
;;                                   (branch#1 (null?))
;;                                   (else-branch
;;                                     (get-entry (assq (car) (get-successors))
;;                                                (cdr)))
;;
;;   (call-graph get-successors) => (cddr)
;;
;;   (call-graph get-entry) => (cond (branch#0 (not))
;;                                   (branch#1 (null?))
;;                                   (else-branch
;;                                     (get-entry (assq (car) (cddr)
;;                                                (cdr))))
;;
;; Now that's conditional recursion! How do we represent such
;; a thing as a graph? Scheme already seems to represent that
;; graph adequately? Conditionals remain conditionals and
;; recursion needs some symbol (or other token) to be represented. Should we
;; start to name conditional branches or should we confine our
;; call graph analyzer: maybe /cond/ shouldn't be considered
;; a function or node on the call graph but we still need to
;; be aware of it. Did we search for that abstract call graph
;; or do we want to predict the call graph for certain values?
;; Maybe we cannot predict the result of a predicate with
;; inherent side effects.
;;
;; Maybe we just want to have a log file representing the call
;; graph? Or we want to have a call graph with call counters?
;; Do we want to have an actual graph (in its textual form) or
;; do we want to create a data graph in memory for later
;; processing? Simple serialization with /write/ doesn't provide
;; any indention and formatting remains a matter of taste. Too
;; many nodes on the call graph isn't helpful.
;;
;; !!!
;; We definitely want to have source code references. While the
;; new memory object layout would allow for another slot for
;; each non immediate doubts remain regarding transformations
;; of the s-expressions and the "resolution" of source code
;; locations.
;; All the call graph log or whatever seems next to useless if
;; we can't refer back to the source code!!!
;;
(sys-load "VSI-core/repl.scm")

(define trace-port #f)

(define (log-trace msg env)
 (with-output-to-port trace-port (lambda()
  (color trace-port 'foreground 'cyan)
  (display "A#")
  (display (length env))
  (display ": ")
  (display msg) 
  (newline))))

;;
;; Environments

(define (add-binding! env sym val)
  (log-trace (format "definition: ~a ~a" sym val) env)
  (cond ((null? env) (%define-global sym val))
        ;; XXX we are the only ones to create environment frames
        ;; thus we can use the regular format? But every frame with
        ;; only one binding has to be a cdrp frame because of the
        ;; ilocs. But it's us to do all the lookups and then we can ...
        ;; What happens if the symbol lookup of CEVAL runs into an
        ;; environment frame with the value in CDR? <- The value of
        ;; the CADR that will be for our frames so they will not be
        ;; classified as cdrp ilocs. We really should redefine /eval/
        ;; to /analyze/ in our top level lookup namespace to then
        ;; use /sys-eval/ where we need it.
        ;; TODO Keep track of bindings and bangs.
        (else (error 'remove cdrp-env-frames-before-doing-this))))

(define (extend-env formals args env)
  (cons (cons formals args) env))

;; TODO Keep track of variable states (location states):
;; constant or modified for now. Don't use /eval/ as we're
;; opting for the abstract call graph where all conditional
;; branches are included.
(define (lookup symbol env) (eval symbol env))

;;
;; Transformers

(define (canonicalize-define expression)
 (if (symbol? (second expression))
   expression
   (let ((symbol   (car (second expression)))
         (formals  (cdr (second expression)))
         (sequence (cddr expression)))
     `(define ,symbol ,(cons 'lambda (cons formals sequence))) )))

;;
;; Call Graph

;; XXX We'd need a well defined node layout.
(define (add-node! parent node)
  (set-cdr! parent (append (cdr parent) (list node))))

(define (call-graph:display call-graph) ;; Printer sketch?
 (cond
  ((closure? call-graph)    (display (procedure-property call-graph 'name)))
  ((not (pair? call-graph)) (display call-graph))
  ((pair? call-graph)       (begin (display "(")
                                   (call-graph:display (car call-graph))
                                   (display " . ")
                                   (call-graph:display (cdr call-graph))
                                   (display  ")"  )) )))
        

; Back links in the graph might allow for easier processing but
; - we would need our own printing routines
; - a lot of rings will need to be cut with some routine to
;   dissolve this directed bidirectional or cyclic graph then.
; (The original intention was to resolve the code back into
;  source code text positions and now the need for bidirectional
;  graphs occurs?)

;;
;; Analyze (another sketch that is)
;;  That much fuel w/o any effect.
;;
;; (Redefinitions might not be supported by SCA and this thing.
;;  Top level preface redefinitions could work out, though.
;;  Maybe it only affects the default bindings of the primary
;;  primitives just because we're ... just like m_exp... was.)

;; Text References
(sys-load "Analyzer/sexp.scm")
(define *last-point* #f)
(define (push-point point) (set! *last-point* point))

;; Shells and Inerts
(define void *unspecified*) ; sugar for now
(define (call-graph exp env)
 (let ((root (cons 'processed-call-graph '()))
       (back '()))
  (analyze! root back exp env)
  root))
(define (analyze exp env)
 (let ((root (cons 'processed-call-graph '()))
       (back '()))
  (analyze! root back exp env)))
(define (recursive? backlist node)
  (assq (car node) backlist))

; Just another ...
(define (analyze! parent-node backlist exp env)
 ;(log-trace (format "expression: ~a" exp) env)
 (cond
  ((symbol? exp)     (lookup exp env))
  ((not (pair? exp)) exp)
  ((pair? exp)       (let* ((det (car exp)) (node (list det)))
  ;; The first constituent of an expression determines the way
  ;; in which it is processed:
  ;; primary primitive, application or syntax transformation
  ;; BTW 5 and (begin 5) are both expressions? Or how do you call
  ;; terminators in Scheme?
  (cond
  ((symbol? det)
   (case det ((define) (add-node! parent-node node)
                       (set! backlist (cons node backlist))
                       (let* ((exp (canonicalize-define exp))
                              (sym (second exp))
                              (val (analyze! node backlist
                                             (third exp) env)))
                         (when (closure? val) ; DIVERSION
                           (set-procedure-property! val 'name sym))
                         (add-binding! env sym val)
                         void))
             ((lambda) (add-node! parent-node node)
                       (set! backlist (cons node backlist))
                       ; (REPL-HERE) XXX this assumes fixed point resolution in the data graph
                       (eval (cons (first exp) (cons (pin-exp (second exp)) (cddr exp))) env)) ; create a closure (token)
             ((if)     (add-node! parent-node node)
                       ; no back link here yet
                       ;(set! backlist (cons node backlist))
                       (add-node! node '...)
                       *unspecified*)
                       ; That would make up for two call graphs!
                       ; We could continue here if not for the
                       ; current way of handling bindings which
                       ; determines what a symbol will be taken for.
                      ;(map (lambda (exp) (analyze! node backlist
                       ;                             exp env))
                       ;     (cdr exp)))
             ((begin)  (let loop ((rest (cdr exp)))
                         (cond
                          ((null? (cdr rest)) ; return value branch
                             (analyze! parent-node backlist (car rest) env))
                          (else
                             (analyze! parent-node backlist (car rest) env)
                             (loop (cdr rest))) )))
             ((sexpin) (push-point (pin-point exp))
                       (log-trace (format "new point: ~a" (pin-point exp)) env)
                       ;; Shouldn't we pin this in the data graph, too?
                       (analyze! parent-node backlist (pin-exp exp) env))
             ((quote)  (REPL-HERE) (second exp)) ;; ??? node or not node
             (else ; other symbols
              (log-trace (format "other symbol lookup: ~a" det) env)
              (analyze! parent-node backlist
                        (cons (lookup det env) (cdr exp)) env))
   )) ; closes symbol? det case
  ((closure? det)
   (let* ((args (map (lambda (exp) (analyze! node backlist exp env))
                     (cdr exp)))
          (src  (procedure-source det))
          (name (procedure-property det 'name)) ;; XXX source ref please
          (formals (second src))
          (seq     (cddr src)))
     (log-trace (format "applying closure: ~a ~a" name exp) env)
     ;; Probably lambda forms should record the original text position
     ;; in the closures to achieve better visualizations. That may even
     ;; make up for the better outer form of a closure.
     (cond
      ;; This check needs to be done before adding the node or the
      ;; predicate becomes too complicated.
      ((recursive? backlist node) (add-node! parent-node node)
                                  (add-node! node 'ooo)
                                  *unspecified*)
      (else
        ;(REPL-HERE)
        (add-node! parent-node node)
        (set! backlist (cons node backlist))
        (analyze! node backlist
                  (cons 'begin seq) (extend-env formals args env)) ))
   )) ; closes closure? det let

;;;; TODO add /procedure?/ here (primitives)

 )))) ; closes cond pair? let
) ; closes analyze



;;
;; Rapit- Scheme Development Loop

;; Call Graph Session
(define trace-port (current-output-port))

(when #f
(repl-clear)
(repl-transcript
(define env '())

(define prog (quote
  (begin (define (a b c) (begin 1 b c))
         (a 'b 'c))))

(call-graph prog env)
(call-graph '(a (a 1 2) (a 3 4)) env) ;; XXX This is reported as recursive!

(define prog
  '(begin (define (a b c) (begin 1 b c))
          (define (b x) (if (a 1 x) 3 4))
          (b  2 5)))
(call-graph prog env)

(define prog
  '(begin (define (b x y) (b 1 x) 3 4) ; There is no sense to this. It just
          (b (b 2 6) 5)))              ; recurses and it was easy to type.
(define c (call-graph prog env))
;(equal? (call-graph (cdr (call-graph prog env)) env) (call-graph prog env))
(call-graph:display c)
)
)

;;
;; PS Dirty Corners of SCM
;;
;; ((quote define) x 5) =>
;; throw catched: undefined: "Unbound variable: x"
;; ((quote cond) (#t #f)) =>
;; throw catched: misc-error: "Wrong type to apply: #t"
;;

(when #f
;; Text Position Session
(repl-transcript
(define p (pin (file-ref "Example.non" 12 34) '(define a (quote (lambda () 1 2 3 4 5) ))))
p
(define g (call-graph `(define x ,p) '() ))
(fr-display *last-point*)
g
)
)
