;;;
;;;  A l g o r i t h m i c
;;;  N o t a t i o n            Syntax Checks
;;;  S c h e m e
;;;

(define all? (lambda ( . args)
  (fold (lambda (a b) (and a b)) (and) args)))

;;;
;;; Predicates
;;;
;;  symbol? should yield true here for pins and quotations

;; Sequence: (begin . expressions)
(define (sequence:syntax? expression environment)
 (proper-list? expression))

;; Quote (quote expression)
(define (quote:syntax? expression environment)
 (and (proper-list? expression)
      (= (length expression) 2) ))

;;;
;;; Environment Accessors

(define (canonicalize-define expression)
 (if (symbol? (second expression))
   expression
   (let* ((symbol   (sexp-car (second expression)))
          (sym-ref  (sexp-ref (second expression)))
          (formals  (sexp-cdr (second expression)))
          (form-ref (sexp-ref formals))
          (sequence (sexp-cddr expression)))
     `(define ,(repin sym-ref symbol)
        ,(cons 'lambda (cons (repin form-ref formals)
                             sequence))) )))

;; Define (define literal-symbol value)
;;        (define (literal-symbol . formals) . sequence)
(define (define:syntax? expression environment)
  ;(newline) (write expression) (newline)
 (and (proper-list? expression)
      (>= (length expression) 3)
      (or (and (symbol? (second expression))
               (= (length expression) 3))
          (and (pair?        (second expression))  ; non-canonical
               (symbol?      (car (second expression)))
               ;(not (null? (cdr (third expression)))) ; non-empty body?
               (let ((can (canonicalize-define expression)))
                 (set-car! expression (car can))
                 (set-cdr! expression (cdr can))
                 (define:syntax? can environment))) )))

;; Set! (set! literal-symbol value)
(define (set-bang:syntax? expression)
 (and (proper-list? expression)
      (= (length expression) 3)
      (symbol? (second expression)) ))

;;;
;;; Conditionals

;; When (when condition . sequence)
(define (when:syntax? expression environment)
 (and (proper-list? expression)
      ;; XXX (when ()) counts as true: they called it: empty combination
      ;; but we don't check for empty combinations yet. It's the silent
      ;; fall through which probably wasn't intended?
      ;; Where (when #t) was accepted. ??? Should we require a non empty
      ;; sequence?
      (>= (length expression) 2)))

;; AND (and . expressions)
;;  OR (or  . expressions)
(define and:syntax? sequence:syntax?)
(define  or:syntax? sequence:syntax?)

;; Cond (cond . conditional-branches)
;; conditional-branch: (predicate sequence)
;; conditional-branch: (else      sequence)
(define (cond:syntax? expression environment)
 (define (branch:syntax? expression)
   (and (proper-list? expression)
        (> (length expression) 1))) ; Deviation from SCM: dirty empty seq
 ;; In fact the actual syntax check should recurse into /cond/ to provide
 ;; better analysis and more exact error reporting on a per branch basis.
 (cut! branch:syntax?  ;; XXX Missing CLR4 autocut for inner defines
 (and (proper-list? expression)
      (> (length expression) 1)
      ;; XXX don't recurse here
      (apply all? (map branch:syntax? (cdr expression))) )))

;; If (if condition then-clause else-clause)
;;    (if condition then-clause)
(define (if:syntax? expression environment)
 (and (proper-list? expression)
      (or (= (length expression) 3)
          (= (length expression) 4)) ))



;;;
;;; Syntax Checking Scheme
;;;

(sys-load "Analyzer/sexp.scm")
(sys-load "Analyzer/environments.scm")


(define (syntax-error msg)
 (display "Syntax error: ")
 (if syntax-check:last-point
   (fr-display syntax-check:last-point)
   (display "unknown source"))
 (display ": ")
 (display msg)
 (newline))

;; (define check-syntax (macro (arguments . cont)
;;  (let ((exp        (first arguments))
;;        (predicate? (second arguments))
;;        (msg        (third arguments)))
;;  `(if (not (,predicate? (sexp-unpin-tree ,exp) '())) ;; XXX empty env
;;     (syntax-error (format "~A: ~A" ,msg ,exp))
;;     (begin ,@cont)) )))

(define syntax-check:last-point #f)
(define (syntax-check exp env)
 (cond
  ((symbol? exp)     (let ((ans (cons (lookup exp env) *unspecified*)))
                      (if (eq? (car ans) *unspecified*)
                       (syntax-error (format "unresolved symbol: ~A" exp))
                       (car ans))))
  ((not (pair? exp)) exp)
  ((pair? exp)       (let ((det (car exp)))
   (cond
    ((symbol? det) (case det

     ((sexpin)
         (set! syntax-check:last-point (pin-point exp))
         (syntax-check (pin-exp exp) env))

     (else (syntax-check (cons (syntax-check det env) (cdr exp)) env))
     )) ; closes symbol case
    ((macro? det)
     (cond
      ((eq? det quote)
         (if (not (quote:syntax? (sexp-unpin-tree exp) env))
           (syntax-error (format "malformed if: ~A" exp))
           (second exp)))

      ((eq? det begin)
         (if (not (sequence:syntax? (sexp-unpin-tree exp) env))
           (syntax-error (format "malformed begin: ~A" exp))
           (let loop ((rest (cdr exp)))
             (cond ((null? (cdr rest)) (syntax-check (car rest) env))
                   (else               (syntax-check (car rest) env)
                                       (loop (cdr rest)) )) )))
      ((eq? det define)
         (if (not (define:syntax? (sexp-unpin-tree exp) env))
           (syntax-error (format "malformed define: ~A" exp))
           (begin (let* ((cexp (canonicalize-define exp))
                         (sym  (second exp))
                         (vale (third  exp)))
                    (syntax-check vale env)
                    (add-binding! sym 'i-have-seen-this-before env)))))
      )) ; closes macro cond
    (else 'TODO))
   )) ; closes pair let det
))

;;;
;;; Articulations
;;;
(define (read-file file-name)
 (when sys-debug-interpreter
   (display "vsi-nature-load: " (current-error-port))
   (display file-name           (current-error-port))
   (newline))
 (let ((expressions (cons 'box '())))
  ;; Read in all expressions.
  (with-input-from-file file-name (lambda ()
   (let loop ((expression (read)) (last expressions))
    (when (not (eof-object? expression))
     (set-cdr! last (cons expression '()))
     (loop (read) (cdr last)) )) ))
  (cdr expressions)))

(define (list-directory path) ;; Viper import & adaption
  (let ((dir (opendir path)))
    (let loop ((entry (readdir dir)))
      (if (eof-object? entry)
          (begin (closedir dir)
                 '())
          (begin (cons entry
                       (loop (readdir dir))) )))))

(define scm-file? (lambda (path)
  (let ((l (string-length path)))
   (and (> l 4)
    (equal? ".scm" (substring path (- l 4) l))) )))
