(sys-load "VSI-core/repl.scm")
(repl-clear)


;;;
;;; VSI aftermath: CLR1 Straight Count
;;; (VSI TRIM/RC - reference counting only memory management)


;;;
;; Introduction
;;
;; Currently VSI uses a special return value flag (CLR2/a) in addition to
;; the regular straight count. While this system has proven to not impose
;; any theoretical difficulties and to be suitable to be applied to every
;; part of the current SCM code base it still seems awkward and one
;; is tempted to just count straight.
;;
;; That presumed straight count is referred to as CLR1 or counting lambda
;; revision 1. This document should clarify its shortcomings or perhaps
;; even allow to find a working model by exploring different variants
;; based on protocols which clearly state when to count stack references.
;;
;; The logical call stack consists of logical stack frames which
;; correspond to the list of arguments and to the variables of functions.
;; It doesn't matter if the actual implementation of the call stack
;; optimizes out the actual address references onto the cell objects and
;; keeps them in the computing registers only as these are considered part
;; of the uppermost stack frame as far as logical references are concerned.
;;
;; Prerequisites
;;
;; 1) We want to have a protocol which clearly states how and when to
;;    declare (= count) stack references without using any knowledge
;;    about the actual call graph as that one is mostly determined by
;;    the source code to be evaluated. Of course each and every known
;;    call graph could be optimized with respect to reference counting?
;;
;; 2) Locality: we want a protocol which doesn't force us to consider
;;    the complete unknown call graph of a program every time we insert
;;    reference declarations into a function (which works best as a
;;    second distinct step during development of a new function).


;;;
;; Reference Counting Primitives (Simplified Model)
;;
;; A simplified model for our reference count without any actual release
;; which we presume to happen when the count reaches zero.
;; Every return value of a generator (some function taking a free cell
;; and tagging it) gets a reference count of one by the generator.

(define (generator data)         (cons 1 data))
(define (minus-reference object) (set-car! object (- (car object) 1)))
(define (plus-reference object)  (set-car! object (+ (car object) 1)))
(define get-value  cdr)
(define set-value! set-cdr!)


;;;
;; CLR1/a
;;
;; We presume all of the function's arguments to be referenced correctly
;; and to be referenced one above because of the generators and to allow
;; for pass through values.
;; We dereference every argument and stack variable before returning
;; except for the return value. In other words the reference count of
;; the return value is raised by one before dereferencing all variables
;; and arguments of the function which includes the return value.
;; 

(define (f a b)
  ; (Some computation using /a/ and /b/.)
  ; Raise the count of the return value.
  (plus-reference a)
  ; Dereference our stack references.
  (minus-reference a)
  (minus-reference b)
  a)
(define (g a)
  ; (Some conditional or computation using /a/.)
  (minus-reference a)
  (generator 12))


(newline) (display "CLR1/a") (newline)
(repl-transcript

(define a (generator 'data))
a
(f (f a (g a)) (g a))

) ; closes repl-transcript

;; What happens here for the described protocol CLR1/a is a miscount into
;; the negative range: in fact when the reference count of /a/ reaches zero
;; it would be reinserted into the free list. Due to other computations it
;; might be reused before it is dereferenced a second time by the call graph
;; of our example. What is the problem with CLR1/a? It doesn't count every
;; stack frame as a reference but it dereferences on every stack frame or
;; function call. That mismatch has been used to craft this example against
;; the presented reference counting protocol. For simpler examples this
;; protocol might still work out. You could try to insert two reference
;; statements into this simple example to compensate for the missing stack
;; reference declarations. But more complex situations might include several
;; stack variables referencing the same object or not by coincidence which
;; might not always be known in advance.


;;;
;; CLR1
;;
;; We reference every argument in a preambel for each function and we
;; dereference every argument and variable in a postambel. The return
;; value's reference count is raised by one before that. Every variable
;; holding a logical reference raises the reference count of the denoted
;; memory object.

(define (f a b)
  ; Preambel
  (plus-reference a)
  (plus-reference b)
  ; (Some computation using /a/ and /b/.)
  ; Raise the count of the return value.
  (plus-reference a)
  ; Postambel
  (minus-reference a)
  (minus-reference b)
  a)
(define (g a)
  ; Preambel
  (plus-reference a)
  ; (Some conditional or computation using /a/.)
  ; Postambel
  (minus-reference a)
  (generator 12))


(newline) (display "CLR1") (newline)
(repl-transcript

(define a (generator 'data))
a
(f (f a (g a)) (g a))

) ; closes repl-transcript

;; This variant of the counting protocol denoted by CLR1 leads to a
;; miscount into the positive range. It would lead to a notable amount
;; of memory leakage. Why is that? When a function adhering to the
;; described protocol receives the return value of another such function
;; it is unaware of the raised count of the return value as it can't
;; distinguish data graph referenced or stack referenced memory objects
;; from memory objects referenced as return values. This is where CLR2/a
;; diverges and declares the return value reference to correspond to a
;; recognizable flag which is one exact half in the reference count of
;; any memory object.


;;;
;; CLR1/b
;;
;; Every call to a function needs to prepare the memory objects to be
;; referenced correctly for that function (including its stack frame
;; references). For the rest every function works like in the CLR1/a:
;; stack frame references have to be removed and thus the reference
;; count of every argument and variable has to be reduced if not for
;; the return value which has to be one above.

(define (f a b)
  ; (Some computation using /a/ and /b/.)
  ; Raise the count of the return value.
  (plus-reference a)
  ; Dereference our stack references.
  (minus-reference a)
  (minus-reference b)
  a)
(define (g a)
  ; (Some conditional or computation using /a/.)
  (minus-reference a)
  (generator 12))


(newline) (display "CLR1/b") (newline)
(repl-transcript

(define (ref object) (plus-reference object) object)
(define a (generator 'data))
a
(g (ref a))
a
(f (ref a) (g (ref a)))
(f (f (ref a) (g (ref a))) (g (ref a)))
(f (f (f (ref a)
         (g (ref a)))
      (g (ref a)))
   (g (ref a)))

) ; closes repl-transcript

;; There is again a miscount into the positive range for this variant called
;; CLR1/b. It's the same mistake as in the CLR1 where the reference count
;; of the return values starts to accumulate. Can we compensate for this
;; without knowing wether a function returns one of its arguments in use
;; as another argument elsewhere or not ? (See also scm-numbers.c for more
;; interesting examples.)


;;;
;; CLR1/c
;; We somehow try to compensate the shortcomings of CLR1/b: we only
;; reference the arguments of the innermost calls. The rest should be
;; handled by the protocol. All innermost calls minus one of them.

(define (f a b)
  ; (Some computation using /a/ and /b/.)
  ; Raise the count of the return value.
  (plus-reference a)
  ; Dereference our stack references.
  (minus-reference a)
  (minus-reference b)
  a)
(define (g a)
  ; (Some conditional or computation using /a/.)
  (minus-reference a)
  (generator 12))


(newline) (display "CLR1/c") (newline)
(repl-transcript

(define a (generator 'data))
a
(begin (plus-reference a)
(g a))
a
(begin (plus-reference a)
(f a (g a)))

(begin (plus-reference a) (plus-reference a)
(f (f a (g a)) (g a)))

(begin (plus-reference a) (plus-reference a) (plus-reference a)
(f (f (f a (g a))
      (g a))
   (g a)))

) ; closes repl-transcript

;; This CLR1/c has been tweaked to work as we definitely want to count
;; straight. Have we found the right protocol which just works? Or did we
;; abuse some property of our limited example? The protocol tells us to
;; count all innermost calls minus one of them. But in fact we're only
;; compensating the effect of the function /g/ which doesn't return its
;; argument. Another prospective function /h/ might return or one of its
;; arguments or a newly generated object depending on the values of
;; its arguments. CLR1/c doesn't really work if not by tweaking forever
;; and ever. But some more thinking and experiments and you might ...

;; (VSI aftermath)

;;; Addendum
;; Hope persists with full control of the actual call stack frames we
;; might be able to do away with the special return value flag. Some
;; self descriptive or known layout might allow to identify managed
;; memory object references from other data. In addition we would need
;; to know wether a reference had already been declared or we would
;; need to count every reference.
