(sys-load "VSI-core/repl.scm")
; (set! repl:color repl:blue-on-white)
(repl-clear)

;;;
;;; VSI/ANS revision of mathematical concepts
;;;

;;;
;; Reminder: Numbers (without their history)
;;
;; Natural Numbers: [1 2 3 4 5 ...[
;;            Zero: 0 (sometimes considered a natural number)
;;        Integers: ]... -3 -2 -1 0 1 2 3 ...[
;;       Rationals: division of integer by integer: n / m
;;
;;    Analysis Reapplied (n and m are natural numbers here)
;;   Convergencies: n / 0 -> infinity
;;                  0 / 0 is undefined
;;                  0 / n = 0
;;                  m / infinity -> 0+0  (!=  n / m / 0 = n * 0/m)
;;                  infinity |+| n = infinity
;;                  n < infinity
;;                  infinity / infinity is undefined
;;                  
;;      Irrational: Roots
;;  Transcendental: pi, e
;;    Real Numbers: Rationals U Irrationals U Transcendental
;;
;; Complex Numbers: i*i = -1 (imaginary unit)
;;
;; --
;; Notation Reminder: closed set: [] open set: ][
;;                    ellipsis means etc (continue like that)
;;                      union of: U
;;                    definition: =
;;                  converges to: ->
;;                some operation: |+|
;;;

;;;
;; Numbers, Digital Representations and Evaluation
;;
;; Numbers as Data Types (of digital computers)
;;
;; Representations of rational numbers
;;  immediate
;;  exact
;;  fraction
;;
;; Representations of (limited) approximations of real and complex numbers
;;  inexact
;;  inexact complex
;;
;; Representations of non numerical values
;;     Convergence: infinity, zero plus zero, zero minus zero
;; Undefined value: not a number
;;
;; The representation of the convergence towards zero is considered to
;; be the same as inexact zero.
;;
;; To /evaluate/ means to calculate the actual value of a formula.
;; Evaluation isn't meant to deliver mathematical properties of
;; an expression nor is it meant to transform an expression even
;; if that is often part of the evaluation.
;;
;; Inexact Operations and Convergencies:
;; The representations of plus and minus infinity can be considered
;; special error return values denoting a numerical overflow. Zero near
;; zero values would then signal a numerical underflow sometimes called
;; numerical overflow, too. In this respect there is another error value
;; called Not a Number or NaN which is returned for undefined operations
;; which don't make sense and can thus be considered an error like
;; dividing zero by zero.
;;
;; Exact operation also use these values for division by 0 and else where
;; they don't necessarily signal an error condition depending on the
;; intentions of the programmer.
;;
;; In addition to these special return values there are similar error
;; conditions which are signalled with a throw: numerical-overflow,
;; out-of-range and wrong-type-arg.
;;;

;;;
;; Operations
;;
;; Mathematical operations are available as functions of evaluation.
;; They take arguments and can be applied to arguments.
;;
;; Reminder: Counting
;;   one two three four
;;   two four six
;;   one two one two
;;   four three two one
;;   tic tac tic tac
;;
;; Folds and Recursions with Natural Numbers:
;;
;; Summation of two numbers /n + m/ means to count /m/ upwards /n/ times or
;; to count /n/ upwards /m/ times which is equal. Summation of /k/ numbers
;; equals the pairwise fold of the result and the next number.
;;
;; Multiplication of two numbers /n * m/ means to sum /n/ with itself
;; /m/ times or to sum /m/ with itself /n/ times which is equal.
;; Multiplication can be folded for /k/ numbers.
;;
;; Exponentiation of two numbers /n/ by /m/ means to multiply /n/ with
;; itself /m/ times.
;;
;;
;; Extension to Integers ...
;;
;; Extension to whatever on whatever
;;
;; /x/ raised to the power of some negative number /y/ is defined to
;; equal the division of one by /x/ raised to the power of /y/ without
;; its sign.
;;
;; /x/ raised to the power of some number /y/ between zero and one is
;; defined to equal the /y/th root of /x/.
;;
;; /x/ raised to the power of some inexact number /y/ greater than one is
;; defined as an approximation between the results of /x/ raised to the
;; power of the next greater and the next lesser integer around /y/.
;;
;; (Hey that even looks good as the graph of a function which allows us
;; to see the beauty of approximation by carbon or chalk.)
;;;


(namespace ; We currently don't want to redefine system variables here
           ; to later on push changes into the core and else.

;;;
;; Sketch of a Mathematical Function of Evaluation
;;  considering convergencies

(define (folding f a rest)
 (cond ((null? rest) a)
       (else (folding (f a (car rest))
                      (cdr rest))) ))

(define (fold operation a . rest) (folding f a rest))

(define (+inf? a) (equal? a +inf.0))
(define (-inf? a) (equal? a -inf.0))
(define (inf?  a) (or (+inf? a) (-inf? a)))

(define (nan? a)
  (not (or (equal? a +nan.0)
           (inf? a))) )

(define (number? a)
  (or (exact? a)
      (and (inexact? a)
           (not (nan? a))) ))

(define (undefined)    +nan.0)
(define (undefined? a) (eqv? a +nan.0))

(define (addition a b)
 (let ((result (+ a b)))
   (cond ((inf? result) (error 'numerical-overflow)) ; REVISED
         (else          result)) ))

(define addition-dispatch (lambda (a b)
 (cond ((and (number? a) (number? b))   (addition a b))
       ((or (+inf? a) (+inf? b))        +inf.0)
       ((or (-inf? a) (-inf? b))        -inf.0)
       ((or (and (+inf? a) (-inf? b))
            (and (-inf? a) (+inf? b)))  (undefined)) )))

(define sum (lambda ( . args)
 (cond ((null? args) 0)
       (else         (apply fold addition-dispatch args))) ))


(display "
;;;
;; Tests
")

(repl-transcript

;; Go check equality before playing with undefined.
(equal? (undefined) (undefined))
(eqv?   (undefined) (undefined))
(eq?    (undefined) (undefined))
(=      (undefined) (undefined))

(newline)
(equal? (undefined) +nan.0)
(eqv?   (undefined) +nan.0)
(eq?    (undefined) +nan.0)
(=      (undefined) +nan.0)

) (newline)


(newline)
(repl-transcript

;; identity
(= (+) 0)

;; Convergence
(newline)
(equal?     ( +  -inf.0 -inf.0) -inf.0)  ;; XXX eqv?
(equal?     ( +  +inf.0 +inf.0) +inf.0)
(undefined? ( +  +inf.0 -inf.0))

(define k (expt 2 1500))
(define l 10000000.123456789)
(+ k l)
(addition k l) ; REVISED, REJECTED

) ; closes repl-transcript
) ; closes namespace


(display "
;;;
;; Harmonization of Symbolic Numericals and Serialization
;;
")

(repl-transcript

(/  1 0)
(/  0 1)
(/ -1 0)
(/  0 0)

(/ 0 0 0) ; for these operations we rely on left to right order of evaluation
(/ 0 1 0)
(/ 1 0 0)

;#(     ; we could use a vector to prevent lookups of values bound to symbols
 1/0
 0/1
-1/0
 0/0

0/0/0  ; these are read in as non numerical symbols
0/1/0
1/0/0
;)
)

;;
;; Outer Form (serialization) of Numericals and Corresponding S-Expressions
;;
;; While the above sequence of expected values  is defined by the
;; introduction to this file one should note the implicit change of the
;; outer form of these border line fractions. That behavior is currently
;; accepted because the structure of the outer forms of numbers in the
;; ALS was (and in ANS currently still is) complex enough to count as
;; abbreviated operations (and all that while one fundamental rule stated:
;; everything which isn't "recognized" as a number is to be considered a
;; non numerical symbol.)
;;
;; Example:
;;
;; VSI> #E1.2
;; 6/5 => 6/5
;;
;; where #Enum equals an (inexact->exact num)
;;
;; Enforced Exactness
;;
;; #Ea   (inexact->exact a)
;; #Ia   (exact->inexact a)
;; a/b   (/ a b)
;;
;; Base Selection:
;;
;; #xa   (string->number (symbol->string a) 16)
;; #ba   (string->number (symbol->string a)  2)
;; #oa   (string->number (symbol->string a)  8)
;; #da   (string->number (symbol->string a) 10) ; default
;;
;; Signs can't easily be expressed as corresponding s-expressions
;;
;; -a    (- a) ; two minus signs (--) will be taken for a non numerical symbol
;; (symbol? -7/-8) => #t
;;
;; +a    a     ; (+ a) => a
;;
;; /number->string/ on the other hand doesn't use the sharp prefix
;; /write/ doesn't use any prefix and uses the default base ten
;;
;; A change of the outer of form of numericals where the inner
;; structure is dissolved into the regular s-expression notation
;; with corresponding abbreviations might or might not turn out
;; to be useful.
;;
;; The difference between numerical symbols (e.g. sequence of digits)
;; and non numerical symbols (e.g. sequence of letters) seems to have
;; been considered an implementation detail in the beginning or in the
;; tradition of list processors. There is no intention to change this.


;;;
;; PS  In English, please! (Might sometimes just mean: screwed it up
;;                          or got it wrong.)
;;
;; 10.000,000
;; 10,000.000 (this might even have been superseeded by the new
;;             kilo Mega, Giga and Tera notation as well as by
;;             various forms denoted as exponential notation)
;;
;;
;; The magnitude of a vector is its number of dimensions where the
;; length of a vector is that infamous application of the dot product.
;; Reapplying this to the implementation means the magnitude should
;; return the number of slots where the length might have been
;; inspired by ...
;;
;; 
;; Radix means root and the exponent in current notation is written
;; above a base (which is nothing but some arbitrary notation then).
;; For integers that base might even equal the n'th root of the
;; result of (expt base n).
;;
;; What was this about then?
;;; VSI Aftermath
