#!/opt/VSI/interpreter --
!#

;;
;; mti's Crypto Sweet: Human Typable Shared Secret (System Scheme)
;; (make-passwd.scm)
;;
;; Functional Programming seemed to be orthogonal to system
;; programming where data is passed in /pipelines/ and similar
;; concepts.
;;
;; This software is in the Public Domain as granted by the original
;; author. You found your copy in the virtual public anyway, right?
;;
;; This implementation creates a human typable shared secret
;; especially for protecting online accounts and it should be used in
;; conjunction with keychains or password managers. It is defined to
;; not use certain special characters which are known to be difficult
;; to find if the keyboard layout in software doesn't match the actual
;; inscriptions on the keys. It is designed to be usable with
;; secondary backups. Its shared secrets have the property of being
;; hard (but not impossible) to communicate by oral speech. They are
;; also hard to memorize especially when you can look at them only for
;; a short time. There might be other algorithms based on the
;; knowledge of the syllable structure of common western (and perhaps
;; other) languages that could produce secure /master passwords/ not
;; be held within keychains and password managers.
;;
;; The magnitude of the space of distinct shared secrets is high
;; enough to be considered secure with current authentication routines
;; in use on the Internet(1). For the original implementation with 65
;; characters and 12 positions it should be: (expt 65 12) which is
;; greater than
;;
;;   (expt 64 12) => (expt (expt 2 6) 12) => (expt 2 72)
;;
;; You can improve it by adding your own preferred special
;; characters. TODO: It should take a command line argument
;; representing the length of the desired shared secret.
;;
;; Please consider changing the shared secrets (Soup of Letters) to
;; your needs: sometimes they force you to use special characters in
;; passwords although real randomness - as employed by this
;; implementation - allows shared secrets without them. Do not change
;; this implementation but add the needed characters to the final
;; shared secret yourself at some random position. A further useful
;; adaption would be insert spaces but beware that they might be
;; difficult to read on secondary backups.
;;
;; (1) Internet isn't considered secure.

;; No abbreviations here: I want to see the characters.
;; Strings & Brass: These can be any characters (code points) but
;; the algorithm is only specified for a set smaller than about 250
;; characters.
(define *characters*
  (append (string->list "ABCDEFGHIJKLMNOPQRSTUVWXYZ")
          (string->list "abcdefghijklmnopqrstuvwxyz")
          (string->list "1234567890")
          (string->list "!,.")))
(define *number-of-characters* (length *characters*))
(define *length-of-password* 12)

(display "make-passwd: number-of-characters: ")
(display *number-of-characters*)
(display "; shared secret size: ")
(display *length-of-password*)
(newline)


(define (throw-dices randomness-port)
  (let throw-dices ((n *length-of-password*))
    (if (<= n 0)
        '()
        (cons (read-byte randomness-port) ; !!! Strings&Brass change
              (throw-dices (- n 1))) )))
(define dices->seed list->byte-string)    ; !!! Strings&Brass

(display "make-passwd: Throwing Dices")
(define randomness-device "/dev/urandom") ; should be non-blocking pool access
;; /call-with-input-file/ opens the input device in text mode but see
;; the description of binary mode: no difference on UNIX systems.  If
;; this changes perhaps due to Unicode system libraries we might get a
;; bias into our random data but as we use the dices only as the seed
;; for the randomizer we might get away with it. Please check your
;; system's source code for further information.
(define my-dices (call-with-input-file randomness-device
                                       throw-dices))
(display " => the seeds have been thrown: #")
(display (length my-dices))
(newline)


(define (get-new-state)
  (if (null? my-dices)
    (display "make-passwd: Oh dear, we failed! --")
    (let ((dice (car my-dices)))
      (set! my-dices (cdr my-dices))
      (seed->random-state (dices->seed (list dice))) )))
(define (x n)
  "This procedure is undocumented."
  (if (<= n 0)
      '()
      (cons (random *number-of-characters* (get-new-state))
            (x (- n 1))) ))
(define (soup n)
  (map (lambda (index)
         (list-ref *characters* index))
       (x n)))

;; This leaves a trace of the newly created secret in memory because
;; it creates a string before displaying. That string remains in
;; memory and will pop up as random data in another process's memory
;; after the interpreter quit.
(define (make-passwd)
  (display "Soup of Letters: ")
  (display (list->string (soup 12)))
  (newline))

;; This secured generator doesn't strive for elegance: the newly
;; created secret exists only as a list in memory, i.e. as a sequence
;; of otherweise urnelated pairs. Their content is destroyed after
;; displaying the list and the secret will remain a secret. (If not
;; for the data-flow (input/output) buffers of the interpreter, the VT
;; and even the OS itself. This is ugly pseudo security but worth not
;; thinking about it - sorry for the inconvenience.
(define (print-passwd-and-destroy)
  (define (destroy-list! list)
    (if (not (null? list))
     (begin (set-car! list #f)
            (destroy-list! (cdr list))) ))
  (display "Letter Hash: ")
  (let ((wannabe-secret (soup 11)))
    (for-each display wannabe-secret)
    (destroy-list! wannabe-secret))
  (newline))

;(make-passwd)
(print-passwd-and-destroy)
