;; You'll need to load this file in the REPL if you want
;; to try the examples and see how "uncorrelated" two
;; dimensions with an effective angle of 90° are.
;;
;; Freeze: Polar Vector Addition (to that vector calculus)
;;
;; NB In the following an attempt is described to add an
;; n-dimensional polar representation for vectors which
;; is by now considered a proof against such a representation
;; or at least against its usefulness.

;;
;; Polar Vectors are described in terms of a vector
;; with the magnitude in its first slot and the angles
;; in the other slots (relative to a full circle of
;; 2*pi clock wise and the preceding slot dimension as 0°
;; (just after 2*pi+(0-0)) as expected.
;;
;;
;; While the conversion from cartesian coordinates to
;; polar normalization works reasonably well (keeping
;; NaN objects in unused dimensions) the reconversion
;; from polar to cartesian fails for a lot of cases but
;; works reliably for too many cases to be completely off.
;;
;; Currently this works well for scaled unit vectors
;; where all components are the same value. There is a
;; need for an arbitrary /c/ in the computation as the
;; angle asin is only defined for the unit circle but as
;; the angle only describes the direction this scaling
;; should not be a problem. The resulting cartesian
;; components maintain equality but proportions are off.
;; (The current algorithm always scales to the correct
;; magnitude.) There are no problems on a plane as long
;; as there is no third coordinate given. Triples with
;; a final value of zero exhibit the same problems as
;; described above while a value of zero in the first
;; position gives correct results. A value of 1 in the
;; first position seems to diverge to (sqrt 2) for
;; equal and growing values of the second and third
;; components.
;;
;; (vpt #(1 2000 2000)) => #(1.41421329720808 1999.99987500009 1999.99987500009)
;; (vpt #(0 2000 2000)) => #(-5.19573633741296e-13 2000.0 2000.0)
;;
;; It's the second time during the "clean room" development
;; of the polar conversion routines that the "magic" value
;; (sqrt 2) appears in an unexpected spot. Last time it should
;; have been (sqrt 3) describing the length of the unit vector
;; with a magnitude of three dimensions and the mistake was
;; caused by using the dot product instead of the final
;; vector-magnitude as that started out from a mere sum of
;; components.
;;
;; (vpt #(2000 0 2000)) => #(2000.0 -3.67394039744206e-13 2000.0)
;; (vpt #(2000 1 2000)) => #(2000.0 0.999999999999947 2000.0);
;;
;; Are these inaccuracies of computation or is the current
;; algorithm only an approximation within certain ranges?
;; That is difficult to tell with approxiating functions
;; called sin and cosin and asin and acos.
;;
;; (define v #(2 300 0.1)) => #<unspecified>
;; (vpt v) => #(1.99995567011565 300.000000295529 0.10000000000891)
;;
;; Reverting that operation [axis-angle-calculation] doesn't seem to
;; be that easy.  Maybe another normalization with the angles
;; describing the angles relative to the planes (instead of the
;; current projection onto a plane) could give better results.
;;
;; Additionally there is a problem with void dimensions where the
;; additional minus sine (/unit-b/) from the previous component
;; would have been the correct value but cos fails which can be
;; considered an error as these should be the same.
;;
;; cos -sin pairs for p1 = #(1 1 0)
;; (cartesian->polar-vector #(1 1 0)) => #(1.4142135623731 5.49778714378214 0.0)
;; (polar->cartesian-vector p1) => #(0.707106781186547 0.707106781186548 1.0 -0.0)
;; the components 2 and 3 ie -sin 7/4 pi and cos 0.0 would need to be equal
;; (prerequisite of the algorithm) to be able to reconstruct the original
;; coordinates. (cos ro = a; -sin ro = b; a = a_1 and b = a_2 for the unscaled
;; reconstructed vector. Perhaps the pair wise reduce requests correlating
;; unfolding or some information is just lost even if the magnitude and the
;; angles of the polar representation do seem to be correct by all means.
;;
;; In the end: this is nothing but an approximating projection which let's
;; the vectors bend. ;-)

;
; a^2 + b^2 = c^2
;
; sin ro  = a
; ro = sin^(-1)(a)

;; (define rec-tangle (%vc-divide %vc-pi 2))
;; (define 2pi (%vc-multiply 2 %vc-pi))

(define (cartesian->polar cartesian)
 (define (quadrant-correction x y angle)
  (cond ((<= y 0)  angle)
        (else      (%vc-minus (%vc-multiply 2 %vc-pi) angle)) ))

 (define (dimensions->angle a b)
  (let ((c (%vc-square-root
            (%vc-plus (%vc-multiply a a) (%vc-multiply b b)))))
   (quadrant-correction a b (%vc-acos (%vc-divide a c)))
 ))

 (list->vector
  (cons (vector-magnitude cartesian)
        (reduce dimensions->angle
                (vector->list cartesian)))) )

;; For some "simple" conversions this is already
;; down to 10^(-6).
(define polar-inexactness (/ 1 (expt 10 12)))

(define (two-pi-periodic angle)
 (inexact->exact
  (rationalize (inexact->exact (%vc-divide angle (%vc-multiply 2 %vc-pi)))
               polar-inexactness)))
;
; ro = acos (/ a c)
; cos ro = a / c
; c * cos ro = a
; c = sqrt ( a^2 + b^2 )
; b = c * sin ro
; c = sqrt ( a^2 + (sin ro)^2 )
; c = sqrt ( (c*cos ro)^2 + (c*sin ro)^2)
; c^2 = c * cos ro * c * cos ro + c * sin ro * c sin ro
; c^2 = c^2 * (cos ro)^2 + c^2 * (sin ro)^2
; 1 = (cos ro)^2 + (sin ro)^2 ???
;
; magnitude = sqrt ( a_1^2 + a_2^2 + a_3^2 + ... + a_n^2)
; a_1^2 + a_2^2 = magnitude^2 - (a_3^2  + ... + a_n^2)
; a_1^2 = magnitude^2 - (a_2^2 + a_3^2 + ... + a_n^2)
;
(define (unit-vector n) (make-vector n 1))
(define (polar-unit-vector n)
  (cartesian->polar (unit-vector n)))

(define (polar magnitude angles)
  (list->vector (cons magnitude angles)))
(define (polar* magnitude . angles)
  (list->vector (cons magnitude angles)))
(define (polar-magnitude polar) (vector-ref polar 0))
(define (polar-angles    polar) (cdr (vector->list polar)))

(define (1D-polar->cartesian-vector polar)
 (let ((magnitude (polar-magnitude polar)))
  (vector magnitude)))

(define (2D-polar->cartesian-vector polar)
 (let ((magnitude (polar-magnitude polar))
       (angles    (cdr (vector->list polar))))
   (vector (* magnitude (cos (first angles)))
           (* magnitude (- (sin (first angles)) )) )))

;;
;; 0° for two dimensions means: take the full length of the
;; last dimension {or the complete magnitude for the first
;; element???}:
;;  cos 0 = 1 => this dimension gets the full length but
;;
;; #(x y z) -> acos x/c acos y/c where c stems from c = sqrt (a^2+b^2)
;;                                           with a = a_n and b = a_(n+1)
;; this corresponds to the following geometric operations
;; scaled projection on the xy-plane and angle to the x axis (acos x/c_1)
;; scaled projection on the yz-plane and angle to the y axis (acos y/c_2)
;;
;; (cartesian->polar #(1 0 0)) => #(1.0 0.0 +nan.0)
;; (polar->cartesian #(1.0 0.0 +nan.0)) => #(1.0 0 0)
;;
;; The first angle component of 0.0 tells us there is no angle
;; between the vector and the x axis. The second angle component
;; is not a valid angle but helps determine the magnitude or
;; number of dimensions of the vector. That second component is
;; currently replaced with 0 and we would expect 1 and 0 as the
;; resulting unit-coordinates (unscaled "directions") as cos 0 = 1
;; and -sin 0 = 1 but we receive 0 and 0 which is correct for this
;; conversion but not what we would expect from that semi-broken
;; algorithm.
;;
;; unit-a should equal the previous unit-b, ie it should be scaled,
;; because we know it denotes the same axis component and we need
;; to maintain the /chain of relations/ to be able to simply scale
;; with the magnitude???
;;
;; unit-a * c = unit-b_n-1
;; c = unit-b_n-1 / unit-a
;;
;; Keeping this chain we eliminate the "error" introduced by
;; scaling the projection to match the unit circle to be able
;; to compute the angle which we wanted to have for the pair
;; wise dimensions.
;;
;; Now we still have to handle void dimensions and NaNs better:
;;
;; VSI> (vpt #(1 0 1))
;; (vpt #(1 0 1)) => unit-a: 1.0
;; unit-b: -0.0
;; unit-a: -1.83697019872103e-16
;; unit-b: 1.0
;; #(1.4142135623731 -0.0 0.0)
;; VSI> (cartesian->polar #(1 0 1))
;; (cartesian->polar #(1 0 1)) => #(1.4142135623731 0.0 4.71238898038469)
;; VSI> (vpt #(2000 1 2000))
;; (vpt #(2000 1 2000)) => unit-a: 0.999999875000023
;; unit-b: 4.99999937559034e-4
;; unit-a: 4.99999937499985e-4
;; unit-b: 0.999999875000023
;; #(1999.9999998819 1.000000000059 2000.0000001181)
;;
;; Maybe we even need a better way to determine the sign
;; (which shouldn't be influenced by the /factor/ at all?)
;;
;; VSI> (vpt #(0 1 1))
;; (vpt #(0 1 1)) => unit-a: -1.83697019872103e-16
;; factor: -5443746451065120.0
;; unit-b: 1.0
;; unit-a: 0.707106781186547
;; factor: -7698620061216700.0
;; unit-b: 0.707106781186548
;; #(1.83697019872103e-16 -1.0 -1.0)
;;
;; That's the fun of implementing a system based on inexactness.
;; How about a /symbolic-sin/ which takes a symbolic /pi/? Then
;; simplify and calculate but Jacal won't reach there.
;;
;; (/ 1 0) => +inf.0
;; (/ 1 (* -2.59786816870648 (expt 10 -16))) => -3849310030608350.0
;;
;; TODO use %vc- operators

(define (polar->cartesian polar)
 (define (angles->unit-coordinates angles)
  (let* ((answer-box (cons #f '())))
   (let loop ((angles angles)
              (last   answer-box)
              (last-b 1))
    (cond
     ((null? angles) (cdr answer-box))
     (else (let* ((angle  (first angles))
                  (unit-a (cos angle))
                  (unit-b (- (sin angle)))
                  (factor (abs (/ last-b unit-a))))

      ;; XXX inexactness harmonization
      ;; This and the factor nan check help with #(1 0 1)
      ;; but for #(1 0 2) the scaling factor is lost. Then
      ;; the presumed rounding error reaches 10^(-6) for
      ;; regular numbers or even 10^(-3) for #(1 20 300)
      (when (not (nan? unit-a))
        (set! unit-a (rationalize unit-a (expt 10 -14))))
      (when (not (nan? unit-b))
        (set! unit-b (rationalize unit-b (expt 10 -14))))
      (set! factor (abs (/ last-b unit-a)))
      
      ;; (cond
      ;;  ((and (zero? unit-a) (zero? unit-b))
      ;;     (set! factor 1)) ; ??? where does this come from ? experience?
      ;;  (else
      ;;   (set! factor (abs (/ last-b unit-a))) ))

      ;; ;; We somehow passed through zero ??? => reset the algorithm
      ;; (when (inf? factor)
      ;;   (set! factor 1)
      ;;   (set! unit-b 1))

      ;; This solves #(0 0 1) but doesn't help with #(1 0 2) as the
      ;; the scaling is gone.
      (when (nan? factor) ;; This dimension lost its scaling info ...
        (set! factor 1))

      (display "unit-a: ") (display unit-a) (newline)       
      (display "factor: ") (display factor) (newline)       
      (display "unit-b: ") (display unit-b) (newline)       

      (set-cdr! last (list (* factor unit-a)))
      (set! last (cdr last))

      (when (null? (cdr angles))
        (set-cdr! last (cons (* factor unit-b) '()))
        (set! last (cdr last)))

      (loop (cdr angles)
            last
            (* factor unit-b)) )))
 )))
 
 (let* ((unit-coordinates
         (map (lambda (x) (if (nan? x) 0 x)) ;; ???
              (angles->unit-coordinates (polar-angles polar))))
        (tmp-vector (list->vector unit-coordinates))
        (tmp-length (vector-magnitude tmp-vector))
        (unit-magnitude-factor (/ (polar-magnitude polar)
                                  tmp-length)))
 (vector-scale unit-magnitude-factor tmp-vector)))
 ;(vector-scale 1 tmp-vector)))
      
(define (vpt vector)
  (polar->cartesian (cartesian->polar vector)))

;;
;; Interestingly enough the current representation doesn't seem
;; to be suited to reliably describe all possible vectors. The
;; pair wise creation of the angles "stumbles" when there is
;; a void dimension in between: #(1 0 1) and #(1 0 2)
;;
;; (vpt #(1 0 1)) => #(1.0 0.0 1.0)
;; (vpt #(1 0 2)) => #(1.58113883008419 0.0 1.58113883008419)
;;
;; (cartesian->polar #(1 0 2)) => #(2.23606797749979 0.0 4.71238898038469)
;;
;; Usually the relation of the components of the vector as sides of
;; a regular circle triangle with a right angle is described by the
;; pair wise angles. That chain is broken with the current pair wise
;; reduction to angles when there is a void dimension. Thus a magnitude
;; plus the pair wise angles between the components of the vector
;; doesn't (seem) to convey enough information to restore the original
;; vector. Maybe void dimensions should be skipped and replaced with 90°
;; but this asks for a special protocol between /cartesian->polar/ and
;; /polar->cartesian/ which isn't considered a general algorithm anymore.
;; Needless to say that such a protocol would limit the usefulness of
;; a polar representation even more: the angles are interdependent but
;; with such a protocol these would be nothing but encoded information?
;;
;; Especially because the usefulness of the current representation is
;; limited: you have two angles: one between x,y and one between y,z
;; which doesn't allow for rotations around the y axis. Well, it
;; should be possible to represent any vector consisting of non void
;; components with the current normalization but it isn't possible to
;; easily calculate that third rotation. But with intact inexactness
;; you're imagination may find other places in wonderland ...

(define (polar-rotate polar-vector . angles)
 (polar (polar-magnitude polar-vector)
        (map %vc-plus (polar-angles polar-vector)
                      angles)))

(define (polar-scale s polar-vector)
 (polar (%vc-multiply s (polar-magnitude polar-vector))
        (polar-angles polar-vector)))

;;
;; Applications / Higher Order Operations
;;

(define (rotate-vector rotation vector)
 (polar->cartesian
  (apply polar-rotate (cartesian->polar vector)
                      rotation)))

;;
;; Additional Comments
;;
;; - Coersion of scalar / vector types is left to the user:
;;   a scalar vector of one dimension is still considered a
;;   vector and not a scalar in the current implementation
;;
;; - There have been and probably are a lot of "problems" with
;;   inherent inaccuracies (or rounding errors) where some
;;   formulas should yield 0,0 but arrive at about 1 * 10^-16 which
;;   is not considered zero by most routines. Harmonization of these
;;   rounding inaccuracies is difficult and not always possible.
;;   Therefore rounding and coping with rounding errors is left as
;;   a responsibility of the programming user.
;;   - Maybe it would be better to express angles as exact fractions
;;     of a rationalized modulo two pi.
;;   - NB 10^16 is in the range of 2^54 which maybe by coincidence
;;     corresponds to the bits available for integers in the current
;;     implementation of SCM on machines with word size of 64 bit.
;;     (but we're talking doubles here and that was 2^56 not 54)
;;
;; - Additionally some shortcomings of the original implementations
;;   of some mathematical functions show up ever again.
;;   - /rationalize/ didn't rationalize inexact numbers correctly
;;   - /asin/ and /acos/ were somehow extended to switch to the
;;     realm of complex numbers where in practice they now have been
;;     changed to exhibit a hopefully more useful behavior.
;;   - sometimes the special +nan.0 is passed on silently sometimes
;;     it is considered an error (/inexact->exact/). That behavior
;;     might make sense when each and every function is considered
;;     on its own but it is daunting for regular programming tasks
;;     where the overall system sometimes surprises with being able
;;     to cope with (accepted) error conditions but then fails for
;;     the next iteration of algorithms. The current number system
;;     consisting of all accessors isn't well harmonized.
;;
