
   ;;   ;;   ;;     Graphical
 ;;     ;; ;;  ;;   Input            Viper System Interface
 ;;  ;; ;; ;;  ;;   Output
   ;;   ;;   ;;                  

                ;;     ANS Color Representation for Algorithmic Processing
 ;;               ;;
 ;;;;;;;;;;;;;;;;;;;;;     Colors are represented as Color Distances
 ;;               ;;        in the RGB color space, i.e. 3D vectors.
                ;;

;;
;; Preambel
;;

(sys-load "lib/vector-geometry.scm")

;; Maths Adapt
(define (color:quotient  a b) (floor (/ a b)))
(define (color:remainder a b) (- a (* b (color:quotient a b))))

;; Sys Adapt
(define color-error error)


;;
;; Basic Color Datatype Accessors
;;

(define (color-normalized? C)
 (and (numerical-vector? C) (let ((Cl (vector->list C)))
                              (and (<= (apply max Cl)  0)
                                   (>= (apply min Cl) -1))) ))

(define (color vis)  ; subject to change: the current quoted list approach imposes limits
 (case (car vis)
 ((rgb)       (list->vector (cdr vis) ))
 ((hsv)       (apply color:hsv->rgb (cdr vis)))
 ((rgb-bytes) (vector-scale -1/255 (list->vector (cdr vis)) ))
 (else        (color-error (format "unknown color format: ~A" vis)))
))

(define (color:red   color) (vector-ref color 0))
(define (color:green color) (vector-ref color 1))
(define (color:blue  color) (vector-ref color 2))


;; Color Equality TO BE REVIEWED
;;
;; (define color:acc (expt 10 -12))
;; (define (color:equal-op? cl1 cl2) (map (lambda (a b) (inexact= color:acc a b)) cl1 cl2))
;; defunct (define (color:equal? . colors) (reduce-predicate color:equal-op? (map vector->list colors)))

(define color:equal? equal?)


;;
;; Color Definitions
;;

;; Descriptive Constant Vector Space Unit & Origin
(define white (color '(rgb -1 -1 -1)))
(define black (color '(rgb  0  0  0)))

;; Descriptive Constant Cubic
(define red   (color '(rgb -1  0  0)))
(define green (color '(rgb  0 -1  0)))
(define blue  (color '(rgb  0  0 -1)))

;; Constant First Order Interpolations
(define cyan    (vector-plus green blue))
(define magenta (vector-plus blue  red))
(define yellow  (vector-plus green red))

(define color:rainbow (list red yellow green cyan blue magenta))

;; Constant Canonical with Postfix
(define   white-color white  )
(define   black-color black  )
(define     red-color red    )
(define   green-color green  )
(define    blue-color blue   )
(define    cyan-color cyan   )
(define magenta-color magenta)
(define  yellow-color yellow )


;; Additional Color Definitions (subject to change)

;; Second Order Interpolations
(define orange-color    (vector-middle red   yellow))
(define lime-color      (vector-middle yellow green))
(define turquoise-color (vector-middle green   blue))
(define aquamarin-color (vector-middle cyan    blue))
(define violet-color    (vector-middle blue magenta))
(define pink-color      (vector-middle magenta  red))

(define vsiorange (color '(rgb-bytes 248 96 24)))
(define vsiblue   (color '(rgb-bytes 24 96 248)))



;;
;; Hue Saturation Value
;;  One loop across six sides of the cube.
;;      Scaling towards the unknown white.
;;              Scaling to the black zero.

(define (color:hsv->rgb h s v) (color-value (color-saturation (color:hue->rgb h) s) v))
(define (color:hue->rgb h)
 (let* ((side  (modulo (inexact->exact (truncate h)) 6))
        (ratio (color:remainder h 1)))
 (case side
 ((0) (vector-ratio     red-color  yellow-color ratio))
 ((1) (vector-ratio  yellow-color   green-color ratio))
 ((2) (vector-ratio   green-color    cyan-color ratio))
 ((3) (vector-ratio    cyan-color    blue-color ratio))
 ((4) (vector-ratio    blue-color magenta-color ratio))
 ((5) (vector-ratio magenta-color     red-color ratio))
 (else (color-error (format "Color hue out of range in color-hsv->rgb ~A" h)))
)))

(define (color-value      color value) (vector-scale value color))
(define (color-saturation color saturation)
 (let* ((wp white-color)
        (bp (vector-scale -1 white-color))
        (c  (vector-minus color wp))
        (d  (vector-scale saturation c))
        (e  (vector-minus d bp )))
    e))

;; from RGB to HSV

(define (color:rainbow-index C)
 (let loop ((idx 0) (r color:rainbow))
  (cond ((null? r)           #f)
        ((color:equal? C (car r)) idx)
        (else               (loop (+ idx 1) (cdr r))) )))

(define (color:rainbow<= C1 C2) (<= (color:rainbow-index C1) (color:rainbow-index C2)))
(define (color:sort-by-rainbow Cl) (sort Cl color:rainbow<=))
(define color:rainbow-index->hue identity)


;; The color /C/ should be normalized by /value/ and /saturation/.
(define (color:rgb-hue C) ; second implementation: the outer form remains "beautyful"
 (cond
 ((apply = (vector->list C))     12) ; default hue
 (else (let* ((Dal (sort (map (lambda (Ln) (cons Ln (vector-distance C Ln)))
                              color:rainbow)
                      (lambda (a b) (<= (cdr a) (cdr b)) )))
              (Dal3 (list-head 3 Dal))
              (Dl (color:sort-by-rainbow (map car Dal3) )))
  ; magenta-red loop illusion
  (when (and (equal? red-color (first Dl)) (not (equal? yellow-color (second Dl))))
    (set! Dl (list (second Dl) (third Dl) (first Dl)))
    (when (not (equal? magenta-color (second Dl))) (log-warn "color:rgb-hue sth strange happened")))
  (cond
  ((color:equal? C (car (first Dal))) ; we hit a point on the rainbow
     (color:rainbow-index->hue (color:rainbow-index C)))
  ((apply = (map cdr Dal3))
     (second Dl))
  (else (set! Dl (color:sort-by-rainbow (list-head 2 (map car Dal))))
        (when (and (equal? red-color (first Dl)) (not (equal? yellow-color (second Dl))))
          (set! Dl (list (second Dl) (first Dl)))
          (when (not (equal? magenta-color (second Dl))) (log-warn "color:rgb-hue unexplainable")))
        (let ((side  (color:rainbow-index (first Dl)))
              (ratio (vector-triangle-ratio (second Dl) C (first Dl))))
          (when (nan? ratio) (set! ratio (vector-line-ratio (first Dl) C))) ;; probably inexact= to a rainbow loop point
          (color:rainbow-index->hue (+ side ratio)) )) )) )))
        

(define (color:rgb-value C) (- (apply min (vector->list C))))
(define (color:rgb-saturation C)
 (let* ((Cl   (vector->list C))
        (Cmin (apply min Cl))
        (Cmax (apply max Cl)))
  (cond
  ((= Cmin 0) 0)
  (else       (- (/ (delta Cmax Cmin) Cmin))) )))

(define (color:rgb->hsv C)
 (let* ((iv  (color:rgb-value      C))
        (Civ (color-value C (/ 1 iv)))
        (is  (color:rgb-saturation Civ)))
  (list 'hsv (color:rgb-hue (color-saturation Civ (/ 1 is))) is iv)))


(when #f ;; RGB<->HSV testing paragraph
(sys-load "VSI-core/repl.scm")

(repl-transcript
 (map color:rgb-hue color:rainbow)
 (color:rgb-hue orange-color)
 (color (list 'hsv (color:rgb-hue orange-color)  1 1))
 (color (list 'hsv (color:rgb-hue magenta) 1 1))
 (color (list 'hsv (color:rgb-hue violet-color)  1 1))
 (define sdv (color-value (color-saturation violet-color 1/2) 1/2))
 (color:rgb-hue violet-color)
 sdv
 (color:rgb-value sdv)
 (color:rgb-saturation sdv)
 (color:rgb->hsv sdv)
 (color:rgb->hsv #(0 -0.5 -0.5))
 (color:rgb->hsv white)
 (color:rgb->hsv black)
 (color:rgb->hsv (vector-ratio white blue 1/4))
 (color:rgb->hsv (vector-ratio black blue 1/4))
 (color '(hsv 600 0 1/2))
 (color:rgb->hsv #(-1/2 -1/2 -1/2))
)

(when #f ; paragraph of not normalized values / extra range TODO
(repl-transcript
(color:rgb->hsv (color-value white -1))
(color:rgb->hsv (color-value white  2))
(color:rgb->hsv red)
(color:rgb->hsv #(256 0 0))
(color:rgb->hsv #(-256 0 0))
(color:rgb->hsv #(-256 3 2))
(color:rgb->hsv #(256 3 2))
)
)
(REPL)
)
