
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; NB Event objects are destroyed at the end of an event arc
;;    in the low level event handlers which created them.
;;    Application code should make a selective copy when if needed.

;; Practical Issue Paragraph
;;
;;; Pointer Device Experience
;;  How was that driver called? Seismir or something which sometimes just ...
;; - Point and Click w/o multitouch works and reports button 1
;; - Multitouch sometimes yes sometimes no reports button 4 press release
;;   for a tap (where taps are actually turned off in the preferences)
;; - Multitouch drag (scrolling action request) is much too sensible
;;   and severely interferes with the tap [if not handled correctly]
;;     down: button 5
;;       up: button 4
;;     left: button 6 
;;    right: button 7 
;;   [the above axes denote user movement] There actually seems to be some
;;   kind of dynamic "speedometer" implemented which smooths fast
;;   multiouch movements lowering the actual "speed" of the action requests.
;;   Therefore a linear scrolling step works astonishingly well in some
;;   test case. Redrawing with a window clear leads to a slight flickering
;;   effect for the simple logo (depending on the actual load). <- That's
;;   still the original IO path of the prototype which carries each request
;;   to the server individually and additionally real flicker is only visible
;;   with flowing debug messages on the terminal. Sequence numbers go up as
;;   hell in the range of thousands for scrolling actions. [<- Motion events
;;   were turned on in the event mask but ignored by the empty override.]
;;
;;   Much too sensible means there's "noise on the wire". All of the above
;;   scrolling action requests might be interleaved with misinterpreted
;;   "noise". [Let's call it inexact scrolling which works great: the
;;   dynamic speed effect above can be completely stopped with multitouch
;;   tap stopping the ever slower scrolling action at once. On the other
;;   hand a really fast movement (long scrolling action) continues (to
;;   send events) even with interleaved clicks which aren't handled like
;;   taps. Maybe there are some guidelines somewhere about actions, events
;;   synthesized events (where autorepeat just bumps in unrecognizably) ...
;;   nothin' is mission critical when you only touch and don't click, is it?]
;;
;;   [There seemed to be some round of adaption vapor during the initial
;;   scrolling action test. ] <- That could not be confirmed by subsequent
;;   testing sessions.
;; 
;;   ADDENDUM There's purposefully some "validator" (or similar) in the
;;   input extension (v1 or v2) which pins buttons 4 to 8 to scrolling
;;   actions especially tuned when actionized from touchypoods ooweee what a ... 
;;
;; - Multitouch & Click produces interleaved events of buttons 1 and 3 
;;
;;; Keyboard Experience
;; - There's no known way to distinguish autorepeat events from regular
;;   key events. There's not even a double down or double up event sequence.
;;   (XKBE should provide the means to turn off the automatic up events)

;;
;; CLASS GRAPH
;;
;; Event
;;  Input Device Event (virtual)
;;                   Key Event (key-press, key-release)
;;                Button Event (button-press, button-release)
;;                Motion Event (pointer-motion, pointer-motion-hint???)
;;    Motion Enter Leave Event (virtual, abbreviated as /mel/)
;;                             Motion Enter Event (enter-notify)
;;                             Motion Leave Event (leave-notify)
;;
;; Event
;;  Expose Event (exposure)
;;  Graphics Exposure Event (???)
;;
;; TODO Other event types with shared byte structures.
;; Focus: focus-in, focus-out
;; Selection events
;;
;; TODO Event -> Structure Notify
;;
;; create-notify, destroy-notify, unmap-notify, map-notify,
;; reparent-notify, configure-notify, gravity-notify,
;; circulate-notify

(define (gio-event-class->sca/x-events class)
 (let ((res '()))
  ;; cfr with display-class/response-reader-class: there we wire the classes to the numerals in the event vector of those symbols.
  ;; Another meta class is out of reach for now and currently it wouldn't even be helpful.
  ;; So the colored wire redundancy pragraphs and those spreads into different files remain for now. 
 (when (has-class? class     key-event-class) (set! res (append '(key-press key-release)               res)))
 (when (has-class? class  button-event-class) (set! res (append '(button-press button-release)         res)))

 ; XXX listening for hints turns off motion events! [CLASS GRAPH REVISION ISSUE]
 ;(when (has-class? class  motion-event-class) (set! res (append '(pointer-motion pointer-motion-hint)  res)))
 (when (has-class? class  motion-event-class) (set! res (append '(pointer-motion )  res)))

 (when (has-class? class     mel-event-class) (set! res (append '(enter-window leave-window)           res)))
 (when (has-class? class  expose-event-class) (set! res (append '(exposure)                            res)))
 ;; XXX structure notify (when (has-class? class destroy-event-class) (set! res (append '(destroy)       res)))

 ;; ??? where does this come from? it's not part of the event flag set in window.class.scm
 ;; (when (is-a? graphics-exposure-event-class class) (set! res (cons 'graphics-exposure res)))

 res
))

;;
;; NB The use of the memoizing accessor /@slot-set!/ in this file assumes
;;    object class compatibility per text location. This should be full-
;;    filled by the current use and class graph.
;;

;; ??? GIO Event Class remains unused for now: there's some slightly confused
;; feeling wrt performance and code granularity where this class wouldn't do
;; any good if not only on paper.
(define gio-event-class (class            (extension display-object-class)
 (slots is-synthetic sequence-number)
(method initialize (display fbw)
 (@slot-set! (this) 'is-synthetic    (bit-and? #"128"     (fixed-call xgenev 'xgenev:code            fbw)))
 (@slot-set! (this) 'sequence-number (byte-string->number (fixed-call xgenev 'xgenev:sequence-number fbw)))
)
(define xgenev (namespace
(define X-generic-event-layout
 (byte-structure-description
  ;; Header
  (code    1)
  (padding 1)
  (sequence-number 2)))
(define-bs-accessors 'xgenev X-generic-event-layout)
)) ; closes namespace xgenev
)) ; closes gio event class



(define input-device-event-class (class   (extension display-object-class)
  (slots code detail
         sequence-number time ;; msec
         root-window event-window child-window
          root-x  root-y
         event-x event-y
         state same-screen)

(method destroy () #t)

(method describe ()
 (let ((nl "\n"))
  (brass-append
      (format " Sequence Number: ~D" (this sequence-number))
   nl (format "      Time Delta: ~,3F seconds" (/ (this time) 1000.0))
   nl (format "     Root Window: ~A (~D,~D)" ((this root-window)  (get-sid)) (this root-x)  (this root-y))
   nl (format "    Event Window: ~A (~D,~D)" ((this event-window) (get-sid)) (this event-x) (this event-y))
   nl (format "    Child Window: ~A" ((this child-window) (get-sid))) 
   nl (format "           State: ~A" (this state))
   nl (format "     Same Screen: ~A" (this same-screen))
   nl)))

(method initialize (display fbw)
 (let* ((riddle  (display riddle))
        (s       (this)))
  ;; TODO If a riddle lookup should ever fail that might as well mean that this event is obsolete
  ;;      and shouldn't be used. There might be conditions where this could happen and that doesn't
  ;;      help error reporting.
  (@slot-set! s 'display      display)
  (@slot-set! s 'code         #! identity !#            (fixed-call xev 'xev-idev:code     fbw))
  (@slot-set! s 'detail       #! identity !#            (fixed-call xev 'xev-idev:detail   fbw))
  (@slot-set! s 'sequence-number (byte-string->number   (fixed-call xev 'xev-idev:sequence-number fbw)))
  (@slot-set! s 'time         (byte-string->number      (fixed-call xev 'xev-idev:time     fbw)))
  (@slot-set! s 'root-window  (riddle (lookup           (fixed-call xev 'xev-idev:root     fbw))))
  (@slot-set! s 'event-window (riddle (lookup           (fixed-call xev 'xev-idev:event    fbw))))
  (@slot-set! s 'child-window (riddle (lookup           (fixed-call xev 'xev-idev:child    fbw))))
  (@slot-set! s 'root-x       (byte-string->number+sign (fixed-call xev 'xev-idev:root-x   fbw)))
  (@slot-set! s 'root-y       (byte-string->number+sign (fixed-call xev 'xev-idev:root-y   fbw)))
  (@slot-set! s 'event-x      (byte-string->number+sign (fixed-call xev 'xev-idev:event-x  fbw)))
  (@slot-set! s 'event-y      (byte-string->number+sign (fixed-call xev 'xev-idev:event-y  fbw)))
  (@slot-set! s 'state        #! identity !#            (fixed-call xev 'xev-idev:state    fbw))
  (@slot-set! s 'same-screen  (byte-string-true?        (fixed-call xev 'xev-idev:same-screen fbw)))))

(define xev (namespace
(define X-input-device-event-layout
 (byte-structure-description
  ;; Input Header
  (code   1)
  (detail 1)
  (sequence-number 2)
  (time   4)
  ;; Windows
  (root    4)
  (event   4)
  (child   4)
  (root-x  2)
  (root-y  2)
  (event-x 2)
  (event-y 2)
  ;; Additional Information
  (state       2)
  (same-screen 1)
  (padding 1)))
(define-bs-accessors 'xev-idev X-input-device-event-layout)
))
)) ; closes input device event class


(define key-event-class (class     (extension input-device-event-class)
  (slots type value keyboard) ; Stable Interface
  (slots with-shift with-capslock with-control with-super with-alt) ; pseudo stable interface
  (slots key-code xdetail keysym-map) ; SCA/X internal interface
  ; The roadmap targets "PC end user application development. That's why
  ; we offer the state of these controls as part of the key event.

(method initialize (display fbw)
 (super display fbw)
 (let* ((type     (byte-string->number (this code)))
        (keyboard (display keyboard)))
  (@slot-set! (this) 'type     (or (and (= 2 type) 'down) (and (= 3 type) 'up)))
  (@slot-set! (this) 'key-code (this detail))
  (@slot-set! (this) 'keysym-map (keyboard (keysym-map (this key-code))))
  (@slot-set! (this) 'xdetail    (keyboard (evaluate-kc (this state) (this key-code))))
  (this (set! 'with-shift    #f))
  (this (set! 'with-capslock #f))
  (this (set! 'with-control  #f))
  (this (set! 'with-super    #f))
  (this (set! 'with-alt      #f))
  (let loop ((ctrls (third (this xdetail))))
   (when (not (null? ctrls))
    (this (set! (case (car ctrls) 
                  ((shift)     'with-shift)
                  ((capslock)  'with-capslock)
                  ((control)   'with-control)
                  ((super)     'with-super)
                  ((alt altgr) 'with-alt))
                #t))
    (loop (cdr ctrls)) ))
  (this (set! 'value (first (this xdetail))))
  (this (set! 'keyboard keyboard))
))

(method describe ()
 (brass-append
  "* Key Event" "\n"
  (super)
  (format "        Key Code: ~A" (this key-code))      "\n"
  (format "            Type: ~A" (this type))          "\n"
  (format "        Keyboard: ~A" ((this keyboard) (sid)))"\n"
  (format "           Value: ~:S" (this value))        "\n"
  (format "   with    shift: ~A" (this with-shift))    "\n"
  (format "   with capslock: ~A" (this with-capslock)) "\n"
  (format "   with  control: ~A" (this with-control))  "\n"
  (format "   with    super: ~A" (this with-super))    "\n" ; combine that with some to end up with "vi-per"
  (format "   with      alt: ~A" (this with-alt))      "\n"
  (format "    SCA/X Detail: ~A" (this xdetail))       "\n"
  (format "      keysym-map: ~A" (this keysym-map))    "\n"
))
))


(define button-event-class (class   (extension input-device-event-class)
  (slots type button)
(method initialize (display fbw)
 (super display fbw)
 (let* ((type   (byte-string->number (this code))))
  (@slot-set! (this) 'type     (or (and (= 4 type) 'down) (and (= 5 type) 'up)))
  ;; Pointer Device Buttons are numbered from 1 onwards.
  (@slot-set! (this) 'button   (byte-string->number (this detail)))
))
(method describe ()
 (brass-append
  "* Pointer Device Button Event" "\n"
  (super)
  (format "          Button: ~A" (this button)) "\n"
  (format "            Type: ~A" (this type))   "\n")
)
))


(define motion-event-class (class   (extension input-device-event-class)
  ; (slots detail)
(method initialize (display fbw)
 (super display fbw)
 (@slot-set! (this) 'detail (if (byte-string-true? (this detail)) 'hint 'normal))
)
(method describe ()
 (brass-append
  "* Pointer Device Motion Event" "\n"
  (super)
  (format "          Detail: ~A" (this detail)) "\n")
)
))

;; Input Device Pointer
;; Motion Enter Leave Event
;; Virtual Class
(define mel-event-class (class            (extension input-device-event-class)
  (slots mode)
(method initialize (display fbw)
 (let* ((riddle (display riddle))
        (s      (this)))
  (@slot-set! s 'display      display)
  (@slot-set! s 'code         #! identity !#            (fixed-call xmelv 'xmelv:code     fbw))
  (@slot-set! s 'detail       #! identity !#            (fixed-call xmelv 'xmelv:detail   fbw))
  (@slot-set! s 'sequence-number (byte-string->number   (fixed-call xmelv 'xmelv:sequence-number fbw)))
  (@slot-set! s 'time         (byte-string->number      (fixed-call xmelv 'xmelv:time     fbw)))
  (@slot-set! s 'root-window  (riddle (lookup           (fixed-call xmelv 'xmelv:root     fbw))))
  (@slot-set! s 'event-window (riddle (lookup           (fixed-call xmelv 'xmelv:event    fbw))))
  (@slot-set! s 'child-window (riddle (lookup           (fixed-call xmelv 'xmelv:child    fbw))))
  (@slot-set! s 'root-x       (byte-string->number+sign (fixed-call xmelv 'xmelv:root-x   fbw)))
  (@slot-set! s 'root-y       (byte-string->number+sign (fixed-call xmelv 'xmelv:root-y   fbw)))
  (@slot-set! s 'event-x      (byte-string->number+sign (fixed-call xmelv 'xmelv:event-x  fbw)))
  (@slot-set! s 'event-y      (byte-string->number+sign (fixed-call xmelv 'xmelv:event-y  fbw)))
  (@slot-set! s 'state        #! identity !#            (fixed-call xmelv 'xmelv:state    fbw))
  ;; TODO mode is one of normal grab ungrab
  (@slot-set! s 'mode         #! identity !#            (fixed-call xmelv 'xmelv:mode     fbw))
  ;; TODO extract some bit flags from the following
  (@slot-set! s 'same-screen  (byte-string-true?        (fixed-call xmelv 'xmelv:same-screen fbw)))
))

(define xmelv (namespace
(define X-mel-event-layout
 (byte-structure-description
  ;; Input Header
  (code   1)
  (detail 1)
  (sequence-number 2)
  (time   4)
  ;; Windows
  (root    4)
  (event   4)
  (child   4)
  (root-x  2)
  (root-y  2)
  (event-x 2)
  (event-y 2)
  ;; Additional Information
  (state       2)
  (mode        1)
  (same-screen 1) ))
(define-bs-accessors 'xmelv X-mel-event-layout)
))
)) ; closes motion enter/leave event class

(define motion-enter-event-class (class (extension mel-event-class)
  (method describe ()
   (brass-append
    "* Pointer Device Motion Enter Event" "\n"
    (super)
    (format "          Detail: ~A" (this mode)) "\n")
  )
))
(define motion-leave-event-class (class (extension mel-event-class)
  (method describe ()
   (brass-append
    "* Pointer Device Motion Leave Event" "\n"
    (super)
    (format "          Detail: ~A" (this mode)) "\n")
  )
))


(define expose-event-class (class (extension display-object-class)
 (slots sequence-number event-window x y width height count)

(method destroy () #t)
(method initialize (display fbw)
  (let ((riddle (display riddle))
        (s      (this)))
  (@slot-set! s 'display  display)
  (@slot-set! s 'sequence-number (byte-string->number (fixed-call xee 'xevent-ee:sequence-number fbw)))
  (@slot-set! s 'event-window    (riddle (lookup (fixed-call xee 'xevent-ee:window          fbw))))
  ; sic! no sign for the following numbers
  (@slot-set! s 'x          (byte-string->number (fixed-call xee 'xevent-ee:x               fbw)))
  (@slot-set! s 'y          (byte-string->number (fixed-call xee 'xevent-ee:y               fbw)))
  (@slot-set! s 'width      (byte-string->number (fixed-call xee 'xevent-ee:width           fbw)))
  (@slot-set! s 'height     (byte-string->number (fixed-call xee 'xevent-ee:height          fbw)))
  (@slot-set! s 'count      (byte-string->number (fixed-call xee 'xevent-ee:count           fbw)))
))

(define xee (namespace (namespace-import sca/x-common)
(define X-expose-event-layout
 (byte-structure-description
  (code    1)
  (padding 1)
  (sequence-number 2)
  (window          4)
  (x               2)
  (y               2)
  (width           2)
  (height          2)
  (count           2)
  (padding        14)))
(define-bs-accessors 'xevent-ee X-expose-event-layout)
)) ; closes xee namespace
)) ; closes expose event class


(define graphics-exposure-event-class (class (extension display-object-class)
 (slots sequence-number drawable x y width height count minor-opcode major-opcode)

(method destroy () #t)
(method initialize (display fbw)
  (let ((riddle (display riddle))
        (s      (this)))
  (@slot-set! s 'display  display)
  (@slot-set! s 'sequence-number (byte-string->number (fixed-call xge 'xevent-ge:sequence-number fbw)))
  (@slot-set! s 'drawable-number (riddle (lookup (fixed-call xge 'xevent-ge:drawable        fbw))))
  ; sic! no sign for the following numbers
  (@slot-set! s 'x          (byte-string->number (fixed-call xge 'xevent-ge:x               fbw)))
  (@slot-set! s 'y          (byte-string->number (fixed-call xge 'xevent-ge:y               fbw)))
  (@slot-set! s 'width      (byte-string->number (fixed-call xge 'xevent-ge:width           fbw)))
  (@slot-set! s 'height     (byte-string->number (fixed-call xge 'xevent-ge:height          fbw)))
  (@slot-set! s 'count      (byte-string->number (fixed-call xge 'xevent-ge:count           fbw)))
  (@slot-set! s 'major-opcode (byte-string->number (fixed-call xge 'xevent-ge:major-opcode  fbw)))
  (@slot-set! s 'minor-opcode (byte-string->number (fixed-call xge 'xevent-ge:minor-opcode  fbw)))
))

(define xge (namespace (namespace-import sca/x-common)
(define X-expose-event-layout
 (byte-structure-description
  (code    1)
  (padding 1)
  (sequence-number 2)
  (drawable        4)
  (x               2)
  (y               2)
  (width           2)
  (height          2)
  (minor-opcode    2)
  (count           2)
  (major-opcode    1)
  (padding        11)))
(define-bs-accessors 'xevent-ge X-expose-event-layout)
)) ; closes xge namespace
)) ; closes graphics exposure event class


;; Structure Notifications Namespace
(define xse (namespace
 (define X-destroy-event-layout
  (byte-structure-description
   (code    1)
   (padding 1)
   (sequence-number 2)
   (event-window    4)
   (destroyed-window 4)
   (padding 20)))
 (define-bs-accessors 'xevent-destroy X-destroy-event-layout)

 (define X-configure-event-layout
  (byte-structure-description
   (code            1)
   (padding         1)
   (sequence-number 2)
   (event-window    4)
   (configured-window 4)
   (above-sibling   4)
   (x               2)
   (y               2)
   (width           2)
   (height          2)
   (border-width    2)
   (override-redirect 1)
   (padding         5)))
 (define-bs-accessors 'xevent-config X-configure-event-layout)
)) ; closes xse namespace


(define destroy-event-class (class (extension display-object-class)
 (slots sequence-number event-window originating-window)

(method destroy () #t)
(method initialize (display fbw)
 (let ((riddle (display riddle))
       (s      (this)))
  (@slot-set! s 'display  display)
  (@slot-set! s 'sequence-number (byte-string->number (fixed-call xse 'xevent-destroy:sequence-number  fbw)))
  (@slot-set! s 'event-window       (riddle (lookup   (fixed-call xse 'xevent-destroy:destroyed-window fbw))))
  (@slot-set! s 'originating-window (riddle (lookup   (fixed-call xse 'xevent-destroy:event-window     fbw))))
  (X-log-debug " a new destroy event " ((this event-window) (get-sid)) " from " ((this originating-window) (get-sid)))
))

)) ; closes destroy event class


;; ??? Maybe we should just punch these values into the window to then ...
(define configure-event-class (class (extension display-object-class)
 (slots sequence-number event-window originating-window)
 (slots x y width height border-width)
 (slots override-redirect above-sibling)

 (method destroy () #t)
 (method initialize (display fbw)
 (let ((riddle (display riddle))
       (s      (this)))
  (@slot-set! s 'display  display)
  (@slot-set! s 'sequence-number (byte-string->number (fixed-call xse 'xevent-config:sequence-number   fbw)))
  (@slot-set! s 'event-window         (riddle (lookup (fixed-call xse 'xevent-config:configured-window fbw))))
  (@slot-set! s 'originating-window   (riddle (lookup (fixed-call xse 'xevent-config:event-window      fbw))))
  (@slot-set! s 'x          (byte-string->number+sign (fixed-call xse 'xevent-config:x                 fbw)))
  (@slot-set! s 'y          (byte-string->number+sign (fixed-call xse 'xevent-config:y                 fbw)))
  (@slot-set! s 'width      (byte-string->number      (fixed-call xse 'xevent-config:width             fbw)))
  (@slot-set! s 'height     (byte-string->number      (fixed-call xse 'xevent-config:height            fbw)))
  (@slot-set! s 'border-width    (byte-string->number (fixed-call xse 'xevent-config:border-width      fbw)))
  (@slot-set! s 'override-redirect (byte-string-true? (fixed-call xse 'xevent-config:override-redirect fbw)))
  (let ((as (fixed-call xse 'xevent-config:above-sibling fbw)))
    (@slot-set! s 'above-sibling (and (byte-string-true? as) (riddle (lookup as)))))
))
)) ; closes configure event class
