
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;;; 
;; See also pp. 39 of the Core Protocol under QueryFont 
;; 
;; Ascent and descent are per character information which isn't well reflected
;; in the diagram below. 
;;
;; Font Wide Bounds: up-left: ( x + min (left-side-bearing), y - max (ascent) ) 
;;                     width: max (right-side-bearing) - max (left-side-bearing)
;;                    height: max (ascent) + max (descent) 
;; 
;; Individual characters may extend beyond those values:
;;                                                There is no sense if not aesthetical sense. 
;; These values serve to determine line height and horizontal character "progression"
;; aka display length (V). There doesn't seem to be any notion of exceptions to that, ie.
;; character combination distance correction (kerning). See also QueryTextExtents (pp. 40) 
;; 
;;  
;;                       x 
;;       character width |---------------|---------|
;;    right side bearing |------------|  |-------| 
;;     left side bearing |--|            |-| 
;;                       |               | 
;;     y-ascent   ascent ----------  XXX ascent is part of the bounding box; it's not above the glyph (but maybe lower than the actual height of the glyph; see above)
;;                       |               | 
;;                       |       X 
;;                       |      X X 
;;                       |     X   X     |   XXXX 
;;                       |    XXXXXXX       X   X 
;;                       |   X       X      X   X 
;;    y         baseline----X---------X--|---XXX---| - - -      y 
;;                       |                      X  
;;     y+descent-1       ------------------- XXX 
;;                       |
;;                       x 
;; bounding box metrics 
;; width  = right-side-bearing - left-side-bearing 
;; height = descent + ascent 
;; ( (x,y) (x+left-side-bearing, y-ascent) ) 
;;
;; start of next character: x+character-width, y (where character width is probably
;;                          different from the bounding box width described above)
;;
;; font size: 20 -> ascent + descent = 20 (pixmap fonts)
;;
;; [ where pt 1/72 inch
;;   where dpi or dots per inch denotes the actual resolution of an output device
;; ]
;;; 

(define font-class (class                    (extension display-resource-class)
 (slots name)
 (slots ascent descent)
 (slots char-metrics) ; vector of such structs

(define char-metric-struct (struct left-bearing right-bearing char-width ascent descent))
(define (decode-char-info bytes)
 (apply struct-from-values char-metric-struct (list
  (byte-string->number+sign (fixed-call xfont 'xcharinfo:left-side-bearing  bytes))
  (byte-string->number+sign (fixed-call xfont 'xcharinfo:right-side-bearing bytes))
  (byte-string->number+sign (fixed-call xfont 'xcharinfo:character-width    bytes))
  (byte-string->number+sign (fixed-call xfont 'xcharinfo:ascent             bytes))
  (byte-string->number+sign (fixed-call xfont 'xcharinfo:descent            bytes))
)))

(method initialize (display name) (super display)
 (this (set! 'name name))
 ;; This might lead to a non fatal error condition???
 ;; Perhaps list-fonts (of the display) should be used to check our argument in the callee?
 (fixed-call xfont 'X-open-font (display connection) (this rid) name)
 (let* ((l (fixed-call xfont 'X-query-font ((this display) connection) (this rid)))
        (font-metrics    (first  l))
        (font-properties (second l))
        (char-info       (third  l)))
   (this (set! 'ascent  (byte-string->number+sign (fixed-call xfont 'xrr-query-font:font-ascent  font-metrics))))
   (this (set! 'descent (byte-string->number+sign (fixed-call xfont 'xrr-query-font:font-descent font-metrics))))
   (this (set! 'char-metrics (list->vector (map decode-char-info char-info))))
 )
)

(method destroy ()
 (fixed-call xfont 'X-close-font ((this display) connection) (this rid))
 (super)
)

(define (char-sequence brass-or-list)
 (cond ((brass? brass-or-list) (brass->list brass-or-list))
       (else                   brass-or-list)))

;; (line-metrics brass) => (values width height)
;; Some home brewn text metric utility. TODO newline and other out of range argument checks
;; (This is meant to be used for single line text labels.)
(method line-metrics (brass-or-list)
 (let* ((l  (map char->integer (char-sequence brass-or-list)))
        ;; The font encoding would be part of the name (X core protocol convention) but
        ;; we ignore it for now:
        ;; The callees should stick to C0 and some ASCII compatible font encoding.
        (m  (this char-metrics))
        (lw (map (lambda (i) (@slot-ref (vector-ref m i) 'char-width)) l))) 
  (values (apply + lw)
          (+ (this ascent) (this descent)))
))

(method char-index-for-x (brass-or-list x)
 (let* ((l  (map char->integer (char-sequence brass-or-list)))
        ;; The font encoding would be part of the name (X core protocol convention) but
        ;; we ignore it for now:
        ;; The callees should stick to C0 and some ASCII compatible font encoding.
        (m  (this char-metrics))
        (lw (map (lambda (i) (@slot-ref (vector-ref m i) 'char-width)) l))) 
  (let loop ((lw lw) (idx 0) (dof 0))
   (cond
    ((or (>= dof x) (null? lw)) idx)
    (else                       (loop (cdr lw) (+ idx 1) (+ dof (car lw))) )))
))


(define xfont (namespace (namespace-import sca/x-common)

(define X-rp-open-font
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2)
  (fid            4)
  (length-of-name 2)
  (padding        2)))
(define-bs-accessors 'xrp-open-font X-rp-open-font)
(define X-open-font:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length X-rp-open-font)))))
  (xrp-open-font:set-opcode! t (byte->byte-string 45))
  t))
(define (X-open-font connection id name)
 (let* (;(nl (brass-length name))    ;; TODO verify C0
        (rb X-open-font:template)
        (bn (brass->string-8 name))
        (rl (+ (byte-string-length rb) (byte-string-length bn)))
        (pad (byte-string-clear! (make-byte-string (- 4 (modulo rl 4))) ))
        (pad (if (= 4 (byte-string-length pad)) (make-byte-string 0) pad))) ; XXX padding issue patch
   (xrp-open-font:set-fid!            rb id)
   (xrp-open-font:set-length-of-name! rb (number->byte-string-2 (byte-string-length bn)))
   (xrp-open-font:set-request-length! rb (number->byte-string-2
                                          (/ (+ rl (byte-string-length pad))
                                             4)))
   (X-request connection (byte-string-adjoin rb bn pad) no-response)
))

(define X-rp-close-font
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2)
  (fid     4)))
(define-bs-accessors 'xrp-close-font X-rp-close-font)
(define X-close-font:template
 (let* ((l (bs-length X-rp-close-font))
        (t (byte-string-clear! (make-byte-string l))))
  (xrp-close-font:set-opcode!         t (byte->byte-string 46))
  (xrp-close-font:set-request-length! t (number->byte-string-2 (/ l 4)))
  t))
(define (X-close-font connection fid)
 (let ((rb X-close-font:template))
  (xrp-close-font:set-fid! rb fid)
  (X-request connection rb no-response)))

(define X-rp-query-font
 (byte-structure-description
  (opcode   1)
  (padding  1)
  (request-length 2)
  (fontable 4)))
(define-bs-accessors 'xrp-query-font X-rp-query-font)
(define X-rr-query-font
 (byte-structure-description
  (reply    1)
  (padding  1)
  (sequence-number 2)
  (reply-length    4)
  (min-bounds     12) ; charinfo
  (padding         4) ; sic!
  (max-bounds     12) ; charinfo
  (padding         4) ; sic!
  (min-char-or-byte2 2)
  (max-char-or-byte2 2)
  (default-char      2)
  (number-of-fontprops 2)
  (draw-direction    1)
  (min-byte1         1)
  (max-byte1         1)
  (all-chars-exist   1)
  (font-ascent       2)
  (font-descent      2)
  (number-of-charinfo  4)))
(define-bs-accessors 'xrr-query-font X-rr-query-font)
(define X-fontprop-layout
 (byte-structure-description
  (name   4)
  (value  4)))
(define-bs-accessors 'xfontprop X-fontprop-layout)
(define X-charinfo-layout
 (byte-structure-description
  (left-side-bearing  2)
  (right-side-bearing 2)
  (character-width    2)
  (ascent             2)
  (descent            2)
  (attributes         2)))
(define-bs-accessors 'xcharinfo X-charinfo-layout)
(define X-query-font:template
 (let ((template (make-byte-string (bs-length X-rp-query-font))))
  (byte-string-clear! template)
  (xrp-query-font:set-opcode!         template (byte->byte-string    47))
  (xrp-query-font:set-request-length! template (number->byte-string-2 2))
  template))
(define (X-query-font connection font-id)
 (let ((rb X-query-font:template))
  (xrp-query-font:set-fontable! rb font-id)
  (X-request connection rb X-query-font-response)))
;;
;; !!! (bs-length X-rr-query-font) => 60
;;
(define (X-query-font-response connection fbw) 
  (let* ((r   fbw)
         (seq-number    (byte-string->number      (xrr-query-font:sequence-number r)))
                                 ; !!! reply-length is usually w/o the regular 32 bytes
         (reply-length  (* 4 (byte-string->number (xrr-query-font:reply-length r))))
         (tail          (X-receive reply-length connection))
                                 ; !!! but we need to access the first 60 bytes as one byte structure
         (r             (byte-string-adjoin r tail))
         (num-fontprops (byte-string->number      (xrr-query-font:number-of-fontprops r)))
         (num-charinfo  (byte-string->number      (xrr-query-font:number-of-charinfo  r))))
   (let* ((fontprops-l (X:read-out-list num-fontprops rl-fontprop #f r (bs-length X-rr-query-font)))
          (charinfo-l  (X:read-out-list num-charinfo  rl-charinfo #f r (+ (bs-length X-rr-query-font)
                                                                          (first fontprops-l)))))
     ;; We just return the plain bytes for now such that the byte structures accessors
     ;; can be used by high level routines to process the data.
     (values (byte-string-select r 0 (bs-length X-rr-query-font))
             (second fontprops-l)
             (second charinfo-l)))
))
(define (rl-fontprop bytes)
  (values (bs-length X-fontprop-layout)
          (byte-string-select bytes 0 (bs-length X-fontprop-layout))))
(define (rl-charinfo bytes)
  (values (bs-length X-charinfo-layout)
          (byte-string-select bytes 0 (bs-length X-charinfo-layout))))
)) ; closes xfont namespace
)) ; closes font class
