
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; A graphics context is considered a structured collection of
;; parameters for drawing operations. This class still hosts the
;; drawing routines which need a GC as an argument in the X core
;; protocol where the actual drawable can be a window or ...

(define graphics-context-class (class        (extension display-resource-class)
 (slots drawable)

(method initialize (drawable) (let ((display (drawable display)))
 (super display)
 (this (set! 'drawable drawable))
 (fixed-call xgc 'X-create-gc (display connection) (this rid) (drawable rid))
))

(method destroy ()
 (fixed-call xgc 'X-free-gc ((this display) connection) (this rid))
 (super) ; presumption: graphics contexts aren't message recipients (else race condition / message misrouting)
)

(method set-font (font)
 (fixed-call xgc 'X-gc-set-font! ((this display) connection) (this rid) (font rid)))

(method set-color (foreground/background color)
 (let ((pixel ((this drawable) (color->pixel color))))

 (case foreground/background
  ((foreground) (fixed-call xgc 'X-gc-set-foreground! ((this display) connection) (this rid) pixel))
  ((background) (fixed-call xgc 'X-gc-set-background! ((this display) connection) (this rid) pixel))
  (else         (error "GIO gc: set-color bad argument"))
)))

(method set-tile (pixmap)
 (fixed-call xgc 'X-gc-set-tile! ((this display) connection) (this rid) (pixmap rid))
)

(define xgc (namespace (namespace-import sca/x-common)

;; p. 140 GC Encoding (Property Value Flags)
(define pc number->byte-string-4)
(define function           (pc #x00000001))
(define plane-mask         (pc #x00000002))
(define foreground         (pc #x00000004))
(define background         (pc #x00000008))
(define line-width         (pc #x00000010))
(define line-style         (pc #x00000020))
(define cap-style          (pc #x00000040))
(define join-style         (pc #x00000080))
(define fill-style         (pc #x00000100))
(define fill-rule          (pc #x00000200))
(define tile               (pc #x00000400))
(define stipple            (pc #x00000800))
(define tile-stipple-x-origin (pc #x00001000))
(define tile-stipple-y-origin (pc #x00002000))
(define font               (pc #x00004000))
(define subwindow-mode     (pc #x00008000))
(define graphics-exposures (pc #x00010000))
(define clip-x-origin      (pc #x00020000))
(define clip-y-origin      (pc #x00040000))
(define clip-mask          (pc #x00080000))
(define dash-offset        (pc #x00100000))
(define dashes             (pc #x00200000))
(define arc-mode           (pc #x00400000))

(define X-rp-gc-creation
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2)
  (cid      4)
  (drawable 4)
  (value-mask 4)))
(define-bs-accessors 'xrp-gc-creation X-rp-gc-creation)
(define X-gc-creation:template
 (let ((template (make-byte-string (bs-length X-rp-gc-creation))))
   (byte-string-clear! template)
   (xrp-gc-creation:set-opcode! template (byte->byte-string 55))
   template))
(define (X-create-gc connection id drawable)
 (let ((t   X-gc-creation:template)
       (con connection))
  (xrp-gc-creation:set-drawable!       t drawable)
  (xrp-gc-creation:set-cid!            t id)
  (xrp-gc-creation:set-request-length! t
   (number->byte-string-2 (/ (bs-length X-rp-gc-creation)
                            4)))
  (X-request con t no-response)
))

(define X-rp-free-gc
 (byte-structure-description
  (opcode    1)
  (padding   1)
  (request-length 2)
  (gcontext  4)))
(define-bs-accessors 'xrp-free-gc X-rp-free-gc)
(define X-free-gc:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-free-gc)))))
  (xrp-free-gc:set-opcode!         template (byte->byte-string    60))
  (xrp-free-gc:set-request-length! template (number->byte-string-2 2))
  template))
(define (X-free-gc connection gc)
 (let ((rb X-free-gc:template))
  (xrp-free-gc:set-gcontext! rb gc)
  (X-request connection rb no-response)))

(define X-rp-change-gc
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2)
  (gid        4)
  (value-mask 4)))
(define-bs-accessors 'xrp-change-gc X-rp-change-gc)
(define X-change-gc:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length X-rp-change-gc)))))
   (xrp-change-gc:set-opcode! t (byte->byte-string 56))
   t))
(define (X-change-gc connection id . values)
  '...)

(define make-change-gc-setter (macro (PROP . CONVERTER)
 `(lambda (connection gid value)
   (let* ((rb    X-change-gc:template)
          (value ,(if (null? CONVERTER) 'value `(,(car CONVERTER) value)))
          (rl    (+ (byte-string-length rb) (byte-string-length value)))
          (pad   (X-pad rl)))
    (xrp-change-gc:set-gid!            rb gid)
    (xrp-change-gc:set-value-mask!     rb ,PROP)
    (xrp-change-gc:set-request-length! rb (number->byte-string-2 (/ (+ rl (byte-string-length pad)) 4)))
    (X-request connection (byte-string-adjoin rb value pad) no-response)))))

(define X-gc-set-tile! (make-change-gc-setter tile))
; ...

(define (X-gc-set-font! connection gid fid)
 (let ((rb X-change-gc:template))
  (xrp-change-gc:set-gid!            rb gid)
  (xrp-change-gc:set-value-mask!     rb font)
  (xrp-change-gc:set-request-length! rb (number->byte-string-2
                                         (/ (+ (byte-string-length rb) (byte-string-length fid))
                                            4)))
  (X-request connection (byte-string-adjoin rb fid) no-response)))

(define (X-gc-set-foreground! connection gid pixel)
 (let ((rb X-change-gc:template))
  (xrp-change-gc:set-gid!            rb gid)
  (xrp-change-gc:set-value-mask!     rb foreground)
  (xrp-change-gc:set-request-length! rb (number->byte-string-2
                                         (/ (+ (byte-string-length rb) (byte-string-length pixel))
                                            4)))
  (X-request connection (byte-string-adjoin rb pixel) no-response)))
(define (X-gc-set-background! connection gid pixel)
 (let ((rb X-change-gc:template))
  (xrp-change-gc:set-gid!            rb gid)
  (xrp-change-gc:set-value-mask!     rb background)
  (xrp-change-gc:set-request-length! rb (number->byte-string-2
                                         (/ (+ (byte-string-length rb) (byte-string-length pixel))
                                            4)))
  (X-request connection (byte-string-adjoin rb pixel) no-response)))
   

)) ; closes xgc namespace



;;; Graphical Output
;;; X core drawing routines

(method polifill-arc (x y width height angle1 angle2)
 (fixed-call xdraw 'X-polifill-arc
   ((this display) connection)
   ((this drawable) rid)
   (this rid)
   x y width height angle1 angle2))

(method image-text-8 (x y brass)
 (when (> (brass-length brass) 0)
  (fixed-call xdraw 'X-image-text-8
   ((this display) connection)
   ((this drawable) rid)
   (this rid)
   x y brass)))

;; (method put-image (dst width height dst-x dst-y pixmap)
;;  (fixed-call xdraw 'X-put-image ((this display) connection)
;;                                 (dst rid)
;;                                 (this rid)
;;                                 width height dst-x dst-y
;;                                 (pixmap (get-pixels)))
;; )

(method put-image (dst width height dst-x dst-y pixmap)
 (fixed-call xshm 'Xshm-put-image ((this display) connection)
                                  (((this display) xext-shared-memory) major-opcode)
                                  (dst rid)
                                  (this rid)
                                  (pixmap width)
                                  (pixmap height)
                                  0 0 width height ; source coords
                                  dst-x dst-y
                                  24 ; XXX depth from format
                                  (pixmap format-id) ; format-id pass through
                                  #f   ; completion event
                                  ((pixmap x-shared-mem) rid)
                                  0)) ; offset
                                  

(define xdraw (namespace (namespace-import sca/x-common)

(define X-arc-layout
 (byte-structure-description
  (x  2)    
  (y  2)
  (width  2)
  (height 2)
  (angle1 2)
  (angle2 2)
))
(define-bs-accessors 'x-arc X-arc-layout)

(define X-rp-polifill-arc
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2) ; 3+3n ???
  (drawable 4)
  (gcontext 4)))
(define-bs-accessors 'xrp-polifill-arc X-rp-polifill-arc)
(define X-polifill-arc:template
 (let ((template (make-byte-string (bs-length X-rp-polifill-arc))))
  (byte-string-clear! template)
  (xrp-polifill-arc:set-opcode! template (byte->byte-string 71))
  template))
(define (X-polifill-arc connection drawable gc x y width height angle1 angle2)
 (let ((t   X-polifill-arc:template)
       (arc (byte-string-clear! (make-byte-string (bs-length X-arc-layout)) ))
       (con connection))
  (x-arc:set-x! arc (number+sign->byte-string-2 x))
  (x-arc:set-y! arc (number+sign->byte-string-2 y))
  (x-arc:set-width!  arc (number->byte-string-2 width))
  (x-arc:set-height! arc (number->byte-string-2 height))
  (x-arc:set-angle1! arc (number+sign->byte-string-2 angle1))
  (x-arc:set-angle2! arc (number+sign->byte-string-2 angle2))
  (xrp-polifill-arc:set-drawable!   t drawable)
  (xrp-polifill-arc:set-gcontext!   t gc)
  (xrp-polifill-arc:set-request-length! t
    (number->byte-string-2
     (/ (+ (bs-length X-rp-polifill-arc)
           (bs-length X-arc-layout))
        4)))
  (X-request con (byte-string-adjoin t arc) no-response) ))

(define X-rp-image-text-8
 (byte-structure-description
  (opcode  1)
  (length-of-string 1)
  (request-length   2)
  (drawable         4)
  (gcontext         4)
  (x     2)
  (y     2)))
(define-bs-accessors 'xrp-image-text8 X-rp-image-text-8)
(define X-image-text-8:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length X-rp-image-text-8)))))
   (xrp-image-text8:set-opcode! t (byte->byte-string 76))
   t))
(define (X-image-text-8 connection drawable gcontext x y brass)
 (let* ((s  (brass->string-8 brass))
        (sl (byte-string-length s))
        (rb X-image-text-8:template)
        (rl (+ sl (byte-string-length rb)))
        (pad (byte-string-clear! (make-byte-string (- 4 (modulo rl 4))) ))
        (pad (if (= 4 (byte-string-length pad)) (make-byte-string 0) pad))) ; XXX padding issue patch
  (xrp-image-text8:set-length-of-string! rb (byte->byte-string sl))
  (xrp-image-text8:set-drawable!         rb drawable)
  (xrp-image-text8:set-gcontext!         rb gcontext)
  (xrp-image-text8:set-x!                rb (number+sign->byte-string-2 x))
  (xrp-image-text8:set-y!                rb (number+sign->byte-string-2 y))
  (xrp-image-text8:set-request-length!   rb (number->byte-string-2 (/ (+ rl (byte-string-length pad))
                                                                      4)))
  (X-request connection (byte-string-adjoin rb s pad) no-response)
))

(define X-rp-put-image
 (byte-structure-description
  (opcode         1)
  (format         1)
  (old-request-length 2)
  (request-length 4)
  (drawable       4)
  (gc             4)
  (width          2)
  (height         2)
  (dst-x          2)
  (dst-y          2)
  (left-pad       1)
  (depth          1)
  (padding        2)))
(define-bs-accessors 'xrp-put-image X-rp-put-image)
(define X-put-image:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-put-image))) ))
   (xrp-put-image:set-opcode!             template (byte->byte-string 72))
   (xrp-put-image:set-old-request-length! template (number->byte-string-2 0)) ; big request
   template))
(define (X-put-image connection drawable gc width height dst-x dst-y pixel)
 (let* ((rb  X-put-image:template)
        (rl  (+ (byte-string-length rb) (padded4 (byte-string-length pixel))))
        (pad (X-pad rl)))
  (xrp-put-image:set-format!         rb (byte->byte-string  2)) ; 0 bitmap 1 xypixmap 2 zpixmap
  (xrp-put-image:set-depth!          rb (byte->byte-string 24)) ; XXX default depth
  (xrp-put-image:set-left-pad!       rb (byte->byte-string  0)) ; 0 for zpixmap pp.55
  (xrp-put-image:set-drawable!       rb drawable)
  (xrp-put-image:set-gc!             rb gc)
  (xrp-put-image:set-dst-x!          rb (number+sign->byte-string-2 dst-x))
  (xrp-put-image:set-dst-y!          rb (number+sign->byte-string-2 dst-y))
  (xrp-put-image:set-width!          rb (number->byte-string-2 width))
  (xrp-put-image:set-height!         rb (number->byte-string-2 height))
  (xrp-put-image:set-request-length! rb (number->byte-string-4 (/ (+ rl (byte-string-length pad)) 4)))
  (X-request connection (byte-string-adjoin rb pixel pad ) no-response)
;; GC components: function, plane-mask, subwindow-mode, clip-x-origin, clip-y-origin, clip-mask
;; GC mode-dependent components: foreground, background
))

)) ; closes xdraw namespace

)) ; closes graphics context class
