
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; NB SCA/X is somehow short before being revised (inner project time)
;;
;;  - interclass namespace issues (depends on namespace-load into a class)
;;    SCA/X reverts to the system environment for some routines until development
;;    is stable enough to encapsulate into namespaces again.
;;  - encoding / decoding stages
;;    With respect to the current experience regarding the protocol encoding
;;    we will revise all routines into encoding stages and communication stages.
;;    Exceptions and quirks of the encoding are now known: the principle of
;;    GIO uses SCA/X but SCA/X doesn't need to know about GIO should be maintained.
;;    Encoding a window object thus remains the identity function with the rid
;;    of the window as its argument which has to be provided by GIO routines but
;;    that access to the rid of the window object doesn't count as encoding.
;;

;;;
;;; SCA/X Transport Layer - Controlling Input Output Path
;;;
;;; Functionality Model with Encapsulation (should be inlined manually to avoid the lookups
;;                                          of the transport structure members in every
;;                                          routine of the interested call graph)

(define X-transport   (struct channel event-queue-box event-queue-tail))
(define X-make-transport (lambda (channel) (let ((q (cons 'event-queue-box '()) ))
                           (struct-from-values X-transport  channel q q)) ))

;; Actually we probably have non blocking IO and send bytes ...
;; Put that open with the flags here, please.
(define X-send        (lambda (transport bytes)  (send-bytes (@slot-ref transport 'channel) bytes)))
(define X-receive     (lambda (length transport) (read-bytes length (@slot-ref transport 'channel))))
(define X-queue-event (lambda (transport event-fbw)
                       (X-log-debug "SCA/X Queuing event #" (length (slot-ref transport 'event-queue-box)))
                       (let ((o (list event-fbw)))
                        (set-cdr! (@slot-ref transport 'event-queue-tail) o)
                        (@slot-set! transport 'event-queue-tail o)) ))
(define X-next-qevent! (lambda (transport)
 (let* ((box  (@slot-ref transport 'event-queue-box))
        (fel  (cdr box)))
  (cond
   ((null? fel)  '())
   (else (let* ((nel (cdr fel)))
     (set-cdr! box nel)
     (when (null? nel) (@slot-set! transport 'event-queue-tail box))
     (car fel) )) ))
))

;; TODO xi support to display the logging channels adequately
(define (print obj port) ;; GIO debug printer
 (cond
 ((is-a? obj input-device-event-class) (display (obj (describe)) port))
 (else                                 (display obj              port))))
(define (with-cooked-mode closure) ;; XXX current xi protocol & shortcoming
 (sys-terminal (tty-cooked-mode))
 (sys-terminal (restore-cursor))
 (closure)
 (sys-terminal (save-cursor)))
 ;(sys-terminal (tty-raw-mode)))
(define (X:log-debug . msg)
 (with-cooked-mode (lambda ()
  (for-each (lambda (m) (print m (current-error-port)))
            msg)
  (newline (current-error-port))) ))
(define X-log-debug X:log-debug)
(define X-log-info  X:log-debug)
(define X-log-warn  X:log-debug)
(define X-log-error X:log-debug)

(define X-min-win-dim    0)
(define X-max-win-dim    (- (expt 2 16) 1))
(define X-min-win-pos (- (- (expt 2 15) 1)))
(define X-max-win-pos    (- (expt 2 15) 1))

(define GIO-windim (macro (D) `(min ,X-max-win-dim (max ,X-min-win-dim ,D)) ))
(define GIO-winpos (macro (P) `(min ,X-max-win-pos (max ,X-min-win-pos ,P)) ))

(define sca/x-common (namespace

(define (X-request connection bytes response-handler)
 ;; !!! It would be easier to calculate and set the request length here.
 ;; Requests which expect a response queue events and replay events later on.
 ;;   We actually finish this request atomically with respect
 ;; to the channel handler invocations. Therefore we should prepare to return
 ;; control to the channel handler which has some kind of priorized path
 ;; to handle the events in our queue. Our return value is the return value of
 ;; the request mapper and it is expected by some method. That will happen before
 ;; control reaches the channel handler and maybe that method would like to handle
 ;; queued events ...
 ;; WE MUST NOT DRAIN EVENTS: at least the mel events are crucial to the input device
 ;;                           operational modes (grabs).
 ;; The application controller could drain events in a controlled manner though.
 ;;
 (X-send connection bytes)
 (when (not (eq? response-handler no-response))
  (let loop () ; read until a response or an error code is found 
   (let* ((fbw (X-receive 32 connection)) ; !!! every response has at least 32 bytes
          (et  (byte-string->number (byte-string-select fbw 0 1))))
    (case et
     ((1)  (response-handler connection fbw))
     ((0)  (error (format "Window Sytem Error X-request ~A" et))) ;; TODO we're missing the error vector here
     (else (X-queue-event connection fbw)
           (loop) ))
 )))
)

(define (no-response connection) #t) ; response handler dummy / could be any unique memory object now


(define one-4 (number->byte-string-4 1))
(define (ones-4! bits)
 (let loop ((i (* 4 8)) (n 0) (bits bits))
  (if (= i 0)
    n
    (loop (- i 1)
          (if (bit-and? one-4 bits) (+ n 1) n)
          (bit-shift-right! bits 1))) ))

#! ROUNDUP X-padding prototypes

GetImage
#define ROUNDUP(nbytes, pad) (((((nbytes) - 1) + (pad)) / (pad)) * (pad))

PutImage, Lynx && RoundUp
#define ROUNDUP(nbytes, pad) (((nbytes) + ((pad) - 1)) & ~(long)((pad) - 1))

!#
(define (round-up num-bytes pad)
 (* (quotient (+ (- num-bytes 1) pad) pad) pad))

(define (round-up/lynx num-bytes pad)
 (byte-string->number
 (bit-and (number->byte-string-8 (+ num-bytes (- pad 1)))
          (bit-not (number->byte-string-8 (- pad 1))))))

(define (padded4 num-bytes) (byte-string->number
  (bit-shift-right (number->byte-string-8 (+ num-bytes 3)) 2)))

(define (X-pad num-bytes) (make-byte-string (- (round-up num-bytes 4) num-bytes)))


(define brass->string-8  brass->c-string/raw) ;; no zero padding?

(define msb network-byte-order) ; TODO put this into the core


;;; X:read-out-list: Read a list of values from a response
;;
;; /length/ number of elements in the list to be returned
;; /converter/ takes a byte string and should return two values
;;    1) the number of bytes processed (next offset)
;;    2) the converted memory object
;; /element-size/ unused XXX remove this one please 
;; /bytes/ is data
;; /offset/ the list data starts at that /offset/ in /bytes/
;;
;; X:read-out-list returns itsself the total number of bytes processed
;; as well as a list of converted memory objects as two such values. 
;;
;; Please note: the converter receives a byte string mapping onto the
;;              encoded list which shouldn't be modified. (Don't use
;;              any byte string routines with a bang on it to keep it
;;              intact for further processing.)
;;;
(define (X:read-out-list length converter element-size bytes offset)
 (let ((size 0)
       (total-length (byte-string-length bytes))
       (res '()))
  (set! res
  (map-in-order (lambda (i)
        (let ((res (converter (byte-string-mapping bytes
                                (+ offset size)
                                (- total-length (+ offset size))))))
          (set! size (+ size (first res)))
          (second res)))
      (iota- length)))
  (values size res)))


))
