
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; The origin of a window is denoted by /+0/     
;;                                     / 0/ in the diagram. 
;; 
;;  Top level window border might not be respected. 
;;    +0 
;;    0   ...._ border-width .......... . . . 
;;        ....|.......................... . . . 
;;        |-- +0 
;;        ....0 
;;        .... 
;;        .... 
;;
;;        ....
;;
;;        ....
;;
;;        .... 
(define window-class (class              (extension display-resource-class)
  ; (slots display rid) <- see display resource class
 (slots x y width height border-width) 
 (slots children)
 (slots activated-events)

 ;; TOC Slots
 ;;     Default Methods
 ;;     Event Method (input events)
 ;;      xwin namespace
 ;;     See gc.class.scm for Output Methods (drawing)

 (method initialize (display-or-window det . args)
 ;; /det/   display or window
 ;; /args/   mapping: rid
 ;;         creation: x y width height border-width ;; see also ConfigureWindow (pp. 23)
  (this (set! 'children '()))
  (this (set! 'activated-events '()))
  (cond
   ((and (viper-object? display-or-window)
         (eq? (class-of (display-or-window)) display-class)
         (byte-string? det)) ;; should be /rid?/
      ; map an existing window
      (let ((display display-or-window))
         (this (set! 'rid det))
        ;; no calls to /super/ here!
        (this (set! 'display display))
        (((this display) riddle) (register-id det this))))

   ((viper-object? display-or-window) (let* ((parent #f) (display #f)
                                             (args (cons det args))
                                             (det display-or-window))
    (cond ((is-a? det display-class) (set! display det)
                                     (set! parent  (det root-window)))
          ((is-a? det window-class)  (set! display (det display))
                                     (set! parent  det))
          (else (error "GIO window class initialize wrong type argument 1")))
                       (super display)
                       (apply (from-namespace xwin 'X-create-window)
                              ((this display) connection)
                              (parent rid)
                              (this rid)
                              args)
                       (this (set! 'x      (first       args)))
                       (this (set! 'y      (second      args)))
                       (this (set! 'width  (third       args)))
                       (this (set! 'height (fourth      args)))
                       (this (set! 'border-width (fifth args)))
                       (parent (add-child this)) ;; ??? do we do this here or ...?
                       ))
   (else (error "GIO window class initialize wrong type argument 2"))
 ))

(method describe () (brass-append
 (format "* Window Origin (~A, ~A) Dimensions ~Ax~A\n" (this x) (this y) (this width) (this height))
 (format "  Activated Events ~A" (this activated-events))
))
  
;; There will be a configure event for each request which confirms the change but
;; we still set the values of the slots now. To reliably retrieve the synchronized values
;; getters should be used with a synchronous request where direct slot access and parallel
;; objects ... but then the window system determines inside or outside and everything else ...
(method set-x (x)
 (fixed-call xwin 'X-config-window-x ((this display) connection) (this rid) (GIO-winpos x))
 (this (set! 'x x)))

(method set-y (y)
 (fixed-call xwin 'X-config-window-y ((this display) connection) (this rid) (GIO-winpos y))
 (this (set! 'y y)))

(method set-width  (w)
 (fixed-call xwin 'X-config-window-width  ((this display) connection) (this rid) (GIO-windim w))
 (this (set! 'width w)))

(method set-height (h)
 (fixed-call xwin 'X-config-window-height ((this display) connection) (this rid) (GIO-windim h))
 (this (set! 'height h)))

(method set-border-width (bw)
 (fixed-call xwin 'X-config-window-border-width ((this display) connection) (this rid) (GIO-windim bw))
 (this (set! 'border-width bw)))


(method color->pixel (color) (((this display) root-visual) (color->pixel color)))
(method get-color->pixel:converter () (((this display) root-visual) (get-color->pixel:converter)))

(method set-background-pixmap (pixmap) ;; defunct???
  (fixed-call xwin 'X-change-window-attributes
              ((this display) connection) (this rid)
              (cons 'background-pixmap (pixmap rid)))
)

(method add-child (w)
 (cond
 ((is-a? w window-class) (this (set! 'children (cons w (this children)) )))
 (else                   (error "GIO window-class add-child wrong type argument"))
))

(method show ()
 (call xwin 'X-map-window ((this display) connection)
                          (this rid)))

(method clear ()
 (fixed-call xwin 'X-clear-area ((this display) connection)
                                (this rid)
                                #f ; no exposures
                                0 0 65535 65535)) ;; ???
                           ;; These coords don't work well with (unattended) window resize
                           ;; actions. Use maximums? 
                                ;; 0 0 (this width) (this height)))

(method close ()
  (fixed-call xwin 'X-destroy-window ((this display) connection) (this rid))
)

;; GIO is slipping away: /destroy/ refers to the local communication object
;; with its rid in the riddle where one would expect to be able to use the
;; method /destroy/ where currently /close/ needs to be used. But maybe
;; it's better like this ... making this synchronous isn't possible with the
;; current event queue without race conditions.
(method destroy-event (event)
 (X-log-debug "destroy event " (this (get-sid)))
 (this (destroy))
)

(method configure-event (event)
 (this (set! 'x      (event x)))
 (this (set! 'y      (event y)))
 (this (set! 'width  (event width)))
 (this (set! 'height (event height)))
 (this (set! 'border-width (event border-width)))
)


(define (debug-event event)
  (when #f (X-log-debug (event (describe))))
)


(method key-event    (event) (debug-event event))
(method button-event (event) (debug-event event))
(method motion-event (event) (debug-event event))
(method enter-event  (event) (debug-event event))
(method leave-event  (event) (debug-event event))

(method expose-event            (event) #t) ;; XXX missing event description <- class graph restructuring
(method graphics-exposure-event (event) #t)

;;; CLEAN THE FOLLOWING OF THE HACKS PLEASE!
(method grab-pointer (time)
 (let ((display (this display))
       (n (namespace (namespace-import xwin)             ;; XXX hack
   (define event-mask (cast:4-2
                             (bit-or ;KeyPress KeyRelease
                              EnterWindow LeaveWindow
                              ButtonPress ButtonRelease  )) )) ))
  (fixed-call xwin 'X-grab-pointer
              (display connection) #t (this rid)
              'asynchronous 'asynchronous (from-namespace n 'event-mask) #"0 0 0 0"
              #"0 0 0 0" time)))
;;;


;; TODO synchronize event mask with activated-events when applicable
;; TODO /active events/ should hold GIO classes not SCA/X symbols
(method activate-events event-classes
 (let ((events (append (flatten! (map gio-event-class->sca/x-events event-classes)) (this activated-events))))
  (fixed-call xwin 'X-change-window-attributes
              ((this display) connection) (this rid)
              (cons 'event-mask events))
  (this (set! 'activated-events events))
))

(method deactivate-events event-classes
 (let ((active-events (this activated-events))
       (events        (flatten! (map gio-event-class->sca/x-events event-classes))))
  ;; That's no good for error reporting but one line only.
  (set! active-events (filter (lambda (e) (not (memq e events))) active-events))
  (fixed-call xwin 'X-change-window-attributes
              ((this display) connection) (this rid)
              (cons 'event-mask active-events))
  (this (set! 'activated-events active-events))
))


;; (method X:change-attributes key-value-list
;;   (apply (once (from-namespace xwin 'X-change-window-attributes))
;;          ((this display) connection) (this rid)
;;          key-value-list)
;; )

(define xwin (namespace (namespace-import sca/x-common)

(define X-rp-create-window
 (byte-structure-description
  ; Default Request Prefix
  (req-type 1) ; opcode
  (depth    1)
  (length   2)
  ; Create Window Arguments
  (window-id 4)
  (parent   4)
  (x      2)
  (y      2)
  (width  2)
  (height 2)
  (border-width 2)

  (class  2)
  (visual 4)
  (mask   4)
  ))
(define-bs-accessors 'xrp-window-creation X-rp-create-window)

(define X-create-window:template
  (let ((template (make-byte-string (bs-length X-rp-create-window))))
    (byte-string-clear! template)
    (xrp-window-creation:set-req-type! template (number->byte-string 1 1))
    (xrp-window-creation:set-depth! template #"0")
    template))

(define cw-back-pixel   (number->byte-string-4 2))
(define cw-border-pixel (number->byte-string-4 8))
(define cw-event-mask   (number->byte-string-4 #x00000800))
;; (define cw-keyboard     (number->byte-string-4 1))
(define cw-keymap       (number->byte-string-4 #x00004000))

(define pc number->byte-string-4)
;; Common Types (p. 117f) SETofEVENT
;; area of interest, input event propagation, event mask
(define KeyPress             (pc #x00000001))
(define KeyRelease           (pc #x00000002))
(define ButtonPress          (pc #x00000004))
(define ButtonRelease        (pc #x00000008))
(define EnterWindow          (pc #x00000010))
(define LeaveWindow          (pc #x00000020))
(define PointerMotion        (pc #x00000040))
(define PointerMotionHint    (pc #x00000080))
(define Button1Motion        (pc #x00000100))
(define Button2Motion        (pc #x00000200))
(define Button3Motion        (pc #x00000400))
(define Button4Motion        (pc #x00000800))
(define Button5Motion        (pc #x00001000))
(define ButtonMotion         (pc #x00002000))
(define KeymapState          (pc #x00004000))
(define Exposure             (pc #x00008000))
(define VisibilityChange     (pc #x00010000))
(define StructureNotify      (pc #x00020000))
(define ResizeRedirect       (pc #x00040000))
(define SubstructureNotify   (pc #x00080000))
(define SubstructureRedirect (pc #x00100000))
(define FocusChange          (pc #x00200000))
(define PropertyChange       (pc #x00400000))
(define ColormapChange       (pc #x00800000))
(define OwnerGrabButton      (pc #x01000000))
;; unused but must be zero       #xFE000000

(define aflag cons)
(define event-flags (list
 (aflag 'key-press             (pc #x00000001))
 (aflag 'key-release           (pc #x00000002))
 (aflag 'button-press          (pc #x00000004))
 (aflag 'button-release        (pc #x00000008))
 (aflag 'enter-window          (pc #x00000010))
 (aflag 'leave-window          (pc #x00000020))
 (aflag 'pointer-motion        (pc #x00000040))
 (aflag 'pointer-motion-hint   (pc #x00000080))
 (aflag 'button1-motion        (pc #x00000100))
 (aflag 'button2-motion        (pc #x00000200))
 (aflag 'button3-motion        (pc #x00000400))
 (aflag 'button4-motion        (pc #x00000800))
 (aflag 'button5-motion        (pc #x00001000))
 (aflag 'button-motion         (pc #x00002000))
 (aflag 'keymap-state          (pc #x00004000))
 (aflag 'exposure              (pc #x00008000))
 (aflag 'visibility-change     (pc #x00010000))
 (aflag 'structure-notify      (pc #x00020000))
 (aflag 'resize-redirect       (pc #x00040000))
 (aflag 'substructure-notify   (pc #x00080000))
 (aflag 'substructure-redirect (pc #x00100000))
 (aflag 'focus-change          (pc #x00200000))
 (aflag 'property-change       (pc #x00400000))
 (aflag 'colormap-change       (pc #x00800000))
 (aflag 'owner-grab-button     (pc #x01000000))
))

;; Window Attributes (pp. 125)
;; see also X-change-window-attributes
(define aflag cons)
(define window-attribute-flags (list
 (aflag 'background-pixmap (pc #x00000001))
 (aflag 'background-pixel  (pc #x00000002))
 (aflag 'border-pixmap     (pc #x00000004))
 (aflag 'border-pixel      (pc #x00000008))
 (aflag 'bit-gravity       (pc #x00000010))
 (aflag 'win-gravity       (pc #x00000020))
 (aflag 'backing-store     (pc #x00000040))
 (aflag 'backing-planes    (pc #x00000080))
 (aflag 'backing-pixel     (pc #x00000100))
 (aflag 'override-redirect (pc #x00000200))
 (aflag 'save-under        (pc #x00000400))
 (aflag 'event-mask        (pc #x00000800))
 (aflag 'do-not-propagate-mask (pc #x00001000))
 (aflag 'colormap          (pc #x00002000))
 (aflag 'cursor            (pc #x00004000))
))

(define (encode-window-class sym)
 (number->byte-string-2
   (case sym
     ((copy-from-parent) 0)
     ((input-output)     1)
     ((input-only)       2)
     (else (error (format "SCA/X mischiefed window class: ~a" sym))) )))

(define (encode-event-symbols . symbols)
 (let ((bitmask (byte-string-clear! (make-byte-string 4))))
  (let loop ((symbols symbols))
   (when (not (null? symbols))
    (let* ((sym (car symbols))
           (flag (assq-ref event-flags sym)))
     (cond (flag (set! bitmask (bit-or bitmask flag))
                 (loop (cdr symbols)))
           (else (error (format "SCA/X mischiefed event symbol ~A" sym))))
  )))
  (bit-or StructureNotify bitmask) ;; XXX always listen for structure notificiations
))

(define (encode-window-attribute-symbols . symbols)
 (let ((bitmask (byte-string-clear! (make-byte-string 4))))
  (let loop ((symbols symbols))
   (when (not (null? symbols))
    (let* ((sym (car symbols))
           (flag (assq-ref window-attribute-flags sym)))
     (cond (flag (set! bitmask (bit-or bitmask flag))
                 (loop (cdr symbols)))
           (else (error (format "SCA/X mischiefed window attribute symbol ~A" sym))))
  )))
  bitmask)
)
(define (encode-window-attributes . key-value-list)
 (let ((mask (apply encode-window-attribute-symbols (map car key-value-list)))
       (rbs  '()))
 (when (> (length key-value-list) 1) ; XXX preliminary
  (error "SCA/X preliminary implementation limit: only one window attribute can be encoded per call"))
 ;; We will need to maintain a fixed order for the final rb where our list
 ;; of rbs isn't self descriptive. Maybe we should just sort the argument list?
 (let loop ((rest key-value-list)) (when (not (null? rest))
  (let ((key-value (car rest)))
   (case (car key-value)
    ((event-mask)            (set! rbs (apply encode-event-symbols (cdr key-value))))
    ((do-not-propagate-mask) (set! rbs (apply encode-event-symbols (cdr key-value))))
    ((background-pixmap)     (set! rbs (cdr key-value)))
    (else                    (error (format "SCA/X encode-window-attribute unsupported attribute ~A" (car key-value))))
 )) ))
 (values mask rbs)
))

(define X-create-window
 (lambda (connection parent-rid id x y width height border-width)
  (let ((t  X-create-window:template))
   (xrp-window-creation:set-x!            t (number+sign->byte-string-2 x))
   (xrp-window-creation:set-y!            t (number+sign->byte-string-2 y))
   (xrp-window-creation:set-width!        t (number->byte-string-2 width))
   (xrp-window-creation:set-height!       t (number->byte-string-2 height))
   (xrp-window-creation:set-border-width! t (number->byte-string-2 border-width))

   ; -Input Output 1- Copy from Parent
   (xrp-window-creation:set-class!        t (number->byte-string-2 0))

   ; Copy from Parent
   (xrp-window-creation:set-visual!       t (number->byte-string-4 0))
   
   (xrp-window-creation:set-window-id!    t id)

   ; The Parent Window
   (xrp-window-creation:set-parent!       t parent-rid)

   ; Mask is empty for now
   (xrp-window-creation:set-mask!         t ; #"0 0 0 0")
                                          (bit-or cw-back-pixel
                                                  cw-border-pixel
                                                  cw-event-mask
                                                  ))
   (xrp-window-creation:set-length!       t
    ;; sic! This calculation is the reflection of the other side.
    (number->byte-string-2 (+ (/ (byte-string-length t) 4)
                              (ones-4! (xrp-window-creation:mask t)))))

   (X-request connection
                 (byte-string-adjoin t #"128 128 128 0"
                                       #"0 0 0 0" ;; XXX don't use byte pixel here
                                       (bit-or StructureNotify))
                 no-response)
)))

(define X-rp-change-window-attributes
 (byte-structure-description
  (opcode     1)
  (padding    1)
  (request-length 2)
  (window     4)
  (value-mask 4)))
(define-bs-accessors 'xrp-change-window-attributes X-rp-change-window-attributes)
(define X-change-window-attributes:template
 (let ((template (make-byte-string (bs-length X-rp-change-window-attributes))))
  (byte-string-clear! template)
  (xrp-change-window-attributes:set-opcode! template (byte->byte-string 2))
  template))
(define (X-change-window-attributes connection window . key-value-list)
 (let* ((rb X-change-window-attributes:template)
        (mask+values (apply encode-window-attributes key-value-list))
        (value-mask  (first  mask+values))
        (tail        (second mask+values))
        (pad         '...)) ; TODO copy from somewhere else, please
   (xrp-change-window-attributes:set-window!         rb window)
   (xrp-change-window-attributes:set-request-length! rb (number->byte-string-2 (+ 3 (/ (byte-string-length tail) 4))))
   (xrp-change-window-attributes:set-value-mask!     rb value-mask)
   (X-request connection (byte-string-adjoin rb tail) no-response)))

(define X-rp-map-window
 (byte-structure-description
  (req-type 1)
  (padding  1)
  (length   2)
  (window   4)))
(define-bs-accessors 'xrp-map X-rp-map-window)
(define X-map-window:template
 (let ((template (make-byte-string (bs-length X-rp-map-window))))
  (byte-string-clear! template)
  (xrp-map:set-req-type! template (number->byte-string 8 1))
  (xrp-map:set-length!   template (number->byte-string-2 2))
  template))
(define X-map-window
 (lambda (connection window-id)
  (let ((t X-map-window:template))
   (xrp-map:set-window! t window-id)
   (X-request connection t no-response) )))

(define X-rp-configure-window
 (byte-structure-description
  (opcode     1)
  (padding    1)
  (request-length 2)
  (window     4)
  (value-mask 2)
  (padding    2)))
(define-bs-accessors 'xrp-configure-window X-rp-configure-window)
(define X-configure-window:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-configure-window)))))
  (xrp-configure-window:set-opcode! template (byte->byte-string 12))
  template))
(define pc number->byte-string-2)
(define X-configure-window:x            (pc #x0001))
(define X-configure-window:y            (pc #x0002))
(define X-configure-window:width        (pc #x0004))
(define X-configure-window:height       (pc #x0008))
(define X-configure-window:border-width (pc #x0010))
(define X-configure-window:sibling      (pc #x0020))
(define X-configure-window:stack-mode   (pc #x0040))
(define (X-configure-window connection window x y width height)
 (let ((rb X-configure-window:template)
       (s   (once (/ (+ (bs-length X-rp-configure-window) 8)) 4)))
  (xrp-configure-window:set-window!         rb window)
  (xrp-configure-window:set-request-length! rb s)
  (xrp-configure-window:set-value-mask!     rb (once (bit-or X-configure-window:x
                                                             X-configure-window:y
                                                             X-configure-window:width
                                                             X-configure-window:height)))
  (X-request connection
             (byte-string-adjoin rb (number+sign->byte-string-2 x)
                                    (number+sign->byte-string-2 y)
                                         (number->byte-string-2 width)
                                         (number->byte-string-2 height))
             no-response)
))
(define make-X-window-config-setter (macro (VAL)
 (let* ((VAL-ENCODING (case VAL ((x y)          number+sign->byte-string-2)
                                ((width height
                                  border-width) number->byte-string-2)
                                (else           (error (format "SCA/X internal error unknown window config: ~A" VAL))) ))
        (VAB          (symbol->brass VAL))
        (VAL-MASK     (brass->symbol (brass-append "X-configure-window:" VAB))))
  `(lambda (connection window ,VAL)
     (let ((rb X-configure-window:template) (pad #"0 0"))
       (xrp-configure-window:set-window!         rb window)
       (xrp-configure-window:set-value-mask!     rb ,VAL-MASK)
       (xrp-configure-window:set-request-length! rb ,(number->byte-string-2 4))
       (X-request connection (byte-string-adjoin rb (,VAL-ENCODING ,VAL) pad) no-response)))
)))
(define X-config-window-x      (make-X-window-config-setter x))
(define X-config-window-y      (make-X-window-config-setter y))
(define X-config-window-width  (make-X-window-config-setter width))
(define X-config-window-height (make-X-window-config-setter height))
(define X-config-window-border-width (make-X-window-config-setter border-width))


(define X-rp-destroy-window (byte-structure-description
 (opcode     1)
 (padding    1)
 (request-length 2)
 (window     4)))
(define-bs-accessors 'xrp-destroy-window X-rp-destroy-window)
(define X-destroy-window:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-destroy-window)))))
  (xrp-destroy-window:set-opcode!         template (byte->byte-string     4))
  (xrp-destroy-window:set-request-length! template (number->byte-string-2 2))
  template))
(define (X-destroy-window connection window)
 (let ((rb X-destroy-window:template))
  (xrp-destroy-window:set-window! rb window)
  (X-request connection rb no-response)))

(define X-rp-clear-area
 (byte-structure-description
  (opcode    1)
  (exposures 1)
  (request-length 2)
  (wid       4)
  (x         2)
  (y         2)
  (width     2)
  (height    2)))
(define-bs-accessors 'xrp-clear-area X-rp-clear-area)
(define X-clear-area:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length X-rp-clear-area)))))
  (xrp-clear-area:set-opcode!         t (byte->byte-string 61))
  (xrp-clear-area:set-request-length! t (number->byte-string-2 4))
  t))
(define (X-clear-area connection wid exposures x y width height)
 (let ((rb X-clear-area:template))
  (xrp-clear-area:set-wid!    rb wid)
  (xrp-clear-area:set-x!      rb (number+sign->byte-string-2 x))
  (xrp-clear-area:set-y!      rb (number+sign->byte-string-2 y))
  (xrp-clear-area:set-width!  rb (number->byte-string-2 width))
  (xrp-clear-area:set-height! rb (number->byte-string-2 height))
  (xrp-clear-area:set-exposures! rb (if exposures #"1" #"0"))
  (X-request connection rb no-response)))

;; GrabPointer (p. 132) 
(define X-rp-grab-pointer
 (byte-structure-description
  (opcode         1)
  (owner-events   1) 
  (request-length 2) 
  (grab-window    4)
  (event-mask     2)
  (pointer-mode   1)
  (keyboard-mode  1)
  (confine-to     4)
  (cursor         4)
  (time           4)))
(define-bs-accessors 'xrp-grab-pointer X-rp-grab-pointer)
(define X-rr-grab-pointer
 (byte-structure-description
  (reply           1)
  (status          1)
  (sequence-number 2)
  (reply-length    4)
  (padding        24)))
(define-bs-accessors 'xrr-grab-pointer X-rr-grab-pointer)
(define X-grab-pointer:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-grab-pointer))) ))
  (xrp-grab-pointer:set-opcode!         template (byte->byte-string    26))
  (xrp-grab-pointer:set-request-length! template (number->byte-string-2 6)) 
  template))
(define (X-grab-pointer connection owner-events grab-window
                        pointer-mode keyboard-mode event-mask confine-to
                        cursor time)
 (let* ((rb X-grab-pointer:template))
  (xrp-grab-pointer:set-owner-events! rb (if owner-events #"1" #"0"))
  (xrp-grab-pointer:set-grab-window!  rb grab-window)
  (xrp-grab-pointer:set-event-mask!   rb event-mask)
  (xrp-grab-pointer:set-confine-to!   rb confine-to) 
  (xrp-grab-pointer:set-cursor!       rb cursor)
  (xrp-grab-pointer:set-time!         rb (number->byte-string-4 time)) 
  (xrp-grab-pointer:set-pointer-mode! rb
   (case pointer-mode ((synchronous) #"0") ((asynchronous) #"1")
    (else (error (format "VSI SCA/X X-grab-pointer bad pointer mode: ~A" pointer-mode))) )) 
  (xrp-grab-pointer:set-keyboard-mode! rb
   (case keyboard-mode ((synchronous) #"0") ((asynchronous) #"1")
    (else (error (format "VSI SCA/X X-grab-pointer bad keyboard mode: ~A" keyboard-mode))) )) 
  (X-request connection rb (lambda (con fbw)
   (let ((seq-number (byte-string->number (xrr-grab-pointer:sequence-number fbw)))
         (status     (byte-string->number (xrr-grab-pointer:status          fbw))))
    (case status
     ((0) 'success)
     ((1) 'already-grabbed)
     ((2) 'invalid-time)
     ((3) 'not-viewable)
     ((4) 'frozen)
     (else status))
  )))
))

;; Ungrab Pointer is in the display class / xserv namespace

)) ; closes xwin namespace

)) ; closes window class
