(define terminal-class (class (extension viper-object-class)
  (slots port resize-listeners)

  ;; The regular utf-8 decoder of VSI-C throws an error: in this class
  ;; we prefer to receive falsity for malformed input sequences.
  ;; (konsole knows about these for "fast newlines" ...)
  (define read-char (from-namespace sys-unicode-namespace 'read-char:utf-8/stream))

  (method initialize ()
    (this (set! 'resize-listeners '()))
    (slot-set! (this) 'port
      (open-file "/dev/tty" "r+0b"))
    (sys-parallel (add-signal-handler SIGWINCH
                                      (this (create-signal-handler))))
  )

  (method create-signal-handler ()
   (lambda ()
    (let ((dim (this (get-dimensions))))
     (for-each (lambda (h) (h (terminal-resize (first dim)
                                               (second dim)
                                               (third dim))))
               (this resize-listeners)) )))

  (method add-resize-listener (listener)
   (this (set! 'resize-listeners (cons listener (this resize-listeners)) )))

  (method play-char-list (char-list)
    (send-bytes (port->fdes (this port))
                (apply byte-string-adjoin
                ; VSI Nature system requirement: UTF-8 TTY
                       (map char->byte-string/utf-8 char-list))))

  (method control (control-characters) ;; XXX not a method
    (display (list->string control-characters) (this port)))

  (method full-reset ()
    (this (control '(#\esc #\c))))

  ;;
  ;; Control Sequence Introduction
  ;; NB There is also some former 8-bit C1 CSI character but
  ;;    this 7-bit C0 setup seems to work well on most terminals.
  (define CSI '(#\esc #\[))

  ;;
  ;; SRM: Send Receive Mode
  ;; turn terminal echo on or off
  (method SRM-echo-off () (this (control (append CSI '(#\1 #\2 #\h)))))
  (method SRM-echo-on  () (this (control (append CSI '(#\1 #\2 #\l)))))

  ;;
  ;; CSI: Cursor Control
  (method position-cursor (x y)
    (let ((column (string->list (number->string x)))
          (line   (string->list (number->string y))))
    (this (control (append CSI line '(#\;) column '(#\H))))))
  (method seq-position-cursor (x y) ;; XXX vtb.class
    (let ((column (string->list (number->string x)))
          (line   (string->list (number->string y))))
    (list-copy (append CSI line '(#\;) column '(#\H)))))

  (method save-cursor ()     (this (control '(#\esc #\7))))
  (method restore-cursor ()  (this (control '(#\esc #\8))))

  (method cursor-home     () (this (control (append CSI '(#\0 #\; #\0 #\H)))))
  (method cursor-up       () (this (control (append CSI '(#\A)))))
  (method cursor-down     () (this (control (append CSI '(#\B)))))
  (method cursor-forward  () (this (control (append CSI '(#\C)))))
  (method cursor-backward () (this (control (append CSI '(#\D)))))

  (method show-cursor () (this (control (append CSI '(#\? #\2 #\5 #\h)))))
  (method hide-cursor () (this (control (append CSI '(#\? #\2 #\5 #\l)))))

  ;;
  ;; CSI: Erase
  (method erase-line    () (this (control (append CSI '(#\2 #\K)))))
  (method erase-display () (this (control (append CSI '(#\2 #\J)))))

  ;;
  ;; CSI: Character Attributes / Colors
  (method inverse () (this (control (append CSI '(#\7 #\m)))))

  (method color (foreground/background color-name)
   (this (control
    (append CSI
      (case foreground/background
            ((foreground)  '(#\3))
             ((background) '(#\4))
             (else (error "terminal: color foreground/background: domain error")))
      (case color-name
            ((black)   '(#\0))
            ((red)     '(#\1))
            ((green)   '(#\2))
            ((yellow)  '(#\3))
            ((blue)    '(#\4))
            ((magenta) '(#\5))
            ((cyan)    '(#\6))
            ((white)   '(#\7))
            ((default original) '(#\8))
            (else (error "terminal: color-name: domain error" color-name)))
      '(#\m) ))))

  (method reset-colors ()
    (this (color 'background 'default))
    (this (color 'foreground 'original)))


  ;;
  ;; CSI: DEC Private / Meta Escape, Alt Escape
  (method CSI-do-send-meta-escape   () (this (control (append CSI '(#\? #\1 #\0 #\3 #\6 #\h)))))
  (method CSI-do-send-alt-escape    () (this (control (append CSI '(#\? #\1 #\0 #\3 #\9 #\h)))))
  (method CSI-dont-send-meta-escape () (this (control (append CSI '(#\? #\1 #\0 #\3 #\6 #\l)))))
  (method CSI-dont-send-alt-escape  () (this (control (append CSI '(#\? #\1 #\0 #\3 #\9 #\l)))))

;; ;; ;;
;; ;; ;; DEFUNCT CSI: Window Control (dtterm&extensions, XTerm)
;; ;; (define (CSI-window-control string)
;; ;;   (append CSI (string->list string)))
;; ;; (define (get-dimensions/CSI terminal)
;; ;;   "Deprecated routine of how not to do it."
;; ;;   (let ((default-values (values #f 80 24)) ; detected? X Y
;; ;;         (response-introduction (append CSI '(#\8 #\;)))
;; ;;         (port (get-port terminal))
;; ;;         (x '()) (y '()))
;; ;;   (define (x-value char)
;; ;;     (cond ((digit? char)
;; ;;             (set! x (append x (list char)))
;; ;;             (x-value (read-char port)))
;; ;;           ((and (eq? char #\t) (not (null? x)))
;; ;;                 #t)
;; ;;           ((and (eq? char #\;) (null? x))
;; ;;                 #f)
;; ;;           (else #f)))
;; ;;   (define (y-value char)
;; ;;     (cond ((digit? char) 
;; ;;              (set! y (append y (list char)))
;; ;;              (y-value (read-char port)))
;; ;;           ((and (eq? char #\;) (not (null? y)))
;; ;;                 #t)
;; ;;           ((and (eq? char #\;) (null? y))
;; ;;                 #f)
;; ;;           (else
;; ;;            ;(my-debug " y-value: not-a-digit " (print-safely char))
;; ;;            #f)))
;; ;;    (control terminal (append CSI (string->list "18t")))
;; ;;    (if (and (begin (sleep 1) #t) ; should we rely on this time interval ???
;; ;;             (char-ready? port)          ; some VTs do not respond to this and
;; ;;             (eat!-char-sequence? response-introduction port) ; we rely on CSI.
;; ;;             (y-value (read-char port))  ; y comes first
;; ;;             (x-value (read-char port))) ; and x is second.
;; ;;        (values #t ; yes, it's true: we detected these values.
;; ;;                (string->number (list->string x))  ; all checks done
;; ;;                (string->number (list->string y))) ;   => this can't fail
;; ;;        default-values))) ; read out failed => default values
;; ;;


  (define (get-dimensions/VSI-C terminal)
    (let ((default-values (values #f 80 24)) ; (not) detected width height 
          (file-descriptor (port->fdes terminal)))
      (let ((answer (%terminal-get-win-size file-descriptor)))
        (if answer
            (receive answer (x y pixel-x pixel-y)
                     (values #!detected:!#  #t
                             #!   width:!#  x
                             #!  height:!#  y ))
            default-values)) ))

  (method get-dimensions () (get-dimensions/VSI-C (this port)))
  (method get-width ()
    (call-with-values (lambda () (this (get-dimensions (this port)))
                      (lambda (detected x y) x))))
  (method get-height ()
    (call-with-values (lambda () (this (get-dimensions (this port))))
                      (lambda (detected x y) y)))


  ;;
  ;; Raw Mode / Sane Mode
  (method tty-raw-mode    () (%terminal-raw-mode    (port->fdes (this port))))
  (method tty-cooked-mode () (%terminal-cooked-mode (port->fdes (this port))))

  ;;
  ;; Beep (n/a or output the control character)
  ;(define (beep terminal) (if #f 'unspecified))

  (method bell () (display #\bel (this port)))

  ;;
  ;; ??? Printing on the Virtual Terminal
  ;; (define-method (display (argument <object>) (terminal <terminal>))
  ;;   (display argument (get-port terminal)))
  ;(define-slot <terminal> 'display ...)


  ;;
  ;; Input & Event Loop
  (define (eat!-char? char port)
    (eq? char (read-char port)))
  (define (eat!-char-sequence? char-list port)
    (if (null? char-list)
        #t
        (if (eat!-char? (car char-list) port)
            (eat!-char-sequence? (cdr char-list) port)
            #f)))

  (define (my-debug . args) #t) ; dummy

  (define (debug-drain port)
    (define (inner-drain)
      (if (not (char-ready? port))
          '()
          (cons (represent-char-safely (read-char port))
                (inner-drain))))
    (reverse! (inner-drain)))
  (define (fast-drain port)
    (if (not (char-ready? port))
        #t
        (begin (read-char port) (fast-drain))))
  ;; (define (drain port) ;; XXX does not drain reliably
  ;;   (map represent-char-safely
  ;;        (string->list (drain-input port))))
  (define unrecognized-input #f)
  (define (convert-csi-input port)
    (let ((char (read-char port)))
      (case char
        ((#\A) 'up)
        ((#\B) 'down)
        ((#\C) 'right)
        ((#\D) 'left)
        ((#\H) 'home)
        ((#\F) 'end)
        ((#\1) (if (eat!-char? #\~ port) 'home #f)) ; vt220 find / linux home
        ((#\2) (if (eat!-char? #\~ port) 'insert-here  #f)) ; numpad enter???
        ((#\3) (if (eat!-char? #\~ port)  #\del        #f))
        ((#\4) (if (eat!-char? #\~ port) 'end #f)) ; vt220 select / linux end
        ((#\5) (if (eat!-char? #\~ port) 'page-up      #f))
        ((#\6) (if (eat!-char? #\~ port) 'page-down    #f))
        (else
         (my-debug "terminal: unrecognized input:" ; XXX drain&dump input here
                   " CSI " (cons (represent-char-safely char)
                                 (debug-drain port)))
         unrecognized-input) )))
  (define (convert-escape-sequence port)
    (let ((char (read-char port)))
      (case char
        ;; NB There is also some former 8-bit C1 CSI character but
        ;;    this 7-bit C0 setup seems to work well on most terminals.
        ((#\[) (convert-csi-input port))
        (else
         (my-debug "terminal: unrecognized input:" ; XXX drain&dump input here
                   " ESC-SEQ " (cons (represent-char-safely char)
                                     (debug-drain port)))
         unrecognized-input) )))
  (define (dispatch-on-escape port)
    (if (not (char-ready? port))
        #\esc ; user typed escape goes to event handler
        (convert-escape-sequence port) ))

  ;; (define (start-event-loop event-handler port)
  ;;  (let ((port port))
  ;;   (let event-loop ((char (read-char port)))
  ;;     ;(my-debug (char->byte char))
  ;;     (and (event-handler (cond
  ;;            ((eof-object? char) 'end-of-file)
  ;;            ((and (char? char) (is-unicode-c0? (char->integer char)))
  ;;               (case char
  ;;                 ((#\esc) (dispatch-on-escape port))
  ;;                 ((#\del) #\backspace) ; delete / backspace
  ;;                 ((#\cr)  'enter) ; XXX check numpad enter input code
  ;;                 (else    char)) )
  ;;            ;(else #f)))
  ;;            (else char ))) ; high code point hack
  ;;     (event-loop (read-char port)) )) ))

  (define (cheapvoc-t-events event-handler port)
   (let ((port port))
    (let ((char (read-char port))
          (bc   #f))
      ;(my-debug (char->byte char))
      ;(and (event-handler (cond
      (set! bc (event-handler
       (cond ((eof-object? char) 'end-of-file)
             ((and (char? char) (is-unicode-c0? (char->integer char)))
                (case char
                  ((#\esc) (dispatch-on-escape port))
                  
                  ;; The following delete / backspace exchange works:
                  ;; the event handler even sees #\del events because
                  ;; delete is transmitted as an escape sequence.
                  ((#\del) #\backspace) ; delete / backspace

                  ((#\cr #\nl)  'enter) ; XXX check numpad enter input code
                                        ; added #\newline to allow for copy
                                        ; & paste into the VTB
                  (else    char)) )
             ;(else #f))))
             (else  char )))) ; high code point hack
      (when (not bc) (sys-control (remove-read-handler port)))
  )))

  ;;
  ;; Key Event Protocol 
  ;;
  ;; Calling Convention: (event-handler key-event)
  ;;
  ;; A key event is either
  ;;  - false (#f) for unrecognized and malformed input,
  ;;  - a character including controls or
  ;;  - a symbol corresponding to the respective keys or 'end-of-file
  ;;    '(enter up down right left home end insert-here
  ;;      end-of-file ;page-up page-down)
  ;;
  ;; Event Loop Control: if the handler returns falsity the loop
  ;; doesn't continue and returns.
  ;;

  ;; (method start-event-loop (event-handler)
  ;;   (start-event-loop event-handler (this port)))
  ;; ;; (method start-event-loop (event-handler)
  ;; ;;    (this (set-event-handler event-handler)))

  (method set-event-handler (event-handler)
    (sys-control (add-read-handler (this port) (lambda ()
     (cheapvoc-t-events event-handler (this port))))))

)) ; closes define terminal-class

