/*
 * Byte Strings: a byte has eight bits.
 *  & byte string mappings (with a little bit more time and take your time)
 */

#include "include/ihub.h"
#include "include/byte-strings.h"

long global_number_of_byte_strings = 0;
long global_number_of_byte_string_mappings = 0;

inline void*
scm_i_byte_string_data (SCM byte_string)
{
  if (BYTE_STRING_MAPPING_P (byte_string)) {
    return (void*)SCM_CELL_OBJECT(byte_string,1);

  } else if (BYTE_STRING_P (byte_string)) {
    return (void*)SCM_CELL_OBJECT_LOC(byte_string,1);

  } else {
    fprintf (stderr, " * VSI-C scm_i_byte_string_data wrong type argument * ");
    abort ();
  }
}

SCM
scm_i_make_byte_string_mapping (void *addr, size_t length)
{
  global_number_of_byte_string_mappings++;
  return scm_cell (scm_tc16_byte_string_mapping,
                   (scm_t_bits) length,
                   (scm_t_bits) addr);
}

int
scm_print_byte_string (SCM exp, SCM port, scm_print_state *pstate)
{
  scm_puts ("#\"", port);
  size_t length = BYTE_STRING_LENGTH (exp);
  char  *data   = (char *) BYTE_STRING_DATA (exp);
  long j;
  for (j = 0; j < length; j++) {
    unsigned char byte = data[j];
    unsigned long ibyte = byte & 255;
    scm_uintprint (ibyte, 10, port); // is this nested data ???

    if (j == length-1) break;

    if ((j+1) % 4 == 0) {
      scm_puts (" ", port);
    }

    /* if ((length > 255)                     \ */
    /*     && (!SCM_WRITINGP (pstate))        \ */
    /*     && (j > 8)) { */
    /*   // perhaps even write shouldn't try to print the whole data */
    /*   // segment or it even should print the byte values??? */
    /*   scm_puts ("...", port); */
    /*   j = length - 8; */
    /* } */
    scm_puts (" ", port);
  }
  scm_puts ("\"", port);
  return (!0);
}

SCM
scm_c_make_byte_string (size_t length_in_bytes)
{
  global_number_of_byte_strings++;

  // Assert: sizeof SCM == 8
  // Num Slots for Bytes: num_slots = (length_in_bytes / 8) + 1
  //    Length in Slot 0: num_slots+1
  // -Alignment to modulo 16: +1- (if mallocs were modulo 16 but that is
  //                               considered a myth right now)
  // !!! SYNC WITH /trim_release_byte_string/ and byte-string.h
  SCM answer = scm_tagged_array (scm_tc7_byte_string,
                                 (length_in_bytes >> 3) + 2);
  SCM_SET_CELL_WORD_0 (answer, (scm_t_bits) length_in_bytes);
  return answer;
}

SCM
scm_make_byte_string_with_copy (void *data_segment, size_t length_in_bytes)
{
  SCM bs = scm_c_make_byte_string (length_in_bytes);
  memcpy (BYTE_STRING_DATA (bs), data_segment, length_in_bytes);
  return bs;
}

SCM
scm_byte_string_equal_p (SCM a, SCM b)
#define FUNC_NAME "equal?(byte-string)"
{
  reference_parameter (a);
  reference_parameter (b);
  SCM answer = SCM_UNSPECIFIED;

  if (BYTE_STRING_P (a) && BYTE_STRING_P (b)) {
    size_t length = BYTE_STRING_LENGTH (a);
    if (length != BYTE_STRING_LENGTH (b)) {
      answer = SCM_BOOL_F;
    } else {
      void* d1 = (void*)BYTE_STRING_DATA (a);
      void* d2 = (void*)BYTE_STRING_DATA (b);
      switch (length) {
      case 8: answer = (*(scm_t_uint64*) d1 == *(scm_t_uint64*) d2)  \
                       ? SCM_BOOL_T : SCM_BOOL_F; break;

      case 4: answer = (*(scm_t_uint32*) d1 == *(scm_t_uint32*) d2)  \
                       ? SCM_BOOL_T : SCM_BOOL_F; break;

      case 2: answer = (*(scm_t_uint16*) d1 == *(scm_t_uint16*) d2)  \
                       ? SCM_BOOL_T : SCM_BOOL_F; break;

      case 1: answer = (*(scm_t_uint8*) d1 == *(scm_t_uint8*) d2)    \
                       ? SCM_BOOL_T : SCM_BOOL_F; break;

      default:
         answer = (!bcmp (d1, d2, length)) ? SCM_BOOL_T : SCM_BOOL_F;
      }
    }
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 3, SCM_BOOL_F, "byte-string");
  }

  dereference_parameter (a);
  dereference_parameter (b);
  return (answer); 
}
#undef FUNC_NAME

SCM_DEFINE (scm_make_byte_string, "make-byte-string", 1, 0, 0,
            (SCM length), "")
#define FUNC_NAME s_scm_make_byte_string
{
  reference_parameter (length);
  size_t my_length; SCM_VALIDATE_ULONG_COPY (1, length, my_length);

  SCM answer = scm_c_make_byte_string (my_length);
  dereference_parameter (length);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_clear_x, "byte-string-clear!", 1, 0, 0,
            (SCM byte_string), "")
#define FUNC_NAME s_scm_byte_string_clear_x
{
  reference_parameter (byte_string);
  SCM answer = SCM_UNSPECIFIED;

  if (BYTE_STRING_P (byte_string)) {
    bzero ((void*)BYTE_STRING_DATA (byte_string),
           BYTE_STRING_LENGTH (byte_string));
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_p, "byte-string?", 1, 0, 0,
            (SCM object), "")
#define FUNC_NAME s_scm_byte_string_p
{
  reference_parameter (object);
  SCM answer = scm_from_bool (BYTE_STRING_P (object));
  dereference_parameter (object);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_mapping_p, "byte-string-mapping?", 1, 0, 0,
            (SCM object), "")
#define FUNC_NAME s_scm_byte_string_mapping_p
{
  reference_parameter (object);
  SCM answer = scm_from_bool (BYTE_STRING_MAPPING_P (object));
  dereference_parameter (object);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_length, "byte-string-length", 1, 0, 0,
            (SCM byte_string), "")
#define FUNC_NAME s_scm_byte_string_length
{
  reference_parameter (byte_string);
  SCM answer = SCM_UNSPECIFIED;

  if (BYTE_STRING_P (byte_string)) {
    answer = scm_from_uint (BYTE_STRING_LENGTH (byte_string));
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  dereference_parameter (byte_string);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_address, "byte-string-address", 1, 0, 0,
            (SCM byte_string), "")
#define FUNC_NAME s_scm_byte_string_address
{
  reference_parameter (byte_string);
  SCM answer = SCM_UNSPECIFIED;

  if (BYTE_STRING_P (byte_string)) {
    void *dat = (BYTE_STRING_DATA (byte_string));
    answer = scm_make_byte_string_with_copy (&dat, sizeof (void*));
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  dereference_parameter (byte_string);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_byte_string_adjoin, "byte-string-adjoin", 1, 0, 1,
            (SCM byte_string, SCM rest), "")
#define FUNC_NAME s_scm_byte_string_adjoin
{
  reference_parameter (byte_string);
  reference_parameter (rest);
  SCM_VALIDATE_REST_ARGUMENT (rest);
  
  if (BYTE_STRING_P (byte_string)) {
    size_t length = BYTE_STRING_LENGTH (byte_string);
    SCM tmp = rest;

    { // Argument scanning block
      int n_arg = 1;

      // Check types and determine overall size.
      while (!scm_is_null (tmp)) {
        SCM current = TMP_CAR (tmp);
        n_arg++;

        if (!BYTE_STRING_P (current)) {
          scm_wrong_type_arg_msg (FUNC_NAME, n_arg, current, "byte-string");
        }
        
        size_t new_length = length + BYTE_STRING_LENGTH (current);
        if (new_length < length)  // overflow
          scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

        length = new_length;
        tmp = TMP_CDR (tmp);
      }
    }

    // XXX validate new_length to be in bounds
    SCM answer = scm_c_make_byte_string (length);

    // Now join the byte sequences into the new buffer.
    { // Main working loop block
      char  *data_offset = BYTE_STRING_DATA (answer);
      size_t last_length = BYTE_STRING_LENGTH (byte_string);

      memcpy (data_offset, (void*)BYTE_STRING_DATA (byte_string), last_length);
      data_offset = &(data_offset[last_length]);

      tmp = rest;
      while (!scm_is_null (tmp)) {
        SCM current = TMP_CAR (tmp);
        last_length = BYTE_STRING_LENGTH (current);
        memcpy (data_offset, (void*)BYTE_STRING_DATA (current), last_length);
        data_offset = &(data_offset[last_length]);
        tmp = TMP_CDR (tmp);
      }
    } // assert: data_offset == data_segment + length

    dereference_parameter (rest);
    dereference_parameter (byte_string);
    return (answer);

  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 0, byte_string, "byte-string");
    return SCM_BOOL_F;
  }
  // never reached
}
#undef FUNC_NAME

SCM scm_c_byte_string_adjoin (SCM lst)
{
  if (scm_is_null (lst)) {
    return scm_c_make_byte_string (0);
  } else {
    return scm_byte_string_adjoin (TMP_CAR (lst), TMP_CDR (lst));
  }
}


SCM_DEFINE (scm_byte_string_copy, "byte-string-copy", 1, 0, 0,
            (SCM byte_string), "")
#define FUNC_NAME s_scm_byte_string_copy
{ // XXX this is just byte-string-adjoin for one argument
  reference_parameter (byte_string);
  SCM answer;

  if (BYTE_STRING_P (byte_string)) {
    answer = \
      scm_make_byte_string_with_copy (BYTE_STRING_DATA (byte_string), \
                                      BYTE_STRING_LENGTH (byte_string));
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 0, byte_string, "byte-string");
    return SCM_BOOL_F;
  }

  dereference_parameter (byte_string);
  return answer;
}
#undef FUNC_NAME


//
// Conversion Matrix Prototypes
// we have: (signed) unsigned, char, short, int, long, long long
// then: pointer (always unsigned), float (???), double (???)
//
// size-tag->byte-string: number -> byte-string
// size-tag->number: byte-string -> number
// (that domain boundary is fluid)
//
// maybe: byte-string->unsigned-number byte-string length
//   and  byte-string->signed-number   byte-string length
//  with some Scheme sugar.
// where supported values for lengths should be: 1, 2, 4, 8

#include "include/byte-strings.i.c"

/* Selector and Setter */
SCM_DEFINE (scm_byte_string_mapping, "byte-string-mapping", 3, 0, 0,
            (SCM byte_string, SCM offset, SCM length), "")
#define FUNC_NAME s_scm_byte_string_mapping
{
  reference_parameter (byte_string);
  reference_parameter (offset);
  reference_parameter (length);

  if (!BYTE_STRING_P (byte_string))
    scm_wrong_type_arg_msg (FUNC_NAME, 0, byte_string, "byte-string");

  size_t o; SCM_VALIDATE_ULONG_COPY (2, offset, o);
  size_t l; SCM_VALIDATE_ULONG_COPY (3, length, l);

  if (o+l > BYTE_STRING_LENGTH (byte_string))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  SCM answer = \
    scm_i_make_byte_string_mapping (o + BYTE_STRING_DATA (byte_string), l);

  dereference_parameter (byte_string);
  dereference_parameter (offset);
  dereference_parameter (length);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_root_mapping, "byte-string-root-mapping", 2, 0, 0,
            (SCM address, SCM length), "")
#define FUNC_NAME s_scm_byte_string_root_mapping
{
  reference_parameter (address);
  reference_parameter (length);

  SCM_VALIDATE_BYTE_STRING_AS (1, address, void*);
  size_t l; SCM_VALIDATE_ULONG_COPY (2, length, l);

  SCM answer = \
    scm_i_make_byte_string_mapping (*((void**)BYTE_STRING_DATA (address)), l);

  dereference_parameter (address);
  dereference_parameter (length);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_select, "byte-string-select", 3, 0, 0,
            (SCM byte_string, SCM offset, SCM length), "")
#define FUNC_NAME s_scm_byte_string_select
{
  reference_parameter (byte_string);
  reference_parameter (offset);
  reference_parameter (length);

  if (!BYTE_STRING_P (byte_string))
    scm_wrong_type_arg_msg (FUNC_NAME, 0, byte_string, "byte-string");

  size_t o; SCM_VALIDATE_ULONG_COPY (2, offset, o);
  size_t l; SCM_VALIDATE_ULONG_COPY (3, length, l);

  if (o+l > BYTE_STRING_LENGTH (byte_string))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  SCM answer = \
    scm_make_byte_string_with_copy (o + BYTE_STRING_DATA (byte_string), l);

  dereference_parameter (byte_string);
  dereference_parameter (offset);
  dereference_parameter (length);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_set_x, "byte-string-set!", 3, 0, 0,
            (SCM byte_target, SCM offset, SCM byte_source), "")
#define FUNC_NAME s_scm_byte_string_set_x
{
  reference_parameter (byte_target);
  reference_parameter (offset);
  reference_parameter (byte_source);

  if (!BYTE_STRING_P (byte_source))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_source, "byte-string");

  if (!BYTE_STRING_P (byte_target))
    scm_wrong_type_arg_msg (FUNC_NAME, 3, byte_target, "byte-string");

  size_t o; SCM_VALIDATE_ULONG_COPY (2, offset, o);
  size_t l = BYTE_STRING_LENGTH (byte_source);

  if (o+l > BYTE_STRING_LENGTH (byte_target))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  memcpy ((void*) (o + BYTE_STRING_DATA (byte_target)),
          (void*) (BYTE_STRING_DATA (byte_source)),
          l);

  dereference_parameter (byte_target);
  dereference_parameter (offset);
  dereference_parameter (byte_source);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#include <endian.h> // <- see also
#include <byteswap.h>
#define SCM_BYTE_SWAP8 bswap_64
#define SCM_BYTE_SWAP4 bswap_32
#define SCM_BYTE_SWAP2 bswap_16
// SCM_BYTE_SWAP1 is still missing ...

SCM_DEFINE (scm_byte_reverse_x, "byte-reverse!", 1, 0, 0, (SCM byte_string), "")
#define FUNC_NAME s_scm_byte_reverse_x
{
  reference_parameter (byte_string);

  if (!BYTE_STRING_P (byte_string)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  size_t length  = BYTE_STRING_LENGTH (byte_string);
  void * data = (void*)BYTE_STRING_DATA (byte_string);
  switch (length) {
  case 8:
    *(scm_t_uint64 *)BYTE_STRING_DATA (byte_string) = \
    SCM_BYTE_SWAP8 (*(scm_t_uint64 *)BYTE_STRING_DATA (byte_string));
    break;
  case 4:
    *(scm_t_uint32 *)BYTE_STRING_DATA (byte_string) = \
    SCM_BYTE_SWAP4 (*(scm_t_uint32 *)BYTE_STRING_DATA (byte_string));
    break;
  case 2:
    *(scm_t_uint16 *)BYTE_STRING_DATA (byte_string) = \
    SCM_BYTE_SWAP2 (*(scm_t_uint16 *)BYTE_STRING_DATA (byte_string));
    break;
  case 1: // identity function
    break;
    
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME


/* Bit Operations */

void
scm_i_bit_not_x (void *data, size_t length)
{
// ??? alignment if ((data % 8) == 0) {
  if (length >= 8) {
    scm_t_uint64 *my_long = (scm_t_uint64 *)data;
    *my_long = ~(*my_long);  // bit wise not (not minus)
    if (length == 8) return;
    scm_i_bit_not_x ((void*)(((unsigned long) data) + 8), \
                     length - 8); return;
  }

  if (length >= 4) {
    scm_t_uint32 *my_int = (scm_t_uint32 *)data;
    *my_int = ~(*my_int);
    length -= 4;
    data = (void*) (((unsigned long) data) + 4);
  }

  if (length >= 2) {
    scm_t_uint16 *my_short = (scm_t_uint16 *)data;
    *my_short = ~(*my_short);
    length -= 2;
    data = (void*) (((unsigned long) data) + 2);
  }
// }

  if (length == 1) {
    scm_t_uint8 *my_char = (scm_t_uint8 *)data;
    *my_char = ~(*my_char);
    length -= 1;
  }

  // length == 0
  return;
}

SCM_DEFINE (scm_bit_not_x, "bit-not!", 1, 0, 0, (SCM byte_string), "")
#define FUNC_NAME s_scm_bit_not_x
{
  reference_parameter (byte_string);

  if (BYTE_STRING_P (byte_string)) {
    scm_i_bit_not_x ((void *)BYTE_STRING_DATA (byte_string),
                     BYTE_STRING_LENGTH (byte_string));
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME


// An algorithmic bit reverse without lookup table. 
// If you ever try to do this never forget:
// you only need one of these to create the lookup
// table and then you wouldn't miss them.
void
scm_i8_bit_reverse (void *data, size_t length)
{
  #define BSBR_T scm_t_uint64
  if (length == 8) {
    register size_t bits = length * 8; // bytes in bits
    register BSBR_T old_word = *(BSBR_T *)data;
    register BSBR_T new_word = 0L;
    while (1) {
      if (old_word & 1L) new_word++; // loop op
      if (--bits == 0) break;   // test
      old_word = old_word >> 1; // step
      new_word = new_word << 1; // step
    }
    *(BSBR_T*)data = new_word;
  }
  return;
  #undef BSBR_T
}
// Copy&Paste Art Work
void scm_i4_bit_reverse (void *data, size_t length) {
  #define BSBR_T scm_t_uint32
  if (length == 4) {
    register size_t bits = length * 8; // bytes in bits
    register BSBR_T old_word = *(BSBR_T *)data;
    register BSBR_T new_word = 0L;
    while (1) {
      if (old_word & 1L) new_word++; // loop op
      if (--bits == 0) break;   // test
      old_word = old_word >> 1; // step
      new_word = new_word << 1; // step
    }
    *(BSBR_T*)data = new_word;
  }
  return;
  #undef BSBR_T
}
void scm_i2_bit_reverse (void *data, size_t length) {
  #define BSBR_T scm_t_uint16
  if (length == 2) {
    register size_t bits = length * 8; // bytes in bits
    register BSBR_T old_word = *(BSBR_T *)data;
    register BSBR_T new_word = 0L;
    while (1) {
      if (old_word & 1L) new_word++; // loop op
      if (--bits == 0) break;   // test
      old_word = old_word >> 1; // step
      new_word = new_word << 1; // step
    }
    *(BSBR_T*)data = new_word;
  }
  return;
  #undef BSBR_T
}
void scm_i1_bit_reverse (void *data, size_t length) {
  #define BSBR_T scm_t_uint8
  if (length == 1) {
    register size_t bits = length * 8; // bytes in bits
    register BSBR_T old_word = *(BSBR_T *)data;
    register BSBR_T new_word = 0L;
    while (1) {
      if (old_word & 1L) new_word++; // loop op
      if (--bits == 0) break;   // test
      old_word = old_word >> 1; // step
      new_word = new_word << 1; // step
    }
    *(BSBR_T*)data = new_word;
  }
  return;
  #undef BSBR_T
}

SCM_DEFINE (scm_bit_reverse_x, "bit-reverse!", 1, 0, 0, (SCM byte_string), "")
#define FUNC_NAME s_scm_bit_reverse_x
{
  reference_parameter (byte_string);

  if (!BYTE_STRING_P (byte_string)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");
  }

  size_t length  = BYTE_STRING_LENGTH (byte_string);
  void * data = (void*)BYTE_STRING_DATA (byte_string);
  switch (length) {
  case 8:
    scm_i8_bit_reverse ((void *)BYTE_STRING_DATA (byte_string), length);
    break;
  case 4:
    scm_i4_bit_reverse ((void *)BYTE_STRING_DATA (byte_string), length);
    break;
  case 2:
    scm_i2_bit_reverse ((void *)BYTE_STRING_DATA (byte_string), length);
    break;
  case 1:
    scm_i1_bit_reverse ((void *)BYTE_STRING_DATA (byte_string), length);
    break;
    
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME

// How do you mirror right to left? Copy&Paste.
SCM_DEFINE (scm_bit_shift_right_x, "bit-shift-right!",
            2, 0, 0, (SCM byte_string, SCM offset), "")
#define FUNC_NAME s_scm_bit_shift_right_x
{
  reference_parameter (byte_string);
  reference_parameter (offset);

  if (!BYTE_STRING_P (byte_string))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");

  unsigned short c_offset; SCM_VALIDATE_USHORT_COPY (2, offset, c_offset);

  size_t length = BYTE_STRING_LENGTH (byte_string);

  // C Bit Shifts fail silently and we don't want that.
  // Perhaps we should return 0?
  if (c_offset >= (length<<3)) // bits and bytes
    scm_out_of_range (FUNC_NAME, offset);

  char *data = (char*)BYTE_STRING_DATA (byte_string);

  #define BOP >>
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = (*(scm_t_uint64*)data) BOP c_offset; break;
  case 4:
    *(scm_t_uint32*)data = (*(scm_t_uint32*)data) BOP c_offset; break;
  case 2:
    *(scm_t_uint16*)data = (*(scm_t_uint16*)data) BOP c_offset; break;
  case 1:
    *(scm_t_uint8*)data  = (*(scm_t_uint8*)data)  BOP c_offset; break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  dereference_parameter (offset);
  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME

// How do you mirror left to right? Copy&Paste.
SCM_DEFINE (scm_bit_shift_left_x, "bit-shift-left!",
            2, 0, 0, (SCM byte_string, SCM offset), "")
#define FUNC_NAME s_scm_bit_shift_left_x
{
  reference_parameter (byte_string);
  reference_parameter (offset);

  if (!BYTE_STRING_P (byte_string))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, byte_string, "byte-string");

  unsigned short c_offset; SCM_VALIDATE_USHORT_COPY (2, offset, c_offset);

  size_t length = BYTE_STRING_LENGTH (byte_string);

  // C Bit Shifts fail silently and we don't want that.
  // Perhaps we should return 0?
  if (c_offset >= (length<<3)) // bits and bytes
    scm_out_of_range (FUNC_NAME, offset);

  char *data = (char*)BYTE_STRING_DATA (byte_string);

  #define BOP << 
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = (*(scm_t_uint64*)data) BOP c_offset; break;
  case 4:
    *(scm_t_uint32*)data = (*(scm_t_uint32*)data) BOP c_offset; break;
  case 2:
    *(scm_t_uint16*)data = (*(scm_t_uint16*)data) BOP c_offset; break;
  case 1:
    *(scm_t_uint8*)data  = (*(scm_t_uint8*)data)  BOP c_offset; break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  dereference_parameter (offset);
  return (rc_return_stack_reference (byte_string));
}
#undef FUNC_NAME

// Change /bit-and/ and then Copy&Paste the following functions:
//  /bit-or/, /xor/
SCM_DEFINE1 (scm_bit_and, "bit-and", scm_tc7_asubr, (SCM bs1, SCM bs2),"")
#define FUNC_NAME s_scm_bit_and
{
  // asubr auto folding identity
  if (SCM_UNBNDP (bs2)) return (scm_byte_string_copy (bs1));
  // clr2/a preambel
  reference_parameter (bs1);
  reference_parameter (bs2);
  // Argument Validation: type
  if (!BYTE_STRING_P (bs1))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, bs1, "byte-string");
  if (!BYTE_STRING_P (bs2))
    scm_wrong_type_arg_msg (FUNC_NAME, 2, bs2, "byte-string");

  // Argument Validation and Initialization: range
  size_t length = BYTE_STRING_LENGTH (bs1);
  if (length != BYTE_STRING_LENGTH (bs2))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  void *d1 = (void*)BYTE_STRING_DATA (bs1);
  void *d2 = (void*)BYTE_STRING_DATA (bs2);

  SCM answer = scm_c_make_byte_string (length);
  char *data = BYTE_STRING_DATA (answer);

  #define BOP &
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = *(scm_t_uint64*)d1 BOP *(scm_t_uint64*)d2; break;
  case 4:
    *(scm_t_uint32*)data = *(scm_t_uint32*)d1 BOP *(scm_t_uint32*)d2; break;
  case 2:
    *(scm_t_uint16*)data = *(scm_t_uint16*)d1 BOP *(scm_t_uint16*)d2; break;
  case 1:
    *(scm_t_uint8*)data  = *(scm_t_uint8*)d1  BOP *(scm_t_uint8*)d2;  break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  // clr2/a postambel
  dereference_parameter (bs1);
  dereference_parameter (bs2);
  return (answer);
}
#undef FUNC_NAME

// Copy&Paste
SCM_DEFINE1 (scm_bit_or, "bit-or", scm_tc7_asubr, (SCM bs1, SCM bs2),"")
#define FUNC_NAME s_scm_bit_or
{
  // asubr auto folding identity
  if (SCM_UNBNDP (bs2)) return (scm_byte_string_copy (bs1));
  // clr2/a preambel
  reference_parameter (bs1);
  reference_parameter (bs2);
  // Argument Validation: type
  if (!BYTE_STRING_P (bs1))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, bs1, "byte-string");
  if (!BYTE_STRING_P (bs2))
    scm_wrong_type_arg_msg (FUNC_NAME, 2, bs2, "byte-string");

  // Argument Validation and Initialization: range
  size_t length = BYTE_STRING_LENGTH (bs1);
  if (length != BYTE_STRING_LENGTH (bs2))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  void *d1 = (void*)BYTE_STRING_DATA (bs1);
  void *d2 = (void*)BYTE_STRING_DATA (bs2);

  SCM answer = scm_c_make_byte_string (length);
  char *data = BYTE_STRING_DATA (answer);

  #define BOP |
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = *(scm_t_uint64*)d1 BOP *(scm_t_uint64*)d2; break;
  case 4:
    *(scm_t_uint32*)data = *(scm_t_uint32*)d1 BOP *(scm_t_uint32*)d2; break;
  case 2:
    *(scm_t_uint16*)data = *(scm_t_uint16*)d1 BOP *(scm_t_uint16*)d2; break;
  case 1:
    *(scm_t_uint8*)data  = *(scm_t_uint8*)d1  BOP *(scm_t_uint8*)d2;  break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  // clr2/a postambel
  dereference_parameter (bs1);
  dereference_parameter (bs2);
  return (answer);
}
#undef FUNC_NAME

// Copy&Paste
SCM_DEFINE1 (scm_bit_xor, "bit-xor", scm_tc7_asubr, (SCM bs1, SCM bs2),"")
#define FUNC_NAME s_scm_bit_xor
{
  // asubr auto folding identity
  if (SCM_UNBNDP (bs2)) return (scm_byte_string_copy (bs1));
  // clr2/a preambel
  reference_parameter (bs1);
  reference_parameter (bs2);
  // Argument Validation: type
  if (!BYTE_STRING_P (bs1))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, bs1, "byte-string");
  if (!BYTE_STRING_P (bs2))
    scm_wrong_type_arg_msg (FUNC_NAME, 2, bs2, "byte-string");

  // Argument Validation and Initialization: range
  size_t length = BYTE_STRING_LENGTH (bs1);
  if (length != BYTE_STRING_LENGTH (bs2))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  void *d1 = (void*)BYTE_STRING_DATA (bs1);
  void *d2 = (void*)BYTE_STRING_DATA (bs2);

  SCM answer = scm_c_make_byte_string (length);
  char *data = BYTE_STRING_DATA (answer);

  #define BOP ^
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = *(scm_t_uint64*)d1 BOP *(scm_t_uint64*)d2; break;
  case 4:
    *(scm_t_uint32*)data = *(scm_t_uint32*)d1 BOP *(scm_t_uint32*)d2; break;
  case 2:
    *(scm_t_uint16*)data = *(scm_t_uint16*)d1 BOP *(scm_t_uint16*)d2; break;
  case 1:
    *(scm_t_uint8*)data  = *(scm_t_uint8*)d1  BOP *(scm_t_uint8*)d2;  break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  // clr2/a postambel
  dereference_parameter (bs1);
  dereference_parameter (bs2);
  return (answer);
}
#undef FUNC_NAME

// Copy&Paste
SCM_DEFINE (scm_bit_modulo, "bit-modulo", 2, 0, 0, (SCM bs1, SCM bs2),"")
#define FUNC_NAME s_scm_bit_modulo
{
  // clr2/a preambel
  reference_parameter (bs1);
  reference_parameter (bs2);
  // Argument Validation: type
  if (!BYTE_STRING_P (bs1))
    scm_wrong_type_arg_msg (FUNC_NAME, 1, bs1, "byte-string");
  if (!BYTE_STRING_P (bs2))
    scm_wrong_type_arg_msg (FUNC_NAME, 2, bs2, "byte-string");

  // Argument Validation and Initialization: range
  size_t length = BYTE_STRING_LENGTH (bs1);
  if (length != BYTE_STRING_LENGTH (bs2))
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);

  void *d1 = (void*)BYTE_STRING_DATA (bs1);
  void *d2 = (void*)BYTE_STRING_DATA (bs2);

  SCM answer = scm_c_make_byte_string (length);
  char *data = BYTE_STRING_DATA (answer);

  #define BOP %
  switch (length) {
  case 8:
    *(scm_t_uint64*)data = *(scm_t_uint64*)d1 BOP *(scm_t_uint64*)d2; break;
  case 4:
    *(scm_t_uint32*)data = *(scm_t_uint32*)d1 BOP *(scm_t_uint32*)d2; break;
  case 2:
    *(scm_t_uint16*)data = *(scm_t_uint16*)d1 BOP *(scm_t_uint16*)d2; break;
  case 1:
    *(scm_t_uint8*)data  = *(scm_t_uint8*)d1  BOP *(scm_t_uint8*)d2;  break;
  default:
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  #undef BOP

  // clr2/a postambel
  dereference_parameter (bs1);
  dereference_parameter (bs2);
  return (answer);
}
#undef FUNC_NAME


/* Machine Dependent Type Sizes */

SCM_DEFINE (scm_sizeof_double, "size-of-double", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_double
{
  RC_RETURN (scm_from_uint (sizeof (double)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_sizeof_float, "size-of-float", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_double
{
  RC_RETURN (scm_from_uint (sizeof (float)));
}
#undef FUNC_NAME

/* SCM_DEFINE (scm_sizeof_vector, "size-of-vector", 0, 0, 0, (), "") */
/* #define FUNC_NAME s_scm_sizeof_vector */
/* { */
/*   RC_RETURN (scm_from_uint (sizeof (vector))); */
/* } */
/* #undef FUNC_NAME */

SCM_DEFINE (scm_sizeof_long_long, "size-of-long-long", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_long_long
{
  RC_RETURN (scm_from_uint (sizeof (long long)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_sizeof_long, "size-of-long", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_long
{
  RC_RETURN (scm_from_uint (sizeof (long)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_sizeof_int, "size-of-int", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_int
{
  RC_RETURN (scm_from_uint (sizeof (int)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_sizeof_short, "size-of-short", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_short
{
  RC_RETURN (scm_from_uint (sizeof (short)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_sizeof_pointer, "size-of-pointer", 0, 0, 0, (), "")
#define FUNC_NAME s_scm_sizeof_pointer
{
  RC_RETURN (scm_from_uint (sizeof (void *)));
}
#undef FUNC_NAME


/* SCM default interface */


inline void
trim_release_byte_string (SCM byte_string)
{ // byte string mappings are handled in trim.c
  global_number_of_byte_strings--;
  // Assert: sizeof SCM == 8
  // !!! See also /scm_c_make_byte_string/
  scm_free_array (((BYTE_STRING_LENGTH (byte_string)) >> 3)+2, byte_string);
  return;
}

void
x_inclusion_byte_strings ()
{
  #include "include/byte-strings.x"
}
