#include "include/config.h"

#include "include/_scm.h"
#include "include/eq.h"
#include "include/list.h"

#include "include/validate.h"
#include "include/pairs.h"
#include "include/alist.h"



/*
 * Association Lists
 */

SCM_DEFINE (scm_acons, "acons", 3, 0, 0,
           (SCM key, SCM value, SCM alist),
	    "Add a new key-value pair to @var{alist}.  A new pair is\n"
	    "created whose car is @var{key} and whose cdr is @var{value}, and the\n"
	    "pair is consed onto @var{alist}, and the new list is returned.  This\n"
	    "function is @emph{not} destructive; @var{alist} is not modified.")
#define FUNC_NAME s_scm_acons
{
  return scm_i_cons (scm_i_cons (key, value), alist);
}
#undef FUNC_NAME


SCM
scm_assq_unchecked_lax (SCM key, SCM ls)
{
  for (; scm_is_pair (ls); ls = TMP_CDR (ls)) {
    SCM tmp = TMP_CAR (ls);

    if (scm_is_pair (tmp) && scm_is_eq (TMP_CAR (tmp), key)) {
      return tmp;
    }
  }

  return SCM_BOOL_F;
}

/*
 * The sloppy set is not used by the core. XXX Purge Scheme level please.
 */
SCM_DEFINE (scm_sloppy_assq, "sloppy-assq", 2, 0, 0,
            (SCM key, SCM alist),
	    "Behaves like @code{assq} but does not do any error checking.\n"
	    "Recommended only for use in VSI internals.")
#define FUNC_NAME s_scm_sloppy_assq
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for (; scm_is_pair (ls); ls = TMP_CDR (ls))
    {
      SCM tmp = TMP_CAR (ls);
      if (scm_is_pair (tmp) && scm_is_eq (TMP_CAR (tmp), key)) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }
  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME



SCM_DEFINE (scm_sloppy_assv, "sloppy-assv", 2, 0, 0,
            (SCM key, SCM alist),
	    "Behaves like @code{assv} but does not do any error checking.\n"
	    "Recommended only for use in VSI internals.")
#define FUNC_NAME s_scm_sloppy_assv
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for (; scm_is_pair (ls); ls = TMP_CDR (ls))
    {
      SCM tmp = TMP_CAR (ls);
      if (scm_is_pair (tmp)
	  && scm_is_true (scm_eqv_p (TMP_CAR (tmp), key))) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_sloppy_assoc, "sloppy-assoc", 2, 0, 0,
            (SCM key, SCM alist),
	    "Behaves like @code{assoc} but does not do any error checking.\n"
	    "Recommended only for use in VSI internals.")
#define FUNC_NAME s_scm_sloppy_assoc
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for (; scm_is_pair (ls); ls = TMP_CDR (ls))
    {
      SCM tmp = TMP_CAR (ls);
      if (scm_is_pair (tmp)
	  && scm_is_true (scm_equal_p (TMP_CAR (tmp), key))) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME




SCM_DEFINE (scm_assq, "assq", 2, 0, 0,
           (SCM key, SCM alist),
	    "@deffnx {Scheme Procedure} assv key alist\n"
	    "@deffnx {Scheme Procedure} assoc key alist\n"
	    "Fetch the entry in @var{alist} that is associated with @var{key}.  To\n"
	    "decide whether the argument @var{key} matches a particular entry in\n"
	    "@var{alist}, @code{assq} compares keys with @code{eq?}, @code{assv}\n"
	    "uses @code{eqv?} and @code{assoc} uses @code{equal?}.  If @var{key}\n"
	    "cannot be found in @var{alist} (according to whichever equality\n"
	    "predicate is in use), then return @code{#f}.  These functions\n"
	    "return the entire alist entry found (i.e. both the key and the value).")
#define FUNC_NAME s_scm_assq
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for(; scm_is_pair (ls); ls = TMP_CDR (ls)) 
    {
      SCM tmp = TMP_CAR (ls);
      SCM_ASSERT_TYPE (scm_is_pair (tmp), alist, SCM_ARG2, FUNC_NAME,
		       "association list");
      if (scm_is_eq (TMP_CAR (tmp), key)) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }
  SCM_ASSERT_TYPE (scm_is_null (ls), alist, SCM_ARG2, FUNC_NAME,
		   "association list");
  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_assv, "assv", 2, 0, 0,
           (SCM key, SCM alist),
	    "Behaves like @code{assq} but uses @code{eqv?} for key comparison.")
#define FUNC_NAME s_scm_assv
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for(; scm_is_pair (ls); ls = TMP_CDR (ls)) 
    {
      SCM tmp = TMP_CAR (ls);
      SCM_ASSERT_TYPE (scm_is_pair (tmp), alist, SCM_ARG2, FUNC_NAME,
		       "association list");
      if (scm_is_true (scm_eqv_p (TMP_CAR (tmp), key))) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }
  SCM_ASSERT_TYPE (scm_is_null (ls), alist, SCM_ARG2, FUNC_NAME,
		   "association list");

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_assoc, "assoc", 2, 0, 0,
           (SCM key, SCM alist),
	    "Behaves like @code{assq} but uses @code{equal?} for key comparison.")
#define FUNC_NAME s_scm_assoc
{
  reference_parameter (key);
  reference_parameter (alist);
  
  SCM ls = alist;
  for(; scm_is_pair (ls); ls = TMP_CDR (ls)) 
    {
      SCM tmp = TMP_CAR (ls);
      SCM_ASSERT_TYPE (scm_is_pair (tmp), alist, SCM_ARG2, FUNC_NAME,
		       "association list");
      if (scm_is_true (scm_equal_p (TMP_CAR (tmp), key))) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
	return tmp;
      }
    }
  SCM_ASSERT_TYPE (scm_is_null (ls), alist, SCM_ARG2, FUNC_NAME,
		   "association list");
  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME




/* Dirk:API2.0:: We should not return #f if the key was not found.  In the
 * current solution we can not distinguish between finding a (key . #f) pair
 * and not finding the key at all.
 *
 * Possible alternative solutions:
 * 1) Remove assq-ref from the API:  assq is sufficient.
 * 2) Signal an error (what error type?) if the key is not found.
 * 3) provide an additional 'default' parameter.
 * 3.1) The default parameter is mandatory.
 * 3.2) The default parameter is optional, but if no default is given and
 *      the key is not found, signal an error (what error type?).
 */
SCM_DEFINE (scm_assq_ref, "assq-ref", 2, 0, 0,
            (SCM alist, SCM key),
	    "@deffnx {Scheme Procedure} assv-ref alist key\n"
	    "@deffnx {Scheme Procedure} assoc-ref alist key\n"
	    "Like @code{assq}, @code{assv} and @code{assoc}, except that only the\n"
	    "value associated with @var{key} in @var{alist} is returned.  These\n"
	    "functions are equivalent to\n\n"
	    "@lisp\n"
	    "(let ((ent (@var{associator} @var{key} @var{alist})))\n"
	    "  (and ent (cdr ent)))\n"
	    "@end lisp\n\n"
	    "where @var{associator} is one of @code{assq}, @code{assv} or @code{assoc}.")
#define FUNC_NAME s_scm_assq_ref
{
  reference_parameter (key);
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assq (key, alist);
  if (scm_is_pair (handle)) {
    SCM answer = TMP_CDR (handle);
    reference_count_return (answer);
    dereference_return_value (handle);
    dereference_parameter (key);
    dereference_parameter (alist);
    return answer;
  }

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_assv_ref, "assv-ref", 2, 0, 0,
            (SCM alist, SCM key),
	    "Behaves like @code{assq-ref} but uses @code{eqv?} for key comparison.")
#define FUNC_NAME s_scm_assv_ref
{
  reference_parameter (key);
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assv (key, alist);
  if (scm_is_pair (handle)) {
    SCM answer = TMP_CDR (handle);
    reference_count_return (answer);
    dereference_return_value (handle);
    dereference_parameter (key);
    dereference_parameter (alist);
    return answer;
  }

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_assoc_ref, "assoc-ref", 2, 0, 0,
            (SCM alist, SCM key),
	    "Behaves like @code{assq-ref} but uses @code{equal?} for key comparison.")
#define FUNC_NAME s_scm_assoc_ref
{
  reference_parameter (key);
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assoc (key, alist);
  if (scm_is_pair (handle)) {
    SCM answer = TMP_CDR (handle);
    reference_count_return (answer);
    dereference_return_value (handle);
    dereference_parameter (key);
    dereference_parameter (alist);
    return answer;
  }

  dereference_parameter (key);
  dereference_parameter (alist);
  return SCM_BOOL_F;
}
#undef FUNC_NAME






SCM_DEFINE (scm_assq_set_x, "assq-set!", 3, 0, 0,
            (SCM alist, SCM key, SCM val),
	    "@deffnx {Scheme Procedure} assv-set! alist key value\n"
	    "@deffnx {Scheme Procedure} assoc-set! alist key value\n"
	    "Reassociate @var{key} in @var{alist} with @var{value}: find any existing\n"
	    "@var{alist} entry for @var{key} and associate it with the new\n"
	    "@var{value}.  If @var{alist} does not contain an entry for @var{key},\n"
	    "add a new one.  Return the (possibly new) alist.\n\n"
	    "These functions do not attempt to verify the structure of @var{alist},\n"
	    "and so may cause unusual results if passed an object that is not an\n"
	    "association list.")
#define FUNC_NAME s_scm_assq_set_x
{
  reference_parameter (alist);
  reference_parameter (key);
  reference_parameter (val);

  SCM handle;

  handle = scm_sloppy_assq (key, alist);
  if (scm_is_pair (handle))
    {
      SCM_SETCDR (handle, val);
      reference_count_return (alist);
      dereference_parameter (alist);
      dereference_parameter (key);
      dereference_parameter (val);
      return alist;
    }
  else
    return scm_acons (rc_Fly (key), rc_Fly (val), rc_Fly (alist));
}
#undef FUNC_NAME

SCM_DEFINE (scm_assv_set_x, "assv-set!", 3, 0, 0,
            (SCM alist, SCM key, SCM val),
	    "Behaves like @code{assq-set!} but uses @code{eqv?} for key comparison.")
#define FUNC_NAME s_scm_assv_set_x
{
  reference_parameter (alist);
  reference_parameter (key);
  reference_parameter (val);

  SCM handle;

  handle = scm_sloppy_assv (key, alist);
  if (scm_is_pair (handle))
    {
      SCM_SETCDR (handle, val);
      reference_count_return (alist);
      dereference_parameter (alist);
      dereference_parameter (key);
      dereference_parameter (val);
      return alist;
    }
  else
    return scm_acons (rc_Fly (key), rc_Fly (val), rc_Fly (alist));
}
#undef FUNC_NAME

SCM_DEFINE (scm_assoc_set_x, "assoc-set!", 3, 0, 0,
            (SCM alist, SCM key, SCM val),
	    "Behaves like @code{assq-set!} but uses @code{equal?} for key comparison.")
#define FUNC_NAME s_scm_assoc_set_x
{
  reference_parameter (alist);
  reference_parameter (key);
  reference_parameter (val);

  SCM handle;

  handle = scm_sloppy_assoc (key, alist);
  if (scm_is_pair (handle))
    {
      SCM_SETCDR (handle, val);
      reference_count_return (alist);
      dereference_parameter (alist);
      dereference_parameter (key);
      dereference_parameter (val);
      return alist;
    }
  else
    return scm_acons (rc_Fly (key), rc_Fly (val), rc_Fly (alist));
}
#undef FUNC_NAME




SCM_DEFINE (scm_assq_remove_x, "assq-remove!", 2, 0, 0,
            (SCM alist, SCM key),
	    "@deffnx {Scheme Procedure} assv-remove! alist key\n"
	    "@deffnx {Scheme Procedure} assoc-remove! alist key\n"
	    "Delete the first entry in @var{alist} associated with @var{key}, and return\n"
	    "the resulting alist.")
#define FUNC_NAME s_scm_assq_remove_x
{
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assq (key, alist);
  if (scm_is_pair (handle))
    alist = scm_delq1_x (handle, alist);

  return rc_return_stack_reference (alist);
}
#undef FUNC_NAME


SCM_DEFINE (scm_assv_remove_x, "assv-remove!", 2, 0, 0,
            (SCM alist, SCM key),
	    "Behaves like @code{assq-remove!} but uses @code{eqv?} for key comparison.")
#define FUNC_NAME s_scm_assv_remove_x
{
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assv (key, alist);
  if (scm_is_pair (handle))
    alist = scm_delq1_x (handle, alist);

  return rc_return_stack_reference (alist);
}
#undef FUNC_NAME


SCM_DEFINE (scm_assoc_remove_x, "assoc-remove!", 2, 0, 0,
            (SCM alist, SCM key),
	    "Behaves like @code{assq-remove!} but uses @code{equal?} for key comparison.")
#define FUNC_NAME s_scm_assoc_remove_x
{
  reference_parameter (alist);

  SCM handle;

  handle = scm_sloppy_assoc (key, alist);
  if (scm_is_pair (handle))
    alist = scm_delq1_x (handle, alist);

  return rc_return_stack_reference (alist);
}
#undef FUNC_NAME






void
scm_init_alist ()
{
#include "include/alist.x"
}
