// Strings & Brass (old io path deprecation)

/* Headers.  */

#define _LARGEFILE64_SOURCE      /* ask for stat64 etc */

#include "include/config.h"

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>  /* for chsize on mingw */

#include <assert.h>

#include "include/_scm.h"
#include "include/eval.h"
#include "include/fports.h"  /* direct access for seek and truncate */
#include "include/smob.h"
#include "include/chars.h"
#include "include/dynwind.h"

#include "include/root.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/validate.h"
#include "include/ports.h"
#include "include/vectors.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_IO_H
#include <io.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

/* Mingw (version 3.4.5, circa 2006) has ftruncate as an alias for chsize
   already, but have this code here in case that wasn't so in past versions,
   or perhaps to help other minimal DOS environments.

   gnulib ftruncate.c has code using fcntl F_CHSIZE and F_FREESP, which
   might be possibilities if we've got other systems without ftruncate.  */

#if HAVE_CHSIZE && ! HAVE_FTRUNCATE
#define ftruncate(fd, size) chsize (fd, size)
#undef HAVE_FTRUNCATE
#define HAVE_FTRUNCATE 1
#endif


/*
 * Port Statistics
 */
long global_number_of_ports = 0;

/* Standard ports --- current input, output, error, and more(!).
 * Stored in gc permanent boxes which were pairs protected
 * from former garbage collection:
 *   the /car/ is /true/
 *   the /cdr/ contains the object
 * 
 * Use /scm_cdr/ and /scm_set_cdr_x/!
 *
 * See also the init routine at the end of this file.
 */

static SCM cur_inport_box = 0;
static SCM cur_outport_box = 0;
static SCM cur_errport_box = 0;
static SCM cur_loadport_box = 0;

SCM_DEFINE (scm_current_input_port, "current-input-port", 0, 0, 0,
	    (),
	    "Return the current input port.  This is the default port used\n"
	    "by many input procedures.  Initially, @code{current-input-port}\n"
	    "returns the @dfn{standard input} in Unix and C terminology.")
#define FUNC_NAME s_scm_current_input_port
{
  return scm_cdr (cur_inport_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_current_output_port, "current-output-port", 0, 0, 0,
	    (),
            "Return the current output port.  This is the default port used\n"
	    "by many output procedures.  Initially,\n"
	    "@code{current-output-port} returns the @dfn{standard output} in\n"
	    "Unix and C terminology.")
#define FUNC_NAME s_scm_current_output_port
{
  return scm_cdr (cur_outport_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_current_error_port, "current-error-port", 0, 0, 0,
           (),
	    "Return the port to which errors and warnings should be sent (the\n"
	    "@dfn{standard error} in Unix and C terminology).")
#define FUNC_NAME s_scm_current_error_port
{
  if (cur_errport_box == 0) {
    fprintf (stderr, "VSI-C: core-ports warning: current-error-port requested before initialization\n");
    return 0;
  }
  return scm_cdr (cur_errport_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_current_load_port, "current-load-port", 0, 0, 0,
	    (),
	    "Return the current-load-port.\n"
            "The load port is used internally by @code{primitive-load}.")
#define FUNC_NAME s_scm_current_load_port
{
  return scm_cdr (cur_loadport_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_current_input_port, "set-current-input-port", 1, 0, 0,
           (SCM port),
	    "@deffnx {Scheme Procedure} set-current-output-port port\n"
	    "@deffnx {Scheme Procedure} set-current-error-port port\n"
	    "Change the ports returned by @code{current-input-port},\n"
	    "@code{current-output-port} and @code{current-error-port}, respectively,\n"
	    "so that they use the supplied @var{port} for input or output.")
#define FUNC_NAME s_scm_set_current_input_port
{
  reference_parameter (port);
  SCM oinp = scm_cdr (cur_inport_box);
  reference_parameter (oinp);

  SCM_VALIDATE_OPINPORT (1, port);

  scm_set_cdr_x (cur_inport_box, port);

  dereference_parameter (port);
  return (rc_return_stack_reference (oinp));
}
#undef FUNC_NAME


SCM_DEFINE (scm_set_current_output_port, "set-current-output-port", 1, 0, 0,
	    (SCM port),
	    "Set the current default output port to @var{port}.")
#define FUNC_NAME s_scm_set_current_output_port
{
  reference_parameter (port);
  SCM ooutp = scm_cdr (cur_outport_box);
  reference_parameter (ooutp);

  SCM_VALIDATE_OPOUTPORT (1, port);

  scm_set_cdr_x (cur_outport_box, port);

  dereference_parameter (port);
  return (rc_return_stack_reference (ooutp));
}
#undef FUNC_NAME


SCM_DEFINE (scm_set_current_error_port, "set-current-error-port", 1, 0, 0,
	    (SCM port),
	    "Set the current default error port to @var{port}.")
#define FUNC_NAME s_scm_set_current_error_port
{
  reference_parameter (port);
  SCM oerrp = scm_cdr (cur_errport_box);
  reference_parameter (oerrp);

  SCM_VALIDATE_OPOUTPORT (1, port);

  scm_set_cdr_x (cur_errport_box, port);

  dereference_parameter (port);
  return (rc_return_stack_reference (oerrp));
}
#undef FUNC_NAME



/* The port kind table --- a dynamically resized array of port types.  */


/* scm_ptobs scm_numptob
 * implement a dynamicly resized array of ptob records.
 * Indexes into this table are used when generating type
 * tags for smobjects (if you know a tag you can get an index and conversely).
 */
scm_t_ptob_descriptor *scm_ptobs;
long scm_numptob;

/*
 * We choose to use an interface similar to the smob interface with
 * fill_input and write as standard fields, passed to the port
 * type constructor, and optional fields set by setters.
 */

static void
flush_port_default (SCM port SCM_UNUSED)
{
}

static void
end_input_default (SCM port SCM_UNUSED, int offset SCM_UNUSED)
{
}

// XXX this is now done by remove_from_port_table
/* void */
/* trim_release_port (SCM port) */
/* { */
/*   void (*release) (SCM) = scm_ptobs[SCM_PTOBNUM (port)].release; */
/*   if (release) release (port); */
/*   return; */
/* } */

scm_t_tag
scm_make_port_type (char *name,
		    int (*fill_input) (SCM port),
		    void (*write) (SCM port, const void *data, size_t size))
{
  char *tmp;
  if (255 <= scm_numptob)
    goto ptoberr;
  SCM_SYSCALL (tmp = (char *) realloc ((char *) scm_ptobs,
				       (1 + scm_numptob)
				       * sizeof (scm_t_ptob_descriptor)));
  if (tmp)
    {
      scm_ptobs = (scm_t_ptob_descriptor *) tmp;

      scm_ptobs[scm_numptob].name = name;
      scm_ptobs[scm_numptob].release = NULL;
      scm_ptobs[scm_numptob].print = scm_port_print;
      scm_ptobs[scm_numptob].equalp = 0;
      scm_ptobs[scm_numptob].close = 0;

      scm_ptobs[scm_numptob].write = write;
      scm_ptobs[scm_numptob].flush = flush_port_default;

      scm_ptobs[scm_numptob].end_input = end_input_default;
      scm_ptobs[scm_numptob].fill_input = fill_input;
      scm_ptobs[scm_numptob].input_waiting = 0;

      scm_ptobs[scm_numptob].seek = 0;
      scm_ptobs[scm_numptob].truncate = 0;

      scm_numptob++;
    }
  if (!tmp)
    {
    ptoberr:
      scm_memory_error ("scm_make_port_type");
    }

  return scm_tc7_port + (scm_numptob - 1) * 256;
}

void
scm_set_port_release (scm_t_tag tc, void (*release) (SCM))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].release = release;
}

void
scm_set_port_print (scm_t_tag tc, int (*print) (SCM exp, SCM port,
					   scm_print_state *pstate))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].print = print;
}

void
scm_set_port_equalp (scm_t_tag tc, SCM (*equalp) (SCM, SCM))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].equalp = equalp;
}

void
scm_set_port_flush (scm_t_tag tc, void (*flush) (SCM port))
{
   scm_ptobs[SCM_TC2PTOBNUM (tc)].flush = flush;
}

void
scm_set_port_end_input (scm_t_tag tc, void (*end_input) (SCM port, int offset))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].end_input = end_input;
}

void
scm_set_port_close (scm_t_tag tc, int (*close) (SCM))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].close = close;
}

void
scm_set_port_seek (scm_t_tag tc, off_t (*seek) (SCM port,
					   off_t OFFSET,
					   int WHENCE))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].seek = seek;
}

void
scm_set_port_truncate (scm_t_tag tc, void (*truncate) (SCM port, off_t length))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].truncate = truncate;
}

void
scm_set_port_input_waiting (scm_t_tag tc, int (*input_waiting) (SCM))
{
  scm_ptobs[SCM_TC2PTOBNUM (tc)].input_waiting = input_waiting;
}



// XXX scm_char_ready_p is in fact a scm_byte_ready_p but we consider this state
// to be consistent for encoded input even if that might not always be true?
SCM_DEFINE (scm_char_ready_p, "char-ready?", 0, 1, 0, 
	    (SCM port),
	    "Return @code{#t} if a character is ready on input @var{port}\n"
	    "and return @code{#f} otherwise.  If @code{char-ready?} returns\n"
	    "@code{#t} then the next @code{read-char} operation on\n"
	    "@var{port} is guaranteed not to hang.  If @var{port} is a file\n"
	    "port at end of file then @code{char-ready?} returns @code{#t}.\n"
	    "\n"
	    "@code{char-ready?} exists to make it possible for a\n"
	    "program to accept characters from interactive ports without\n"
	    "getting stuck waiting for input.  Any input editors associated\n"
	    "with such ports must make sure that characters whose existence\n"
	    "has been asserted by @code{char-ready?} cannot be rubbed out.\n"
	    "If @code{char-ready?} were to return @code{#f} at end of file,\n"
	    "a port at end of file would be indistinguishable from an\n"
	    "interactive port that has no ready characters.")
#define FUNC_NAME s_scm_char_ready_p
{
  reference_parameter (port);

  scm_t_port *pt;

  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
    reference_parameter (port);
  } else {
    SCM_VALIDATE_OPINPORT (1, port);
  }

  pt = SCM_PTAB_ENTRY (port);

  /* if the current read buffer is filled, or the
     last pushed-back char has been read and the saved buffer is
     filled, result is true.  */
  if (pt->read_pos < pt->read_end 
      || (pt->read_buf == pt->putback_buf
	  && pt->saved_read_pos < pt->saved_read_end)) {
    dereference_parameter (port);
    return SCM_BOOL_T;
  } else {
    scm_t_ptob_descriptor *ptob = &scm_ptobs[SCM_PTOBNUM (port)];
    SCM answer;  
    if (ptob->input_waiting)
      answer = scm_from_bool(ptob->input_waiting (port));
    else
      answer = SCM_BOOL_T;
    dereference_parameter (port);
    return (answer);
  }
}
#undef FUNC_NAME

/* move up to read_len chars from port's putback and/or read buffers
   into memory starting at dest.  returns the number of chars moved.  */
size_t scm_take_from_input_buffers (SCM port, char *dest, size_t read_len)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  size_t chars_read = 0;
  size_t from_buf = min (pt->read_end - pt->read_pos, read_len);

  if (from_buf > 0)
    {
      memcpy (dest, pt->read_pos, from_buf);
      pt->read_pos += from_buf;
      chars_read += from_buf;
      read_len -= from_buf;
      dest += from_buf;
    }

  /* if putback was active, try the real input buffer too.  */
  if (pt->read_buf == pt->putback_buf)
    {
      from_buf = min (pt->saved_read_end - pt->saved_read_pos, read_len);
      if (from_buf > 0)
	{
	  memcpy (dest, pt->saved_read_pos, from_buf);
	  pt->saved_read_pos += from_buf;
	  chars_read += from_buf;
	}
    }
  return chars_read;
}

// crap
/* /\* Clear a port's read buffers, returning the contents.  *\/ */
/* SCM_DEFINE (scm_drain_input, "drain-input", 1, 0, 0, (SCM port), "") */
/* #define FUNC_NAME s_scm_drain_input */
/* { */
/*   reference_parameter (port); */

/*   SCM result; */
/*   char *data; */
/*   scm_t_port *pt; */
/*   long count; */

/*   SCM_VALIDATE_OPINPORT (1, port); */
/*   pt = SCM_PTAB_ENTRY (port); */

/*   count = pt->read_end - pt->read_pos; */
/*   if (pt->read_buf == pt->putback_buf) */
/*     count += pt->saved_read_end - pt->saved_read_pos; */

/*   result = scm_i_make_string (count, &data); */
/*   reference_parameter (result); */
/*   scm_take_from_input_buffers (port, data, count); // XXX LOL just an input buffer flush */

/*   dereference_parameter (port); */
/*   return (rc_return_stack_reference (result)); */
/* } */
/* #undef FUNC_NAME */


/* The port table --- an array of pointers to ports.  */

void
scm_i_set_filename (SCM port, SCM name)
{
  scm_t_port *entry = SCM_PTAB_ENTRY (port);
  heap_reference   (entry, name);
  heap_dereference (entry, entry->file_name);
  entry->file_name = name;
  return;
}

scm_t_port **scm_i_port_table = NULL;

long scm_i_port_table_size = 0;	/* Number of ports in scm_i_port_table.  */
long scm_i_port_table_room = 20;	/* Size of the array.  */


SCM
scm_new_port_table_entry (scm_t_tag tag)
#define FUNC_NAME "scm_new_port_table_entry"
{
  SCM z = scm_cell ((scm_t_tag) tag, SCM_UNPACK (SCM_EOL), SCM_UNPACK (SCM_EOL));
  scm_t_port *entry = (scm_t_port *) scm_gc_calloc (sizeof (scm_t_port), "port");
  if (scm_i_port_table_size == scm_i_port_table_room)
    {
      void *newt = scm_realloc ((char *) scm_i_port_table,
				(size_t) (sizeof (scm_t_port *)
					  * scm_i_port_table_room * 2));
      scm_i_port_table = (scm_t_port **) newt;
      scm_i_port_table_room *= 2;
    }

  entry->entry = scm_i_port_table_size;

  entry->file_name = SCM_BOOL_F;
  entry->rw_active = SCM_PORT_NEITHER;

  scm_i_port_table[scm_i_port_table_size] = entry;
  scm_i_port_table_size++;

  entry->port = z;
  heap_reference (scm_i_port_table, z);
  // SCM_SET_CELL_TYPE(z, tag);
  SCM_SETPTAB_ENTRY(z, entry);
  
  global_number_of_ports++;
  return z;
}
#undef FUNC_NAME


/* Remove a port from the table and destroy it.  */

void
scm_remove_from_port_table (SCM port)
#define FUNC_NAME "scm_remove_from_port_table"
{
  reference_parameter (port);

  scm_t_port *p = SCM_PTAB_ENTRY (port);
  long i = p->entry;
  void (*release) (SCM) = scm_ptobs[SCM_PTOBNUM (port)].release;
  SCM_SET_FILENAME (port, SCM_BOOL_F);

  if (i >= scm_i_port_table_size)
    SCM_MISC_ERROR ("Port not in table: ~S", scm_list_1 (port));

  if (release)
    release (port);

  if (p->putback_buf)
    scm_gc_free (p->putback_buf, p->putback_buf_size, "putback buffer");
  scm_gc_free (p, sizeof (scm_t_port), "port");
  /* Since we have just freed slot i we can shrink the table by moving
     the last entry to that slot... */
  if (i < scm_i_port_table_size - 1)
    {
      scm_i_port_table[i] = scm_i_port_table[scm_i_port_table_size - 1];
      scm_i_port_table[i]->entry = i;
    }
  SCM_SETPTAB_ENTRY (port, 0);
  scm_i_port_table_size--;

  heap_dereference (scm_i_port_table, port);
  dereference_parameter (port);

  return;
}
#undef FUNC_NAME


#ifdef GUILE_DEBUG
/* Functions for debugging.  */

SCM_DEFINE (scm_pt_size, "pt-size", 0, 0, 0,
            (),
	    "Return the number of ports in the port table.  @code{pt-size}\n"
	    "is only included in @code{--enable-guile-debug} builds.")
#define FUNC_NAME s_scm_pt_size
{
  return scm_from_int (scm_i_port_table_size);
}
#undef FUNC_NAME

SCM_DEFINE (scm_pt_member, "pt-member", 1, 0, 0,
            (SCM index),
	    "Return the port at @var{index} in the port table.\n"
	    "@code{pt-member} is only included in\n"
	    "@code{--enable-guile-debug} builds.")
#define FUNC_NAME s_scm_pt_member
{
  size_t i = scm_to_size_t (index);
  if (i >= scm_i_port_table_size)
    return SCM_BOOL_F;
  else
    return scm_i_port_table[i]->port;
}
#undef FUNC_NAME
#endif

void
scm_port_non_buffer (scm_t_port *pt)
{
  pt->read_pos = pt->read_buf = pt->read_end = &pt->shortbuf;
  pt->write_buf = pt->write_pos = &pt->shortbuf;
  pt->read_buf_size = pt->write_buf_size = 1;
  pt->write_end = pt->write_buf + pt->write_buf_size;
}



/* Retrieving a port's mode.  */

/* Return the flags that characterize a port based on the mode
 * string used to open a file for that port.
 *
 * See PORT FLAGS in scm.h
 */

static long
scm_i_mode_bits_n (const char *modes, size_t n)
{
  return (SCM_OPN
	  | (memchr (modes, 'r', n) || memchr (modes, '+', n) ? SCM_RDNG : 0)
	  | (   memchr (modes, 'w', n)
	     || memchr (modes, 'a', n)
	     || memchr (modes, '+', n) ? SCM_WRTNG : 0)
	  | (memchr (modes, '0', n) ? SCM_BUF0 : 0)
	  | (memchr (modes, 'l', n) ? SCM_BUFLINE : 0));
}

long
scm_mode_bits (char *modes)
{
  return scm_i_mode_bits_n (modes, strlen (modes));
}

long
scm_i_mode_bits (SCM modes)
{
  reference_parameter (modes);
  long bits;

  if (!SCM_BRASSP (modes))
    scm_wrong_type_arg_msg (NULL, 0, modes, "brass");

  SCM bodes = scm_to_sen (modes);
  reference_parameter (bodes);
  bits = scm_i_mode_bits_n (BYTE_STRING_DATA (bodes),
			    BYTE_STRING_LENGTH (bodes));
  dereference_parameter (bodes);
  dereference_parameter (modes);
  return bits;
}

/* Return the mode flags from an open port.
 * Some modes such as "append" are only used when opening
 * a file and are not returned here.  */

SCM_DEFINE (scm_port_mode, "port-mode", 1, 0, 0,
           (SCM port),
	    "Return the port modes associated with the open port @var{port}.\n"
	    "These will not necessarily be identical to the modes used when\n"
	    "the port was opened, since modes such as \"append\" which are\n"
	    "used only during port creation are not retained.")
#define FUNC_NAME s_scm_port_mode
{
  reference_parameter (port);

  char modes[4];
  modes[0] = '\0';

  SCM_VALIDATE_OPPORT (1, port);
  if (SCM_CELL_WORD_0 (port) & SCM_RDNG) {
    if (SCM_CELL_WORD_0 (port) & SCM_WRTNG)
      strcpy (modes, "r+");
    else
      strcpy (modes, "r");
  }
  else if (SCM_CELL_WORD_0 (port) & SCM_WRTNG)
    strcpy (modes, "w");
  if (SCM_CELL_WORD_0 (port) & SCM_BUF0)
    strcat (modes, "0");

  dereference_parameter (port);
  return scm_from_sen (modes, strlen (modes));
}
#undef FUNC_NAME



/* Closing ports.  */

/* scm_close_port
 * Call the close operation on a port object. 
 * see also scm_close.
 */
SCM_DEFINE (scm_close_port, "close-port", 1, 0, 0,
           (SCM port),
	    "Close the specified port object.  Return @code{#t} if it\n"
	    "successfully closes a port or @code{#f} if it was already\n"
	    "closed.  An exception may be raised if an error occurs, for\n"
	    "example when flushing buffered output.  See also @ref{Ports and\n"
	    "File Descriptors, close}, for a procedure which can close file\n"
	    "descriptors.")
#define FUNC_NAME s_scm_close_port
{
  reference_parameter (port);

  size_t i;
  int rv;

  SCM_VALIDATE_PORT (1, port);
  if (SCM_CLOSEDP (port)) {
    dereference_parameter (port);
    return SCM_BOOL_F;
  }
  i = SCM_PTOBNUM (port);
  if (scm_ptobs[i].close)
    rv = (scm_ptobs[i].close) (port);
  else
    rv = 0;
  scm_remove_from_port_table (port);
  SCM_CLR_PORT_OPEN_FLAG (port);

  dereference_parameter (port);
  return scm_from_bool (rv >= 0);
}
#undef FUNC_NAME

SCM_DEFINE (scm_close_input_port, "close-input-port", 1, 0, 0,
           (SCM port),
	    "Close the specified input port object.  The routine has no effect if\n"
	    "the file has already been closed.  An exception may be raised if an\n"
	    "error occurs.  The value returned is unspecified.\n\n"
	    "See also @ref{Ports and File Descriptors, close}, for a procedure\n"
	    "which can close file descriptors.")
#define FUNC_NAME s_scm_close_input_port
{
  reference_parameter (port);
  SCM_VALIDATE_INPUT_PORT (1, port);
  scm_close_port (rc_Fly (port));
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_close_output_port, "close-output-port", 1, 0, 0,
           (SCM port),
	    "Close the specified output port object.  The routine has no effect if\n"
	    "the file has already been closed.  An exception may be raised if an\n"
	    "error occurs.  The value returned is unspecified.\n\n"
	    "See also @ref{Ports and File Descriptors, close}, for a procedure\n"
	    "which can close file descriptors.")
#define FUNC_NAME s_scm_close_output_port
{
  reference_parameter (port);

  SCM_VALIDATE_OUTPUT_PORT (1, port);
  scm_close_port (rc_Fly (port));

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (scm_sys_ports, "sys-ports", 0, 0, 0,
	    (),
	    "Returns the port table as a list.") 
#define FUNC_NAME s_scm_sys_ports
{
  SCM answer_box = scm_cons (SCM_BOOL_F, SCM_EOL);
  SCM last       = answer_box;
  reference_parameter (answer_box);
  reference_parameter (last);

  long i;
  for (i = 0; i < scm_i_port_table_size; i++) {
    SCM port = scm_i_port_table[i]->port;
    SCM_SETCDR (last, scm_cons (port, SCM_EOL));
    last = FLY_CDR (last);
  }
  
  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (last);
  return (answer_box);
}
#undef FUNC_NAME



SCM_DEFINE (scm_input_port_p, "input-port?", 1, 0, 0,
           (SCM x),
	    "Return @code{#t} if @var{x} is an input port, otherwise return\n"
	    "@code{#f}.  Any object satisfying this predicate also satisfies\n"
	    "@code{port?}.")
#define FUNC_NAME s_scm_input_port_p
{
  reference_parameter (x);
  SCM answer = scm_from_bool (SCM_INPUT_PORT_P (x));
  dereference_parameter (x);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_output_port_p, "output-port?", 1, 0, 0,
           (SCM x),
	    "Return @code{#t} if @var{x} is an output port, otherwise return\n"
	    "@code{#f}.  Any object satisfying this predicate also satisfies\n"
	    "@code{port?}.")
#define FUNC_NAME s_scm_output_port_p
{
  reference_parameter (x);
  SCM answer = scm_from_bool (SCM_OUTPUT_PORT_P (x));
  dereference_parameter (x);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_port_p, "port?", 1, 0, 0,
	    (SCM x),
	    "Return a boolean indicating whether @var{x} is a port.\n"
	    "Equivalent to @code{(or (input-port? @var{x}) (output-port?\n"
	    "@var{x}))}.")
#define FUNC_NAME s_scm_port_p
{
  reference_parameter (x);
  SCM answer = scm_from_bool (SCM_PORTP (x));
  dereference_parameter (x);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_port_closed_p, "port-closed?", 1, 0, 0,
           (SCM port),
	    "Return @code{#t} if @var{port} is closed or @code{#f} if it is\n"
	    "open.")
#define FUNC_NAME s_scm_port_closed_p
{
  reference_parameter (port);
  SCM_VALIDATE_PORT (1, port);
  SCM answer = scm_from_bool (!SCM_OPPORTP (port));
  dereference_parameter (port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_eof_object_p, "eof-object?", 1, 0, 0,
           (SCM x),
	    "Return @code{#t} if @var{x} is an end-of-file object; otherwise\n"
	    "return @code{#f}.")
#define FUNC_NAME s_scm_eof_object_p
{
  reference_parameter (x);
  SCM answer = scm_from_bool(SCM_EOF_OBJECT_P (x));
  dereference_parameter (x);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_force_output, "force-output", 0, 1, 0,
           (SCM port),
	    "Flush the specified output port, or the current output port if @var{port}\n"
	    "is omitted.  The current output buffer contents are passed to the\n"
	    "underlying port implementation (e.g., in the case of fports, the\n"
	    "data will be written to the file and the output buffer will be cleared.)\n"
	    "It has no effect on an unbuffered port.\n\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_force_output
{
  if (SCM_UNBNDP (port)) {
    port = scm_current_output_port ();
    reference_parameter (port);
  } else {
    reference_parameter (port);
    SCM_VALIDATE_OPOUTPORT (1, port);
  }
  scm_flush (port);
  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_flush_all_ports, "flush-all-ports", 0, 0, 0,
            (),
	    "Equivalent to calling @code{force-output} on\n"
	    "all open output ports.  The return value is unspecified.")
#define FUNC_NAME s_scm_flush_all_ports
{
  size_t i;

  for (i = 0; i < scm_i_port_table_size; i++)
    {
      if (SCM_OPOUTPORTP (scm_i_port_table[i]->port))
	scm_flush (scm_i_port_table[i]->port);
    }
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_read_char, "read-char", 0, 1, 0,
           (SCM port),
	    "Return the next character available from @var{port}, updating\n"
	    "@var{port} to point to the following character.  If no more\n"
	    "characters are available, the end-of-file object is returned.")
#define FUNC_NAME s_scm_read_char
{
  reference_parameter (port);
  SCM c;
  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
    reference_parameter (port);
  }
  SCM_VALIDATE_OPINPORT (1, port);

  c = scm_getc (port);

  dereference_parameter (port);
  return c;
}
#undef FUNC_NAME

#include <error.h>
SCM_DEFINE (scm_read_byte, "read-byte", 0, 1, 0, (SCM port), "")
#define FUNC_NAME s_scm_read_byte
{
  reference_parameter (port);
  int c;
  if (SCM_NUMBERP (port)) {
    char c[2];
    int answer;
    int fd = scm_to_int (port); 
  again:
    answer = read (fd, &c, 1);
    if (answer == 1) {
      dereference_parameter (port);
      return (scm_from_uchar (c[0]));

    } else if (answer == 0) {
      dereference_parameter (port);
      return SCM_EOF_VAL;

    } else if (answer == -1) {
      if (errno == EINTR)  goto again;
      if (global_debug_interpreter) {
        // defunct error (0, errno, "scm_read_byte: read error");
        fprintf (stderr, "scm_read_byte: read error | %s\n", strerror (errno));
      }
    }
    dereference_parameter (port);
    return (SCM_BOOL_F);

  } else {
    if (SCM_UNBNDP (port)) {
      port = scm_current_input_port ();
      reference_parameter (port);
    }
    SCM_VALIDATE_OPINPORT (1, port);
    c = scm_i_read_byte (port);

    dereference_parameter (port);
    if (EOF == c)
      return SCM_EOF_VAL;
    return scm_from_uchar (c);
  }
}
#undef FUNC_NAME

#include "include/byte-strings.h"
SCM_DEFINE (scm_read_bytes, "read-bytes",
            1, 1, 0, (SCM number_bytes, SCM port), "")
#define FUNC_NAME s_scm_read_bytes
{
  reference_parameter (port);
  reference_parameter (number_bytes);

  if (!SCM_NUMBERP (number_bytes)) {
    dereference_parameter (port);
    dereference_parameter (number_bytes);
    return SCM_BOOL_F;  // XXX CLR/3 error object
  }

  size_t length = scm_to_ulong (number_bytes);
  void *buffer = scm_malloc (length); // XXX use byte string instead of malloc
  if (!buffer) {
  error:
    dereference_parameter (port);
    dereference_parameter (number_bytes);
    return SCM_BOOL_F;  // XXX CLR/3 error object
  }

  if (SCM_NUMBERP (port)) {
    int fd = scm_to_int (port); 
    ssize_t answer;

    answer = read (fd, buffer, length);
    if (answer > 0 && answer < length) {
      length = answer;
      buffer = scm_realloc (buffer, length);
      if (!buffer) goto error; // XXX
    }

    if (answer == length) {
      dereference_parameter (port);
      dereference_parameter (number_bytes);
      SCM ret = scm_make_byte_string_with_copy (buffer, length);
      free (buffer);
      return (ret);

    } else if (answer == 0) {
      free (buffer);
      dereference_parameter (port);
      dereference_parameter (number_bytes);
      return SCM_EOF_VAL;

    } else if (answer == -1) {
      // if (errno == EINTR)  goto again;
      if (global_debug_interpreter) {
        // defunct error (0, errno, "scm_read_byte: read error");
        fprintf (stderr, "scm_read_bytes: read error | %s\n", strerror (errno));
      }
      free (buffer);
      goto error;
    }
    free (buffer);
    dereference_parameter (port);
    dereference_parameter (number_bytes);
    return (SCM_BOOL_F);

  } else {
    if (SCM_UNBNDP (port)) {
      port = scm_current_input_port ();
      reference_parameter (port);
    }
    SCM_VALIDATE_OPINPORT (2, port); // XXX we lose one malloc by throw here
    size_t answer = scm_c_read (port, buffer, length); // XXX probably throws

    if (answer != length) {
      free (buffer);
      dereference_parameter (port);
      dereference_parameter (number_bytes);
      return SCM_BOOL_F;  // XXX CLR3 error object
    }

    dereference_parameter (port);
    dereference_parameter (number_bytes);
    SCM ret = scm_make_byte_string_with_copy (buffer, length);
    free (buffer);
    return (ret);
  }
}
#undef FUNC_NAME

SCM_DEFINE (scm_write_bytes, "write-bytes",
            1, 1, 0, (SCM byte_string, SCM port), "")
#define FUNC_NAME s_scm_write_bytes
{
  reference_parameter (port);
  reference_parameter (byte_string);

  if (!BYTE_STRING_P (byte_string)) {
    dereference_parameter (port);
    dereference_parameter (byte_string);
    return SCM_BOOL_F;  // XXX CLR/3 error object
  }

  size_t length = BYTE_STRING_LENGTH (byte_string);
  void *buffer  = (void*)BYTE_STRING_DATA (byte_string);

  if (SCM_NUMBERP (port)) {
    int fd = scm_to_int (port); 
    ssize_t answer;

    answer = write (fd, buffer, length);

    if (answer >= 0) {
      dereference_parameter (port);
      dereference_parameter (byte_string);
      return (scm_from_long (answer));

    } else if (answer == -1) {
      // if (errno == EINTR)  goto again;
      if (global_debug_interpreter) {
        // defunct error (0, errno, "scm_write_bytes: read error");
        fprintf (stderr, "scm_write_bytes: write error | %s\n",
                 strerror (errno));
      }
    }
    dereference_parameter (port);
    dereference_parameter (byte_string);
    return (SCM_BOOL_F);

  } else { // port is port
    if (SCM_UNBNDP (port)) {
      port = scm_current_output_port ();
      reference_parameter (port);
    }
    SCM_VALIDATE_OPOUTPORT (2, port);
    scm_c_write (port, buffer, length);

    dereference_parameter (port);
    dereference_parameter (byte_string);
    return scm_from_ulong (length);
  }
}
#undef FUNC_NAME

/* this should only be called when the read buffer is empty.  it
   tries to refill the read buffer.  it returns the first char from
   the port, which is either EOF or *(pt->read_pos).  */
int
scm_fill_input (SCM port)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  assert (pt->read_pos == pt->read_end);

  if (pt->read_buf == pt->putback_buf)
    {
      /* finished reading put-back chars.  */
      pt->read_buf = pt->saved_read_buf;
      pt->read_pos = pt->saved_read_pos;
      pt->read_end = pt->saved_read_end;
      pt->read_buf_size = pt->saved_read_buf_size;
      if (pt->read_pos < pt->read_end)
	return *(pt->read_pos);
    }
  return scm_ptobs[SCM_PTOBNUM (port)].fill_input (port);
}


/* Inlined Port I/O. */

inline int
scm_i_read_byte (SCM port)
{
  int c;
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  if (pt->rw_active == SCM_PORT_WRITE)
    /* may be marginally faster than calling scm_flush.  */
    scm_ptobs[SCM_PTOBNUM (port)].flush (port); // ???

  if (pt->rw_random)
    pt->rw_active = SCM_PORT_READ;

  if (pt->read_pos >= pt->read_end)
    {
      if (scm_fill_input (port) == EOF)
	return EOF;
    }

  c = *(pt->read_pos++);

  return c;
}

SCM
scm_getc (SCM port)
{
  SCM c = utf8_read_char (port); // XXX port encoding

  if (SCM_EOF_OBJECT_P (c)) {
    return (c);
  }

  switch (SCM_CHAR (c))
    {
      case '\a':
        break;
      case '\b':
        SCM_DECCOL (port);
        break;
      case '\n':
        SCM_INCLINE (port);
        break;
      case '\r':
        SCM_ZEROCOL (port);
        break;
      case '\t':
        SCM_TABCOL (port);
        break;
      default:
        SCM_INCCOL (port);
        break;
    }

  return c;
}

inline void
scm_putc (char c, SCM port)
{
  SCM_ASSERT_TYPE (SCM_OPOUTPORTP (port), port, 0, NULL, "output port");
  scm_lfwrite (&c, 1, port);
}

inline void
scm_puts (const char *s, SCM port)
{
  SCM_ASSERT_TYPE (SCM_OPOUTPORTP (port), port, 0, NULL, "output port");
  scm_lfwrite (s, strlen (s), port);
}


/* scm_lfwrite
 *
 * This function differs from scm_c_write; it updates port line and
 * column. */

// Wow: printer wheels in the port area sighted! ;-)

void
scm_lfwrite_br (SCM br, SCM port)
{
  reference_parameter (port);
  reference_parameter (br);

  SCM bs = scm_to_pen (br);
  reference_parameter (bs);
  scm_lfwrite (BYTE_STRING_DATA (bs), BYTE_STRING_LENGTH (bs), port);

  dereference_parameter (port);
  dereference_parameter (bs);
  dereference_parameter (br);
  return;
}

void
scm_lfwrite_usv (usv_t *chars, size_t len, SCM port)
{
  reference_parameter (port);

  size_t idx = 0;
  SCM buf = scm_cons (SCM_BOOL_F, SCM_EOL);
  reference_parameter (buf);
  SCM last = buf;

  for (idx = 0; idx < len; idx++) { // XXX port encoding
    SCM_SETCDR (last, scm_i_cons (utf8_usv_to_bs (chars[idx]), SCM_EOL));
    last = TMP_CDR (last);
  }
  SCM tmp = scm_c_byte_string_adjoin (TMP_CDR (buf));
  reference_parameter (tmp);
  scm_lfwrite (BYTE_STRING_DATA (tmp), BYTE_STRING_LENGTH (tmp), port);

  dereference_parameter (tmp);
  dereference_parameter (buf);
  dereference_parameter (port);
  return;
}

void 
scm_lfwrite (const char *ptr, size_t size, SCM port)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  scm_t_ptob_descriptor *ptob = &scm_ptobs[SCM_PTOBNUM (port)];

  if (pt->rw_active == SCM_PORT_READ)
    scm_end_input (port);

  ptob->write (port, ptr, size);

  for (; size; ptr++, size--) {
    if (*ptr == '\a') {
    }
    else if (*ptr == '\b') {
      SCM_DECCOL(port);
    }
    else if (*ptr == '\n') {
      SCM_INCLINE(port);
    }
    else if (*ptr == '\r') {
      SCM_ZEROCOL(port);
    }
    else if (*ptr == '\t') {
      SCM_TABCOL(port);
    }
    else {
      SCM_INCCOL(port);
    }
  }

  if (pt->rw_random)
    pt->rw_active = SCM_PORT_WRITE;
}

/* scm_c_read
 *
 * Used by an application to read arbitrary number of bytes from an
 * SCM port.  Same semantics as libc read, except that scm_c_read only
 * returns less than SIZE bytes if at end-of-file.
 *
 * Warning: Doesn't update port line and column counts!  */

/* This structure, and the following swap_buffer function, are used
   for temporarily swapping a port's own read buffer, and the buffer
   that the caller of scm_c_read provides. */
struct port_and_swap_buffer
{
  scm_t_port *pt;
  unsigned char *buffer;
  size_t size;
};

static void
swap_buffer (void *data)
{
  struct port_and_swap_buffer *psb = (struct port_and_swap_buffer *) data;
  unsigned char *old_buf = psb->pt->read_buf;
  size_t old_size = psb->pt->read_buf_size;

  /* Make the port use (buffer, size) from the struct. */
  psb->pt->read_pos = psb->pt->read_buf = psb->pt->read_end = psb->buffer;
  psb->pt->read_buf_size = psb->size;

  /* Save the port's old (buffer, size) in the struct. */
  psb->buffer = old_buf;
  psb->size = old_size;
}

size_t
scm_c_read (SCM port, void *buffer, size_t size)
#define FUNC_NAME "scm_c_read"
{
  scm_t_port *pt;
  size_t n_read = 0, n_available;
  struct port_and_swap_buffer psb;

  SCM_VALIDATE_OPINPORT (1, port);

  pt = SCM_PTAB_ENTRY (port);
  if (pt->rw_active == SCM_PORT_WRITE)
    scm_ptobs[SCM_PTOBNUM (port)].flush (port);

  if (pt->rw_random)
    pt->rw_active = SCM_PORT_READ;

  /* Take bytes first from the port's read buffer. */
  if (pt->read_pos < pt->read_end)
    {
      n_available = min (size, pt->read_end - pt->read_pos);
      memcpy (buffer, pt->read_pos, n_available);
      buffer = (char *) buffer + n_available;
      pt->read_pos += n_available;
      n_read += n_available;
      size -= n_available;
    }

  /* Avoid the scm_dynwind_* costs if we now have enough data. */
  if (size == 0)
    return n_read;

  /* Now we will call scm_fill_input repeatedly until we have read the
     requested number of bytes.  (Note that a single scm_fill_input
     call does not guarantee to fill the whole of the port's read
     buffer.) */
  if (pt->read_buf_size <= 1)
    {
      /* The port that we are reading from is unbuffered - i.e. does
	 not have its own persistent buffer - but we have a buffer,
	 provided by our caller, that is the right size for the data
	 that is wanted.  For the following scm_fill_input calls,
	 therefore, we use the buffer in hand as the port's read
	 buffer.

	 We need to make sure that the port's normal (1 byte) buffer
	 is reinstated in case one of the scm_fill_input () calls
	 throws an exception; we use the scm_dynwind_* API to achieve
	 that. */
      psb.pt = pt;
      psb.buffer = buffer;
      psb.size = size;

      SCM dw_token = scm_dynwind_begin (SCM_F_DYNWIND_REWINDABLE);
      scm_dynwind_rewind_handler (swap_buffer, &psb, SCM_F_WIND_EXPLICITLY);
      scm_dynwind_unwind_handler (swap_buffer, &psb, SCM_F_WIND_EXPLICITLY);

      /* Call scm_fill_input until we have all the bytes that we need,
	 or we hit EOF. */
      while (pt->read_buf_size && (scm_fill_input (port) != EOF))
	{
	  pt->read_buf_size -= (pt->read_end - pt->read_pos);
	  pt->read_pos = pt->read_buf = pt->read_end;
	}
      n_read += pt->read_buf - (unsigned char *) buffer;

      /* Reinstate the port's normal buffer. */
      scm_dynwind_end (dw_token);
    }
  else
    {
      /* The port has its own buffer.  It is important that we use it,
	 even if it happens to be smaller than our caller's buffer, so
	 that a custom port implementation's entry points (in
	 particular, fill_input) can rely on the buffer always being
	 the same as they first set up. */
      while (size && (scm_fill_input (port) != EOF))
	{
	  n_available = min (size, pt->read_end - pt->read_pos);
	  memcpy (buffer, pt->read_pos, n_available);
	  buffer = (char *) buffer + n_available;
	  pt->read_pos += n_available;
	  n_read += n_available;
	  size -= n_available;
	} 
    }

  return n_read;
}
#undef FUNC_NAME

/* scm_c_write
 *
 * Used by an application to write arbitrary number of bytes to an SCM
 * port.  Similar semantics as libc write.  However, unlike libc
 * write, scm_c_write writes the requested number of bytes and has no
 * return value.
 *
 * Warning: Doesn't update port line and column counts!
 */

void
scm_c_write (SCM port, const void *ptr, size_t size)
#define FUNC_NAME "scm_c_write"
{
  scm_t_port *pt;
  scm_t_ptob_descriptor *ptob;

  SCM_VALIDATE_OPOUTPORT (1, port);

  pt = SCM_PTAB_ENTRY (port);
  ptob = &scm_ptobs[SCM_PTOBNUM (port)];

  if (pt->rw_active == SCM_PORT_READ)
    scm_end_input (port);

  ptob->write (port, ptr, size);

  if (pt->rw_random)
    pt->rw_active = SCM_PORT_WRITE;
}
#undef FUNC_NAME

void 
scm_flush (SCM port)
{
  long i = SCM_PTOBNUM (port);
  (scm_ptobs[i].flush) (port);
}

void
scm_end_input (SCM port)
{
  long offset;
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  if (pt->read_buf == pt->putback_buf)
    {
      offset = pt->read_end - pt->read_pos;
      pt->read_buf = pt->saved_read_buf;
      pt->read_pos = pt->saved_read_pos;
      pt->read_end = pt->saved_read_end;
      pt->read_buf_size = pt->saved_read_buf_size;
    }
  else
    offset = 0;

  scm_ptobs[SCM_PTOBNUM (port)].end_input (port, offset);
}



void 
scm_unget_byte (int c, SCM port)
#define FUNC_NAME "scm_ungetc"
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  if (pt->read_buf == pt->putback_buf)
    /* already using the put-back buffer.  */
    {
      /* enlarge putback_buf if necessary.  */
      if (pt->read_end == pt->read_buf + pt->read_buf_size
	  && pt->read_buf == pt->read_pos)
	{
	  size_t new_size = pt->read_buf_size * 2;
	  unsigned char *tmp = (unsigned char *)
	    scm_gc_realloc (pt->putback_buf, pt->read_buf_size, new_size,
			    "putback buffer");

	  pt->read_pos = pt->read_buf = pt->putback_buf = tmp;
	  pt->read_end = pt->read_buf + pt->read_buf_size;
	  pt->read_buf_size = pt->putback_buf_size = new_size;
	}

      /* shift any existing bytes to buffer + 1.  */
      if (pt->read_pos == pt->read_end)
	pt->read_end = pt->read_buf + 1;
      else if (pt->read_pos != pt->read_buf + 1)
	{
	  int count = pt->read_end - pt->read_pos;

	  memmove (pt->read_buf + 1, pt->read_pos, count);
	  pt->read_end = pt->read_buf + 1 + count;
	}

      pt->read_pos = pt->read_buf;
    }
  else
    /* switch to the put-back buffer.  */
    {
      if (pt->putback_buf == NULL)
	{
	  pt->putback_buf
	    = (unsigned char *) scm_gc_malloc (SCM_INITIAL_PUTBACK_BUF_SIZE,
					       "putback buffer");
	  pt->putback_buf_size = SCM_INITIAL_PUTBACK_BUF_SIZE;
	}

      pt->saved_read_buf = pt->read_buf;
      pt->saved_read_pos = pt->read_pos;
      pt->saved_read_end = pt->read_end;
      pt->saved_read_buf_size = pt->read_buf_size;

      pt->read_pos = pt->read_buf = pt->putback_buf;
      pt->read_end = pt->read_buf + 1;
      pt->read_buf_size = pt->putback_buf_size;
    }

  *pt->read_buf = c;

  if (pt->rw_random)
    pt->rw_active = SCM_PORT_READ;

  if (c == '\n')
    {
      /* What should col be in this case?
       * We'll leave it at -1.
       */
      SCM_LINUM (port) -= 1;
    }
  else
    SCM_COL(port) -= 1;
}
#undef FUNC_NAME


void 
scm_ungets (const char *s, int n, SCM port)
{
  while (n--)
    scm_unget_byte (s[n], port);
}

void
scm_ungetc (SCM chr, SCM port)
{
  SCM buf = utf8_usv_to_bs (SCM_CHAR (chr)); // XXX port encoding
  reference_parameter (buf);
  scm_ungets (BYTE_STRING_DATA (buf), BYTE_STRING_LENGTH (buf), port);
  dereference_parameter (buf);
}


SCM_DEFINE (scm_peek_char, "peek-char", 0, 1, 0,
           (SCM port),
	    "Return the next character available from @var{port},\n"
	    "@emph{without} updating @var{port} to point to the following\n"
	    "character.  If no more characters are available, the\n"
	    "end-of-file object is returned.\n"
	    "\n"
	    "The value returned by\n"
	    "a call to @code{peek-char} is the same as the value that would\n"
	    "have been returned by a call to @code{read-char} on the same\n"
	    "port.  The only difference is that the very next call to\n"
	    "@code{read-char} or @code{peek-char} on that @var{port} will\n"
	    "return the value returned by the preceding call to\n"
	    "@code{peek-char}.  In particular, a call to @code{peek-char} on\n"
	    "an interactive port will hang waiting for input whenever a call\n"
	    "to @code{read-char} would have hung.")
#define FUNC_NAME s_scm_peek_char
{
  reference_parameter (port);
  SCM c;
  int column;

  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
    reference_parameter (port);
  } else {
    SCM_VALIDATE_OPINPORT (1, port);
  }
  column = SCM_COL(port);
  c = scm_getc (port);
  if (SCM_EOF_OBJECT_P (c)) {
    dereference_parameter (port);
    return SCM_EOF_VAL;
  }

  scm_ungetc (c, port);
  SCM_COL(port) = column;

  dereference_parameter (port);
  return c;
}
#undef FUNC_NAME

SCM_DEFINE (scm_unread_char, "unread-char", 1, 1, 0,
            (SCM cobj, SCM port),
	    "Place @var{char} in @var{port} so that it will be read by the\n"
	    "next read operation.  If called multiple times, the unread characters\n"
	    "will be read again in last-in first-out order.  If @var{port} is\n"
	    "not supplied, the current input port is used.")
#define FUNC_NAME s_scm_unread_char
{
  reference_parameter (cobj);
  reference_parameter (port);

  SCM_VALIDATE_CHAR (1, cobj);
  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
    reference_parameter (port);
  } else {
    SCM_VALIDATE_OPINPORT (2, port);
  }

  scm_ungetc (cobj, port);

  dereference_parameter (port);
  return (rc_return_stack_reference (cobj));
}
#undef FUNC_NAME

SCM_DEFINE (scm_unread_string, "unread-string", 2, 0, 0,
            (SCM str, SCM port),
	    "Place the string @var{str} in @var{port} so that its characters will be\n"
	    "read in subsequent read operations.  If called multiple times, the\n"
	    "unread characters will be read again in last-in first-out order.  If\n"
	    "@var{port} is not supplied, the current-input-port is used.")
#define FUNC_NAME s_scm_unread_string
{
  reference_parameter (str);
  reference_parameter (port);

  SCM_VALIDATE_BRASS (1, str);

  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
    reference_parameter (port);
  } else {
    SCM_VALIDATE_OPINPORT (2, port);
  }

  SCM b = scm_to_pen (str);
  reference_parameter (b);
  scm_ungets (BYTE_STRING_DATA (str), BYTE_STRING_LENGTH (str), port);
  
  dereference_parameter (b);
  dereference_parameter (port);
  return (rc_return_stack_reference (str));
}
#undef FUNC_NAME

SCM_DEFINE (scm_seek, "seek", 3, 0, 0,
            (SCM fd_port, SCM offset, SCM whence), "")
#define FUNC_NAME s_scm_seek
{
  reference_parameter (fd_port);
  reference_parameter (offset);
  reference_parameter (whence);

  int how;

  how = scm_to_int (whence);
  if (how != SEEK_SET && how != SEEK_CUR && how != SEEK_END)
    SCM_OUT_OF_RANGE (3, whence);

  if (SCM_OPFPORTP (fd_port))
    {
      /* go direct to fport code to allow 64-bit offsets */
      SCM answer = scm_i_fport_seek (fd_port, offset, how);
      reference_count_return (answer);
      dereference_parameter (fd_port);
      dereference_parameter (offset);
      dereference_parameter (whence);
      return (answer);
    }
  else if (SCM_OPPORTP (fd_port))
    {
      scm_t_ptob_descriptor *ptob = scm_ptobs + SCM_PTOBNUM (fd_port);
      off_t off = scm_to_off_t (offset);
      off_t rv;

      if (!ptob->seek)
	SCM_MISC_ERROR ("port is not seekable", 
                        scm_cons (fd_port, SCM_EOL));
      else
	rv = ptob->seek (fd_port, off, how);
      dereference_parameter (fd_port);
      dereference_parameter (offset);
      dereference_parameter (whence);
      return scm_from_off_t (rv);
    }
  else /* file descriptor?.  */
    {
      off_t_or_off64_t off = scm_to_off_t_or_off64_t (offset);
      off_t_or_off64_t rv;
      rv = lseek_or_lseek64 (scm_to_int (fd_port), off, how);
      dereference_parameter (fd_port);
      dereference_parameter (offset);
      dereference_parameter (whence);
      if (rv == -1)
	SCM_SYSERROR;
      return scm_from_off_t_or_off64_t (rv);
    }
}
#undef FUNC_NAME

#ifndef O_BINARY
#define O_BINARY 0
#endif

SCM_DEFINE (scm_truncate_file, "truncate-file", 1, 1, 0,
            (SCM object, SCM length),
	    "Truncate @var{file} to @var{length} bytes.  @var{file} can be a\n"
	    "filename string, a port object, or an integer file descriptor.\n"
	    "The return value is unspecified.\n"
	    "\n"
	    "For a port or file descriptor @var{length} can be omitted, in\n"
	    "which case the file is truncated at the current position (per\n"
	    "@code{ftell} above).\n"
	    "\n"
	    "On most systems a file can be extended by giving a length\n"
	    "greater than the current size, but this is not mandatory in the\n"
	    "POSIX standard.")
#define FUNC_NAME s_scm_truncate_file
{
  reference_parameter (object);
  reference_parameter (length);
  int rv;

  /* "object" can be a port, fdes or filename.

     Negative "length" makes no sense, but it's left to truncate() or
     ftruncate() to give back an error for that (normally EINVAL).
     */

  if (SCM_UNBNDP (length))
    {
      /* must supply length if object is a filename.  */
      if (SCM_BRASSP (object))
        SCM_MISC_ERROR("must supply length if OBJECT is a filename", SCM_EOL);
      
      length = scm_seek (object, SCM_INUM0, scm_from_int (SEEK_CUR));
      reference_parameter (length);
    }

  if (scm_is_integer (object))
    {
      off_t_or_off64_t c_length = scm_to_off_t_or_off64_t (length);
      SCM_SYSCALL (rv = ftruncate_or_ftruncate64 (scm_to_int (object),
                                                  c_length));
    }
  else if (SCM_OPOUTFPORTP (object))
    {
      /* go direct to fport code to allow 64-bit offsets */
      rv = scm_i_fport_truncate (object, length);
    }
  else if (SCM_OPOUTPORTP (object))
    {
      off_t c_length = scm_to_off_t (length);
      scm_t_port *pt = SCM_PTAB_ENTRY (object);
      scm_t_ptob_descriptor *ptob = scm_ptobs + SCM_PTOBNUM (object);
      
      if (!ptob->truncate)
	SCM_MISC_ERROR ("port is not truncatable", SCM_EOL);
      if (pt->rw_active == SCM_PORT_READ)
	scm_end_input (object);
      else if (pt->rw_active == SCM_PORT_WRITE)
	ptob->flush (object);
      
      ptob->truncate (object, c_length);
      rv = 0;
    }
  else if (SCM_BRASSP (object))
    {
      off_t_or_off64_t c_length = scm_to_off_t_or_off64_t (length);
      SCM b = scm_to_sen (object);
      reference_parameter (b);
      char *str = BYTE_STRING_DATA (object);
      int eno;
      SCM_SYSCALL (rv = truncate_or_truncate64 (str, c_length));
      eno = errno;
      errno = eno;
      dereference_parameter (b);
    }
  else
    {
      scm_wrong_type_arg (FUNC_NAME, 1, object);
    }

  dereference_parameter (object);
  dereference_parameter (length);
  if (rv == -1)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_port_line, "port-line", 1, 0, 0,
            (SCM port),
	    "Return the current line number for @var{port}.\n"
	    "\n"
	    "The first line of a file is 0.  But you might want to add 1\n"
	    "when printing line numbers, since starting from 1 is\n"
	    "traditional in error messages, and likely to be more natural to\n"
	    "non-programmers.")
#define FUNC_NAME s_scm_port_line
{
  reference_parameter (port);
  SCM_VALIDATE_OPENPORT (1, port);

  SCM answer = scm_from_long (SCM_LINUM (port));

  reference_count_return (answer);
  dereference_parameter (port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_port_line_x, "set-port-line!", 2, 0, 0,
            (SCM port, SCM line),
	    "Set the current line number for @var{port} to @var{line}.  The\n"
	    "first line of a file is 0.")
#define FUNC_NAME s_scm_set_port_line_x
{
  reference_parameter (port);
  reference_parameter (line);

  SCM_VALIDATE_OPENPORT (1, port);

  SCM_PTAB_ENTRY (port)->line_number = scm_to_long (line);

  dereference_parameter (port);
  dereference_parameter (line);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_port_column, "port-column", 1, 0, 0,
            (SCM port),
	    "Return the current column number of @var{port}.\n"
	    "If the number is\n"
	    "unknown, the result is #f.  Otherwise, the result is a 0-origin integer\n"
	    "- i.e. the first character of the first line is line 0, column 0.\n"
	    "(However, when you display a file position, for example in an error\n"
	    "message, we recommend you add 1 to get 1-origin integers.  This is\n"
	    "because lines and column numbers traditionally start with 1, and that is\n"
	    "what non-programmers will find most natural.)")
#define FUNC_NAME s_scm_port_column
{
  reference_parameter (port);

  SCM_VALIDATE_OPENPORT (1, port);

  SCM answer = scm_from_int (SCM_COL (port));

  reference_count_return (answer);
  dereference_parameter (port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_port_column_x, "set-port-column!", 2, 0, 0,
            (SCM port, SCM column),
	    "Set the current column of @var{port}.  Before reading the first\n"
	    "character on a line the column should be 0.")
#define FUNC_NAME s_scm_set_port_column_x
{
  reference_parameter (port);
  reference_parameter (column);

  SCM_VALIDATE_OPENPORT (1, port);

  SCM_PTAB_ENTRY (port)->column_number = scm_to_int (column);

  dereference_parameter (column);
  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_port_filename, "port-filename", 1, 0, 0,
            (SCM port),
	    "Return the filename associated with @var{port}.  This function returns\n"
	    "the strings \"standard input\", \"standard output\" and \"standard error\"\n"
	    "when called on the current input, output and error ports respectively.")
#define FUNC_NAME s_scm_port_filename
{
  reference_parameter (port);

  SCM_VALIDATE_OPENPORT (1, port);

  SCM answer = SCM_FILENAME (port);

  reference_count_return (answer);
  dereference_parameter (port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_port_filename_x, "set-port-filename!", 2, 0, 0,
            (SCM port, SCM filename),
	    "Change the filename associated with @var{port}, using the current input\n"
	    "port if none is specified.  Note that this does not change the port's\n"
	    "source of data, but only the value that is returned by\n"
	    "@code{port-filename} and reported in diagnostic output.")
#define FUNC_NAME s_scm_set_port_filename_x
{
  reference_parameter (port);
  reference_parameter (filename);

  SCM_VALIDATE_OPENPORT (1, port);

  /* We allow the user to set the filename to whatever he likes.  */
  SCM_SET_FILENAME (port, filename);

  dereference_parameter (port);
  dereference_parameter (filename);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

void
scm_print_port_mode (SCM exp, SCM port)
{
  scm_puts (SCM_CLOSEDP (exp)
	    ? "closed: "
	    : (SCM_RDNG & SCM_CELL_WORD_0 (exp)
	       ? (SCM_WRTNG & SCM_CELL_WORD_0 (exp)
		  ? "input-output: "
		  : "input: ")
	       : (SCM_WRTNG & SCM_CELL_WORD_0 (exp)
		  ? "output: "
		  : "bogus: ")),
	    port);
}

int
scm_port_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  char *type = SCM_PTOBNAME (SCM_PTOBNUM (exp));
  if (!type)
    type = "port";
  scm_puts ("#<", port);
  scm_print_port_mode (exp, port);
  scm_puts (type, port);
  scm_putc (' ', port);
  scm_uintprint (SCM_CELL_WORD_1 (exp), 16, port);
  scm_putc ('>', port);
  return 1;
}

void
scm_ports_prehistory ()
{
  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace:"
                     " scm_ports_prehistory\n");

  scm_numptob = 0;
  scm_ptobs = (scm_t_ptob_descriptor *) scm_malloc (sizeof (scm_t_ptob_descriptor));
}



/* Void ports.   */

scm_t_bits scm_tc16_void_port = 0;

static int fill_input_void_port (SCM port SCM_UNUSED)
{
  return EOF;
}

static void
write_void_port (SCM port SCM_UNUSED,
		 const void *data SCM_UNUSED,
		 size_t size SCM_UNUSED)
{
}

static SCM
scm_i_void_port (long mode_bits)
{
  SCM answer = scm_new_port_table_entry (scm_tc16_void_port);
  scm_t_port * pt = SCM_PTAB_ENTRY(answer);

  scm_port_non_buffer (pt);
  
  SCM_SETSTREAM (answer, 0);
  SCM_SET_CELL_WORD_0 (answer, mode_bits); // flags
  return answer;
}

SCM
scm_void_port (char *mode_str)
{
  return scm_i_void_port (scm_mode_bits (mode_str));
}

SCM_DEFINE (scm_sys_make_void_port, "%make-void-port", 1, 0, 0,
            (SCM mode),
	    "Create and return a new void port.  A void port acts like\n"
	    "@file{/dev/null}.  The @var{mode} argument\n"
	    "specifies the input/output modes for this port: see the\n"
	    "documentation for @code{open-file} in @ref{File Ports}.")
#define FUNC_NAME s_scm_sys_make_void_port
{
  reference_parameter (mode);
  SCM answer = scm_i_void_port (scm_i_mode_bits (mode));
  dereference_parameter (mode);
  return answer;
}
#undef FUNC_NAME


/* Initialization.  */

void
scm_init_ports ()
{

  /* Initialise the list of ports.  */
  scm_i_port_table = (scm_t_port **)
    scm_malloc (sizeof (scm_t_port *) * scm_i_port_table_room);
  if (!scm_i_port_table) {
    fprintf (stderr, " * VSI port table allocation failed * ");
    abort ();
  }

  scm_tc16_void_port = scm_make_port_type ("void", fill_input_void_port, 
					   write_void_port);

  cur_inport_box = scm_cons (SCM_BOOL_T, SCM_BOOL_F);
  cur_outport_box = scm_cons (SCM_BOOL_T, SCM_BOOL_F);
  cur_errport_box = scm_cons (SCM_BOOL_T, SCM_BOOL_F);
  cur_loadport_box = scm_cons (SCM_BOOL_T, SCM_BOOL_F);

  rc_declare_data_graph_root (cur_inport_box);
  rc_declare_data_graph_root (cur_outport_box);
  rc_declare_data_graph_root (cur_errport_box);
  rc_declare_data_graph_root (cur_loadport_box);
}

void
x_inclusion_ports ()
{
  /* lseek() symbols.  */
  scm_c_define ("SEEK_SET", scm_from_int (SEEK_SET));
  scm_c_define ("SEEK_CUR", scm_from_int (SEEK_CUR));
  scm_c_define ("SEEK_END", scm_from_int (SEEK_END));

#include "include/ports.x"
}
