#include "include/config.h"

#include "include/_scm.h"

#include <gmp.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "include/smob.h"
#include "include/numbers.h"
#include "include/feature.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/vectors.h"

#include "include/validate.h"
#include "include/random.h"

/* Original Commentary: */

/*
 * A plugin interface for RNGs
 *
 * Using this interface, it is possible for the application to tell
 * VSI to use a different RNG.  This is desirable if it is
 * necessary to use the same RNG everywhere in the application in
 * order to prevent interference, if the application uses RNG
 * hardware, or if the application has special demands on the RNG.
 *
 * Look in random.h and how the default generator is "plugged in" in
 * scm_init_random().
 */

scm_t_rng scm_the_rng;


/*
 * The prepackaged RNG
 *
 * This is the MWC (Multiply With Carry) random number generator
 * described by George Marsaglia at the Department of Statistics and
 * Supercomputer Computations Research Institute, The Florida State
 * University (http://stat.fsu.edu/~geo).
 *
 * It uses 64 bits, has a period of 4578426017172946943 (4.6e18), and
 * passes all tests in the DIEHARD test suite
 * (http://stat.fsu.edu/~geo/diehard.html)
 *
 */
/* end of original commentary */

/*
 * NB (from someone else): There have been some changes to that original
 * version due to the introduction of reference counting and the transition
 * to byte strings and brass. No idea if it's still in specs, uknow?
 *  Furthermore the random sphere functions have been removed.
 */

#define A 2131995753UL

#ifndef M_PI
#define M_PI 3.14159265359
#endif

unsigned long
scm_i_uniform32 (scm_t_i_rstate *state)
{
  scm_t_uint64 x = (scm_t_uint64) A * state->w + state->c;
  scm_t_uint32 w = x & 0xffffffffUL;
  state->w = w;
  state->c = x >> 32L;
  return w;
}

void
scm_i_init_rstate (scm_t_i_rstate *state, const char *seed, int n)
{
  scm_t_uint32 w = 0L;
  scm_t_uint32 c = 0L;
  int i, m;
  for (i = 0; i < n; ++i)
    {
      m = i % 8;
      if (m < 4)
	w += seed[i] << (8 * m);
      else
        c += seed[i] << (8 * (m - 4));
    }
  if ((w == 0 && c == 0) || (w == -1 && c == A - 1))
    ++c;
  state->w = w;
  state->c = c;
}

scm_t_i_rstate *
scm_i_copy_rstate (scm_t_i_rstate *state)
{
  scm_t_rstate *new_state = scm_malloc (scm_the_rng.rstate_size);
  return memcpy (new_state, state, scm_the_rng.rstate_size);
}


/*
 * Random number library functions
 */

scm_t_rstate *
scm_c_make_rstate (const char *seed, int n)
{
  scm_t_rstate *state = scm_malloc (scm_the_rng.rstate_size);
  state->reserved0 = 0;
  scm_the_rng.init_rstate (state, seed, n);
  return state;
}


scm_t_rstate *
scm_c_default_rstate ()
#define FUNC_NAME "scm_c_default_rstate"
{
  SCM state = SCM_VARIABLE_REF (scm_var_random_state);
  if (!SCM_RSTATEP (state))
    SCM_MISC_ERROR ("*random-state* contains bogus random state", SCM_EOL);
  return SCM_RSTATE (state);
}
#undef FUNC_NAME


inline double
scm_c_uniform01 (scm_t_rstate *state)
{
  double x = (double) scm_the_rng.random_bits (state) / (double) 0xffffffffUL;
  return ((x + (double) scm_the_rng.random_bits (state))
	  / (double) 0xffffffffUL);
}

double
scm_c_normal01 (scm_t_rstate *state)
{
  if (state->reserved0)
    {
      state->reserved0 = 0;
      return state->reserved1;
    }
  else
    {
      double r, a, n;
      
      r = sqrt (-2.0 * log (scm_c_uniform01 (state)));
      a = 2.0 * M_PI * scm_c_uniform01 (state);
      
      n = r * sin (a);
      state->reserved1 = r * cos (a);
      state->reserved0 = 1;
      
      return n;
    }
}

double
scm_c_exp1 (scm_t_rstate *state)
{
  return - log (scm_c_uniform01 (state));
}

unsigned char scm_masktab[256];

static inline scm_t_uint32
scm_i_mask32 (scm_t_uint32 m)
{
  return (m < 0x100
	  ? scm_masktab[m]
	  : (m < 0x10000
	     ? scm_masktab[m >> 8] << 8 | 0xff
	     : (m < 0x1000000
		? scm_masktab[m >> 16] << 16 | 0xffff
		: scm_masktab[m >> 24] << 24 | 0xffffff)));
}

static scm_t_uint32
scm_c_random32 (scm_t_rstate *state, scm_t_uint32 m)
{
  scm_t_uint32 r, mask = scm_i_mask32 (m);
  while ((r = scm_the_rng.random_bits (state) & mask) >= m);
  return r;
}

/* Returns 32 random bits. */
unsigned long
scm_c_random (scm_t_rstate *state, unsigned long m)
{
  return scm_c_random32 (state, (scm_t_uint32)m);
}

scm_t_uint64
scm_c_random64 (scm_t_rstate *state, scm_t_uint64 m)
{
  scm_t_uint64 r;
  scm_t_uint32 mask;

  if (m <= SCM_T_UINT32_MAX)
    return scm_c_random32 (state, (scm_t_uint32) m);
  
  mask = scm_i_mask32 (m >> 32);
  while ((r = ((scm_t_uint64) (scm_the_rng.random_bits (state) & mask) << 32)
          | scm_the_rng.random_bits (state)) >= m)
    ;
  return r;
}

/*
  SCM scm_c_random_bignum (scm_t_rstate *state, SCM m)

  Takes a random state (source of random bits) and a bignum m.
  Returns a bignum b, 0 <= b < m.

  It does this by allocating a bignum b with as many base 65536 digits
  as m, filling b with random bits (in 32 bit chunks) up to the most
  significant 1 in m, and, finally checking if the resultant b is too
  large (>= m).  If too large, we simply repeat the process again.  (It
  is important to throw away all generated random bits if b >= m,
  otherwise we'll end up with a distorted distribution.)

*/

SCM
scm_c_random_bignum (scm_t_rstate *state, SCM m)
{
  reference_parameter (m);

  SCM result = scm_i_mkbig ();
  reference_parameter (result);

  const size_t m_bits = mpz_sizeinbase (SCM_I_BIG_MPZ (m), 2);
  /* how many bits would only partially fill the last u32? */
  const size_t end_bits = m_bits % (sizeof (scm_t_uint32) * SCM_CHAR_BIT);
  scm_t_uint32 *random_chunks = NULL;
  const scm_t_uint32 num_full_chunks =
    m_bits / (sizeof (scm_t_uint32) * SCM_CHAR_BIT);
  const scm_t_uint32 num_chunks = num_full_chunks + ((end_bits) ? 1 : 0);

  /* we know the result will be this big */
  mpz_realloc2 (SCM_I_BIG_MPZ (result), m_bits);

  random_chunks =
    (scm_t_uint32 *) scm_gc_calloc (num_chunks * sizeof (scm_t_uint32),
                                     "random bignum chunks");

  do
    {
      scm_t_uint32 *current_chunk = random_chunks + (num_chunks - 1);
      scm_t_uint32 chunks_left = num_chunks;

      mpz_set_ui (SCM_I_BIG_MPZ (result), 0);
      
      if (end_bits)
        {
          /* generate a mask with ones in the end_bits position, i.e. if
             end_bits is 3, then we'd have a mask of ...0000000111 */
          const unsigned long rndbits = scm_the_rng.random_bits (state);
          int rshift = (sizeof (scm_t_uint32) * SCM_CHAR_BIT) - end_bits;
          scm_t_uint32 mask = 0xffffffff >> rshift;
          scm_t_uint32 highest_bits = ((scm_t_uint32) rndbits) & mask;
          *current_chunk-- = highest_bits;
          chunks_left--;
        }
      
      while (chunks_left)
        {
          /* now fill in the remaining scm_t_uint32 sized chunks */
          *current_chunk-- = scm_the_rng.random_bits (state);
          chunks_left--;
        }
      mpz_import (SCM_I_BIG_MPZ (result),
                  num_chunks,
                  -1,
                  sizeof (scm_t_uint32),
                  0,
                  0,
                  random_chunks);
      /* if result >= m, regenerate it (it is important to regenerate
	 all bits in order not to get a distorted distribution) */
    } while (mpz_cmp (SCM_I_BIG_MPZ (result), SCM_I_BIG_MPZ (m)) >= 0);
  scm_gc_free (random_chunks,
               num_chunks * sizeof (scm_t_uint32),
               "random bignum chunks");

  SCM answer = scm_i_normbig (result);
  reference_count_return (answer);
  dereference_parameter (result);
  dereference_parameter (m);
  return (answer);
}

/*
 * Scheme level representation of random states.
 */
 
scm_t_tag scm_tc16_rstate;

static SCM
make_rstate (scm_t_rstate *state)
{
  SCM_RETURN_NEWSMOB (scm_tc16_rstate, state);
}


/*
 * Scheme level interface.
 */

SCM_GLOBAL_VARIABLE_INIT (scm_var_random_state, "*random-state*", scm_seed_to_random_state (scm_from_locale_string ("URL:http://stat.fsu.edu/~geo/diehard.html")));

SCM_DEFINE (scm_random, "random", 1, 1, 0, 
            (SCM n, SCM state),
            "Return a number in [0, N).\n"
            "\n"
            "Accepts a positive integer or real n and returns a\n"
            "number of the same type between zero (inclusive) and\n"
            "N (exclusive). The values returned have a uniform\n"
            "distribution.\n"
            "\n"
            "The optional argument @var{state} must be of the type produced\n"
	    "by @code{seed->random-state}. It defaults to the value of the\n"
	    "variable @var{*random-state*}. This object is used to maintain\n"
	    "the state of the pseudo-random-number generator and is altered\n"
	    "as a side effect of the random operation.")
#define FUNC_NAME s_scm_random
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (2, state);
  reference_parameter (state);

  SCM answer = SCM_UNSPECIFIED;

  if (SCM_I_INUMP (n))
    {
      unsigned long m = (unsigned long) SCM_I_INUM (n);
      SCM_ASSERT_RANGE (1, n, SCM_I_INUM (n) > 0);
#if SCM_SIZEOF_UNSIGNED_LONG <= 4
      answer = scm_from_uint32 (scm_c_random (SCM_RSTATE (state),
                                              (scm_t_uint32) m));
#elif SCM_SIZEOF_UNSIGNED_LONG <= 8
      answer = scm_from_uint64 (scm_c_random64 (SCM_RSTATE (state),
                                                 (scm_t_uint64) m));
#else
#error "Cannot deal with this platform's unsigned long size"
#endif
      dereference_parameter (state);
      return (answer);
    }
  SCM_VALIDATE_NIM (1, n);
  reference_parameter (n);  // late reference

  if (SCM_REALP (n)) {
    answer = scm_from_double (SCM_REAL_VALUE (n)
                              * scm_c_uniform01 (SCM_RSTATE (state)));
    dereference_parameter (n);
    dereference_parameter (state);
    return (answer);
  }

  if (!SCM_BIGP (n))
    SCM_WRONG_TYPE_ARG (1, n);

  answer = scm_c_random_bignum (SCM_RSTATE (state), n);
  reference_count_return (answer);
  dereference_parameter (n);
  dereference_parameter (state);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_copy_random_state, "copy-random-state", 0, 1, 0, 
            (SCM state),
            "Return a copy of the random state @var{state}.")
#define FUNC_NAME s_scm_copy_random_state
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (1, state);
  reference_parameter (state);

  SCM answer = make_rstate (scm_the_rng.copy_rstate (SCM_RSTATE (state)));

  dereference_parameter (state);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_seed_to_random_state, "seed->random-state", 1, 0, 0, 
            (SCM seed),
            "Return a new random state using @var{seed}.")
#define FUNC_NAME s_scm_seed_to_random_state
{
  reference_parameter (seed);

  SCM res;
  SCM beed;

  if (SCM_NUMBERP (seed)) {
    beed = scm_to_sen (scm_number_to_brass (seed, SCM_UNDEFINED));

  } else if (BYTE_STRING_P (seed)) {
    beed = seed;

  } else {
    SCM_VALIDATE_BRASS (1, seed);
    
    beed = scm_brass_to_c0_byte_string (seed);
    if (scm_is_false (beed)) {
      scm_out_of_range (FUNC_NAME, seed);
    }
  }
  reference_parameter (beed);

  res = make_rstate (scm_c_make_rstate (BYTE_STRING_DATA (beed),
					BYTE_STRING_LENGTH (beed)));

  dereference_parameter (seed);
  dereference_parameter (beed);
  return res;
}
#undef FUNC_NAME

SCM_DEFINE (scm_random_uniform, "random:uniform", 0, 1, 0, 
            (SCM state),
	    "Return a uniformly distributed inexact real random number in\n"
	    "[0,1).")
#define FUNC_NAME s_scm_random_uniform
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (1, state);
  reference_parameter (state);

  SCM answer = scm_from_double (scm_c_uniform01 (SCM_RSTATE (state)));

  dereference_parameter (state);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_random_normal, "random:normal", 0, 1, 0, 
            (SCM state),
	    "Return an inexact real in a normal distribution.  The\n"
	    "distribution used has mean 0 and standard deviation 1.  For a\n"
	    "normal distribution with mean m and standard deviation d use\n"
	    "@code{(+ m (* d (random:normal)))}.")
#define FUNC_NAME s_scm_random_normal
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (1, state);
  reference_parameter (state);

  SCM answer = scm_from_double (scm_c_normal01 (SCM_RSTATE (state)));

  dereference_parameter (state);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_random_normal_vector_x, "random:normal-vector!", 1, 1, 0, 
            (SCM v, SCM state),
            "Fills vect with inexact real random numbers that are\n"
            "independent and standard normally distributed\n"
            "(i.e., with mean 0 and variance 1).")
#define FUNC_NAME s_scm_random_normal_vector_x
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (2, state);

  if (scm_is_vector (v))
    {
      reference_parameter (v);
      reference_parameter (state);

      int i;
      int len = SCM_SIMPLE_VECTOR_LENGTH (v);
      SCM element = scm_from_double (scm_c_normal01 (SCM_RSTATE (state)));
      for (i = 0; i < len; i++)
        SCM_SIMPLE_VECTOR_SET (v, i, element);

      dereference_parameter (v);
      dereference_parameter (state);
    }
  else
    {
      scm_wrong_type_arg_msg (NULL, 0, v, "vector");
    }

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_random_exp, "random:exp", 0, 1, 0, 
            (SCM state),
	    "Return an inexact real in an exponential distribution with mean\n"
	    "1.  For an exponential distribution with mean u use (* u\n"
	    "(random:exp)).")
#define FUNC_NAME s_scm_random_exp
{
  if (SCM_UNBNDP (state))
    state = SCM_VARIABLE_REF (scm_var_random_state);
  SCM_VALIDATE_RSTATE (1, state);

  reference_parameter (state);
  SCM answer = scm_from_double (scm_c_exp1 (SCM_RSTATE (state)));
  dereference_parameter (state);

  return (answer);
}
#undef FUNC_NAME


static void
rstate_release (SCM rstate)
{
  free (SCM_RSTATE (rstate));
  return;
}

void
scm_init_random ()
{
  int i, m;
  /* plug in default RNG */
  scm_t_rng rng =
  {
    sizeof (scm_t_i_rstate),
    (unsigned long (*)()) scm_i_uniform32,
    (void (*)())          scm_i_init_rstate,
    (scm_t_rstate *(*)())    scm_i_copy_rstate
  };
  scm_the_rng = rng;
  
  scm_tc16_rstate = scm_make_smob_type ("random-state", 0);
  scm_set_smob_release (scm_tc16_rstate, 2, rstate_release);

  for (m = 1; m <= 0x100; m <<= 1)
    for (i = m >> 1; i < m; ++i)
      scm_masktab[i] = m - 1;
}

void x_inclusion_random ()
{
  scm_add_feature ("random"); /* XXX */

#include "include/random.x"
}
