#include "config.h"

#include "include/_scm.h"
#include "include/eval.h"
#include "include/feature.h"
#include "include/vectors.h"
#include "include/dynwind.h"

#include "include/validate.h"
#include "include/sort.h"

static scm_t_trampoline_2
compare_function (SCM less, unsigned int arg_nr, const char* fname)
{
  const scm_t_trampoline_2 cmp = scm_trampoline_2 (less);
  SCM_ASSERT_TYPE (cmp != NULL, less, arg_nr, fname, "less predicate");
  return cmp;
}


SCM_DEFINE (scm_restricted_vector_sort_x, "restricted-vector-sort!", 4, 0, 0,
            (SCM vec, SCM less, SCM startpos, SCM endpos),
	    "Sort the vector @var{vec}, using @var{less} for comparing\n"
	    "the vector elements.  @var{startpos} (inclusively) and\n"
	    "@var{endpos} (exclusively) delimit\n"
	    "the range of the vector which gets sorted.  The return value\n"
	    "is not specified.")
#define FUNC_NAME s_scm_restricted_vector_sort_x
{
  reference_parameter (vec);
  reference_parameter (less);
  reference_parameter (startpos);
  reference_parameter (endpos);

  const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
  size_t vlen, spos, len;
  ssize_t vinc;
  SCM *velts;

  velts = scm_vector_writable_elements (vec, &vlen, &vinc);
  spos = scm_to_unsigned_integer (startpos, 0, vlen);
  len = scm_to_unsigned_integer (endpos, spos, vlen) - spos;

  quicksort (velts + spos*vinc, len, vinc, cmp, less);

  dereference_parameter (vec);
  dereference_parameter (less);
  dereference_parameter (startpos);
  dereference_parameter (endpos);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


/* (sorted? sequence less?)
 * is true when sequence is a list (x0 x1 ... xm) or a vector #(x0 ... xm)
 * such that for all 1 <= i <= m,
 * (not (less? (list-ref list i) (list-ref list (- i 1)))). */
SCM_DEFINE (scm_sorted_p, "sorted?", 2, 0, 0,
            (SCM items, SCM less),
	    "Return @code{#t} iff @var{items} is a list or a vector such that\n"
	    "for all 1 <= i <= m, the predicate @var{less} returns true when\n"
	    "applied to all elements i - 1 and i")
#define FUNC_NAME s_scm_sorted_p
{
  reference_parameter (items);
  reference_parameter (less);

  const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
  long len, j;			/* list/vector length, temp j */
  SCM item, rest;		/* rest of items loop variable */

  if (scm_is_null (items)) {
    dereference_parameter (items);
    dereference_parameter (less);
    return SCM_BOOL_T;
  }

  if (scm_is_pair (items)) {
    len = scm_ilength (items); /* also checks that it's a pure list */
    SCM_ASSERT_RANGE (1, items, len >= 0);
    if (len <= 1)
      return SCM_BOOL_T;

    item = TMP_CAR (items);
    rest = TMP_CDR (items);
    j = len - 1;
    while (j > 0) {
      SCM tmp = (*cmp) (less, TMP_CAR (rest), item);
      if (scm_is_true (tmp) ) {
        dereference_return_value (tmp);
        dereference_parameter (items);
        dereference_parameter (less);
        return SCM_BOOL_F;
      } else {
        dereference_return_value (tmp);
        item = TMP_CAR (rest);
        rest = TMP_CDR (rest);
        j--;
      }
    }
    dereference_parameter (items);
    dereference_parameter (less);
    return SCM_BOOL_T;

  } else if (scm_is_vector (items)) {
    size_t i, len;
    ssize_t inc;
    const SCM *elts;
    SCM result = SCM_BOOL_T;

    elts = scm_vector_elements (items, &len, &inc);

    for (i = 1; i < len; i++, elts += inc) {
      SCM tmp = (*cmp) (less, elts[inc], elts[0]);
      if (scm_is_true (tmp)) {
        dereference_return_value (tmp);
        result = SCM_BOOL_F;
        break;
      }
      dereference_return_value (tmp);
    }

    dereference_parameter (items);
    dereference_parameter (less);
    return result;

  } else {
    SCM_WRONG_TYPE_ARG (1, items);
  }

  dereference_parameter (items);
  dereference_parameter (less);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


/* (merge a b less?)
   takes two lists a and b such that (sorted? a less?) and (sorted? b less?)
   and returns a new list in which the elements of a and b have been stably
   interleaved so that (sorted? (merge a b less?) less?).
   Note:  this does _not_ accept vectors. */
SCM_DEFINE (scm_merge, "merge", 3, 0, 0, 
            (SCM alist, SCM blist, SCM less),
	    "Merge two already sorted lists into one.\n"
	    "Given two lists @var{alist} and @var{blist}, such that\n"
	    "@code{(sorted? alist less?)} and @code{(sorted? blist less?)},\n"
	    "return a new list in which the elements of @var{alist} and\n"
	    "@var{blist} have been stably interleaved so that\n"
	    "@code{(sorted? (merge alist blist less?) less?)}.\n"
	    "Note:  this does _not_ accept vectors.")
#define FUNC_NAME s_scm_merge
{
  reference_parameter (alist);
  reference_parameter (blist);
  reference_parameter (less);

  SCM build;

  if (scm_is_null (alist)) {
    dereference_parameter (alist);
    dereference_parameter (less);
    return (rc_return_stack_reference (blist));

  } else if (scm_is_null (blist)) {
    dereference_parameter (blist);
    dereference_parameter (less);
    return (rc_return_stack_reference (alist));

  } else {
    const scm_t_trampoline_2 cmp = compare_function (less, 3, FUNC_NAME);
    long alen, blen;		/* list lengths */
    SCM last;
    SCM cmp_result = SCM_UNSPECIFIED;

    SCM_VALIDATE_NONEMPTYLIST_COPYLEN (1, alist, alen);
    SCM_VALIDATE_NONEMPTYLIST_COPYLEN (2, blist, blen);

    cmp_result = (*cmp) (less, TMP_CAR (blist), TMP_CAR (alist));
    if (scm_is_true (cmp_result)) {
      build = scm_cons (TMP_CAR (blist), SCM_EOL);
      reference_parameter (build);
      blist = FLY_CDR (blist);
      blen--;
    } else {
      build = scm_cons (TMP_CAR (alist), SCM_EOL);
      reference_parameter (build);
      alist = FLY_CDR (alist);
      alen--;
    }
    dereference_return_value (cmp_result);
    last = build;

    // RC checkpoint: alist, blist, build
    //  /last/ is list referenced, /cmp_result/ availabe

    while ((alen > 0) && (blen > 0)) {
      cmp_result = (*cmp) (less, TMP_CAR (blist), TMP_CAR (alist));
      if (scm_is_true (cmp_result)) {
        SCM_SETCDR (last, scm_cons (TMP_CAR (blist), SCM_EOL));
        blist = FLY_CDR (blist);
        blen--;
      } else {
        SCM_SETCDR (last, scm_cons (TMP_CAR (alist), SCM_EOL));
        alist = FLY_CDR (alist);
        alen--;
      }
      dereference_return_value (cmp_result);
      last = TMP_CDR (last);
    }
    if ((alen > 0) && (blen == 0))
      SCM_SETCDR (last, alist);
    else if ((alen == 0) && (blen > 0))
      SCM_SETCDR (last, blist);
  }
  dereference_parameter (alist);
  dereference_parameter (blist);
  dereference_parameter (less);
  return (rc_return_stack_reference (build));
}
#undef FUNC_NAME


static SCM 
scm_merge_list_x (SCM alist, SCM blist,
		  long alen, long blen,
		  scm_t_trampoline_2 cmp, SCM less)
{
  reference_parameter (alist);
  reference_parameter (blist);
  reference_parameter (less);

  SCM build, last;

  if (scm_is_null (alist)) {
    dereference_parameter (alist);
    dereference_parameter (less);
    return (rc_return_stack_reference (blist));

  } else if (scm_is_null (blist)) {
    dereference_parameter (blist);
    dereference_parameter (less);
    return (rc_return_stack_reference (alist));

  } else {
    if (sharp_scm_is_true ((*cmp) (less, TMP_CAR (blist), TMP_CAR (alist)))) {
      build = blist;
      reference_parameter (build);
      blist = FLY_CDR (blist);
      blen--;
    } else {
      build = alist;
      reference_parameter (build);
      alist = FLY_CDR (alist);
      alen--;
    }
    last = build;

    while ((alen > 0) && (blen > 0)) {
      if (sharp_scm_is_true ((*cmp) (less, TMP_CAR (blist),
                                           TMP_CAR (alist)))) {
        SCM_SETCDR (last, blist);
        blist = FLY_CDR (blist);
        blen--;

      } else {
        SCM_SETCDR (last, alist);
        alist = FLY_CDR (alist);
        alen--;
      }
      last = TMP_CDR (last);
    }
    if ((alen > 0) && (blen == 0))
      SCM_SETCDR (last, alist);
    else if ((alen == 0) && (blen > 0))
      SCM_SETCDR (last, blist);
    }

  dereference_parameter (alist);
  dereference_parameter (blist);
  dereference_parameter (less);
  return (rc_return_stack_reference (build));

}  /* scm_merge_list_x */


SCM_DEFINE (scm_merge_x, "merge!", 3, 0, 0, 
            (SCM alist, SCM blist, SCM less),
	    "Takes two lists @var{alist} and @var{blist} such that\n"
	    "@code{(sorted? alist less?)} and @code{(sorted? blist less?)} and\n"
	    "returns a new list in which the elements of @var{alist} and\n"
	    "@var{blist} have been stably interleaved so that\n"
	    " @code{(sorted? (merge alist blist less?) less?)}.\n"
	    "This is the destructive variant of @code{merge}\n"
	    "Note:  this does _not_ accept vectors.")
#define FUNC_NAME s_scm_merge_x
{
  reference_parameter (alist);
  reference_parameter (blist);
  reference_parameter (less);

  if (scm_is_null (alist)) {
    dereference_parameter (alist);
    dereference_parameter (less);
    return (rc_return_stack_reference (blist));

  } else if (scm_is_null (blist)) {
    dereference_parameter (blist);
    dereference_parameter (less);
    return (rc_return_stack_reference (alist));

  } else {
    const scm_t_trampoline_2 cmp = compare_function (less, 3, FUNC_NAME);
    long alen, blen;		/* list lengths */
    SCM_VALIDATE_NONEMPTYLIST_COPYLEN (1, alist, alen);
    SCM_VALIDATE_NONEMPTYLIST_COPYLEN (2, blist, blen);
    return scm_merge_list_x (rc_Fly (alist), rc_Fly (blist),
                             alen, blen, cmp, rc_Fly (less));
  }
}
#undef FUNC_NAME


/* This merge sort algorithm is same as slib's by Richard A. O'Keefe.
   The algorithm is stable. We also tried to use the algorithm used by
   scsh's merge-sort but that algorithm showed to not be stable, even
   though it claimed to be.
*/
static SCM 
scm_merge_list_step (SCM * seq, scm_t_trampoline_2 cmp, SCM less, long n)
{
  // RC: /less/ and /*seq/ need to be referenced elsewhere
  //     /*seq/ may change value and will be re-referenced accordingly
  // the return value of this function has the return value flag set
  SCM a, b;

  if (n > 2) {
    long mid = n / 2;
    a = scm_merge_list_step (seq, cmp, less, mid);
    b = scm_merge_list_step (seq, cmp, less, n - mid);
    return scm_merge_list_x (a, b, mid, n - mid, cmp, less);

  } else if (n == 2) {
    SCM p = *seq;
    SCM rest = TMP_CDR (*seq);
    SCM x = TMP_CAR  (*seq);
    SCM y = TMP_CADR (*seq);
    reference_parameter (p);
    reference_parameter (rest);
    reference_parameter (x);
    reference_parameter (y);

    SCM tmp = TMP_CDR (rest);
    reference_parameter (tmp);  // RC exchange: counts for *seq
    dereference_parameter (*seq);
    *seq = tmp;

    SCM_SETCDR (rest, SCM_EOL);

    tmp = (*cmp) (less, y, x);
    if (scm_is_true (tmp)) {
      SCM_SETCAR (p, y);
      SCM_SETCAR (rest, x);
    }

    dereference_return_value (tmp);
    dereference_parameter (rest);
    dereference_parameter (x);
    dereference_parameter (y);
    return (rc_return_stack_reference (p));

  } else if (n == 1) {
    SCM p = *seq;
    reference_parameter (p);
    SCM tmp = TMP_CDR (p);
    reference_parameter (tmp);  // RC exchange: counts for *seq
    dereference_parameter (*seq);
    *seq = tmp;
                                   
    SCM_SETCDR (p, SCM_EOL);
    return (rc_return_stack_reference (p));

  } else
    return SCM_EOL;
}				/* scm_merge_list_step */


SCM_DEFINE (scm_sort_x, "sort!", 2, 0, 0, 
            (SCM items, SCM less),
	    "Sort the sequence @var{items}, which may be a list or a\n"
	    "vector.  @var{less} is used for comparing the sequence\n"
	    "elements.  The sorting is destructive, that means that the\n"
	    "input sequence is modified to produce the sorted result.\n"
	    "This is not a stable sort.")
#define FUNC_NAME s_scm_sort_x
{
  reference_parameter (items);
  reference_parameter (less);

  long len;			/* list/vector length */
  if (scm_is_null (items)) {
    dereference_parameter (less);
    return SCM_EOL;
  }

  SCM answer;
  if (scm_is_pair (items)) {
    const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
    SCM_VALIDATE_LIST_COPYLEN (1, items, len);

    answer = scm_merge_list_step (&items, cmp, less, len);

    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);

  } else if (scm_is_vector (items)) {
    scm_restricted_vector_sort_x (items,
                                  less,
                                  scm_from_int (0),
                                  scm_vector_length (items));
    dereference_parameter (less);
    return (rc_return_stack_reference (items));

  } else
    SCM_WRONG_TYPE_ARG (1, items);
}
#undef FUNC_NAME


SCM_DEFINE (scm_sort, "sort", 2, 0, 0, 
            (SCM items, SCM less),
	    "Sort the sequence @var{items}, which may be a list or a\n"
	    "vector.  @var{less} is used for comparing the sequence\n"
	    "elements.  This is not a stable sort.")
#define FUNC_NAME s_scm_sort
{
  reference_parameter (items);
  reference_parameter (less);

  if (scm_is_null (items)) {
    dereference_parameter (less);
    return (SCM_EOL);
  }

  SCM answer;
  if (scm_is_pair (items)) {
    answer = scm_sort_x (scm_list_copy (items), less);
    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);

  } else if (scm_is_vector (items)) {
    answer = scm_sort_x (scm_vector_copy (items), less);
    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);
  } else {
    SCM_WRONG_TYPE_ARG (1, items);
  }
}
#undef FUNC_NAME


static void
scm_merge_vector_x (SCM *vec,
		    SCM *temp,
		    scm_t_trampoline_2 cmp,
		    SCM less,
		    size_t low,
		    size_t mid,
		    size_t high,
		    ssize_t inc)
{
  size_t it;	     	/* Index for temp vector */
  size_t i1 = low;      /* Index for lower vector segment */
  size_t i2 = mid + 1; 	/* Index for upper vector segment */

#define VEC(i) vec[(i)*inc]

  /* Copy while both segments contain more characters */
  for (it = low; (i1 <= mid) && (i2 <= high); ++it)
    {
      SCM tmp = (*cmp) (less, VEC(i2), VEC(i1));
      reference_parameter (tmp);

      if (scm_is_true (tmp))
	temp[it] = VEC(i2++);
      else
	temp[it] = VEC(i1++);

      dereference_parameter (tmp);
    }

  {
    /* Copy while first segment contains more characters */
    while (i1 <= mid)
      temp[it++] = VEC(i1++);

    /* Copy while second segment contains more characters */
    while (i2 <= high)
      temp[it++] = VEC(i2++);

    /* Copy back from temp to vp */
    for (it = low; it <= high; it++)
      VEC(it) = temp[it];
  }
} 	        		/* scm_merge_vector_x */


static void
scm_merge_vector_step (SCM *vec,
		       SCM *temp,
		       scm_t_trampoline_2 cmp,
		       SCM less,
		       size_t low,
		       size_t high,
		       ssize_t inc)
{
  if (high > low)
    {
      size_t mid = (low + high) / 2;
      scm_merge_vector_step (vec, temp, cmp, less, low, mid, inc);
      scm_merge_vector_step (vec, temp, cmp, less, mid+1, high, inc);
      scm_merge_vector_x (vec, temp, cmp, less, low, mid, high, inc);
    }
}				/* scm_merge_vector_step */


SCM_DEFINE (scm_stable_sort_x, "stable-sort!", 2, 0, 0, 
            (SCM items, SCM less),
	    "Sort the sequence @var{items}, which may be a list or a\n"
	    "vector. @var{less} is used for comparing the sequence elements.\n"
	    "The sorting is destructive, that means that the input sequence\n"
	    "is modified to produce the sorted result.\n"
	    "This is a stable sort.")
#define FUNC_NAME s_scm_stable_sort_x
{
  reference_parameter (items);
  reference_parameter (less);

  const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
  long len;			/* list/vector length */

  if (scm_is_null (items)) {
    dereference_parameter (less);
    return items;
  }

  SCM answer;
  if (scm_is_pair (items)) {
    SCM_VALIDATE_LIST_COPYLEN (1, items, len);

    answer = scm_merge_list_step (&items, cmp, less, len);

    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);

  } else if (scm_is_vector (items)) {
    SCM temp, *temp_elts, *vec_elts;
    size_t len;
    ssize_t inc;
      
    vec_elts = scm_vector_writable_elements (items, &len, &inc);
    temp = scm_c_make_vector (len, SCM_UNDEFINED);
    reference_parameter (temp);
    temp_elts = scm_vector_writable_elements (temp, NULL, NULL);

    scm_merge_vector_step (vec_elts, temp_elts, cmp, less, 0, len-1, inc);
    
    /* We ungracefully clear the temporary vector as it didn't keep
       track of its references. */
    size_t tmp = 0;
    for ( ; tmp < len; tmp++) {
      temp_elts[tmp] = SCM_UNSPECIFIED;
    }

    dereference_parameter (temp);
    dereference_parameter (less);
    return (rc_return_stack_reference (items));

  } else
    SCM_WRONG_TYPE_ARG (1, items);
}
#undef FUNC_NAME


SCM_DEFINE (scm_stable_sort, "stable-sort", 2, 0, 0, 
            (SCM items, SCM less),
	    "Sort the sequence @var{items}, which may be a list or a\n"
	    "vector. @var{less} is used for comparing the sequence elements.\n"
	    "This is a stable sort.")
#define FUNC_NAME s_scm_stable_sort
{
  reference_parameter (items);
  reference_parameter (less);

  if (scm_is_null (items)) {
    dereference_parameter (less);
    return SCM_EOL;
  }

  SCM answer;
  if (scm_is_pair (items)) {
    answer = scm_stable_sort_x (scm_list_copy (items), less);
    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);

  } else if (scm_is_vector (items)) {
    answer = scm_stable_sort_x (scm_vector_copy (items), less);
    reference_count_return (answer);
    dereference_parameter (items);
    dereference_parameter (less);
    return (answer);

  } else {
    SCM_WRONG_TYPE_ARG (1, items);
  }
}
#undef FUNC_NAME


SCM_DEFINE (scm_sort_list_x, "sort-list!", 2, 0, 0, 
            (SCM items, SCM less),
	    "Sort the list @var{items}, using @var{less} for comparing the\n"
	    "list elements. The sorting is destructive, that means that the\n"
	    "input list is modified to produce the sorted result.\n"
	    "This is a stable sort.")
#define FUNC_NAME s_scm_sort_list_x
{
  reference_parameter (items);
  reference_parameter (less);

  const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
  long len;
  SCM_VALIDATE_LIST_COPYLEN (1, items, len);

  SCM answer = scm_merge_list_step (&items, cmp, less, len);

  reference_count_return (answer);
  dereference_parameter (items);
  dereference_parameter (less);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_sort_list, "sort-list", 2, 0, 0, 
	    (SCM items, SCM less),
	    "Sort the list @var{items}, using @var{less} for comparing the\n"
	    "list elements. This is a stable sort.")
#define FUNC_NAME s_scm_sort_list
{
  reference_parameter (items);
  reference_parameter (less);

  const scm_t_trampoline_2 cmp = compare_function (less, 2, FUNC_NAME);
  long len;

  SCM_VALIDATE_LIST_COPYLEN (1, items, len);

  items = rc_X (scm_list_copy (rc_Fly (items)));
  SCM answer = scm_merge_list_step (&items, cmp, less, len);
  
  reference_count_return (answer);
  dereference_parameter (items);
  dereference_parameter (less);
  return (answer);
}
#undef FUNC_NAME


void
scm_init_sort ()
{
#include "include/sort.x"

  scm_add_feature ("sort");
}
