// Strings & Brass

#include "include/config.h"

#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"
#include "include/ioext.h"
#include "include/fports.h"
#include "include/feature.h"
#include "include/ports.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/validate.h"
#include "include/dynwind.h"
#include "include/iselect.h"

#include <fcntl.h>

#ifdef HAVE_IO_H
#include <io.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


SCM_DEFINE (scm_ftell, "ftell", 1, 0, 0, 
            (SCM fd_port),
	    "Return an integer representing the current position of\n"
	    "@var{fd/port}, measured from the beginning.  Equivalent to:\n"
	    "\n"
	    "@lisp\n"
	    "(seek port 0 SEEK_CUR)\n"
	    "@end lisp")
#define FUNC_NAME s_scm_ftell
{
  reference_parameter (fd_port);
  SCM answer = scm_seek (fd_port, SCM_INUM0, scm_from_int (SEEK_CUR));
  dereference_parameter (fd_port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_dup_to_fdes, "dup->fdes", 1, 1, 0, 
            (SCM fd_or_port, SCM fd),
	    "Return a new integer file descriptor referring to the open file\n"
	    "designated by @var{fd_or_port}, which must be either an open\n"
	    "file port or a file descriptor.")
#define FUNC_NAME s_scm_dup_to_fdes
{
  reference_parameter (fd_or_port);
  reference_parameter (fd);
  int oldfd, newfd, rv;

  if (scm_is_integer (fd_or_port)) {
    oldfd = scm_to_int (fd_or_port);
  } else {
    SCM_VALIDATE_OPFPORT (1, fd_or_port);
    oldfd = SCM_FPORT_FDES (fd_or_port);
  }

  if (SCM_UNBNDP (fd)) {
    newfd = dup (oldfd);
    if (newfd == -1)
      SCM_SYSERROR;
    fd = scm_from_int (newfd);
    reference_parameter (fd);
  } else {
    newfd = scm_to_int (fd);
    if (oldfd != newfd) {
      rv = dup2 (oldfd, newfd);
      if (rv == -1)
        SCM_SYSERROR;
    }
  }
  dereference_parameter (fd_or_port);
  dereference_parameter (fd);
  return fd;
}
#undef FUNC_NAME


SCM_DEFINE (scm_dup2, "dup2", 2, 0, 0, 
            (SCM oldfd, SCM newfd),
	    "A simple wrapper for the @code{dup2} system call.\n"
	    "Copies the file descriptor @var{oldfd} to descriptor\n"
	    "number @var{newfd}, replacing the previous meaning\n"
	    "of @var{newfd}.  Both @var{oldfd} and @var{newfd} must\n"
	    "be integers.\n"
	    "Unlike for dup->fdes or primitive-move->fdes, no attempt\n"
	    "is made to move away ports which are using @var{newfd}.\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_dup2
{
  reference_parameter (oldfd);
  reference_parameter (newfd);
  int c_oldfd;
  int c_newfd;
  int rv;

  c_oldfd = scm_to_int (oldfd);
  c_newfd = scm_to_int (newfd);
  rv = dup2 (c_oldfd, c_newfd);
  if (rv == -1)
    SCM_SYSERROR;

  dereference_parameter (oldfd);
  dereference_parameter (newfd);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_fileno, "fileno", 1, 0, 0, (SCM port), "")
#define FUNC_NAME s_scm_fileno
{
  reference_parameter (port);
  SCM_VALIDATE_OPFPORT (1, port);
  SCM answer = scm_from_int (SCM_FPORT_FDES (port));
  dereference_parameter (port);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_isatty_p, "isatty?", 1, 0, 0, 
            (SCM port),
	    "Return @code{#t} if @var{port} is using a serial non--file\n"
	    "device, otherwise @code{#f}.")
#define FUNC_NAME s_scm_isatty_p
{
  reference_parameter (port);
  int rv;

  if (!SCM_OPFPORTP (port))
    return SCM_BOOL_F;
  
  rv = isatty (SCM_FPORT_FDES (port));
  dereference_parameter (port);
  return  scm_from_bool(rv);
}
#undef FUNC_NAME



SCM_DEFINE (scm_fdopen, "fdopen", 2, 0, 0,
            (SCM fdes, SCM modes),
	    "Return a new port based on the file descriptor @var{fdes}.\n"
	    "Modes are given by the string @var{modes}.  The revealed count\n"
	    "of the port is initialized to zero.  The modes string is the\n"
	    "same as that accepted by @ref{File Ports, open-file}.")
#define FUNC_NAME s_scm_fdopen
{
  reference_parameter (fdes);
  reference_parameter (modes);
  SCM answer = scm_i_fdes_to_port (scm_to_int (fdes),
                                   scm_i_mode_bits (modes), SCM_BOOL_F);
  reference_count_return (answer);
  dereference_parameter (fdes);
  dereference_parameter (modes);
  return (answer);
}
#undef FUNC_NAME



/* Move a port's underlying file descriptor to a given value.
 * Returns  #f if fdes is already the given value.
 *          #t if fdes moved. 
 * MOVE->FDES is implemented in Scheme and calls this primitive.
 */
SCM_DEFINE (scm_primitive_move_to_fdes, "primitive-move->fdes", 2, 0, 0,
            (SCM port, SCM fd),
	    "Moves the underlying file descriptor for @var{port} to the integer\n"
	    "value @var{fdes} without changing the revealed count of @var{port}.\n"
	    "Any other ports already using this descriptor will be automatically\n"
	    "shifted to new descriptors and their revealed counts reset to zero.\n"
	    "The return value is @code{#f} if the file descriptor already had the\n"
	    "required value or @code{#t} if it was moved.")
#define FUNC_NAME s_scm_primitive_move_to_fdes
{
  reference_parameter (port);
  reference_parameter (fd);

  scm_t_fport *stream;
  int old_fd;
  int new_fd;
  int rv;

  SCM_VALIDATE_OPFPORT (1, port);
  stream = SCM_FSTREAM (port);
  old_fd = stream->fdes;
  new_fd = scm_to_int (fd);
  if  (old_fd == new_fd) {
    dereference_parameter (port);
    dereference_parameter (fd);
    return SCM_BOOL_F;
  }
  rv = dup2 (old_fd, new_fd);
  if (rv == -1)
    SCM_SYSERROR;
  stream->fdes = new_fd;
  SCM_SYSCALL (close (old_fd));  
  dereference_parameter (port);
  dereference_parameter (fd);
  return SCM_BOOL_T;
}
#undef FUNC_NAME

/* Return a list of ports using a given file descriptor.  */
SCM_DEFINE (scm_fdes_to_ports, "fdes->ports", 1, 0, 0, 
           (SCM fd),
	    "Return a list of existing ports which have @var{fdes} as an\n"
	    "underlying file descriptor, without changing their revealed\n"
	    "counts.")
#define FUNC_NAME s_scm_fdes_to_ports
{
  reference_parameter (fd);

  SCM result = SCM_EOL;
  int int_fd;
  long i;

  int_fd = scm_to_int (fd);

  for (i = 0; i < scm_i_port_table_size; i++) {
    if (SCM_OPFPORTP (scm_i_port_table[i]->port)
        && ((scm_t_fport *) scm_i_port_table[i]->stream)->fdes == int_fd)
      result = scm_cons (scm_i_port_table[i]->port, result);
  }
  dereference_parameter (fd);
  return result;
}
#undef FUNC_NAME    


void 
scm_init_ioext ()
{
#include "include/ioext.x"
}
