// Strings & Brass

#include "include/config.h"

#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"
#include "include/eval.h"
#include "include/chars.h"
#include "include/fports.h"
#include "include/root.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/vectors.h"

#include "include/validate.h"
#include "include/vports.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif



/* {Ports - soft ports}
 * 
 */


static scm_t_tag scm_tc16_sfport = 0;


static void
sf_flush (SCM port)
{
  reference_parameter (port);

  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  SCM stream = SCM_PACK (pt->stream);
  reference_parameter (stream);

  if (pt->write_pos > pt->write_buf) {
    /* write the byte. */
    rc_Void (scm_call_1 (TMP_SIMPLE_VECTOR_REF (stream, 0),
                         SCM_MAKE_CHAR (*pt->write_buf)));
    pt->write_pos = pt->write_buf;
  
    /* flush the output.  */
    {
      SCM f = TMP_SIMPLE_VECTOR_REF (stream, 2);

      if (scm_is_true (f))
        rc_Void (scm_call_0 (f));
    }
  }
  dereference_parameter (stream);
  dereference_parameter (port);
}

static void
sf_write (SCM port, const void *data, size_t size)
{
  reference_parameter (port);
  SCM p = SCM_PACK (SCM_STREAM (port));
  reference_parameter (p);

  rc_Void (scm_call_1 (TMP_SIMPLE_VECTOR_REF (p, 1),
                       scm_make_byte_string_with_copy ((char *) data, size)));
  dereference_parameter (p);
  dereference_parameter (port);
}

/* calling the flush proc (element 2) is in case old code needs it,
   but perhaps softports could the use port buffer in the same way as
   fports.  */

/* places a single char in the input buffer.  */
static int 
sf_fill_input (SCM port)
{
  reference_parameter (port);

  SCM p = SCM_PACK (SCM_STREAM (port));
  reference_parameter (p);
  SCM answer;

  answer = scm_call_0 (TMP_SIMPLE_VECTOR_REF (p, 3)); /* get char.  */
  reference_parameter (answer);
  if (scm_is_false (answer) || SCM_EOF_OBJECT_P (answer)) {
    dereference_parameter (answer);
    dereference_parameter (port);
    dereference_parameter (p);
    return EOF;
  }
  SCM_ASSERT (SCM_CHARP (answer), answer, SCM_ARG1, "sf_fill_input");
  {
    scm_t_port *pt = SCM_PTAB_ENTRY (port);    

    *pt->read_buf = SCM_CHAR (answer);
    pt->read_pos = pt->read_buf;
    pt->read_end = pt->read_buf + 1;

    dereference_parameter (p);
    dereference_parameter (answer);
    dereference_parameter (port);
    return *pt->read_buf;
  }
}


static int 
sf_close (SCM port)
{
  reference_parameter (port);
  SCM p = SCM_PACK (SCM_STREAM (port));
  reference_parameter (p);
  SCM f = TMP_SIMPLE_VECTOR_REF (p, 4);
  reference_parameter (f);
  if (scm_is_false (f))
    return 0;
  SCM answer = scm_call_0 (f);
  reference_count_return (answer);
  errno = 0;
  dereference_parameter (f);
  dereference_parameter (p);
  dereference_parameter (port);
  return sharp_scm_is_false (answer) ? EOF : 0;
}


static int 
sf_input_waiting (SCM port)
{
  reference_parameter (port);
  SCM p = SCM_PACK (SCM_STREAM (port));
  reference_parameter (p);
  if (SCM_SIMPLE_VECTOR_LENGTH (p) >= 6)
    {
      SCM f = TMP_SIMPLE_VECTOR_REF (p, 5);
      reference_parameter (f);
      if (scm_is_true (f)) {
        int answer = scm_to_int (scm_call_0 (f));
        dereference_parameter (f);
        dereference_parameter (p);
        dereference_parameter (port);
        return answer;
      }
    }
  /* Default is such that char-ready? for soft ports returns #t, as it
     did before this extension was implemented. */
  return 1;
}



SCM_DEFINE (scm_make_soft_port, "make-soft-port", 2, 0, 0,
           (SCM pv, SCM modes),
	    "Return a port capable of receiving or delivering characters as\n"
	    "specified by the @var{modes} string (@pxref{File Ports,\n"
	    "open-file}).  @var{pv} must be a vector of length 5 or 6.  Its\n"
	    "components are as follows:\n"
	    "\n"
	    "@enumerate 0\n"
	    "@item\n"
	    "procedure accepting one character for output\n"
	    "@item\n"
	    "procedure accepting a string for output\n"
	    "@item\n"
	    "thunk for flushing output\n"
	    "@item\n"
	    "thunk for getting one character\n"
	    "@item\n"
	    "thunk for closing port (not by garbage collection)\n"
	    "@item\n"
	    "(if present and not @code{#f}) thunk for computing the number of\n"
	    "characters that can be read from the port without blocking.\n"
	    "@end enumerate\n"
	    "\n"
	    "For an output-only port only elements 0, 1, 2, and 4 need be\n"
	    "procedures.  For an input-only port only elements 3 and 4 need\n"
	    "be procedures.  Thunks 2 and 4 can instead be @code{#f} if\n"
	    "there is no useful operation for them to perform.\n"
	    "\n"
	    "If thunk 3 returns @code{#f} or an @code{eof-object}\n"
	    "(@pxref{Input, eof-object?, ,r5rs, The Revised^5 Report on\n"
	    "Scheme}) it indicates that the port has reached end-of-file.\n"
	    "For example:\n"
	    "\n"
	    "@lisp\n"
	    "(define stdout (current-output-port))\n"
	    "(define p (make-soft-port\n"
	    "           (vector\n"
	    "            (lambda (c) (write c stdout))\n"
	    "            (lambda (s) (display s stdout))\n"
	    "            (lambda () (display \".\" stdout))\n"
	    "            (lambda () (char-upcase (read-char)))\n"
	    "            (lambda () (display \"@@\" stdout)))\n"
	    "           \"rw\"))\n"
	    "\n"
	    "(write p p) @result{} #<input-output: soft 8081e20>\n"
	    "@end lisp")
#define FUNC_NAME s_scm_make_soft_port
{
  reference_parameter (pv);
  reference_parameter (modes);

  int vlen;
  scm_t_port *pt;
  SCM z;

  SCM_VALIDATE_VECTOR (1, pv);
  vlen = SCM_SIMPLE_VECTOR_LENGTH (pv);
  SCM_ASSERT ((vlen == 5) || (vlen == 6), pv, 1, FUNC_NAME);
  SCM_VALIDATE_BRASS (2, modes);
  
  z = scm_new_port_table_entry (scm_tc16_sfport);
  pt = SCM_PTAB_ENTRY (z);
  scm_port_non_buffer (pt);
  SCM_SET_CELL_WORD_0 (z, scm_i_mode_bits (modes)); // flags

  reference_parameter (z);
  SCM_SETSTREAM (z, SCM_UNPACK (pv));
  reference (z, pv);

  dereference_parameter (pv);
  dereference_parameter (modes);
  return (rc_return_stack_reference (z));
}
#undef FUNC_NAME

static void
vport_release (SCM vport)
{
  dereference (vport, SCM_PACK (SCM_STREAM (vport)));
  SCM_SETSTREAM (vport, SCM_UNPACK (SCM_UNSPECIFIED));
  return;
}

static scm_t_tag
scm_make_sfptob ()
{
  scm_t_tag tc = scm_make_port_type ("soft", sf_fill_input, sf_write);

  scm_set_port_release (tc, vport_release);
  scm_set_port_flush (tc, sf_flush);
  scm_set_port_close (tc, sf_close);
  scm_set_port_input_waiting (tc, sf_input_waiting);

  return tc;
}

void
scm_init_vports ()
{
  scm_tc16_sfport = scm_make_sfptob ();
}

void
x_inclusion_vports ()
{
#include "include/vports.x"
}
