// Strings & Brass

#include "include/config.h"

#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"
#include "include/dynwind.h"
#include "include/pairs.h"
#include "include/brass.h"
#include "include/throw.h"
#include "include/trim.h"

#include "include/validate.h"
#include "include/error.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif



/* {Errors and Exceptional Conditions}
 */


/* C interface to scm_error_scm.  */
void
scm_error (SCM key, const char *subr, const char *message, SCM args, SCM rest)
{
  if (global_debug_interpreter) {
    // No SCM routines here.
    fprintf (stderr, "VSI-C error routine: subr: %s message: %s",
                     subr, message);
    // abort ();
  }
  scm_error_scm 
    (key,
     (subr == NULL) ? SCM_BOOL_F : scm_from_cen (subr),
     (message == NULL) ? SCM_BOOL_F : scm_from_cen (message),
     args, rest);
}

/* All errors should pass through here.  */
SCM_DEFINE (scm_error_scm, "scm-error", 5, 0, 0, 
           (SCM key, SCM subr, SCM message, SCM args, SCM data),
	    "Raise an error with key @var{key}.  @var{subr} can be a string\n"
	    "naming the procedure associated with the error, or @code{#f}.\n"
	    "@var{message} is the error message string, possibly containing\n"
	    "@code{~S} and @code{~A} escapes.  When an error is reported,\n"
	    "these are replaced by formatting the corresponding members of\n"
	    "@var{args}: @code{~A} (was @code{%s} in older versions of\n"
	    "Guile) formats using @code{display} and @code{~S} (was\n"
	    "@code{%S}) formats using @code{write}.  @var{data} is a list or\n"
	    "@code{#f} depending on @var{key}: if @var{key} is\n"
	    "@code{system-error} then it should be a list containing the\n"
	    "Unix @code{errno} value; If @var{key} is @code{signal} then it\n"
	    "should be a list containing the Unix signal number; If\n"
	    "@var{key} is @code{out-of-range} or @code{wrong-type-arg},\n"
            "it is a list containing the bad value; otherwise\n"
	    "it will usually be @code{#f}.")
#define FUNC_NAME s_scm_error_scm
{
  scm_ithrow (key, scm_list_4 (subr, message, args, data), 1);
  
  /* No return, but just in case: */
  fprintf (stderr, "VSI-C: error: scm_ithrow returned!\n");
  exit (1);
}
#undef FUNC_NAME

#define SCM_I_STRERROR(err) strerror (err)
#define SCM_I_ERRNO() errno

SCM_DEFINE (scm_strerror, "strerror", 1, 0, 0, 
            (SCM err),
	    "Return the Unix error message corresponding to @var{err}, which\n"
	    "must be an integer value.")
#define FUNC_NAME s_scm_strerror
{
  reference_parameter (err);

  SCM ret;
  char * err_str = SCM_I_STRERROR (scm_to_int (err));

  ret = scm_from_sen (err_str, strlen (err_str));
  reference_count_return (ret);

  dereference_parameter (err);
  return (ret);
}
#undef FUNC_NAME

SCM_GLOBAL_SYMBOL (scm_system_error_key, "system-error");
void
scm_syserror (const char *subr)
{
  SCM err = scm_from_int (SCM_I_ERRNO ());
  scm_error (scm_system_error_key,
	     subr,
	     "~A",
	     scm_cons (scm_strerror (err), SCM_EOL),
	     scm_cons (err, SCM_EOL));
}

void
scm_syserror_msg (const char *subr, const char *message, SCM args, int eno)
{
  scm_error (scm_system_error_key,
	     subr,
	     message,
	     args,
	     scm_cons (scm_from_int (eno), SCM_EOL));
}

SCM_GLOBAL_SYMBOL (scm_num_overflow_key, "numerical-overflow");
void
scm_num_overflow (const char *subr)
{
  scm_error (scm_num_overflow_key,
	     subr,
	     "Numerical overflow",
	     SCM_BOOL_F,
	     SCM_BOOL_F);
}

SCM_GLOBAL_SYMBOL (scm_out_of_range_key, "out-of-range");
void
scm_out_of_range (const char *subr, SCM bad_value)
{
  scm_error (scm_out_of_range_key,
	     subr,
	     "Value out of range: ~S",
             scm_list_1 (bad_value),
             scm_list_1 (bad_value));
}

void
scm_out_of_range_pos (const char *subr, SCM bad_value, SCM pos)
{
  scm_error (scm_out_of_range_key,
	     subr,
	     "Argument ~A out of range: ~S",
             scm_list_2 (pos, bad_value),
	     scm_list_1 (bad_value));
}


SCM_GLOBAL_SYMBOL (scm_args_number_key, "wrong-number-of-args");
void
scm_wrong_num_args (SCM proc)
{
  scm_error (scm_args_number_key,
	     NULL,
	     "Wrong number of arguments to ~A",
	     scm_list_1 (proc),
	     SCM_BOOL_F);
}


void
scm_error_num_args_subr (const char *subr)
{
  scm_error (scm_args_number_key,
	     NULL,
	     "Wrong number of arguments to ~A",
	     scm_list_1 (scm_from_cen (subr)),
	     SCM_BOOL_F);
}


SCM_GLOBAL_SYMBOL (scm_arg_type_key, "wrong-type-arg");
void
scm_wrong_type_arg (const char *subr, int pos, SCM bad_value)
{
  scm_error (scm_arg_type_key,
	     subr,
	     (pos == 0) ? "Wrong type: ~S"
	     : "Wrong type argument in position ~A: ~S",
	     (pos == 0) ? scm_list_1 (bad_value)
	     : scm_list_2 (scm_from_int (pos), bad_value),
	     scm_list_1 (bad_value));
}

void
scm_wrong_type_arg_msg (const char *subr, int pos,
                        SCM bad_value, const char *szMessage)
{
  SCM msg = scm_from_cen (szMessage);

  if (pos == 0) {
    scm_error (scm_arg_type_key,
               subr, "Wrong type (expecting ~A): ~S",
               scm_list_2 (msg, bad_value),
               scm_list_1 (bad_value));
  } else {
    scm_error (scm_arg_type_key,
               subr,
               "Wrong type argument in position ~A (expecting ~A): ~S",
               scm_list_3 (scm_from_int (pos), msg, bad_value),
               scm_list_1 (bad_value));
  }
}


// XXX this duplicates out-of-memory which should be replaced with this key
SCM_GLOBAL_SYMBOL (scm_memory_alloc_key, "memory-allocation-error");
void
scm_memory_error (const char *subr)
{
  fprintf (stderr, " * VSI-C FATAL: memory error in %s *\n", subr);
  abort ();
}

SCM_GLOBAL_SYMBOL (scm_misc_error_key, "misc-error");
void
scm_misc_error (const char *subr, const char *message, SCM args)
{
  scm_error (scm_misc_error_key, subr, message, args, SCM_BOOL_F);
}

SCM_GLOBAL_SYMBOL (scm_out_of_memory_key, "out-of-memory");
void
scm_out_of_memory ()
{
  scm_error (scm_out_of_memory_key,
	     NULL,
             "memory backup segment has been opened - trying to continue",
             SCM_BOOL_F,
             SCM_BOOL_F);
}

void
scm_encoding_error (const char *details) // TODO
{
  scm_error (scm_system_error_key,
	     NULL,
	     details,
	     SCM_BOOL_F,
             SCM_BOOL_F);
}

void
scm_i_display_error (SCM unused, SCM port, SCM subr,
                     SCM message, SCM args, SCM rest)
{
  reference_parameter (unused);
  reference_parameter (port);
  reference_parameter (subr);
  reference_parameter (message);
  reference_parameter (args);
  reference_parameter (rest);

  scm_puts    ("VSI: ",    port);
  scm_simple_format (port, message, args);

  scm_puts    (" (Issued subsystem: ",    port);
  scm_display (subr,           port);
  scm_puts    (")", port);

  /* scm_puts    (" | rest: ",    port); */
  /* scm_display (rest,           port); */

  scm_newline (      port         );

  dereference_parameter (unused);
  dereference_parameter (port);
  dereference_parameter (subr);
  dereference_parameter (message);
  dereference_parameter (args);
  dereference_parameter (rest);
  return;
}

void
scm_init_error ()
{
#include "./posix-cpp_err_symbols.c"
#include "include/error.x"
}
