// Strings & Brass

#include "include/config.h"

#include "include/_scm.h"
#include "include/eval.h"
#include "include/list.h"
#include "include/throw.h"
#include "include/macros.h"
#include "include/smob.h"
#include "include/procprop.h"
#include "include/alist.h"
#include "include/strports.h"
#include "include/read.h"
#include "include/feature.h"
#include "include/dynwind.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/brass.h"

#include "include/validate.h"
#include "include/debug.h"


SCM_SYMBOL (scm_sym_procname, "procname");
SCM_SYMBOL (scm_sym_dots, "...");
SCM_SYMBOL (scm_sym_source, "source");


#include "include/variable.h"
#include "include/procs.h"

#include <string.h>  // strlen


SCM
scm_reverse_lookup (SCM env, SCM data)
{
  while (scm_is_pair (env) && scm_is_pair (TMP_CAR (env)))
    {
      SCM names = TMP_CAAR (env);
      SCM values = TMP_CDAR (env);
      while (scm_is_pair (names))
	{
	  if (scm_is_eq (TMP_CAR (values), data))
	    return SCM_CAR (names);
	  names = TMP_CDR (names);
	  values = TMP_CDR (values);
	}
      if (!scm_is_null (names) && scm_is_eq (values, data))
	return names;
      env = TMP_CDR (env);
    }
  return SCM_BOOL_F;
}

SCM_DEFINE (scm_procedure_name, "procedure-name", 1, 0, 0, 
            (SCM proc),
	    "Return the name of the procedure @var{proc}")
#define FUNC_NAME s_scm_procedure_name
{
  reference_parameter (proc);

  SCM_VALIDATE_PROC (1, proc);
  switch (SCM_TYP7 (proc)) {
  case scm_tcs_subrs:
    return SCM_SNAME (proc);
  default:
    {
      // XXX defunct; see interpreter-nature.c
      SCM name = scm_reverse_lookup (SCM_ENV (proc), proc);
      reference_parameter (name);
      dereference_parameter (proc);
      return (rc_return_stack_reference (name));
    }
  }
}
#undef FUNC_NAME

SCM_DEFINE (scm_procedure_source, "procedure-source", 1, 0, 0, 
            (SCM proc),
	    "Return the source of the procedure @var{proc}.")
#define FUNC_NAME s_scm_procedure_source
{
  reference_parameter (proc);
  SCM answer = SCM_BOOL_F;

  SCM_VALIDATE_NIM (1, proc);
 again:
  switch (SCM_TYP7 (proc)) {
  case scm_tcs_closures:
    {
      const SCM formals = SCM_CLOSURE_FORMALS (proc);
      const SCM body = SCM_CLOSURE_BODY (proc);
      // const SCM src = scm_source_property (body, scm_sym_copy);

      /* defunct ??? */
      /* if (scm_is_true (src)) {  // redundant? */
      /*   answer = scm_cons2 (scm_sym_lambda, formals, src); */
      /*   reference_count_return (answer); */
      /*   dereference_parameter (proc); */
      /*   return (answer); */

      /* } else { */
        const SCM env = SCM_EXTEND_ENV (formals, SCM_EOL, SCM_ENV (proc));
        answer = scm_cons2 (scm_sym_lambda,
                            scm_i_finite_list_copy (formals),
                            scm_i_unmemocopy_body (body, env));
        reference_count_return (answer);
        dereference_parameter (proc);
        return (answer);
      /* } */
    }
  case scm_tcs_subrs:
  procprop:
    answer = scm_procedure_property (proc, scm_sym_source);
    reference_count_return (answer);
    dereference_parameter (proc);
    return (answer);
  default:
    ;
  }
  SCM_WRONG_TYPE_ARG (1, proc);
  dereference_parameter (proc);
  return SCM_BOOL_F; /* not reached */
}
#undef FUNC_NAME

SCM_DEFINE (scm_procedure_code, "procedure-code", 1, 0, 0, (SCM proc), "")
#define FUNC_NAME s_scm_procedure_code
{
  reference_parameter (proc);
  SCM answer = SCM_BOOL_F;

  SCM_VALIDATE_NIM (1, proc);
 again:
  switch (SCM_TYP7 (proc)) {
  case scm_tcs_closures:
    {
      const SCM formals = SCM_CLOSURE_FORMALS (proc);
      const SCM body = SCM_CLOSURE_BODY (proc);
        answer = scm_cons2 (scm_sym_lambda, formals, body);
        reference_count_return (answer);
        dereference_parameter (proc);
        return (answer);
    }
  case scm_tcs_subrs:
  procprop:
    dereference_parameter (proc);
    return (SCM_BOOL_F);
  default:
    ;
  }
  SCM_WRONG_TYPE_ARG (1, proc);
  dereference_parameter (proc);
  return SCM_BOOL_F; /* not reached */
}
#undef FUNC_NAME

SCM_DEFINE (scm_procedure_environment, "procedure-environment", 1, 0, 0, 
            (SCM proc),
	    "Return the environment of the procedure @var{proc}.")
#define FUNC_NAME s_scm_procedure_environment
{
  reference_parameter (proc);
  SCM_VALIDATE_NIM (1, proc);

  SCM answer = SCM_UNDEFINED;
  switch (SCM_TYP7 (proc)) {
  case scm_tcs_closures:
    answer = SCM_ENV (proc); break;
  case scm_tcs_subrs:
    answer = SCM_EOL; break;
  default:
    SCM_WRONG_TYPE_ARG (1, proc);
    /* not reached */
  }
  reference_count_return (answer);
  dereference_parameter (proc);
  return answer;
}
#undef FUNC_NAME



/*
 * Statistics
 *
 * TODO This should be in debug-sysinfo.c but x_inclusion and
 * corresponding SCM_DEFINE statements need to reside in the same
 * file.
 */

#include "include/ihub.h"

static SCM scm_global_smob_counts ()
{
  long i; 
  SCM stats = SCM_EOL;

  /* The tag type of smob number zero is used for free cells. */
  for (i = scm_numsmob - 1; i >= 1; i--) {
    /* This currently recreates these strings every time this is called. */
    SCM scm_name = 0;
    SCM count    = scm_from_long (scm_smobs[i].number_of_objects);

    if (scm_smobs[i].name != NULL)
      scm_name = scm_from_sen (scm_smobs[i].name, strlen (scm_smobs[i].name));
    else
      scm_name = scm_from_long (i);

    stats = scm_acons (scm_name, count, stats);
  }

  return stats;
}

SCM_DEFINE (scm_global_object_counts, "%global-object-counts", 0, 0, 0,
            (),
            "Returns the numbers of memory objects.")
{
  return scm_list_n (
     scm_from_ulong (global_cell_segments),
     scm_from_ulong (global_double_cell_segments),

     scm_from_long (global_number_of_pairs),

     scm_from_long (global_number_of_vectors),
     scm_from_long (global_number_of_brass),
     scm_from_long (global_number_of_variables),
     scm_from_ulong (global_number_of_symbols_created),

     scm_from_long (global_number_of_closures),
     scm_from_long (global_number_of_byte_strings),
     scm_from_long (global_number_of_byte_string_mappings),

     scm_from_long (global_number_of_ports),

     scm_from_long (global_number_of_reals),
     scm_from_long (global_number_of_complex),
     scm_from_long (global_number_of_bignums),
     scm_from_long (global_number_of_fractions),

     scm_global_smob_counts (),

     SCM_UNDEFINED /* end of arguments (protocol) */
     );
}

/* Color Code Profiling and Call Counters */
long global_color_counter_green  = 0;
long global_color_counter_red    = 0;
long global_color_counter_blu    = 0;
long global_color_counter_violet = 0;
long global_color_counter_yellow = 0;

SCM_DEFINE (scm_global_call_counters, "%global-call-counters", 0, 0, 0,
            (),
            "Returns the current values of the call counters.")
{
  return scm_list_n (
     scm_from_ulong (global_cells_requested),
     scm_from_ulong (global_double_cells_requested),
     scm_from_ulong (global_counter_reinsertions),
     scm_from_ulong (global_counter_drops),
     scm_from_long (global_color_counter_green),
     scm_from_long (global_color_counter_blu),
     scm_from_long (global_color_counter_red),
     scm_from_long (global_color_counter_violet),
     scm_from_long (global_color_counter_yellow),
     scm_from_ulong (global_counter_non_phalanx),
     scm_from_ulong (global_counter_objprop_fuel),
     scm_from_ulong (global_counter_procprop_fuel),
     scm_from_ulong (global_counter_debug_fuel),
     scm_from_ulong (0L),
     scm_from_ulong (global_counter_dynwind_fuel),
     scm_from_ulong (0L),
     scm_from_ulong (global_counter_ceval_fuel),

     SCM_UNDEFINED /* end of arguments (protocol) */
     );
}



void
scm_init_debug ()
{
}

void
x_inclusion_debug ()
{
  scm_add_feature ("debug-extensions"); /* XXX */

  scm_c_define ("SCM_IM_AND",    SCM_IM_AND);
  scm_c_define ("SCM_IM_BEGIN",  SCM_IM_BEGIN);
  scm_c_define ("SCM_IM_CASE",   SCM_IM_CASE);
  scm_c_define ("SCM_IM_COND",   SCM_IM_COND);
  scm_c_define ("SCM_IM_DO",     SCM_IM_DO);
  scm_c_define ("SCM_IM_IF",     SCM_IM_IF);
  scm_c_define ("SCM_IM_LAMBDA", SCM_IM_LAMBDA);
  scm_c_define ("SCM_IM_LET",    SCM_IM_LET);
  scm_c_define ("SCM_IM_LETSTAR", SCM_IM_LETSTAR);
  scm_c_define ("SCM_IM_LETREC", SCM_IM_LETREC);
  scm_c_define ("SCM_IM_OR",     SCM_IM_OR);
  scm_c_define ("SCM_IM_QUOTE",  SCM_IM_QUOTE);
  scm_c_define ("SCM_IM_SET_X",  SCM_IM_SET_X);
  scm_c_define ("SCM_IM_DEFINE", SCM_IM_DEFINE);
  scm_c_define ("SCM_IM_APPLY",  SCM_IM_APPLY);
  scm_c_define ("SCM_IM_CONT",   SCM_IM_CONT);
  scm_c_define ("SCM_IM_DISPATCH", SCM_IM_DISPATCH);

#include "include/debug.x"
}
