/* classes: h_files */

#ifndef SCM_RANDOM_H
#define SCM_RANDOM_H
#include "include/__scm.h"


/*
 * A plugin interface for RNGs
 *
 * Using this interface, it is possible for the application to tell
 * VSI to use a different RNG.  This is desirable if it is
 * necessary to use the same RNG everywhere in the application in
 * order to prevent interference, if the application uses RNG
 * hardware, or if the application has special demands on the RNG.
 *
 * Look how the default generator is "plugged in" in scm_init_random().
 */

typedef struct scm_t_rstate {
  int reserved0;
  double reserved1;
  /* Custom fields follow here */
} scm_t_rstate;

typedef struct scm_t_rng {
  size_t rstate_size;				    /* size of random state */
  /* Though this returns an unsigned long, it's only 32 bits of randomness. */
  unsigned long (*random_bits) (scm_t_rstate *state); /* gives 32 random bits */
  void (*init_rstate) (scm_t_rstate *state, const char *seed, int n);
  scm_t_rstate *(*copy_rstate) (scm_t_rstate *state);
} scm_t_rng;

scm_t_rng scm_the_rng;


/*
 * Default RNG
 */
typedef struct scm_t_i_rstate {
  scm_t_rstate rstate;
  unsigned long w;
  unsigned long c;
} scm_t_i_rstate;

/* Though this returns an unsigned long, it's only 32 bits of randomness. */
unsigned long scm_i_uniform32 (scm_t_i_rstate *);
void scm_i_init_rstate (scm_t_i_rstate *, const char *seed, int n);
scm_t_i_rstate *scm_i_copy_rstate (scm_t_i_rstate *);


/*
 * Random number library functions
 */
scm_t_rstate *scm_c_make_rstate (const char *, int);
scm_t_rstate *scm_c_default_rstate (void);
#define scm_c_uniform32(RSTATE) scm_the_rng.random_bits (RSTATE)
double scm_c_uniform01 (scm_t_rstate *);
double scm_c_normal01 (scm_t_rstate *);
double scm_c_exp1 (scm_t_rstate *);
/* Though this returns an unsigned long, it's only 32 bits of randomness. */
unsigned long scm_c_random (scm_t_rstate *, unsigned long m);
/* This one returns 64 bits of randomness. */
scm_t_uint64 scm_c_random64 (scm_t_rstate *state, scm_t_uint64 m);
SCM scm_c_random_bignum (scm_t_rstate *, SCM m);


/*
 * Scheme level interface
 */
scm_t_tag scm_tc16_rstate;
#define SCM_RSTATEP(obj) SCM_SMOB_PREDICATE (scm_tc16_rstate, obj)
#define SCM_RSTATE(obj)  ((scm_t_rstate *) SCM_SMOB_DATA (obj))

unsigned char scm_masktab[256];

SCM scm_var_random_state;
SCM scm_random (SCM n, SCM state);
SCM scm_copy_random_state (SCM state);
SCM scm_seed_to_random_state (SCM seed);
SCM scm_random_uniform (SCM state);
SCM scm_random_normal (SCM state);
SCM scm_random_normal_vector_x (SCM v, SCM state);
SCM scm_random_exp (SCM state);
void scm_init_random (void);
void x_inclusion_random (void);

#endif  /* SCM_RANDOM_H */
