/* classes: h_files */

#ifndef SCM_TAGS_H
#define SCM_TAGS_H
/** This file defines the format of SCM values and cons pairs.
 ** It is here that tag bits are assigned for various purposes.
 **/

/* picks up scmconfig.h too */
#include "include/__scm.h"

#if HAVE_INTTYPES_H
# include <inttypes.h>  /* for INTPTR_MAX and friends */
#else
# if HAVE_STDINT_H
#  include <stdint.h>   /* for INTPTR_MAX and friends */
# endif
#endif



/*
 *
 * For the representation of scheme objects and their handling, VSI-C provides
 * two types:  scm_t_bits and SCM.
 *
 * - scm_t_bits values can hold bit patterns of non-objects and objects:
 *
 *   Non-objects -- in this case the value may not be changed into a SCM value
 *   in any way.
 *
 *   Objects -- in this case the value may be changed into a SCM value using
 *   the SCM_PACK macro.
 *
 * - SCM values can hold proper scheme objects only.  They can be changed into
 *   a scm_t_bits value using the SCM_UNPACK macro.
 *
 * When working in the domain of scm_t_bits values, programmers must keep
 * track of any scm_t_bits value they create that is not a proper scheme
 * object.  This makes sure that in the domain of SCM values developers can
 * rely on the fact that they are dealing with proper scheme objects only.
 * Thus, the distinction between scm_t_bits and SCM values helps to identify
 * those parts of the code where special care has to be taken to not create
 * bad SCM values.
 */

//* * For dealing with the bit level representation of scheme objects we define */
/*  * scm_t_bits: */
/*  *\/ */
/* /\* On Solaris 7 and 8, /usr/include/sys/int_limits.h defines */
/*    INTPTR_MAX and UINTPTR_MAX to empty, INTPTR_MIN is not defined. */
/*    To avoid uintptr_t and intptr_t in this case we require */
/*    UINTPTR_MAX-0 != 0 etc.  *\/ */
/* #if SCM_SIZEOF_INTPTR_T != 0 && defined(INTPTR_MAX) && defined(INTPTR_MIN) \ */
/*   && INTPTR_MAX-0 != 0 && INTPTR_MIN-0 != 0 \ */
/*   && SCM_SIZEOF_UINTPTR_T != 0 && defined(UINTPTR_MAX) && UINTPTR_MAX-0 != 0 */

/* typedef intptr_t scm_t_signed_bits; */
/* #define SCM_T_SIGNED_BITS_MAX INTPTR_MAX */
/* #define SCM_T_SIGNED_BITS_MIN INTPTR_MIN */
/* typedef uintptr_t scm_t_bits; */
/* #define SIZEOF_SCM_T_BITS SCM_SIZEOF_UINTPTR_T */
/* #define SCM_T_BITS_MAX UINTPTR_MAX */

/* #else */

/* typedef signed long scm_t_signed_bits; */
/* #define SCM_T_SIGNED_BITS_MAX LONG_MAX */
/* #define SCM_T_SIGNED_BITS_MIN LONG_MIN */
/* typedef unsigned long scm_t_bits; */
/* #define SIZEOF_SCM_T_BITS SCM_SIZEOF_UNSIGNED_LONG */
/* #define SCM_T_BITS_MAX ULONG_MAX */
/* typedef unsigned int scm_t_tag; */

/* #endif */

/* /\* But as external interface, we define SCM, which may, according to the */
/*  * desired level of type checking, be defined in several ways: */
/*  *\/ */
/* #if (SCM_DEBUG_TYPING_STRICTNESS == 2) */
/*     typedef union { struct { scm_t_bits n; } n; } SCM; */
/*     static SCM scm_pack(scm_t_bits b) { SCM s; s.n.n = b; return s; } */
/* #   define SCM_UNPACK(x) ((x).n.n) */
/* #   define SCM_PACK(x) (scm_pack ((scm_t_bits) (x))) */
/* #elif (SCM_DEBUG_TYPING_STRICTNESS == 1) */
/* /\* This is the default, which provides an intermediate level of compile time */
/*  * type checking while still resulting in very efficient code. */
/*  *\/ */
/*    typedef struct scm_unused_struct * SCM; */


/* #   define SCM_UNPACK(x) ((scm_t_bits) (x)) */

/* /\* */
/*   There is no typechecking on SCM_PACK, since all kinds of types */
/*   (unsigned long, void*) go in SCM_PACK */
/*  *\/ */
/* #   define SCM_PACK(x) ((SCM) (x)) */

/* #else */
/* /\* This should be used as a fall back solution for machines on which casting */
/*  * to a pointer may lead to loss of bit information, e. g. in the three least */
/*  * significant bits. */
/*  *\/ */
/*     typedef scm_t_bits SCM; */
/* #   define SCM_UNPACK(x) (x) */
/* #   define SCM_PACK(x) ((SCM) (x)) */
/* #endif */

typedef uint32_t usv_t;

#define SCM_T_SIGNED_BITS_MAX LONG_MAX
#define SCM_T_SIGNED_BITS_MIN LONG_MIN
typedef signed   long scm_t_signed_bits;

typedef unsigned long scm_t_bits;
#define SIZEOF_SCM_T_BITS SCM_SIZEOF_UNSIGNED_LONG
#define SCM_T_BITS_MAX ULONG_MAX
typedef unsigned long scm_t_bits;

typedef unsigned long scm_t_tag;
typedef signed   long scm_t_refcount;
typedef struct scm_i_cell_head { scm_t_tag tag;
                                 scm_t_refcount reference_count;
                                 scm_t_bits a; } scm_t;
typedef scm_t_bits SCM;
// typedef void* SCM;
#define SCM2PTR(x)  ((void*) x)
#define PTR2SCM(x)  ((SCM) x)
#define SCM_UNPACK(x) ((scm_t_bits) x)
#define SCM_PACK(x)  ((SCM) x)


/* SCM values can not be compared by using the operator ==.  Use the following
 * macro instead, which is the equivalent of the scheme predicate 'eq?'.
 */
#define scm_is_eq(x, y) (SCM_UNPACK (x) == SCM_UNPACK (y))


/*
 * Summary of type code abbreviations
 *
 * tc2 b1 (= tc3)
 * 00  0: pointer to a cell (scm_tc3_cons)
 * scm_tc2_int: only 2 bits for the tag and two tc3 tags (see also numbers.h)
 * 01  0: immediate (scm_tc3_int_1)
 * 11  0: immediate (scm_tc3_int_2)
 * 10  0: immediate (scm_tc3_imm24, scm_tc_8; 24=32-8; 64-8=56)
 * 00  1: currently unused (was tc3_struct)
 * 01  1: tc3_closure
 * 10  1: tag tc7
 * 11  1: tag tc7
 *
 * tc7 (non immediates):
 *   See below for the list of types.
 *
 * the eigth bit of non immediates is currently unused
 * tc8 (immediates): (See scm_tc8_tags)
 * 00000 100: special objects, flags
 * 00001 100: characters
 * 00010 100: immediate symbols
 * 00011 100: ilocs (internal references into environments)
 *
 * tc16 (for tc7==scm_tc7_smob):
 *   There is one smob type, namely scm_tc_free_cell, for which it assumes
 *   that it is declared first and thus gets a known-in-advance tc16-code.
 *   Numbers and ports have other distinct tc16 ranges.
 */

/*
 * Viper System Interface: System Information (debug-sysinfo.c)
 *
 * * Data Type Bit Masks *
 *                    8 tc7  tc3
 *      scm_tc3_cons: 00 0000 000
 *   scm_tc3_closure: 00 0000 011
 * 
 *     scm_tc3_tc7_1: 00 0000 101
 *     scm_tc3_tc7_2: 00 0000 111
 * 
 *    scm_tc7_symbol: 00 0000 101
 *  scm_tc7_variable: 00 0000 111
 *    scm_tc7_vector: 00 0001 101
 *      scm_tc7_port: 00 1111 101
 * 
 * tc16:scm_tc7_smob: 00 1111 111
 */


/* Checking if an SCM variable holds an immediate or a non-immediate object. */
#define SCM_IMP(x) 		(6 & SCM_UNPACK (x))
#define SCM_NIMP(x) 		(!SCM_IMP (x))

/* Checking if a SCM variable holds an immediate integer: See numbers.h for
 * the definition of the following macros: SCM_I_FIXNUM_BIT,
 * SCM_MOST_POSITIVE_FIXNUM, SCM_I_INUMP, SCM_I_MAKINUM, SCM_I_INUM.  */

/*
 * XXX These C literal constants are treated as int?
 * We hope everyone did the casting correctly.
 */

/* A pair is a pointer to a non-immediate, i.e. cell containing an immediate
 * or another pointer. Other types based on cells need a type tag to be
 * recognizable as such.
 */
#define SCM_I_CONSP(x)  (!SCM_IMP (x) && ((1 & SCM_CELL_TYPE (x)) == 0))


/* Definitions for tc2: */

#define scm_tc2_int              2


/* Definitions for tc3: (these are bit masks, not numbers) */

#define SCM_ITAG3(x) 		 (7 & SCM_UNPACK (x))
#define SCM_TYP3(x) 		 (7 & SCM_CELL_TYPE (x))

#define scm_tc3_cons	 	 0
#define scm_tc3_unused    	 1  // unused
#define scm_tc3_int_1		 (scm_tc2_int + 0)
#define scm_tc3_closure		 3
#define scm_tc3_imm24		 4
#define scm_tc3_tc7_1		 5
#define scm_tc3_int_2		 (scm_tc2_int + 4)
#define scm_tc3_tc7_2		 7


/* Definitions for tc7: (these are bit masks, not numbers) */

#define SCM_ITAG7(x) 		(127 & SCM_UNPACK (x))
#define SCM_TYP7(x) 		(0x7f &        SCM_CELL_TYPE (x))

#define scm_tc7_symbol		5
#define scm_tc7_variable        7

#define scm_tc7_vector		13
#define scm_tc7_unused_s1       21
#define scm_tc7_number		23
#define scm_tc7_unused_s2       39

#define scm_tc7_brass           31

#define scm_tc7_byte_string     15   // 0b00001111
#define scm_tc16_byte_string_mapping   0b100001111
#define scm_tc7_unused_1        29
#define scm_tc7_unused_2	37
#define scm_tc7_unused_3	45
#define scm_tc7_unused_4	47
#define scm_tc7_unused_5	53
#define scm_tc7_unused_6	55
#define scm_tc7_unused_7	71
#define scm_tc7_unused_8	77
#define scm_tc7_unused_9	79

#define scm_tc7_dsubr		61  // floating point "primitives"
#define scm_tc7_subr_n		63  // n arguments
#define scm_tc7_rpsubr		69  // auto-folding (2 args like reduce ???)
#define scm_tc7_subr_0		85  // no arguments
#define scm_tc7_subr_1		87  // one argument
#define scm_tc7_cxr		93  // CXRs (cddar etc)
#define scm_tc7_subr_3		95  // three arguments
#define scm_tc7_subr_2		101 // two arguments
#define scm_tc7_asubr		103 // auto-folding of arguments: see scm_sum
#define scm_tc7_subr_1o		109 // one ?+? optional ???
#define scm_tc7_subr_2o		111 // two ?+? optional ???
#define scm_tc7_lsubr_2		117 // two arguments + one ?rest? argument
#define scm_tc7_lsubr		119 // one ?rest? argument

/* There are 256 port subtypes.  */
#define scm_tc7_port		125

/* There are 256 smob subtypes.  [**] If you change scm_tc7_smob, you must
 * also change the places it is hard coded in this file and possibly others.
 * Dirk:FIXME:: Any hard coded reference to scm_tc7_smob must be replaced by a
 * symbolic reference.  */
#define scm_tc7_smob		127 /* DO NOT CHANGE [**] */


/* Definitions for tc16: (the subtype numbers are assigned dynamically) */
#define SCM_TYP16(x) 		(0xffff & SCM_CELL_TYPE (x))
#define SCM_TYP16_PREDICATE(tag, x) (!SCM_IMP (x) && SCM_TYP16 (x) == (tag))


/* Here is the first smob subtype.  */

/* scm_tc_free_cell is the 0th smob type.  We place this in free cells to tell
 * the conservative marker not to trace it.  */
#define scm_tc_free_cell	(scm_tc7_smob + 0 * 256L)



/* {Immediate Values}
 */

enum scm_tc8_tags
{
  scm_tc8_flag = scm_tc3_imm24 + 0x00,  /* special objects ('flags') */
  scm_tc8_char = scm_tc3_imm24 + 0x08,  /* characters */
  scm_tc8_isym = scm_tc3_imm24 + 0x10,  /* immediate symbols ('isyms') */
  scm_tc8_iloc = scm_tc3_imm24 + 0x18   /* evaluator byte codes ('ilocs') */
};

#define SCM_ITAG8(X)		(SCM_UNPACK (X) & 0xff)
#define SCM_MAKE_ITAG8(X, TAG)	SCM_PACK (((X) << 8) + TAG)
// #define SCM_MAKE_ITAG8(X, TAG)	SCM_PACK (((X) << 8) | TAG)
#define SCM_ITAG8_DATA(X)	(SCM_UNPACK (X) >> 8)



/* Flags (special objects).  The indices of the flags must agree with the
 * declarations in print.c: iflagnames.  */

#define SCM_IFLAGP(n)    (SCM_ITAG8 (n) == scm_tc8_flag)
#define SCM_MAKIFLAG(n)  SCM_MAKE_ITAG8 ((n), scm_tc8_flag)
#define SCM_IFLAGNUM(n)  (SCM_ITAG8_DATA (n))

#define SCM_BOOL_F		SCM_MAKIFLAG (0)
#define SCM_BOOL_T 		SCM_MAKIFLAG (1)
#define SCM_UNDEFINED	 	SCM_MAKIFLAG (2)
#define SCM_EOF_VAL 		SCM_MAKIFLAG (3)
#define SCM_EOL			SCM_MAKIFLAG (4)
#define SCM_UNSPECIFIED		SCM_MAKIFLAG (5)

#define SCM_CUBIC		SCM_MAKIFLAG (6)  // unused

#define SCM_UNBNDP(x)		(scm_is_eq ((x), SCM_UNDEFINED))
#define SCM_UNDEFINED_P(x)      (scm_is_eq ((x), SCM_UNDEFINED))



/* Evaluator byte codes ('immediate symbols').  These constants are used only
 * in eval but their values have to be allocated here.  The indices of the
 * SCM_IM_ symbols must agree with the declarations in print.c:
 * scm_isymnames.  */

#define SCM_ISYMP(n) 		(SCM_ITAG8 (n) == scm_tc8_isym)
#define SCM_MAKISYM(n) 		SCM_MAKE_ITAG8 ((n), scm_tc8_isym)

#define SCM_IM_AND              SCM_MAKISYM (0)
#define SCM_IM_BEGIN            SCM_MAKISYM (1)
#define SCM_IM_CASE             SCM_MAKISYM (2)
#define SCM_IM_COND             SCM_MAKISYM (3)
#define SCM_IM_DO               SCM_MAKISYM (4) // unused
#define SCM_IM_IF               SCM_MAKISYM (5)
#define SCM_IM_LAMBDA           SCM_MAKISYM (6)
#define SCM_IM_LET              SCM_MAKISYM (7)
#define SCM_IM_LETSTAR          SCM_MAKISYM (8)
#define SCM_IM_LETREC           SCM_MAKISYM (9)
#define SCM_IM_OR               SCM_MAKISYM (10)
#define SCM_IM_QUOTE            SCM_MAKISYM (11)
#define SCM_IM_SET_X            SCM_MAKISYM (12)
#define SCM_IM_DEFINE           SCM_MAKISYM (13)
#define SCM_IM_APPLY		SCM_MAKISYM (14)
#define SCM_IM_CONT		SCM_MAKISYM (15) //unused,reserved:ShallowLambda
#define SCM_IM_DISPATCH		SCM_MAKISYM (16)
#define SCM_IM_SLOT_REF		SCM_MAKISYM (17) // unused, reserved
#define SCM_IM_SLOT_SET_X	SCM_MAKISYM (18) // unused, reserved
#define SCM_IM_DELAY		SCM_MAKISYM (19) // unused
#define SCM_IM_FUTURE		SCM_MAKISYM (20) // unused
#define SCM_IM_CALL_WITH_VALUES SCM_MAKISYM (21) // unused
#define SCM_IM_ELSE             SCM_MAKISYM (22)
#define SCM_IM_ARROW            SCM_MAKISYM (23)
#define SCM_IM_NIL_COND         SCM_MAKISYM (24)  /* Multi-language support */
// #define SCM_IM_BIND             SCM_MAKISYM (25)  /* Multi-language support */



/* Dispatching aids:

   When switching on SCM_TYP7 of a SCM value, use these fake case
   labels to catch types that use fewer than 7 bits for tagging.  */

/* For cons pairs with immediate values in the CAR
 */

#define scm_tcs_cons_imcar \
       scm_tc2_int + 0:   case scm_tc2_int + 4:   case scm_tc3_imm24 + 0:\
  case scm_tc2_int + 8:   case scm_tc2_int + 12:  case scm_tc3_imm24 + 8:\
  case scm_tc2_int + 16:  case scm_tc2_int + 20:  case scm_tc3_imm24 + 16:\
  case scm_tc2_int + 24:  case scm_tc2_int + 28:  case scm_tc3_imm24 + 24:\
  case scm_tc2_int + 32:  case scm_tc2_int + 36:  case scm_tc3_imm24 + 32:\
  case scm_tc2_int + 40:  case scm_tc2_int + 44:  case scm_tc3_imm24 + 40:\
  case scm_tc2_int + 48:  case scm_tc2_int + 52:  case scm_tc3_imm24 + 48:\
  case scm_tc2_int + 56:  case scm_tc2_int + 60:  case scm_tc3_imm24 + 56:\
  case scm_tc2_int + 64:  case scm_tc2_int + 68:  case scm_tc3_imm24 + 64:\
  case scm_tc2_int + 72:  case scm_tc2_int + 76:  case scm_tc3_imm24 + 72:\
  case scm_tc2_int + 80:  case scm_tc2_int + 84:  case scm_tc3_imm24 + 80:\
  case scm_tc2_int + 88:  case scm_tc2_int + 92:  case scm_tc3_imm24 + 88:\
  case scm_tc2_int + 96:  case scm_tc2_int + 100: case scm_tc3_imm24 + 96:\
  case scm_tc2_int + 104: case scm_tc2_int + 108: case scm_tc3_imm24 + 104:\
  case scm_tc2_int + 112: case scm_tc2_int + 116: case scm_tc3_imm24 + 112:\
  case scm_tc2_int + 120: case scm_tc2_int + 124: case scm_tc3_imm24 + 120

/* For cons pairs with non-immediate values in the SCM_CAR
 */
#define scm_tcs_cons_nimcar \
       scm_tc3_cons + 0:\
  case scm_tc3_cons + 8:\
  case scm_tc3_cons + 16:\
  case scm_tc3_cons + 24:\
  case scm_tc3_cons + 32:\
  case scm_tc3_cons + 40:\
  case scm_tc3_cons + 48:\
  case scm_tc3_cons + 56:\
  case scm_tc3_cons + 64:\
  case scm_tc3_cons + 72:\
  case scm_tc3_cons + 80:\
  case scm_tc3_cons + 88:\
  case scm_tc3_cons + 96:\
  case scm_tc3_cons + 104:\
  case scm_tc3_cons + 112:\
  case scm_tc3_cons + 120

/* For structs DEPRECATED
 */
/* #define scm_tcs_struct \ */
/*        scm_tc3_struct + 0:\ */
/*   case scm_tc3_struct + 8:\ */
/*   case scm_tc3_struct + 16:\ */
/*   case scm_tc3_struct + 24:\ */
/*   case scm_tc3_struct + 32:\ */
/*   case scm_tc3_struct + 40:\ */
/*   case scm_tc3_struct + 48:\ */
/*   case scm_tc3_struct + 56:\ */
/*   case scm_tc3_struct + 64:\ */
/*   case scm_tc3_struct + 72:\ */
/*   case scm_tc3_struct + 80:\ */
/*   case scm_tc3_struct + 88:\ */
/*   case scm_tc3_struct + 96:\ */
/*   case scm_tc3_struct + 104:\ */
/*   case scm_tc3_struct + 112:\ */
/*   case scm_tc3_struct + 120 */

/* For closures
 */
#define scm_tcs_closures \
       scm_tc3_closure + 0:\
  case scm_tc3_closure + 8:\
  case scm_tc3_closure + 16:\
  case scm_tc3_closure + 24:\
  case scm_tc3_closure + 32:\
  case scm_tc3_closure + 40:\
  case scm_tc3_closure + 48:\
  case scm_tc3_closure + 56:\
  case scm_tc3_closure + 64:\
  case scm_tc3_closure + 72:\
  case scm_tc3_closure + 80:\
  case scm_tc3_closure + 88:\
  case scm_tc3_closure + 96:\
  case scm_tc3_closure + 104:\
  case scm_tc3_closure + 112:\
  case scm_tc3_closure + 120

/* For subrs
 */
#define scm_tcs_subrs \
       scm_tc7_asubr:\
  case scm_tc7_subr_0:\
  case scm_tc7_subr_1:\
  case scm_tc7_dsubr:\
  case scm_tc7_cxr:\
  case scm_tc7_subr_3:\
  case scm_tc7_subr_2:\
  case scm_tc7_subr_n:\
  case scm_tc7_rpsubr:\
  case scm_tc7_subr_1o:\
  case scm_tc7_subr_2o:\
  case scm_tc7_lsubr_2:\
  case scm_tc7_lsubr


#endif  /* SCM_TAGS_H */
