/* Include the headers for just about everything.
   We call all their initialization functions.  */

#include "include/config.h"

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <gmp.h>

#include "include/_scm.h"

#include "include/alist.h"
#include "include/boolean.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/chars.h"
#include "include/debug.h"
#include "include/dictionary.h"
#include "include/dynwind.h"
#include "include/eq.h"
#include "include/error.h"
#include "include/eval.h"
#include "include/evalext.h"
#include "include/feature.h"
#include "include/filesys.h"
#include "include/fports.h"
#include "include/trim.h"
#include "include/gsubr.h"
#include "include/iselect.h"
#include "include/ioext.h"
#include "include/list.h"
#include "include/load.h"
#include "include/macros.h"
#include "include/nature.h"
#include "include/net_db.h"
#include "include/numbers.h"
#include "include/pairs.h"
#include "include/ports.h"
#include "include/posix.h"
#ifdef HAVE_REGCOMP
#include "include/regex-posix.h"
#endif
#include "include/print.h"
#include "include/procprop.h"
#include "include/procs.h"
#include "include/random.h"
#include "include/read.h"
#include "include/signals.h"
#include "include/script.h"
#include "include/simpos.h"
#include "include/smob.h"
#include "include/socket.h"
#include "include/sort.h"
#include "include/stime.h"
#include "include/srfi-1.h"
#include "include/strports.h"
#include "include/symbols.h"
#include "include/throw.h"
#include "include/variable.h"
#include "include/vectors.h"
#include "include/version.h"
#include "include/vports.h"
#include "include/terminal-support.h"
#include "include/gio.h"

#include "include/init.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif





/* initializing standard and current I/O ports */

typedef struct
{
  int fdes;
  char *mode;
  char *name;
} stream_body_data;

/* proc to be called in scope of exception handler stream_handler. */
static SCM
stream_body (void *data)
{
  stream_body_data *body_data = (stream_body_data *) data;
  SCM port = scm_fdes_to_port (body_data->fdes, body_data->mode,
			       scm_from_sen (body_data->name, strlen (body_data->name)));

  RC_RETURN (port);
}

/* exception handler for stream_body.  */
static SCM
stream_handler (void *data SCM_UNUSED,
		SCM tag SCM_UNUSED,
		SCM throw_args SCM_UNUSED)
{
  return SCM_BOOL_F;
}

/* Convert a file descriptor to a port, using scm_fdes_to_port.
   - NAME is a C string, not a SCM string
   - catch exceptions: allow VSI to be able to start up even
   if it has been handed bogus stdin/stdout/stderr.  replace the
   bad ports with void ports.  */
static SCM
scm_standard_stream_to_port (int fdes, char *mode, char *name)
{
  SCM port;
  stream_body_data body_data;

  body_data.fdes = fdes;
  body_data.mode = mode;
  body_data.name = name;
  port = scm_internal_catch (SCM_BOOL_T, stream_body, &body_data, 
			     stream_handler, NULL);
  if (scm_is_false (port))
    port = scm_void_port (mode);
  return port;
}

/* Create standard ports from stdin, stdout, and stderr.  */
static void
scm_init_standard_ports ()
{
  /* From the SCSH manual:

     It can be useful to turn I/O buffering off in some cases, for
     example when an I/O stream is to be shared by multiple
     subprocesses.  For this reason, scsh allocates an unbuffered port
     for file descriptor 0 at start-up time.

     Because shells frequently share stdin with subprocesses, if the
     shell does buffered reads, it might ``steal'' input intended for
     a subprocess.  For this reason, all shells, including sh, csh,
     and scsh, read stdin unbuffered.  Applications that can tolerate
     buffered input on stdin can reset \ex{(current-input-port)} to
     block buffering for higher performance.

     Yes, maybe terminals aren't just streams ...
   */

  scm_set_current_input_port 
    (scm_standard_stream_to_port (0, 
				  isatty (0) ? "r0" : "r",
				  "standard input"));
  scm_set_current_output_port
    (scm_standard_stream_to_port (1,
				  isatty (1) ? "w0" : "w",
				  "standard output"));
  scm_set_current_error_port
    (scm_standard_stream_to_port (2,
				  isatty (2) ? "w0" : "w",
				  "standard error"));
}



/* Loading the startup Scheme files.  */

#include "include/debug.h" /* to use debug_print_system_info */

void
scm_load_startup_files ()
{
  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace:"
             " init: scm_load_startup_files\n");

  /* We want a path only containing directories from GUILE_LOAD_PATH,
     SCM_SITE_DIR and SCM_LIBRARY_DIR when searching for the site init
     file, so we do this before loading Ice-9.  */
  SCM init_path =
    scm_sys_search_load_path (scm_from_cen ("init-9.scm"));
  reference_parameter (init_path);

  if (global_debug_interpreter)
    fprintf (stderr, "init: scm_load_startup_files: init-9.scm\n");

  /* Load the init-9.scm file.  */
  if (scm_is_true (init_path)) {
    scm_primitive_load (init_path);
    if (global_debug_interpreter)
      fprintf (stderr, "init: scm_load_startup_files: done\n");
    dereference_parameter (init_path);
  }
  else {
    scm_shell_usage (0, "\n\n * VSI: FATAL ERROR: init-9.scm not found *\n");
    debug_print_system_info (NULL);
    fprintf (stderr, "\n * VSI: FATAL ERROR: init-9.scm not found *\n\n");
    exit (1);
  }
}


/* The main init code.  */

#ifdef _UNICOS
typedef int setjmp_type;
#else
typedef long setjmp_type;
#endif

/* All the data needed to invoke the main function.  */
struct main_func_closure
{
  /* the function to call */
  void (*main_func)(void *closure, int argc, char **argv);
  void *closure;		/* dummy data to pass it */
  int argc;
  char **argv;			/* the argument list it should receive */
};

static void *invoke_main_func(void *body_data);


/* Fire up the Scheme interpreter.

   Call MAIN_FUNC, passing it CLOSURE, ARGC, and ARGV.  MAIN_FUNC
   should do all the work of the program (initializing other packages,
   reading user input, etc.) before returning.  When MAIN_FUNC
   returns, call exit (0); this function never returns.  If you want
   some other exit value, MAIN_FUNC may call exit itself.

   scm_boot_strap arranges for program-arguments to return the strings
   given by ARGC and ARGV.  If MAIN_FUNC modifies ARGC/ARGV, should
   call scm_set_program_arguments with the final list, so Scheme code
   will know which arguments have been processed.

   scm_boot_strap establishes a catch-all catch handler which prints an
   error message and exits the process. This means that the interpreter
   exits in a coherent way when system errors occur and the user isn't
   prepared to handle it.  If the user doesn't like this behavior,
   they can establish their own universal catcher to shadow this one.

   Why must the caller do all the real work from MAIN_FUNC?  The
   garbage collector assumes that all local variables of type SCM will
   be above scm_boot_strap's stack frame on the stack.  If you try to
   manipulate SCM values after this function returns, it's the luck of
   the draw whether the GC will be able to find the objects you
   allocate.  So, scm_boot_strap function exits, rather than
   returning, to discourage people from making that mistake.  */


void
scm_boot_strap (int argc, char ** argv, void (*main_func) (), void *closure)
{
  void *res;
  struct main_func_closure c;

  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace: init.c: scm_boot_strap\n");

  c.main_func = main_func;
  c.closure = closure;
  c.argc = argc;
  c.argv = argv;

  // no boot strap no more res = scm_with_thread (invoke_main_func, &c);

  /* If the caller doesn't want this, they should exit from main_func
     themselves.
  */
  if (res == NULL)
    exit (EXIT_FAILURE);
  else
    exit (0);
}

static void *
invoke_main_func (void *body_data)
{
  struct main_func_closure *closure = (struct main_func_closure *) body_data;

  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace: init.c: invoke_main_func\n");

  scm_set_program_arguments (closure->argc, closure->argv, 0);
  (*closure->main_func) (closure->closure, closure->argc, closure->argv);

  // scm_restore_signals ();

  /* Indicate success by returning non-NULL.
   */
  return (void *)1;
}

int global_scm_initialized = 0;

static void
cleanup_for_exit ()
{
  scm_flush_all_ports ();
}

void
scm_init_interpreter ()
{
  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace:"
                     " init.c: scm_init_interpreter\n");


  if (global_scm_initialized) /* XXX is this stupid? */
    return;

  if (sizeof (mpz_t) > (3 * sizeof (scm_t_bits)))
    {
      fprintf (stderr,
               "GMP's mpz_t must fit into a double_cell,"
               "but doesn't seem to here.\n");
    }


  scm_init_storage ();
  scm_init_nature (); // System Dictionary
  scm_smob_init ();
  scm_ports_prehistory ();

  scm_init_dynwind ();

  scm_init_fports    ();
  scm_init_strports  ();
  scm_init_macros    ();
  scm_init_numbers   ();
  scm_init_ports     ();
#ifdef HAVE_POSIX
  scm_init_filesys   ();
#endif
  scm_init_regex_posix ();
  scm_init_load      ();   /* XXX creates bindings */
  scm_init_stime     ();
  scm_init_throw     ();
  scm_init_vports    ();
  scm_init_eval      ();
  scm_init_debug     ();  /* Requires macro smobs */
  scm_init_random    ();
  scm_init_standard_ports ();  /* Requires fports */
  scm_init_script    ();

  scm_init_gio       ();

  /* Now fill the System Environment */
  scm_eks_inclusions ();

  global_scm_initialized = 1;
}

/*
 * Initialization of
 * System Enironment Bindings
 */
void
scm_eks_inclusions ()
{
  /* TODO rename /scm_init/ procedures to /x_inclusion/ */

  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace: scm_eks_inclusions\n");

  x_inclusion_feature (); /* purge this please */
  x_inclusion_gsubr   ();
  x_inclusion_debug   ();
  scm_init_error      ();
  x_inclusion_eval    ();

  /* Primitives */
  x_inclusion_storage ();
  x_inclusion_nature  ();

  x_inclusion_load   ();
  x_inclusion_print  ();
  x_inclusion_read   ();

  x_inclusion_symbols ();
  x_inclusion_macros  ();
  x_inclusion_numbers ();
  scm_init_boolean    ();
  scm_init_pairs      ();
  x_inclusion_procs   ();
  scm_init_list       ();
  x_inclusion_equality();
  scm_init_alist      ();
  scm_init_sort       ();
  scm_init_variable   ();

  scm_init_chars ();
  x_inclusion_vectors ();
  x_inclusion_throw   ();
  x_inclusion_dynwind ();
  x_inclusion_ports   ();

  // x_inclusion_dictionary ();

  scm_init_evalext    ();
  scm_init_ioext      ();

  /* Extra Primitives */
  x_inclusion_stime   ();
  x_inclusion_version ();
  x_inclusion_fports  ();
  x_inclusion_vports  ();
  x_inclusion_strports ();
  x_inclusion_random  ();

  scm_init_simpos     ();
  x_inclusion_script  ();
#ifdef HAVE_POSIX
  x_inclusion_filesys ();
  scm_init_posix      ();
#endif
  x_inclusion_posix_signals ();
  x_inclusion_regex   ();
  scm_init_terminal_support ();
  
#ifdef HAVE_NETWORKING
  scm_init_net_db ();
  scm_init_socket ();
#endif

  x_inclusion_procprop ();
  x_inclusion_list_processing ();  // former srfi-1
  x_inclusion_gio ();
  x_inclusion_byte_strings ();
  x_inclusion_brass ();
}
