// Strings & Brass

#include "include/config.h"

#include <string.h>
#include <stdio.h>

#include "include/_scm.h"
#include "include/libpath.h"
#include "include/fports.h"
#include "include/read.h"
#include "include/eval.h"
#include "include/throw.h"
#include "include/alist.h"
#include "include/dynwind.h"
#include "include/root.h"
#include "include/byte-strings.h"
#include "include/brass.h"
#include "include/chars.h"

#include "include/validate.h"
#include "include/load.h"

#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#ifndef R_OK
#define R_OK 4
#endif


/* Loading a file, given an absolute filename.  */

SCM_DEFINE (scm_primitive_load, "primitive-load", 1, 0, 0, 
           (SCM filename),
	    "Load the file named @var{filename} and evaluate its contents in\n"
	    "the top-level environment. The load paths are not searched;\n"
	    "@var{filename} must either be a full pathname or be a pathname\n"
	    "relative to the current directory.")
#define FUNC_NAME s_scm_primitive_load
{
  reference_parameter (filename);

  SCM_VALIDATE_BRASS (1, filename);

  if (global_debug_interpreter) {
    SCM bilename = scm_to_sen (filename);
    reference_parameter (bilename);
    fprintf (stderr, "init: scm_primitive_load: %s\n",
             BYTE_STRING_DATA (bilename));
    dereference_parameter (bilename);
  }

  { /* scope */
    SCM port = scm_open_file (filename, scm_from_cen ("r"));
    reference_parameter (port);

    while (1)
      {
	SCM form;
        form = scm_read (port);

	if (SCM_EOF_OBJECT_P (form))
	  break;

	rc_Void (scm_primitive_eval_x (form));
      }

    scm_close_port (port); dereference_parameter (port);
  }
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM
scm_c_primitive_load (const char *filename)
{
  return scm_primitive_load (scm_from_cen (filename));
}


/* Builtin path to scheme library files. */
#ifdef SCM_PKGDATA_DIR
SCM_DEFINE (scm_sys_package_data_dir, "%package-data-dir", 0, 0, 0, 
            (),
	    "Return the name of the directory where Scheme packages, modules and\n"
	    "libraries are kept.")
#define FUNC_NAME s_scm_sys_package_data_dir
{
  return scm_from_cen (SCM_PKGDATA_DIR);
}
#undef FUNC_NAME
#endif /* SCM_PKGDATA_DIR */

#ifdef SCM_LIBRARY_DIR
SCM_DEFINE (scm_sys_library_dir, "%library-dir", 0,0,0,
            (),
	    "Return the directory where the Scheme library files are installed.")
#define FUNC_NAME s_scm_sys_library_dir
{
  return scm_from_cen (SCM_LIBRARY_DIR);
}
#undef FUNC_NAME
#endif /* SCM_LIBRARY_DIR */

#ifdef SCM_SITE_DIR
SCM_DEFINE (scm_sys_site_dir, "%site-dir", 0,0,0,
            (),
	    "Return the directory where the site files are installed.")
#define FUNC_NAME s_scm_sys_site_dir
{
  return scm_from_cen (SCM_SITE_DIR);
}
#undef FUNC_NAME
#endif /* SCM_SITE_DIR */




/* Initializing the load path, and searching it.  */

/* List of names of directories we search for files to load.  */
SCM *scm_loc_load_path = NULL;


/* Initialize the global variable %load-path, given the value of the
   SCM_SITE_DIR and SCM_LIBRARY_DIR preprocessor symbols. */
SCM
scm_initial_load_path ()
{
  char *env;
  SCM path = SCM_EOL;

  path = scm_list_1 (scm_from_cen (SCM_SITE_DIR));
  /* path = scm_list_4 (scm_from_cen ("."), */
  /*                    scm_from_cen (SCM_SITE_DIR), */
  /*       	     scm_from_cen (SCM_LIBRARY_DIR), */
  /*       	     scm_from_cen (SCM_PKGDATA_DIR)); */
  return path;
}

/* Utility functions for assembling C strings in a buffer.
 */

struct stringbuf {
  char *buf, *ptr;
  size_t buf_len;
};

static void
stringbuf_free (void *data)
{
  struct stringbuf *buf = (struct stringbuf *)data;
  free (buf->buf);
}

static void
stringbuf_grow (struct stringbuf *buf)
{
  size_t ptroff = buf->ptr - buf->buf;
  buf->buf_len *= 2; 
  buf->buf = scm_realloc (buf->buf, buf->buf_len);
  buf->ptr = buf->buf + ptroff;
}

static void
stringbuf_cat_locale_string (struct stringbuf *buf, const char* str)
{
  size_t max_len = buf->buf_len - (buf->ptr - buf->buf) - 1;
  size_t len = strlen (str);
  if (len > max_len)
    {
      /* buffer is too small, double its size and try again. 
       */
      stringbuf_grow (buf);
      stringbuf_cat_locale_string (buf, str);
    }
  else
    {
      /* string fits, terminate it and check for embedded '\0'.
       */
      memcpy (buf->ptr, str, len);
      buf->ptr[len] = '\0';
      if (strlen (buf->ptr) != len)
	scm_misc_error (NULL,
			"string contains #\\nul character: ~S",
			scm_list_1 (scm_from_cen (str)));
      buf->ptr += len;
    }
}

static void
stringbuf_cat (struct stringbuf *buf, char *str)
{
  size_t max_len = buf->buf_len - (buf->ptr - buf->buf) - 1;
  size_t len = strlen (str);
  if (len > max_len)
    {
      /* buffer is too small, double its size and try again. 
       */
      stringbuf_grow (buf);
      stringbuf_cat (buf, str);
    }
  else
    {
      /* string fits, copy it into buffer.
       */
      strcpy (buf->ptr, str);
      buf->ptr += len;
    }
}

  
/* Search PATH for a directory containing a file named FILENAME.
   The file must be readable, and not a directory.
   If we find one, return its full filename; otherwise, return #f.
   If FILENAME is absolute, return it unchanged.
   If given, EXTENSIONS is a list of strings; for each directory 
   in PATH, we search for FILENAME concatenated with each EXTENSION.  */
SCM_DEFINE (scm_search_path, "search-path", 2, 0, 0,
           (SCM path, SCM filename),
	    "Search @var{path} for a directory containing a file named\n"
	    "@var{filename}. The file must be readable, and not a directory.\n"
	    "If we find one, return its full filename; otherwise, return\n"
	    "@code{#f}.  If @var{filename} is absolute, return it unchanged.\n"
	    "If given, @var{extensions} is a list of strings; for each\n"
	    "directory in @var{path}, we search for @var{filename}\n"
	    "concatenated with each @var{extension}.")
#define FUNC_NAME s_scm_search_path
{
  reference_parameter (path);
  reference_parameter (filename);

  SCM bilename = scm_to_sen (filename);
  reference_parameter (bilename);
  struct stringbuf buf;
  char *filename_chars;
  size_t filename_len;
  SCM result = SCM_BOOL_F;

  filename_chars = BYTE_STRING_DATA (bilename);
  filename_len = BYTE_STRING_LENGTH (bilename);

  /* If FILENAME is absolute, return it unchanged.  */
  if (filename_len >= 1 && filename_chars[0] == '/') {
    dereference_parameter (bilename);
    dereference_parameter (path);
    return (rc_return_stack_reference (filename));
  }


  buf.buf_len = 512;
  buf.buf = scm_malloc (buf.buf_len);

  /* Try every path element.
   */
  for (; scm_is_pair (path); path = FLY_CDR (path)) {
    SCM sir = TMP_CAR (path);
    SCM bir = scm_to_sen (sir);
    reference_parameter (bir);
    const char *dir = BYTE_STRING_DATA (bir);

    buf.ptr = buf.buf;
    stringbuf_cat_locale_string (&buf, dir);
	
    /* Concatenate the path name and the filename. */
      
    if ((buf.ptr > buf.buf) && (buf.ptr[-1] != '/'))
      stringbuf_cat (&buf, "/");

    stringbuf_cat (&buf, filename_chars);

    struct stat mode;
	  
    /* If the file exists at all, we should return it.  If the
       file is inaccessible, then that's an error.  */

    dereference_parameter (bir);
    if (stat (buf.buf, &mode) == 0 && ! (mode.st_mode & S_IFDIR)) {
      result = scm_from_sen (buf.buf, strlen (buf.buf));
      reference_parameter (result);
      goto end;
    }
  }
      
  if (!scm_is_null (path))
    scm_wrong_type_arg_msg (NULL, 0, path, "proper list");

 end:
  free (buf.buf);
  dereference_parameter (bilename);
  dereference_parameter (path);
  dereference_parameter (filename);
  return (rc_return_stack_reference (result));
}
#undef FUNC_NAME


/* Search %load-path for a directory containing a file named FILENAME.
   The file must be readable, and not a directory.
   If we find one, return its full filename; otherwise, return #f.
   If FILENAME is absolute, return it unchanged.  */
SCM_DEFINE (scm_sys_search_load_path, "%search-load-path", 1, 0, 0, 
	    (SCM filename),
	    "Search @var{%load-path} for the file named @var{filename},\n"
	    "which must be readable by the current user.  If @var{filename}\n"
	    "is found in the list of paths to search or is an absolute\n"
	    "pathname, return its full pathname.  Otherwise, return\n"
	    "@code{#f}.  Filenames may have any of the optional extensions\n"
	    "in the @code{%load-extensions} list; @code{%search-load-path}\n"
	    "will try each extension automatically.")
#define FUNC_NAME s_scm_sys_search_load_path
{
  SCM path = *scm_loc_load_path;
  SCM_VALIDATE_BRASS (1, filename);

  if (scm_ilength (path) < 0)
    SCM_MISC_ERROR ("%load-path is not a proper list", SCM_EOL);
  return scm_search_path (path, filename);
}
#undef FUNC_NAME


SCM_DEFINE (scm_primitive_load_path, "primitive-load-path", 1, 0, 0, 
	    (SCM filename),
	    "Search @var{%load-path} for the file named @var{filename} and\n"
	    "load it into the top-level environment.  If @var{filename} is a\n"
	    "relative pathname and is not found in the list of search paths,\n"
	    "an error is signalled.")
#define FUNC_NAME s_scm_primitive_load_path
{
  SCM full_filename;

  full_filename = scm_sys_search_load_path (filename);

  if (scm_is_false (full_filename))
    SCM_MISC_ERROR ("Unable to find file ~S in load path",
		    scm_list_1 (filename));

  return scm_primitive_load (full_filename);
}
#undef FUNC_NAME

SCM
scm_c_primitive_load_path (const char *filename)
{
  return scm_primitive_load_path (scm_from_cen (filename));
}


void
scm_init_load ()
{
  scm_loc_load_path = SCM_VARIABLE_LOC (scm_c_define ("%load-path", SCM_EOL));

  *scm_loc_load_path = scm_initial_load_path ();
  heap_reference (scm_loc_load_path, *scm_loc_load_path); // XXX helper ref
  /* scm_c_define("current-reader", the_reader); ??? purge*/
}

void
x_inclusion_load ()
{
#include "include/load.x"
}
