#include "include/config.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#include "include/_scm.h"
#include "include/chars.h"
#include "include/eval.h"
#include "include/alist.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/brass.h"
#include "include/strports.h"
#include "include/vectors.h"
#include "include/validate.h"
#include "include/byte-strings.h"

#include "include/read.h"



SCM_GLOBAL_SYMBOL (scm_sym_dot, ".");

/*
  Give meaningful error messages for errors

  We use the format

  FILE:LINE:COL: MESSAGE
  This happened in ....

 */


void
scm_i_input_error (char const *function,
		   SCM port, const char *message, SCM arg)
{
  reference_parameter (port);
  reference_parameter (arg);

  SCM fn = (SCM_BRASSP (SCM_FILENAME(port))
	    ? SCM_FILENAME(port)
	    : scm_from_cen ("#<unknown port>"));

  SCM string_port = scm_open_output_string ();
  reference_parameter (string_port);

  SCM string = SCM_EOL;
  scm_simple_format (string_port,
		     scm_from_cen ("~A:~S:~S: ~A"),
		     scm_list_4 (fn,
				 scm_from_long (SCM_LINUM (port) + 1),
				 scm_from_int (SCM_COL (port) + 1),
				 scm_from_cen (message)));
    
  string = scm_from_pen (scm_get_output_string (string_port));
  scm_close_output_port (string_port);
  dereference_parameter (string_port);

  scm_error_scm (scm_from_locale_symbol ("read-error"),
		 function ? scm_from_cen (function) : SCM_BOOL_F,
		 string,
		 arg,
		 SCM_BOOL_F);
}



/* Token readers.  */


/* Size of the C buffer used to read symbols and numbers.  */
#define READER_BUFFER_SIZE            128

/* Size of the C buffer used to read strings.  */
#define READER_STRING_BUFFER_SIZE     512

/* The maximum size of Scheme character names.  */
#define READER_CHAR_NAME_MAX_SIZE      50


#define SCC SCM_MAKE_CHAR

/* `isblank' is only in C99.  */
#define CHAR_IS_BLANK(_chr)					\
  (((_chr) == SCC (' ')) || ((_chr) == SCC ('\t')) || ((_chr) == SCC ('\n'))	\
   || ((_chr) == SCC ('\f')) || ((_chr) == SCC ('\r')))


/* R5RS one-character delimiters (see section 7.1.1, ``Lexical
   structure'').  */
#define CHAR_IS_R5RS_DELIMITER(c)				\
  (CHAR_IS_BLANK (c)						\
   || (c == SCC (')')) || (c == SCC ('(')) || (c == SCC (';')) || (c == SCC ('"')))

#define CHAR_IS_DELIMITER  CHAR_IS_R5RS_DELIMITER

/* Exponent markers, as defined in section 7.1.1 of R5RS, ``Lexical
   Structure''.  */
#define CHAR_IS_EXPONENT_MARKER(_chr)				\
  (((_chr) == SCC ('e')) || ((_chr) == SCC ('s')) || ((_chr) == SCC ('f'))	\
   || ((_chr) == SCC ('d')) || ((_chr) == SCC ('l')))

/* An inlinable version of `scm_c_downcase ()'.  */
#define CHAR_DOWNCASE(_chr)				\
  (((_chr) <= UCHAR_MAX) ? tolower (_chr) : (_chr))  // XXX byte magic


#ifndef HAVE_DECL_STRNCASECMP
extern int strncasecmp (char const *s1, char const *s2, size_t n);
#endif

#ifndef HAVE_STRNCASECMP
static int
strncasecmp (const char *s1, const char *s2, size_t len2)
{
  while (*s1 && *s2 && len2 > 0)
    {
      int c1 = *s1, c2 = *s2;

      if (CHAR_DOWNCASE (c1) != CHAR_DOWNCASE (c2))
	return 0;
      else
	{
	  ++s1;
	  ++s2;
	  --len2;
	}
    }
  return !(*s1 || *s2 || len2 > 0);
}
#endif


/* Read an SCSH block comment.  */
static inline SCM scm_read_scsh_block_comment (SCM chr, SCM port);

/* Read from PORT until a delimiter (e.g., a whitespace) is read.  Return
   zero if the whole token fits in BUF, non-zero otherwise.  */
static inline int
read_token (SCM port, usv_t *buf, size_t buf_size, size_t *read) // buf_size in bytes!
{
  reference_parameter (port);

  *read = 0;

  while (*read < buf_size) {
    SCM chr;

    chr = scm_getc (port);
    // chr = (SCM_CASE_INSENSITIVE_P ? CHAR_DOWNCASE (chr) : chr);

    if (SCM_EOF_OBJECT_P (chr)) {
      dereference_parameter (port);
      return 0;

    } else if (CHAR_IS_DELIMITER (chr)) {
      scm_ungetc (chr, port);
      dereference_parameter (port);
      return 0;

    } else {
      *buf = SCM_CHAR (chr);
      buf++, (*read)++;
    }
  }

  dereference_parameter (port);
  return 1;
}


/* Skip whitespace from PORT and return the first non-whitespace character
   read.  Raise an error on end-of-file.  */
static int
flush_ws (SCM port, const char *eoferr)
{
  reference_parameter (port);

  SCM c;
  while (1) {
   switch (c = scm_getc (port)) {
   case SCM_EOF_VAL:
   goteof:
     if (eoferr) {
       scm_i_input_error (eoferr,
                          port,
                          "end of file",
                          SCM_EOL);
     }
     dereference_parameter (port);
     return c;

   case SCC(';'):
   lp:
     switch (c = scm_getc (port)) {
	  case SCM_EOF_VAL: goto goteof;
	  default:  goto lp;
	  case SCM_LINE_INCREMENTORS:  // case after default ???
	    break;
     }
     break;

   case SCC('#'):
     switch (c = scm_getc (port)) {
     case SCM_EOF_VAL:
       eoferr = "read_sharp";
       goto goteof;
     case SCC('!'):
       scm_read_scsh_block_comment (c, port);
       break;
     default:
       scm_ungetc (c, port);
       dereference_parameter (port);
       return SCC('#');
     }
     break;

   case SCM_LINE_INCREMENTORS:
   case SCM_SINGLE_SPACES:
   case SCC('\t'):
     break;

   default:
     dereference_parameter (port);
     return c;
   }
  }

  dereference_parameter (port);
  return 0;
}



/* Token readers.  */

static SCM scm_read_expression (SCM port);
static SCM scm_read_sharp (SCM chr, SCM port);


static SCM
scm_read_sexp (SCM chr, SCM port)
#define FUNC_NAME "scm_i_lreadparen"
{
  reference_parameter (port);

  SCM c;
  SCM tmp;
  SCM tl, ans = SCM_EOL;
  SCM tl2 = SCM_EOL, ans2 = SCM_EOL, copy = SCM_BOOL_F;
  static const SCM terminating_char = SCC(')');

  /* Need to capture line and column numbers here. */
  long line = SCM_LINUM (port);
  int column = SCM_COL (port) - 1;


  c = flush_ws (port, FUNC_NAME);
  if (terminating_char == c) {
    dereference_parameter (port);
    return SCM_EOL;
  }

  scm_ungetc (c, port);

  tmp = scm_read_expression (port);
  reference_parameter (tmp);

  if (scm_is_eq (scm_sym_dot, tmp))
    {
      ans = scm_read_expression (port);
      reference_parameter (ans);
      if (terminating_char != (c = flush_ws (port, FUNC_NAME)))
	scm_i_input_error (FUNC_NAME, port, "missing close paren",
			   SCM_EOL);

      reference_count_return (ans);
      dereference_parameter (ans);
      dereference_parameter (tmp);
      dereference_parameter (port);
      return ans;
    }

  /* Build the head of the list structure. */
  ans = tl = scm_cons (tmp, SCM_EOL);
  reference_parameter (tl);
  reference_parameter (ans);

  // reference checkpoint: port, tmp (+list-count), ans, tl

  /* if (SCM_COPY_SOURCE_P) */
  /*   ans2 = tl2 = scm_cons (scm_is_pair (tmp) */
  /*       		   ? copy */
  /*       		   : tmp, */
  /*       		   SCM_EOL); */

  reference_parameter (ans2);
  reference_parameter (tl2);
  // reference checkpoint: port, tmp (+list-count), ans, tl, ans2, tl2

  while (terminating_char != (c = flush_ws (port, FUNC_NAME))) {
    SCM new_tail;

    scm_ungetc (c, port);

    dereference_parameter (tmp);
    tmp = scm_read_expression (port);
    reference_parameter (tmp);

    if (scm_is_eq (scm_sym_dot, tmp)) {

      dereference_parameter (tmp);
      tmp = scm_read_expression (port);
      reference_parameter (tmp);

      SCM_SETCDR (tl, tmp);

      /* if (SCM_COPY_SOURCE_P) */
      /*   SCM_SETCDR (tl2, scm_cons (scm_is_pair (tmp) ? copy : tmp, */
      /*                              SCM_EOL)); */

      c = flush_ws (port, FUNC_NAME);
      if (terminating_char != c)
        scm_i_input_error (FUNC_NAME, port,
                           "in pair: missing close paren", SCM_EOL);
      goto exit;
    }

    new_tail = scm_cons (tmp, SCM_EOL);
    reference_parameter (new_tail);

    SCM_SETCDR (tl, new_tail);

    dereference_parameter (tl);
    tl = new_tail;
    reference_parameter (tl);

    /* if (SCM_COPY_SOURCE_P) { */
    /*   SCM new_tail2 = scm_cons (scm_is_pair (tmp) */
    /*                             ? copy */
    /*                             : tmp, SCM_EOL); */
    /*   reference_parameter (new_tail2); */
    /*   SCM_SETCDR (tl2, new_tail2); */

    /*   dereference_parameter (tl2); */
    /*   tl2 = new_tail2; */
    /*   reference_parameter (tl2); */

    /*   dereference_parameter (new_tail2); */
    /* } */
    dereference_parameter (new_tail);
  }

 exit:

  reference_count_return (ans);
  dereference_parameter (tmp);
  dereference_parameter (ans);
  dereference_parameter (tl);
  dereference_parameter (ans2);
  dereference_parameter (tl2);
  dereference_parameter (copy);  // bogus (always SCM_BOOL_F)
  dereference_parameter (port);
  return ans;
}
#undef FUNC_NAME

static SCM
scm_read_brass (SCM chr, SCM port)
#define FUNC_NAME "scm_lreadr"
{
  reference_parameter (port);

  SCM str = SCM_BOOL_F;
  usv_t c_str[READER_STRING_BUFFER_SIZE];
  unsigned c_str_len = 0;
  SCM c;

  while (SCC('"') != (c = scm_getc (port)))
    {
      if (c == SCM_EOF_VAL)
	str_eof: scm_i_input_error (FUNC_NAME, port,
				    "end of file in brass constant",
				    SCM_EOL);

      if (c_str_len + 1 >= sizeof (c_str))
	{
	  SCM chunk;

	  if (str == SCM_BOOL_F) {
	    str = scm_c_make_brass (0);
            reference_parameter (str);
          }

	  chunk = scm_c_make_brass_with_copy (c_str, c_str_len);
	  str = rc_X (scm_brass_append (scm_list_2 (rc_Fly (str), chunk)));

	  c_str_len = 0;
	}

      if (c == SCC('\\')) {
	switch (c = scm_getc (port))
	  {
	  case SCM_EOF_VAL: goto str_eof;
	  case SCC('"'):  break;
	  case SCC('\\'): break;
  	  case SCC('\n'): continue;
	  case SCC('0'):  c = SCC('\0'); break;
	  case SCC('f'):  c = SCC('\f'); break;
	  case SCC('n'):  c = SCC('\n'); break;
	  case SCC('r'):  c = SCC('\r'); break;
	  case SCC('t'):  c = SCC('\t'); break;
	  case SCC('a'):  c = SCC('\007'); break;  // ???
	  case SCC('v'):  c = SCC('\v'); break;
	  case SCC('x'): {
            SCM sa, sb;
            sa = scm_getc (port);
            if (sa == SCM_EOF_VAL) goto str_eof;
            sb = scm_getc (port);
            if (sb == SCM_EOF_VAL) goto str_eof;
            usv_t a = SCM_CHAR(a);
            usv_t b = SCM_CHAR(b);
            if      ('0' <= a && a <= '9') a -= '0';
            else if ('A' <= a && a <= 'F') a = a - 'A' + 10;
            else if ('a' <= a && a <= 'f') a = a - 'a' + 10;
            else goto bad_escaped;
            if      ('0' <= b && b <= '9') b -= '0';
            else if ('A' <= b && b <= 'F') b = b - 'A' + 10;
            else if ('a' <= b && b <= 'f') b = b - 'a' + 10;
            else goto bad_escaped;
            c = SCM_MAKE_CHAR(a * 16 + b);
            break;
          }
	  default:
	  bad_escaped:
	    scm_i_input_error (FUNC_NAME, port,
			       "illegal character in escape sequence: ~S",
			       scm_list_1 (c));
	  }
      }
      c_str[c_str_len++] = SCM_CHAR(c);
    }

  if (c_str_len > 0) {
    SCM chunk;

    chunk = scm_c_make_brass_with_copy (c_str, c_str_len);
    if (str == SCM_BOOL_F) {
      str = chunk;
      reference_parameter (str);
    } else {
      str = rc_X (scm_brass_append (scm_list_2 (rc_Fly (str), chunk)));
    }
  } else {
    str = (str == SCM_BOOL_F) ? scm_c_make_brass (0) : str;
    reference_parameter (str);
  }

  reference_count_return (str);
  dereference_parameter (str);
  dereference_parameter (port);
  return str;
}
#undef FUNC_NAME


static SCM
scm_read_number (SCM chr, SCM port)
{
  reference_parameter (port);

  SCM result, str = SCM_EOL;
  usv_t buffer[READER_BUFFER_SIZE];
  size_t read;
  int overflow = 0;

  scm_ungetc (chr, port);
  do
    {
      overflow = read_token (port, buffer, sizeof (buffer), &read);

      if ((overflow) || (scm_is_pair (str))) {
        // XXX sys / port encoding
	str = scm_cons (scm_c_make_brass_with_copy (buffer, read), str);
      }
    }
  while (overflow);

  if (scm_is_pair (str))
    {
      reference_parameter (str);
      /* The slow path.  */

      str = rc_X (scm_brass_append \
                  (scm_reverse_x (rc_Fly (str), SCM_EOL)));
      result = scm_brass_to_number (str, SCM_UNDEFINED);
      reference_parameter (result);
      if (!scm_is_true (result)) {
        dereference_parameter (result);
	/* Return a symbol instead of a number.  */
	result = scm_brass_to_symbol (str);
        reference_parameter (result);
      }
    }
  else
    {
      SCM br = scm_c_make_brass_with_copy (buffer, read);
      reference_parameter (br);
      result = scm_brass_to_number (br, SCM_UNDEFINED);
      reference_parameter (result);
      if (!scm_is_true (result)) {
        dereference_parameter (result);
	/* Return a symbol instead of a number.  */
	result = scm_brass_to_symbol (br);
        reference_parameter (result);
      }
      dereference_parameter (br);
    }

  reference_count_return (result);
  dereference_parameter (str);
  dereference_parameter (result);
  dereference_parameter (port);
  return result;
}

static SCM
scm_read_mixed_case_symbol (SCM chr, SCM port)
{
  reference_parameter (port);

  SCM result, str = SCM_EOL;
  int overflow = 0, ends_with_colon = 0;
  usv_t buffer[READER_BUFFER_SIZE];
  size_t read = 0;
  int postfix = 0;  // transitional

  scm_ungetc (chr, port);
  do
    {
      overflow = read_token (port, buffer, sizeof (buffer), &read);

      if (read > 0)
	ends_with_colon = (buffer[read - 1] == ':');

      if ((overflow) || (scm_is_pair (str)))
	str = scm_cons (scm_c_make_brass_with_copy (buffer, read), str);
    }
  while (overflow);

  if (scm_is_pair (str))
    {
      reference_parameter (str);
      str = rc_X (scm_brass_append \
                  (scm_reverse_x (rc_Fly (str), SCM_EOL)));

      result = scm_brass_to_symbol (str);
      reference_parameter (result);
      dereference_parameter (str);
    }
  else
    {
      result = scm_c_make_brass_with_copy (buffer, read);
      reference_parameter (result);
    }

  reference_count_return (result);
  dereference_parameter (result);
  dereference_parameter (port);
  return scm_brass_to_symbol (result);
}

static SCM
scm_read_number_and_radix (SCM chr, SCM port)
#define FUNC_NAME "scm_lreadr"
{
  reference_parameter (port);

  SCM result, str = SCM_EOL;
  size_t read;
  usv_t buffer[READER_BUFFER_SIZE];
  unsigned int radix;
  int overflow = 0;

  switch (chr)
    {
    case SCC('B'):
    case SCC('b'):
      radix = 2;
      break;

    case SCC('o'):
    case SCC('O'):
      radix = 8;
      break;

    case SCC('d'):
    case SCC('D'):
      radix = 10;
      break;

    case SCC('x'):
    case SCC('X'):
      radix = 16;
      break;

    default:
      scm_ungetc (chr, port);
      scm_ungetc (SCC('#'), port);
      radix = 10;
    }

  do {
    overflow = read_token (port, buffer, sizeof (buffer), &read);
      
    if ((overflow) || (scm_is_pair (str))) {
      str = scm_cons (scm_c_make_brass_with_copy (buffer, read), str);
    }
  } while (overflow);

  if (scm_is_pair (str)) {
    reference_parameter (str);
    str = rc_X (scm_brass_append \
                (scm_reverse_x (rc_Fly (str), SCM_EOL)));
    result = scm_brass_to_number (str, scm_from_uint (radix));
  } else {
    SCM br = scm_c_make_brass_with_copy (buffer, read);
    reference_parameter (br);
    result = scm_brass_to_number (br, scm_from_uint (radix));
    dereference_parameter (br); // sic! no symbol here: we error here in # 
  }
  reference_parameter (result);

  if (scm_is_true (result)) {
    reference_count_return (result);
    dereference_parameter (str);
    dereference_parameter (result);
    dereference_parameter (port);
    return result;
  }

  scm_i_input_error (FUNC_NAME, port, "unknown sharp # object", SCM_EOL);

  return SCM_BOOL_F;
}
#undef FUNC_NAME

static SCM
scm_read_quote (SCM chr, SCM port)
{
  reference_parameter (port);

  SCM p;
  long line = SCM_LINUM (port);
  int column = SCM_COL (port) - 1;

  switch (chr)
    {
    case SCC('`'):
      p = scm_sym_quasiquote;
      break;

    case SCC('\''):
      p = scm_sym_quote;
      break;

    case SCC(','):
      {
	SCM c;

	c = scm_getc (port);
	if (SCC('@') == c) {
	  p = scm_sym_uq_splicing;

	} else {
          scm_ungetc (c, port);
          p = scm_sym_unquote;
        }
	break;
      }

    default:
      fprintf (stderr, "%s: unhandled quote character (%lux)\n",
	       "scm_read_quote", chr);
      abort (); // ???
    }
  reference_parameter (p);

  p = rc_X (scm_cons2 (rc_Fly (p), scm_read_expression (port), SCM_EOL));

  reference_count_return (p);
  dereference_parameter (p);
  dereference_parameter (port);
  return p;
}

static inline SCM
scm_read_semicolon_comment (SCM chr, SCM port)
{
  reference_parameter (port);

  SCM c;

  for (c = scm_getc (port);
       (c != SCM_EOF_VAL) && (c != SCC('\n'));
       c = scm_getc (port));

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}


/* Sharp readers, i.e. readers called after a `#' sign has been read.  */

static SCM
scm_read_boolean (int chr, SCM port)
{
  reference_parameter (port);

  switch (chr)
    {
    case SCC('t'):
    case SCC('T'):
      dereference_parameter (port);
      return SCM_BOOL_T;

    case SCC('f'):
    case SCC('F'):
      dereference_parameter (port);
      return SCM_BOOL_F;
    }

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}

static SCM
scm_read_character (SCM chr, SCM port)
#define FUNC_NAME "scm_lreadr"
{
  reference_parameter (port);

  SCM c;
  usv_t charname[READER_CHAR_NAME_MAX_SIZE];
  size_t charname_len;

  if (read_token (port, charname, sizeof (charname), &charname_len))
    goto char_error;

  if (charname_len == 0)
    {
      chr = scm_getc (port);
      if (chr == SCM_EOF_VAL)
	scm_i_input_error (FUNC_NAME, port, "unexpected end of file "
			   "while reading character", SCM_EOL);

      /* CHR must be a token delimiter, like a whitespace.  */
      dereference_parameter (port);
      return (chr);
    }

  if (charname_len == 1) {
    dereference_parameter (port);
    return SCM_MAKE_UC_CHAR (charname[0]);
  }

  if (*charname >= '0' && *charname < '8')
    {
      /* Somehow deprecated number in charname notation but outside of
         C0 we need to write some so we read some. */
      SCM p = scm_brass_to_number ( \
                scm_c_make_brass_with_copy (charname, charname_len), scm_from_uint (8));
      if (SCM_I_INUMP (p)) {
        dereference_parameter (port);
	return SCM_MAKE_UC_CHAR (SCM_I_INUM (p));
      } else {
        dereference_parameter (p);
      }
    }

  /* Have A nice copy! ;-) */
  SCM aans = scm_charname_lookup ( \
               scm_c_make_brass_with_copy (charname, charname_len));
  if (!scm_is_false (aans)) {
    dereference_parameter (port);
    return aans;
  }

 char_error:
  scm_i_input_error (FUNC_NAME, port, "unknown character name ~a",
		     scm_list_1 (scm_c_make_brass_with_copy (charname,
                                                             charname_len)));

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

static inline SCM
scm_read_vector (SCM chr, SCM port)
{
  /* Note: We call `scm_read_sexp ()' rather than READER here in order to
     guarantee that it's going to do what we want.  After all, this is an
     implementation detail of `scm_read_vector ()', not a desirable
     property.  */
  return (scm_vector (scm_read_sexp (chr, port)));
}

static inline SCM
scm_read_scsh_block_comment (SCM chr, SCM port)
{
  reference_parameter (port);

  int bang_seen = 0;

  for (;;)
    {
      SCM c = scm_getc (port);

      if (c == SCM_EOF_VAL)
	scm_i_input_error ("skip_block_comment", port,
			   "unterminated `#! ... !#' comment", SCM_EOL);

      if (c == SCC('!'))
	bang_seen = 1;
      else if (c == SCC('#') && bang_seen)
	break;
      else
	bang_seen = 0;
    }

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}

/* static int */
/* byte_digit (int char) */
/* { // unicode C0 hexadecimal digit reader */
/*   if (char >= 0x30 && char <= 0x39) // 0-9 */
/*     return (char - 0x30); */

/*   if (char >= 0x41 && char <= 0x46) // A-F */
/*     return (char - 0x41 + 10); */

/*   if (char >= 0x61 && char <= 0x66) // a-f */
/*     return (char - 0x61 + 10); */

/*   return (-1); */
/* } */

static SCM
scm_read_byte_string (SCM chr, SCM port) // XXX preliminary
#define FUNC_NAME "scm_read_byte_string"
{ // XXX todo RC and other things
  reference_parameter (port);

  const SCM terminating_char = SCC('"');
  size_t buffer_length = 128;
  size_t buffer_growth = 1024;
  size_t length = 0;
  SCM number;
  char byte;
  char *buffer = scm_malloc (buffer_length); // XXX use a byte string
  int c;

  // XXX todo dynwind free

  while (1) {
    byte = 0;
    c = flush_ws (port, FUNC_NAME);
    if (terminating_char == c) {
      if (length == 0) goto broken_byte_string;
      buffer = scm_realloc (buffer, length);
      if (!buffer) goto broken_byte_string;
      dereference_parameter (port);
      SCM answer = scm_make_byte_string_with_copy ((void*)buffer, length);
      free (buffer);
      return answer;
    }

    number = scm_read_number (c, port);
  
    if (SCM_I_INUMP (number)) {
      long num = SCM_I_INUM (number);
      if (num >= 0 && num <= 255) {
        if (length == buffer_length) {
          buffer_length += buffer_growth;
          buffer = scm_realloc (buffer, buffer_length);
          if (!buffer) goto broken_byte_string;
        }
        byte = (char) num;
        buffer[length] = byte;
        length++;
        continue;
      }
    }
  broken_byte_string:
    scm_i_input_error (FUNC_NAME, port,
                       "mischiefed byte-string: ~S",
                       scm_list_1 (c));
  }
  // never reached
}
#undef FUNC_NAME

/* The reader for the sharp `#' character.  It basically dispatches reads
   among the above token readers.   */
static SCM
scm_read_sharp (SCM chr, SCM port)
#define FUNC_NAME "scm_lreadr"
{
  reference_parameter (port);

  chr = scm_getc (port);

  // Dispatch Counting
  rc_Fly (port);

  switch (chr)
    {
    case SCC('\\'):
      return (scm_read_character (chr, port));
    case SCC('('):
      return (scm_read_vector (chr, port));
    case SCC('f'):
    case SCC('t'):
    case SCC('T'):
    case SCC('F'):
      /* This one may return either a boolean or an SRFI-4 vector.  */
      return (scm_read_boolean (chr, port));
    case SCC('b'):
    case SCC('B'):
    case SCC('o'):
    case SCC('O'):
    case SCC('d'):
    case SCC('D'):
    case SCC('x'):
    case SCC('X'):
    case SCC('I'):
    case SCC('E'):
      return (scm_read_number_and_radix (chr, port));
    case SCC('"'):
      return (scm_read_byte_string (chr, port));
    case SCC('!'):
      return (scm_read_scsh_block_comment (chr, port));
    default:
      scm_i_input_error (FUNC_NAME, port, "Unknown # object: ~S",
                         scm_list_1 (chr));
    }

  dereference_return_value (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

static SCM
scm_read_expression (SCM port)
#define FUNC_NAME "scm_read_expression"
{

  while (1) {
    SCM chr;

    reference_parameter (port);

    chr = scm_getc (port);

    rc_Fly (port);  // dispatch counting

    switch (chr) {
    case SCM_WHITE_SPACES:
    case SCM_LINE_INCREMENTORS:
      break;
    case SCC(';'):
      (void) scm_read_semicolon_comment (chr, port);
      break;
    case SCC('('):
      return (scm_read_sexp (chr, port));
    case SCC('"'):
      return (scm_read_brass (chr, port));
    case SCC('\''):
    case SCC('`'):
    case SCC(','):
      return (scm_read_quote (chr, port));
    case SCC('#'):
      {
        reference_parameter (port);
        SCM result;
        result = scm_read_sharp (chr, port); reference_parameter (result);
        if (scm_is_eq (result, SCM_UNSPECIFIED)) {
          /* We read a comment or some such.  */
          rc_Fly (port);
          break;
        } else {
          reference_count_return (result);
          dereference_parameter (port);
          dereference_parameter (result);
          return result;
        }
      }
    case SCC(')'):
      scm_i_input_error (FUNC_NAME, port, "unexpected banana\")\"", SCM_EOL);
      break;
    case SCM_EOF_VAL:
      dereference_return_value (port);
      return SCM_EOF_VAL;

    default:
      {
        usv_t c = SCM_CHAR (chr);
        if (((c >= '0') && (c <= '9'))
            || (strchr ("+-.", (char)c)))
          return (scm_read_number (chr, port));
        else
          return (scm_read_mixed_case_symbol (chr, port));
      }
    }
  }
}
#undef FUNC_NAME


/* Actual reader.  */

SCM_DEFINE (scm_read, "read", 0, 1, 0, 
            (SCM port),
	    "Read an s-expression from the input port @var{port}, or from\n"
	    "the current input port if @var{port} is not specified.\n"
	    "Any whitespace before the next token is discarded.")
#define FUNC_NAME s_scm_read
{

  SCM c;

  if (SCM_UNBNDP (port)) {
    port = scm_current_input_port ();
  }
  SCM_VALIDATE_OPINPORT (1, port);
  reference_parameter (port);

  c = flush_ws (port, (char *) NULL);
  if (SCM_EOF_OBJECT_P (c)) {
    dereference_parameter (port);
    return SCM_EOF_VAL;
  }
  scm_ungetc (c, port);

  return (scm_read_expression (rc_Fly (port)));
}
#undef FUNC_NAME


void
x_inclusion_read ()
{
  scm_c_define ("sys-boxed:undefined",  scm_cons (SCM_BOOL_F, SCM_UNDEFINED));
  scm_c_define ("sys-boxed:unspecified",scm_cons (SCM_BOOL_F, SCM_UNSPECIFIED));
  scm_c_define ("sys-boxed:eof",        scm_cons (SCM_BOOL_F, SCM_EOF_VAL));
#include "include/read.x"
}
