/* More List Processing
 */

#include "include/config.h"
#include "include/ihub.h"

#include "include/srfi-1.h"

#define srfi1_ilength scm_ilength   // XXX replace name in code

SCM
equal_trampoline (SCM proc, SCM arg1, SCM arg2)
{
  return scm_equal_p (arg1, arg2);
}

#include <stdio.h>

void scm_i_copy_list_n (int count, SCM list, SCM *head, SCM *tail);

inline void
scm_i_copy_list_n (int count, SCM list, SCM *head, SCM *tail)
{
  int length = 0;
  reference_parameter (list);

  if (count == 0) {
    dereference_parameter (list);
    *head = SCM_EOL;
    *tail = SCM_EOL;
  }

  if (!scm_is_pair (list)) {
    reference_count_return (list);  // XXX improper list extension
    dereference_parameter (list);
    *head = list;  // XXX improper list extension
    *tail = SCM_EOL;
    return;
  }

  *head = scm_i_cons (TMP_CAR (list), SCM_EOL);
  *tail = *head;
  reference_parameter (*head);
  reference_parameter (*tail);
  length++;

  SCM current = TMP_CDR (list);
  while (scm_is_pair (current) && length < count) {
    scm_i_set_cdr_x (*tail, scm_i_cons (TMP_CAR (current), SCM_EOL));
    *tail = fly_cdr (*tail);
    current = TMP_CDR (current);
    length++;
  }
  if (!scm_is_pair (current) && !scm_is_null (current))
    scm_i_set_cdr_x (*tail, current);  // XXX improper list extension

  rc_return_stack_reference (*head);
  dereference_parameter (*tail);  // list referenced
  dereference_parameter (list);
}


SCM_DEFINE (scm_srfi1_concatenate, "concatenate", 1, 0, 0,
            (SCM lstlst),
	    "Construct a list by appending all lists in @var{lstlst}.\n"
	    "\n"
	    "@code{concatenate} is the same as @code{(apply append\n"
	    "@var{lstlst})}.  It exists because some Scheme implementations\n"
	    "have a limit on the number of arguments a function takes, which\n"
	    "the @code{apply} might exceed.  In VSI there is no such\n"
	    "limit.")
#define FUNC_NAME s_scm_srfi1_concatenate
{
  SCM_VALIDATE_LIST (SCM_ARG1, lstlst);
  return scm_append (lstlst);
}
#undef FUNC_NAME


SCM_DEFINE (scm_srfi1_concatenate_x, "concatenate!", 1, 0, 0,
            (SCM lstlst),
	    "Construct a list by appending all lists in @var{lstlst}.  Those\n"
	    "lists may be modified to produce the result.\n"
	    "\n"
	    "@code{concatenate!} is the same as @code{(apply append!\n"
	    "@var{lstlst})}.  It exists because some Scheme implementations\n"
	    "have a limit on the number of arguments a function takes, which\n"
	    "the @code{apply} might exceed.  In VSI there is no such\n"
	    "limit.")
#define FUNC_NAME s_scm_srfi1_concatenate
{
  SCM_VALIDATE_LIST (SCM_ARG1, lstlst);
  return scm_append_x (lstlst);
}
#undef FUNC_NAME


SCM
scm_i_list_contains_p (SCM pred, scm_t_trampoline_2 equal_p,
                       SCM object, SCM list)
{ // RC: lax
  while (1) {
    if (!scm_is_pair (list))
      return SCM_BOOL_F;
    if (scm_is_true (equal_p (pred, TMP_CAR (list), object)))
      return list;
    list = TMP_CDR (list);
  }
  // (instruction pointer) control can't reach here
}

SCM_DEFINE (scm_list_contains_p, "list-contains?", 3, 0, 0,
	    (SCM pred, SCM object, SCM list), "")
#define FUNC_NAME s_scm_list_contains_p
{
  reference_parameter (pred);
  reference_parameter (object);
  reference_parameter (list);

  SCM answer = SCM_BOOL_F;
  scm_t_trampoline_2 equal_p;
  if (SCM_UNBNDP (pred)) {
    equal_p = equal_trampoline;
  } else {
    equal_p = scm_trampoline_2 (pred);
    SCM_ASSERT (equal_p, pred, SCM_ARG2, FUNC_NAME);
  }
  answer = scm_i_list_contains_p (pred, equal_p, object, list);

  reference_count_return (answer);
  reference_parameter (pred);
  reference_parameter (object);
  reference_parameter (list);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_fold, "fold", 3, 0, 0,
            (SCM proc, SCM init, SCM list1),
	    "Documentational")
#define FUNC_NAME s_scm_fold
{
  SCM lst;
  int argnum;
  reference_parameter (proc);
  reference_parameter (init);
  reference_parameter (list1);

  scm_t_trampoline_2 proc_tramp = scm_trampoline_2 (proc);
  SCM_ASSERT (proc_tramp, proc, SCM_ARG1, FUNC_NAME);

  // Viper Argument Order & Calling Conventions
  for ( ; scm_is_pair (list1); list1 = FLY_CDR (list1)) {
    init = rc_X (proc_tramp (proc, rc_Fly (init), TMP_CAR (list1)));
  }

  /* check below that list1 is a proper list, and done */
  lst = list1;
  argnum = 2;
    
  SCM_ASSERT_TYPE (scm_is_null (lst), lst, argnum, FUNC_NAME, "list");

  dereference_parameter (proc);
  dereference_parameter (list1);
  return (rc_return_stack_reference (init));
}
#undef FUNC_NAME


/* Note: Currently, scm_map applies PROC to the argument list(s)
   sequentially, starting with the first element(s).  This is used in
   the Scheme procedure `map-in-order', which guarantees sequential
   behaviour, is implemented using scm_map.  If the behaviour changes,
   we need to update `map-in-order'.
*/

SCM_REGISTER_PROC (s_srfi1_map_in_order, "map-in-order", 2, 0, 1, scm_map);

SCM_DEFINE (scm_srfi1_member, "member", 2, 1, 0,
           (SCM x, SCM list, SCM pred),
	    "Return the first sublist of @var{lst} whose @sc{car} is equal\n"
	    "to @var{x}.  If @var{x} does not appear in @var{lst}, return\n"
	    "@code{#f}.\n"
	    "\n"
	    "Equality is determined by @code{equal?}, or by the equality\n"
	    "predicate @var{=} if given.  @var{=} is called @code{(= @var{x}\n"
	    "elem)}, ie.@: with the given @var{x} first, so for example to\n"
	    "find the first element greater than 5,\n"
	    "\n"
	    "@example\n"
	    "(member 5 '(3 5 1 7 2 9) <) @result{} (7 2 9)\n"
	    "@end example\n"
	    "\n"
	    "This version of @code{member} extends the core @code{member} by\n"
	    "accepting an equality predicate.")
#define FUNC_NAME s_scm_srfi1_member
{
  reference_parameter (x);
  reference_parameter (list);
  reference_parameter (pred);

  scm_t_trampoline_2 equal_p;
  SCM_VALIDATE_LIST (2, list);
  if (SCM_UNBNDP (pred)) {
    equal_p = equal_trampoline;
  } else {
    equal_p = scm_trampoline_2 (pred);
    SCM_ASSERT (equal_p, pred, 3, FUNC_NAME);
  }

  SCM lst    = list;  // RC: list referenced
  SCM answer = SCM_BOOL_F;
  for (; !scm_is_null (lst); lst = TMP_CDR (lst)) {
    if (sharp_scm_is_true (equal_p (pred, x, TMP_CAR (lst)))) {
      answer = lst;
      // RC return reference declaration sequence starts here
      reference_count_return (answer);
      break;
    }
  }
  // RC return reference declaration sequence started above
  dereference_parameter (x);
  dereference_parameter (list);
  dereference_parameter (pred);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_assoc, "assoc", 2, 1, 0,
	    (SCM key, SCM alist, SCM pred),
	    "Behaves like @code{assq} but uses third argument @var{pred?}\n"
	    "for key comparison.  If @var{pred?} is not supplied,\n"
	    "@code{equal?} is used.  (Extended from R5RS.)\n")
#define FUNC_NAME s_scm_srfi1_assoc
{
  reference_parameter (key);
  reference_parameter (alist);
  reference_parameter (pred);

  SCM ls = alist;
  scm_t_trampoline_2 equal_p;
  if (SCM_UNBNDP (pred)) {
    equal_p = equal_trampoline;
  } else {
    equal_p = scm_trampoline_2 (pred);
    SCM_ASSERT (equal_p, pred, 3, FUNC_NAME);
  }

  for(; scm_is_pair (ls); ls = TMP_CDR (ls)) 
    {
      SCM tmp = TMP_CAR (ls);
      SCM_ASSERT_TYPE (scm_is_pair (tmp), alist, SCM_ARG2, FUNC_NAME,
		       "association list");
      if (scm_is_true (equal_p (pred, key, TMP_CAR (tmp)))) {
        reference_count_return (tmp);
        dereference_parameter (key);
        dereference_parameter (alist);
        dereference_parameter (pred);
	return tmp;
      }
    }
  SCM_ASSERT_TYPE (scm_is_null (ls), alist, SCM_ARG2, FUNC_NAME,
		   "association list");
  dereference_parameter (key);
  dereference_parameter (alist);
  dereference_parameter (pred);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_srfi1_not_pair_p, "not-pair?", 1, 0, 0,
            (SCM obj),
	    "Return @code{#t} is @var{obj} is not a pair, @code{#f}\n"
	    "otherwise.\n"
	    "\n"
	    "This is shorthand notation @code{(not (pair?  @var{obj}))} and\n"
	    "is supposed to be used for end-of-list checking in contexts\n"
	    "where dotted lists are allowed.")
#define FUNC_NAME s_scm_srfi1_not_pair_p
{
  return scm_from_bool (! scm_is_pair (obj));
}
#undef FUNC_NAME


SCM_DEFINE (scm_srfi1_tenth, "tenth", 1, 0, 0,
            (SCM lst),
	    "Return the tenth element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_tenth
{
  return scm_list_ref (lst, scm_from_int (9));
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_ninth, "ninth", 1, 0, 0,
            (SCM lst),
	    "Return the ninth element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_ninth
{
  return scm_list_ref (lst, scm_from_int (8));
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_eighth, "eighth", 1, 0, 0,
            (SCM lst),
	    "Return the eighth element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_eighth
{
  return scm_list_ref (lst, SCM_I_MAKINUM (7));
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_seventh, "seventh", 1, 0, 0,
            (SCM lst),
	    "Return the seventh element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_seventh
{
  return scm_list_ref (lst, scm_from_int (6));
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_sixth, "sixth", 1, 0, 0,
            (SCM lst),
	    "Return the sixth element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_sixth
{
  return scm_list_ref (lst, scm_from_int (5));
}
#undef FUNC_NAME

SCM_DEFINE (scm_srfi1_fifth, "fifth", 1, 0, 0,
            (SCM lst),
	    "Return the fifth element of @var{lst}.")
#define FUNC_NAME s_scm_srfi1_fifth
{
  return scm_list_ref (lst, SCM_I_MAKINUM (4));
}
#undef FUNC_NAME


void
x_inclusion_list_processing (void)
{
#include "include/srfi-1.x"
}

/* End of srfi-1.c.  */
