#include "config.h"

#include <assert.h>

#include "include/_scm.h"
#include "include/eval.h"
#include "include/alist.h"
#include "include/ports.h"
#include "include/smob.h"

#include "include/dynwind.h"


scm_t_tag tc16_frame;
scm_t_tag tc16_winder;

/* {Dynamic wind}
 
   Things that can be on the wind list:

   #<frame>
   #<winder>
   (enter-proc . leave-proc)     dynamic-wind
   (tag . jmpbuf)                catch
   (tag . pre-unwind-data)       throw-handler / lazy-catch
     tag is either a symbol or a boolean

*/

SCM global_dynwind_box = SCM_UNSPECIFIED;


unsigned long global_counter_dynwind_fuel = 0;

SCM_DEFINE (scm_dynamic_wind, "dynamic-wind", 3, 0, 0,
           (SCM in_guard, SCM thunk, SCM out_guard),
	    "All three arguments must be 0-argument procedures.\n"
	    "@var{in_guard} is called, then @var{thunk}, then\n"
	    "@var{out_guard}.\n"
	    "\n"
	    "If, any time during the execution of @var{thunk}, the\n"
	    "continuation of the @code{dynamic_wind} expression is escaped\n"
	    "non-locally, @var{out_guard} is called.  If the continuation of\n"
	    "the dynamic-wind is re-entered, @var{in_guard} is called.  Thus\n"
	    "@var{in_guard} and @var{out_guard} may be called any number of\n"
	    "times.\n"
	    "@lisp\n"
	    "(define x 'normal-binding)\n"
	    "@result{} x\n"
	    "(define a-cont  (call-with-current-continuation\n"
	    "		  (lambda (escape)\n"
	    "		     (let ((old-x x))\n"
	    "		       (dynamic-wind\n"
	    "			  ;; in-guard:\n"
	    "			  ;;\n"
	    "			  (lambda () (set! x 'special-binding))\n"
	    "\n"
	    "			  ;; thunk\n"
	    "			  ;;\n"
	    "		 	  (lambda () (display x) (newline)\n"
	    "				     (call-with-current-continuation escape)\n"
	    "				     (display x) (newline)\n"
	    "				     x)\n"
	    "\n"
	    "			  ;; out-guard:\n"
	    "			  ;;\n"
	    "			  (lambda () (set! x old-x)))))))\n"
	    "\n"
	    ";; Prints:\n"
	    "special-binding\n"
	    ";; Evaluates to:\n"
	    "@result{} a-cont\n"
	    "x\n"
	    "@result{} normal-binding\n"
	    "(a-cont #f)\n"
	    ";; Prints:\n"
	    "special-binding\n"
	    ";; Evaluates to:\n"
	    "@result{} a-cont  ;; the value of the (define a-cont...)\n"
	    "x\n"
	    "@result{} normal-binding\n"
	    "a-cont\n"
	    "@result{} special-binding\n"
	    "@end lisp")
#define FUNC_NAME s_scm_dynamic_wind
{
  reference_parameter (in_guard);
  reference_parameter (thunk);
  reference_parameter (out_guard);

  SCM ans, old_winds;
  SCM_ASSERT (scm_is_true (scm_thunk_p (out_guard)),
	      out_guard,
	      SCM_ARG3, FUNC_NAME);

  scm_call_0 (in_guard);

  old_winds = scm_i_dynwinds ();
  reference_parameter (old_winds);

  global_counter_dynwind_fuel++; 
  scm_i_set_dynwinds (scm_acons (in_guard, out_guard, old_winds));

  ans = scm_call_0 (thunk);
  reference_parameter (ans);

  scm_i_set_dynwinds (old_winds);
  dereference_parameter (old_winds);

  scm_call_0 (out_guard);

  dereference_parameter (in_guard);
  dereference_parameter (thunk);
  dereference_parameter (out_guard);
  return (rc_return_stack_reference (ans));
}
#undef FUNC_NAME

SCM
scm_dynwind_begin (scm_t_dynwind_flags flags)
{
  SCM f;
  SCM_NEWSMOB (f, tc16_frame, 0);
  if (flags & SCM_F_DYNWIND_REWINDABLE)
    SCM_SET_SMOB_FLAGS (f, FRAME_F_REWINDABLE);
  scm_i_set_dynwinds (scm_cons (f, scm_i_dynwinds ()));

  return f;
}

void
scm_dynwind_end (SCM token)
{
  reference_parameter (token);

  SCM winds;

  /* Unwind upto and including the next frame entry.  We can only
     encounter #<winder> entries on the way.
   */

  winds = scm_i_dynwinds ();
  reference_parameter (winds);
  // SCM prev = winds;

  while (scm_is_pair (winds)) {
    SCM entry = TMP_CAR (winds);
    reference_parameter (entry);

    if (FRAME_P (entry)) {
      if (entry == token) {
        winds = FLY_CDR (winds);
        scm_i_set_dynwinds (winds);

        dereference_parameter (entry);
        dereference_parameter (winds);
        dereference_parameter (token);
        return;

      } else  {
        /* seatbelt for correctly paired begin/end usage */
        fprintf (stderr, " * you didn't call me * ");
        abort ();
      }
    }

    winds = FLY_CDR (winds);
    scm_i_set_dynwinds (winds);

    /* assert (WINDER_P (entry)); */
    if (!WINDER_P (entry)) {
      if (scm_is_pair (entry)) {
        fprintf (stderr, "scm_dynwind_end: a pair: %p\n",
                 (void *) entry);
      } else {
        fprintf (stderr, "scm_dynwind_end: WARNING:"
                 " wrong type element: %p\n",
                 (void *) entry);
      }
    } else if (!WINDER_REWIND_P (entry) && WINDER_EXPLICIT_P (entry)) {
      WINDER_PROC(entry) (WINDER_DATA (entry));
    }
    dereference_parameter (entry);
  }

  fprintf (stderr, "\nVSI-C: scm_dynwind_end: ERROR: out of dynwinds\n");
  abort ();
}

void
scm_dynwind_unwind_handler (void (*proc) (void *), void *data,
			    scm_t_wind_flags flags)
{
  SCM w;
  SCM_NEWSMOB2 (w, tc16_winder,	(scm_t_bits) proc, (scm_t_bits) data);
  if (flags & SCM_F_WIND_EXPLICITLY)
    SCM_SET_SMOB_FLAGS (w, WINDER_F_EXPLICIT);
  scm_i_set_dynwinds (scm_cons (w, scm_i_dynwinds ()));
}

void
scm_dynwind_rewind_handler (void (*proc) (void *), void *data,
			    scm_t_wind_flags flags)
{
  SCM w;
  SCM_NEWSMOB2 (w, tc16_winder,	(scm_t_bits) proc, (scm_t_bits) data);
  SCM_SET_SMOB_FLAGS (w, WINDER_F_REWIND);
  scm_i_set_dynwinds (scm_cons (w, scm_i_dynwinds ()));
  if (flags & SCM_F_WIND_EXPLICITLY)
    proc (data);
}


void
scm_dynwind_free (void *mem)
{
  scm_dynwind_unwind_handler (free, mem, SCM_F_WIND_EXPLICITLY);
}

#ifdef GUILE_DEBUG
SCM_DEFINE (scm_wind_chain, "wind-chain", 0, 0, 0, 
            (),
	    "Return the current wind chain. The wind chain contains all\n"
	    "information required by @code{dynamic-wind} to call its\n"
	    "argument thunks when entering/exiting its scope.")
#define FUNC_NAME s_scm_wind_chain
{
  return scm_i_dynwinds ();
}
#undef FUNC_NAME
#endif

void
scm_swap_bindings (SCM vars, SCM vals)
{
  reference_parameter (vars);
  reference_parameter (vals);

  SCM tmp;
  while (SCM_NIMP (vals))
    {
      tmp = SCM_VARIABLE_REF (TMP_CAR (vars));
      SCM_VARIABLE_SET (TMP_CAR (vars), TMP_CAR (vals));
      SCM_SETCAR (vals, tmp);
      vars = FLY_CDR (vars);
      vals = FLY_CDR (vals);
    }

  dereference_parameter (vars);
  dereference_parameter (vals);
}

void
scm_dowinds (SCM to, long delta)
{
  reference_parameter (to);

  SCM tmp = SCM_UNSPECIFIED;

 tail:
  dereference_parameter (tmp);
  tmp = scm_i_dynwinds ();
  reference_parameter (tmp);

  if (delta < 0) {
    fprintf (stderr, "VSI: broken wind delta\n");
    abort ();

  } else if (!scm_is_eq (to, tmp)) {
    SCM wind;
    SCM wind_elt;
    SCM wind_key;

    wind = scm_i_dynwinds ();
    reference_parameter (wind);

    wind_elt = TMP_CAR (wind);
    scm_i_set_dynwinds (TMP_CDR (wind));

    if (FRAME_P (wind_elt)) {
      /* Nothing to do. */

    } else if (WINDER_P (wind_elt)) {
      if (!WINDER_REWIND_P (wind_elt))
        WINDER_PROC (wind_elt) (WINDER_DATA (wind_elt));

    } else {
      // ??? Check this please.
      wind_key = TMP_CAR (wind_elt);
      if (SCM_NIMP (wind_key)) {
        if (scm_is_pair (wind_key)) {
          if (SCM_VARIABLEP (TMP_CAR (wind_key))) {
            reference_parameter (wind_key);
            reference_parameter (wind_elt);

            scm_swap_bindings (wind_key, TMP_CDR (wind_elt));

            dereference_parameter (wind_key);
            dereference_parameter (wind_elt);
          }
        } else if (SCM_TYP3 (wind_key) == scm_tc3_closure) {
          rc_Void (scm_call_0 (TMP_CDR (wind_elt)));
        }
      }
    }

    dereference_parameter (wind);

    delta--;
    goto tail;		/* scm_dowinds(to, delta-1); */
  }

  dereference_parameter (tmp);
  dereference_parameter (to);
  return;
}

/* Debugging Routines */
SCM
debug_dynwinds ()
{
  debug_trim_assert_list (global_dynwinds_box);
  return (scm_cdr (global_dynwinds_box));
}


long *global_winder_count = NULL;
long *global_frame_count  = NULL;

inline void
trim_release_winder (SCM winder)
{
  if (WINDER_MARK_P (winder)) {
    dereference (winder, (SCM) WINDER_DATA (winder));
  }
  (*global_winder_count)--;
  return;
}

void
scm_init_dynwind ()
{
  global_dynwinds_box = scm_cons (SCM_BOOL_F, SCM_EOL);
  scm_permanent_object (global_dynwinds_box);

  tc16_frame = scm_make_smob_type ("frame", 0);
  tc16_winder = scm_make_smob_type ("winder", 0);

  global_frame_count  = \
    &(scm_smobs[SCM_TC2SMOBNUM (tc16_frame)].number_of_objects);
  global_winder_count = \
    &(scm_smobs[SCM_TC2SMOBNUM (tc16_winder)].number_of_objects);
}

void
x_inclusion_dynwind ()
{
#include "include/dynwind.x"
}
