#include "include/config.h"

#include <stdio.h>
#include "include/_scm.h"
#include "include/smob.h"
#include "include/alist.h"
#include "include/eval.h"
#include "include/eq.h"
#include "include/dynwind.h"
#include "include/debug.h"
#include "include/ports.h"
#include "include/validate.h"
#include "include/throw.h"
#include "include/init.h"

scm_t_tag tc16_jmpbuffer;

static int
jmpbuffer_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  scm_puts ("#<jmpbuffer ", port);
  scm_puts (JBACTIVE(exp) ? "(active) " : "(inactive) ", port);
  scm_uintprint((scm_t_bits) JBJMPBUF (exp), 16, port);
  scm_putc ('>', port);
  return 1 ;
}

static SCM
make_jmpbuf (void)
{
  SCM answer;
  SCM_NEWSMOB2 (answer, tc16_jmpbuffer, 0, 0);
  SETJBJMPBUF(answer, (scm_i_jmp_buf *)0);
  DEACTIVATEJB(answer);
  return answer;
}


/* scm_c_catch (the guts of catch) */

struct jmp_buf_and_retval	/* use only on the stack, in scm_catch */
{
  scm_i_jmp_buf buf;		/* must be first */
  SCM throw_tag;
  SCM retval;
};

/* These are the structures we use to store pre-unwind handling (aka
   "lazy") information for a regular catch, and put on the wind list
   for a "lazy" catch.  They store the pre-unwind handler function to
   call, and the data pointer to pass through to it.  It's not a
   Scheme closure, but it is a function with data, so the term
   "closure" is appropriate in its broader sense.

   (We don't need anything like this to run the normal (post-unwind)
   catch handler, because the same C frame runs both the body and the
   handler.)  */

struct pre_unwind_data {
  scm_t_catch_handler handler;
  void *handler_data;
  int running;
  int lazy_catch_p;
};


/* scm_c_catch is the guts of catch.  It handles all the mechanics of
   setting up a catch target, invoking the catch body, and perhaps
   invoking the handler if the body does a throw.

   TAG is the catch tag.  Typically, this is a symbol, but this
   function doesn't actually care about that.

   BODY is a pointer to a C function which runs the body of the catch;
   this is the code you can throw from.  We call it like this:
      BODY (BODY_DATA)
   where:
      BODY_DATA is just the BODY_DATA argument we received; we pass it
	 through to BODY as its first argument.  The caller can make
	 BODY_DATA point to anything useful that BODY might need.

   HANDLER is a pointer to a C function to deal with a throw to TAG,
   should one occur.  We call it like this:
      HANDLER (HANDLER_DATA, THROWN_TAG, THROW_ARGS)
   where
      HANDLER_DATA is the HANDLER_DATA argument we recevied; it's the
         same idea as BODY_DATA above.
      THROWN_TAG is the tag that the user threw to; usually this is
         TAG, but it could be something else if TAG was #t (i.e., a
         catch-all), or the user threw to a jmpbuf.
      THROW_ARGS is the list of arguments the user passed to the THROW
         function, after the tag.

   BODY_DATA is just a pointer we pass through to BODY.  HANDLER_DATA
   is just a pointer we pass through to HANDLER.  We don't actually
   use either of those pointers otherwise ourselves.  The idea is
   that, if our caller wants to communicate something to BODY or
   HANDLER, it can pass a pointer to it as MUMBLE_DATA, which BODY and
   HANDLER can then use.  Think of it as a way to make BODY and
   HANDLER closures, not just functions; MUMBLE_DATA points to the
   enclosed variables.


   XXX DEPRECATION / If the following pattern exists purge it, please.

   Of course, it's up to the caller to make sure that any data a
   MUMBLE_DATA needs is protected from GC.  A common way to do this is
   to make MUMBLE_DATA a pointer to data stored in an automatic
   structure variable; since the collector must scan the stack for
   references anyway, this assures that any references in MUMBLE_DATA
   will be found.  */

SCM
scm_c_catch (SCM tag,
	     scm_t_catch_body body, void *body_data,
	     scm_t_catch_handler handler, void *handler_data,
	     scm_t_catch_handler pre_unwind_handler,
             void *pre_unwind_handler_data)
{
  reference_parameter (tag);

  struct jmp_buf_and_retval jbr;
  jbr.throw_tag = SCM_UNSPECIFIED;
  jbr.retval    = SCM_UNSPECIFIED; // XXX that's unspecified in a wide sense
  SCM jmpbuf;
  SCM answer;
  struct pre_unwind_data pre_unwind;

  jmpbuf = make_jmpbuf (); reference_parameter (jmpbuf);
  answer = SCM_EOL;

  global_counter_dynwind_fuel++;
  scm_i_set_dynwinds (scm_acons (tag, jmpbuf, scm_i_dynwinds ()));
  dereference_parameter (jmpbuf); // jmpbuf is data graph referenced now (CLR3)
  SETJBJMPBUF(jmpbuf, &jbr);
  // SCM_SETJBDFRAME(jmpbuf, scm_i_last_debug_frame ());

  pre_unwind.handler = pre_unwind_handler;
  pre_unwind.handler_data = pre_unwind_handler_data;
  pre_unwind.running = 0;
  pre_unwind.lazy_catch_p = 0;
  SCM_SETJBPREUNWIND(jmpbuf, &pre_unwind);

  // RC checkpoint: tag, jmpbuf (data graph referenced / CLR3)
  // CLR3: /jmpbuf/ will be dereferenced by an unwind
  //       /tag/ will end up with our stale stack reference but
  //         it is presumed to be a symbol and thus a global
  //  This happened for cee.scm: jmpbuffers accumulate when a
  // long jump occurs "over" our catch stack frame here. Future
  // stack frame might includes bits with hints to what dereference
  // as part of a dynamic wind. On the other hand dynamic winds
  // rise the complexity in a lot of spots and the current forms
  // make it even more complicated in Scheme.
  // XXX jmpbuffers still accumulate
 
  if (SCM_I_SETJMP (jbr.buf) == 0) { // regular return
    ACTIVATEJB (jmpbuf);
    answer = body (body_data);
    reference_parameter (answer);
    DEACTIVATEJB (jmpbuf);
    scm_i_set_dynwinds (SCM_CDR (scm_i_dynwinds ()));

  } else {                       // non local return
    SCM throw_tag;
    SCM throw_args;

    DEACTIVATEJB (jmpbuf);
    scm_i_set_dynwinds (SCM_CDR (scm_i_dynwinds ()));
    throw_args = jbr.retval;    // initialized with data in ithrow
    throw_tag = jbr.throw_tag;  // initialized with data in ithrow

    reference_parameter (throw_tag);
    reference_parameter (throw_args);
    heap_dereference (jbr, jbr.retval);     // referenced by ithrow
    heap_dereference (jbr, jbr.throw_tag);  // referenced by ithrow
    jbr.throw_tag = SCM_EOL;
    jbr.retval = SCM_EOL;

    answer = handler (handler_data, throw_tag, throw_args);

    reference_parameter (answer);
    dereference_parameter (throw_tag);
    dereference_parameter (throw_args);
  }

  dereference_parameter (tag);
  return (rc_return_stack_reference (answer));;
}

SCM
scm_internal_catch (SCM tag,
		    scm_t_catch_body body, void *body_data,
		    scm_t_catch_handler handler, void *handler_data)
{
  return scm_c_catch(tag,
		     body, body_data,
		     handler, handler_data,
		     NULL, NULL);
}



/* The smob tag for pre_unwind_data smobs.  */
scm_t_tag tc16_pre_unwind_data;

/* Strictly speaking, we could just pass a zero for our print
   function, because we don't need to print them.  They should never
   appear in normal data structures, only in the wind list.  However,
   it might be nice for debugging someday... */
static int
pre_unwind_data_print (SCM closure, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  struct pre_unwind_data *c = (struct pre_unwind_data *) SCM_CELL_WORD_1 (closure);
  char buf[200];

  sprintf (buf, "#<pre-unwind-data 0x%lx 0x%lx>",
	   (long) c->handler, (long) c->handler_data);
  scm_puts (buf, port);

  return 1;
}


/* Given a pointer to a pre_unwind_data structure, return a smob for it,
   suitable for inclusion in the wind list.  */
static SCM
make_pre_unwind_data (struct pre_unwind_data *c)
{
  SCM_RETURN_NEWSMOB (tc16_pre_unwind_data, c);
}


SCM
scm_c_with_throw_handler (SCM tag,
			  scm_t_catch_body body,
			  void *body_data,
			  scm_t_catch_handler handler,
			  void *handler_data,
			  int lazy_catch_p)
{
  reference_parameter (tag);

  SCM pre_unwind, answer;
  struct pre_unwind_data c;

  c.handler = handler;
  c.handler_data = handler_data;
  c.running = 0;
  c.lazy_catch_p = lazy_catch_p;
  pre_unwind = make_pre_unwind_data (&c); // RC pass through

  global_counter_dynwind_fuel++;
  scm_i_set_dynwinds (scm_acons (tag, pre_unwind, scm_i_dynwinds ()));

  answer = (*body) (body_data);
  reference_parameter (answer);

  scm_i_set_dynwinds (SCM_CDR (scm_i_dynwinds ()));

  dereference_parameter (tag);
  return (rc_return_stack_reference (answer));
}

/* Exactly like scm_internal_catch, except:
   - It does not unwind the stack (this is the major difference).
   - The handler is not allowed to return.  */
SCM
scm_internal_lazy_catch (SCM tag, scm_t_catch_body body, void *body_data, scm_t_catch_handler handler, void *handler_data)
{
  return scm_c_with_throw_handler (tag, body, body_data,
                                   handler, handler_data, 1);
}


static SCM
ss_handler (void *data SCM_UNUSED, SCM tag, SCM throw_args)
{ // XXX compatibility procedure
  return (scm_throw (tag, throw_args));
}

struct cwss_data
{
  SCM tag;
  scm_t_catch_body body;
  void *data;
};

static SCM
cwss_body (void *data)
{
  struct cwss_data *d = data;
  RC_RETURN (scm_internal_lazy_catch (d->tag, d->body, d->data,
                                      ss_handler, NULL));
}

SCM
scm_internal_stack_catch (SCM tag,
			  scm_t_catch_body body,
			  void *body_data,
			  scm_t_catch_handler handler,
			  void *handler_data)
{
  reference_parameter (tag);

  struct cwss_data d;
  d.tag = tag;
  d.body = body;
  d.data = body_data;
  SCM answer = \
    scm_internal_catch (tag, cwss_body, &d, handler, handler_data);
  reference_parameter (answer);

  dereference_parameter (tag);
  return (rc_return_stack_reference (answer));
}



/* body and handler functions for use with any of the above catch variants */

/* This is a body function you can pass to scm_internal_catch if you
   want the body to be like Scheme's `catch' --- a thunk.

   BODY_DATA is a pointer to a scm_body_thunk_data structure, which
   contains the Scheme procedure to invoke as the body, and the tag
   we're catching.  */

SCM
scm_body_thunk (void *body_data)
{
  struct scm_body_thunk_data *c = (struct scm_body_thunk_data *) body_data;

  heap_reference (c, c->body_proc);
  SCM answer = scm_call_0 (c->body_proc);
  heap_dereference (c, c->body_proc);

  return (answer);
}


/* This is a handler function you can pass to scm_internal_catch if
   you want the handler to act like Scheme's catch: (throw TAG ARGS ...)
   applies a handler procedure to (TAG ARGS ...).

   If the user does a throw to this catch, this function runs a
   handler procedure written in Scheme.  HANDLER_DATA is a pointer to
   an SCM variable holding the Scheme procedure object to invoke.  It
   ought to be a pointer to an automatic variable (i.e., one living on
   the stack), or the procedure object should be otherwise protected
   from GC.  */
SCM
scm_handle_by_proc (void *handler_data, SCM tag, SCM throw_args)
{
  SCM *handler_proc_p = (SCM *) handler_data;

  RC_RETURN (scm_apply_1 (*handler_proc_p, tag, throw_args));
}

/* SCM_HANDLE_BY_PROC_CATCHING_ALL is like SCM_HANDLE_BY_PROC but
   catches all throws that the handler might emit itself.  The handler
   used for these `secondary' throws is SCM_HANDLE_BY_MESSAGE_NO_EXIT.  */

struct hbpca_data {
  SCM proc;
  SCM args;
};

static SCM
hbpca_body (void *body_data)
{
  struct hbpca_data *data = (struct hbpca_data *)body_data;
  RC_RETURN (scm_apply_0 (data->proc, data->args));
}

SCM
scm_handle_by_proc_catching_all (void *handler_data, SCM tag, SCM throw_args)
{
  reference_parameter (tag);
  reference_parameter (throw_args);

  SCM *handler_proc_p = (SCM *) handler_data;
  struct hbpca_data data;
  data.proc = *handler_proc_p;
  data.args = scm_cons (tag, throw_args);
  heap_reference (data, data.args);

  SCM answer = scm_internal_catch (SCM_BOOL_T,
                                   hbpca_body, &data,
                                   scm_handle_by_message_noexit, NULL);
  reference_parameter (answer);
  heap_dereference (data, data.args);
  dereference_parameter (tag);
  dereference_parameter (throw_args);
  return (rc_return_stack_reference (answer));
}

/* Derive the an exit status from the arguments to (quit ...).  */
int
scm_exit_status (SCM args)
{
  reference_parameter (args);

  int answer = 0;

  if (!scm_is_null (args)) {
    SCM cqa = SCM_CAR (args);
      
    if (scm_is_integer (cqa))
      answer = (scm_to_int (cqa));
    else if (scm_is_false (cqa))
      answer = 1;
    }

  dereference_parameter (args);
  return answer;
}
	

static void
handler_message (void *handler_data, SCM tag, SCM args)
{
  reference_parameter (tag);
  reference_parameter (args);

  char *prog_name = (char *) handler_data;
  SCM p = scm_current_error_port ();
  reference_parameter (p);

  if (scm_ilength (args) == 4) {
    SCM subr    = TMP_CAR (args);
    SCM message = TMP_CADR (args);
    SCM parts   = TMP_CADDR (args);
    SCM rest    = TMP_CADDDR (args);

    scm_i_display_error (SCM_UNDEFINED, p, subr, message, parts, rest);

  } else {
    if (! prog_name)
      prog_name = "vsi";

    scm_puts (prog_name, p);
    scm_puts (": ", p);

    scm_puts ("uncaught throw to ", p);
    scm_prin1 (tag, p, 0);
    scm_puts (": ", p);
    scm_prin1 (args, p, 1);
    scm_putc ('\n', p);
  }

  dereference_parameter (p);
  dereference_parameter (tag);
  dereference_parameter (args);
}


/* This is a handler function to use if you want scheme to print a
   message and die.  Useful for dealing with throws to uncaught keys
   at the top level.

   At boot time, we establish a catch-all that uses this as its handler.
   1) If the user wants something different, they can use (catch #t
   ...) to do what they like.
   2) Outside the context of a read-eval-print loop, there isn't
   anything else good to do; VSI-C should not assume the existence
   of a read-eval-print loop.
   3) Given that we shouldn't do anything complex, it's much more
   robust to do it in C code.

   HANDLER_DATA, if non-zero, is assumed to be a char * pointing to a
   message header to print; if zero, we use "guile" instead.  That
   text is followed by a colon, then the message described by ARGS.  */

/* Dirk:FIXME:: The name of the function should make clear that the
 * application gets terminated.
 */

SCM
scm_handle_by_message (void *handler_data, SCM tag, SCM args)
{
  reference_parameter (tag);
  reference_parameter (args);

  if (scm_is_true (scm_eq_p (tag, scm_from_locale_symbol ("quit"))))
    exit (scm_exit_status (args));

  handler_message (handler_data, tag, args);
  // scm_i_pthread_exit (NULL);
  exit (13);  // XXX some some

  /* this point not reached, but suppress gcc warning about no return value
     in case scm_i_pthread_exit isn't marked as "noreturn" (which seemed not
     to be the case on cygwin for instance) */
  return SCM_BOOL_F;
}


/* This is just like scm_handle_by_message, but it doesn't exit; it
   just returns #f.  It's useful in cases where you don't really know
   enough about the body to handle things in a better way, but don't
   want to let throws fall off the bottom of the wind list.  */
SCM
scm_handle_by_message_noexit (void *handler_data, SCM tag, SCM args)
{
  reference_parameter (tag);
  reference_parameter (args);

  if (scm_is_true (scm_eq_p (tag, scm_from_locale_symbol ("quit")))) {
    dereference_parameter (tag);
    exit (scm_exit_status (rc_Fly (args)));
  }

  handler_message (handler_data, tag, args);

  dereference_parameter (tag);
  dereference_parameter (args);
  return SCM_BOOL_F;
}


SCM
scm_handle_by_throw (void *handler_data SCM_UNUSED, SCM tag, SCM args)
{
  scm_ithrow (tag, args, 1);
  return SCM_UNSPECIFIED;  /* never returns */
}



/* the Scheme-visible CATCH, WITH-THROW-HANDLER and LAZY-CATCH functions */

SCM_DEFINE (scm_catch_with_pre_unwind_handler, "catch", 3, 1, 0,
	    (SCM key, SCM thunk, SCM handler, SCM pre_unwind_handler),
	    "Invoke @var{thunk} in the dynamic context of @var{handler} for\n"
	    "exceptions matching @var{key}.  If thunk throws to the symbol\n"
	    "@var{key}, then @var{handler} is invoked this way:\n"
	    "@lisp\n"
	    "(handler key args ...)\n"
	    "@end lisp\n"
	    "\n"
	    "@var{key} is a symbol or @code{#t}.\n"
	    "\n"
	    "@var{thunk} takes no arguments.  If @var{thunk} returns\n"
	    "normally, that is the return value of @code{catch}.\n"
	    "\n"
	    "Handler is invoked outside the scope of its own @code{catch}.\n"
	    "If @var{handler} again throws to the same key, a new handler\n"
	    "from further up the call chain is invoked.\n"
	    "\n"
	    "If the key is @code{#t}, then a throw to @emph{any} symbol will\n"
	    "match this call to @code{catch}.\n"
	    "\n"
	    "If a @var{pre-unwind-handler} is given and @var{thunk} throws\n"
	    "an exception that matches @var{key}, Guile calls the\n"
	    "@var{pre-unwind-handler} before unwinding the dynamic state and\n"
	    "invoking the main @var{handler}.  @var{pre-unwind-handler} should\n"
	    "be a procedure with the same signature as @var{handler}, that\n"
	    "is @code{(lambda (key . args))}.  It is typically used to save\n"
	    "the stack at the point where the exception occurred, but can also\n"
	    "query other parts of the dynamic state at that point, such as\n"
	    "fluid values. XXX fluids are gone\n"
	    "\n"
	    "A @var{pre-unwind-handler} can exit either normally or non-locally.\n"
	    "If it exits normally, Guile unwinds the stack and dynamic context\n"
	    "and then calls the normal (third argument) handler.  If it exits\n"
	    "non-locally, that exit determines the continuation.")
#define FUNC_NAME s_scm_catch_with_pre_unwind_handler
{
  reference_parameter (key);
  reference_parameter (thunk);
  reference_parameter (handler);
  reference_parameter (pre_unwind_handler);

  struct scm_body_thunk_data c;

  SCM_ASSERT (scm_is_symbol (key) || scm_is_eq (key, SCM_BOOL_T),
	      key, SCM_ARG1, FUNC_NAME);

  c.tag = key;
  c.body_proc = thunk;

  /* scm_c_catch takes care of all the mechanics of setting up a catch
     key; we tell it to call scm_body_thunk to run the body, and
     scm_handle_by_proc to deal with any throws to this catch.  The
     former receives a pointer to c, telling it how to behave.  The
     latter receives a pointer to HANDLER, so it knows who to
     call.  */
  SCM answer = \
    scm_c_catch (key, scm_body_thunk, &c, 
                 scm_handle_by_proc, &handler,
                 SCM_UNBNDP (pre_unwind_handler) ? NULL : scm_handle_by_proc,
                 &pre_unwind_handler);
  reference_parameter (answer);

  dereference_parameter (key);
  dereference_parameter (thunk);
  dereference_parameter (handler);
  dereference_parameter (pre_unwind_handler);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME

/* The following function exists to provide backwards compatibility
   for the C scm_catch API.  Otherwise we could just change
   "scm_catch_with_pre_unwind_handler" above to "scm_catch". */
SCM
scm_catch (SCM key, SCM thunk, SCM handler)
{
  return scm_catch_with_pre_unwind_handler (key, thunk, handler, SCM_UNDEFINED);
}


SCM_DEFINE (scm_with_throw_handler, "with-throw-handler", 3, 0, 0,
	    (SCM key, SCM thunk, SCM handler),
	    "Add @var{handler} to the dynamic context as a throw handler\n"
	    "for key @var{key}, then invoke @var{thunk}.")
#define FUNC_NAME s_scm_with_throw_handler
{
  reference_parameter (key);
  reference_parameter (thunk);
  reference_parameter (handler);

  struct scm_body_thunk_data c;

  SCM_ASSERT (scm_is_symbol (key) || scm_is_eq (key, SCM_BOOL_T),
	      key, SCM_ARG1, FUNC_NAME);

  c.tag = key;
  c.body_proc = thunk;

  /* scm_c_with_throw_handler takes care of the mechanics of setting
     up a throw handler; we tell it to call scm_body_thunk to run the
     body, and scm_handle_by_proc to deal with any throws to this
     handler.  The former receives a pointer to c, telling it how to
     behave.  The latter receives a pointer to HANDLER, so it knows
     who to call.  */
  SCM answer = scm_c_with_throw_handler (key,
                                         scm_body_thunk, &c, 
                                         scm_handle_by_proc, &handler,
                                         0);
  reference_parameter (answer);
  dereference_parameter (key);
  dereference_parameter (thunk);
  dereference_parameter (handler);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME

SCM_DEFINE (scm_lazy_catch, "lazy-catch", 3, 0, 0,
	    (SCM key, SCM thunk, SCM handler),
	    "This behaves exactly like @code{catch}, except that it does\n"
	    "not unwind the stack before invoking @var{handler}.\n"
	    "If the @var{handler} procedure returns normally, Guile\n"
	    "rethrows the same exception again to the next innermost catch,\n"
	    "lazy-catch or throw handler.  If the @var{handler} exits\n"
	    "non-locally, that exit determines the continuation.")
#define FUNC_NAME s_scm_lazy_catch
{
  reference_parameter (key);
  reference_parameter (thunk);
  reference_parameter (handler);

  struct scm_body_thunk_data c;

  SCM_ASSERT (scm_is_symbol (key) || scm_is_eq (key, SCM_BOOL_T),
	      key, SCM_ARG1, FUNC_NAME);

  c.tag = key;
  c.body_proc = thunk;
  heap_reference (c, key);
  heap_reference (c, thunk);

  /* scm_internal_lazy_catch takes care of all the mechanics of
     setting up a lazy catch key; we tell it to call scm_body_thunk to
     run the body, and scm_handle_by_proc to deal with any throws to
     this catch.  The former receives a pointer to c, telling it how
     to behave.  The latter receives a pointer to HANDLER, so it knows
     who to call.  */
  SCM answer = scm_internal_lazy_catch (rc_Fly (key),
                                        scm_body_thunk, &c, 
                                        scm_handle_by_proc, &handler);
  reference_parameter (answer);
  heap_dereference (c, key);
  heap_dereference (c, thunk);
  dereference_parameter (key);
  dereference_parameter (thunk);
  dereference_parameter (handler);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME



/* throwing */

static void toggle_pre_unwind_running (void *data)
{
  struct pre_unwind_data *pre_unwind = (struct pre_unwind_data *)data;
  pre_unwind->running = !pre_unwind->running;
}

SCM_DEFINE (scm_throw, "throw", 1, 0, 1,
           (SCM key, SCM args),
	    "Invoke the catch form matching @var{key}, passing @var{args} to the\n"
	    "@var{handler}.  \n\n"
	    "@var{key} is a symbol.  It will match catches of the same symbol or of\n"
	    "@code{#t}.\n\n"
	    "If there is no handler at all, Guile prints an error and then exits.")
#define FUNC_NAME s_scm_throw
{
  SCM_VALIDATE_SYMBOL (1, key);
  return scm_ithrow (key, args, 1);
}
#undef FUNC_NAME

SCM
scm_ithrow (SCM key, SCM args, int noreturn SCM_UNUSED)
{
  reference_parameter (key);
  reference_parameter (args);

  SCM jmpbuf = SCM_UNDEFINED;
  SCM wind_goal;

  SCM dynpair = SCM_UNDEFINED;
  SCM winds = SCM_UNSPECIFIED;

 rethrow:  // RC: key args
  winds = scm_i_dynwinds ();
  reference_parameter (winds);

  /* Search the wind list for an appropriate catch.
     "Waiter, please bring us the wind list." */
  for (;
       scm_is_pair (winds);
       winds = FLY_CDR (winds)) {

    dynpair = TMP_CAR (winds);
    if (scm_is_pair (dynpair)) {
      SCM this_key = TMP_CAR (dynpair);

      if (scm_is_eq (this_key, SCM_BOOL_T) || scm_is_eq (this_key, key)) {
        jmpbuf = TMP_CDR (dynpair);

        if (!SCM_PRE_UNWIND_DATA_P (jmpbuf))
          break;
        else {
          struct pre_unwind_data *c =
            (struct pre_unwind_data *) SCM_CELL_WORD_1 (jmpbuf);
          if (!c->running)
            break;
        }
      }
    }
  }

  /* If we didn't find anything, print a message and abort the process
     right here.  If you don't want this, establish a catch-all around
     any code that might throw up. */
  if (scm_is_null (winds)) {
      scm_handle_by_message (NULL, rc_Fly (key), rc_Fly (args));
      abort ();
  }

  /* If the wind list is malformed, bail.  */
  if (!scm_is_pair (winds)) {
    fprintf (stderr, " * VSI: malformed winding list * ");
    abort ();
  }

  reference_parameter (jmpbuf);
  dereference_parameter (winds);  // variable not used anymore

  // RC Checkpoint: key args jmpbuf

  wind_goal = scm_i_dynwinds (); 
  reference_parameter (wind_goal);
  for ( ;
       (!scm_is_pair (TMP_CAR (wind_goal))
	|| !scm_is_eq (TMP_CDAR (wind_goal), jmpbuf));
       wind_goal = FLY_CDR (wind_goal))
    ;

  // RC Checkpoint: key args jmpbuf wind_goal

  /* Is this a throw handler (or lazy catch)?  In a wind list entry
     for a throw handler or lazy catch, the key is bound to a
     pre_unwind_data smob, not a jmpbuf.  */
  if (SCM_PRE_UNWIND_DATA_P (jmpbuf)) {  // unused ???
    struct pre_unwind_data *c =
      (struct pre_unwind_data *) SCM_CELL_WORD_1 (jmpbuf);
    SCM handle = SCM_UNSPECIFIED;
    SCM answer;

    /* For old-style lazy-catch behaviour, we unwind the dynamic
       context before invoking the handler. */
    if (c->lazy_catch_p) {
      scm_dowinds (wind_goal, (scm_ilength (scm_i_dynwinds ())
                               - scm_ilength (wind_goal)));
      handle = scm_i_dynwinds ();
      reference_parameter (handle);
      scm_i_set_dynwinds (TMP_CDR (handle));
    }

    // RC inner checkpoint: key args jmpbuf wind_goal handle

    /* Call the handler, with framing to set the pre-unwind
       structure's running field while the handler is running, so we
       can avoid recursing into the same handler again.  Note that
       if the handler returns normally, the running flag stays
       set until some kind of non-local jump occurs. */
    (void) scm_dynwind_begin (SCM_F_DYNWIND_REWINDABLE);
    fprintf (stderr, \
             "\nVSI-C: scm_i_throw: WARNING: dynwind_begin w/o token\n");
    scm_dynwind_rewind_handler (toggle_pre_unwind_running,
                                c,
                                SCM_F_WIND_EXPLICITLY);
    scm_dynwind_unwind_handler (toggle_pre_unwind_running, c, 0);
    answer = (c->handler) (c->handler_data, key, args);
    reference_parameter (answer);

    /* There is deliberately no scm_dynwind_end call here.  This
       means that the unwind handler (toggle_pre_unwind_running)
       stays in place until a non-local exit occurs, and will then
       reset the pre-unwind structure's running flag.  For sample
       code where this makes a difference, see the "again but with
       two chained throw handlers" test case in exceptions.test.  */

    /* If the handler returns, rethrow the same key and args. */
    dereference_parameter (answer);
    dereference_parameter (handle);
    dereference_parameter (wind_goal);
    dereference_parameter (jmpbuf);
    goto rethrow;  // RC: references on key and args
  

  /* Otherwise, it's a normal catch.  */
  } else if (SCM_JMPBUFP (jmpbuf)) {
    struct pre_unwind_data * pre_unwind;
    struct jmp_buf_and_retval * jbr;

    /* Before unwinding anything, run the pre-unwind handler if
       there is one, and if it isn't already running. */
    pre_unwind = SCM_JBPREUNWIND (jmpbuf);
    if (pre_unwind->handler && !pre_unwind->running) {
      /* Use framing to detect and avoid possible reentry into
         this handler, which could otherwise cause an infinite
         loop. */
      SCM handler_token = scm_dynwind_begin (SCM_F_DYNWIND_REWINDABLE);
      scm_dynwind_rewind_handler (toggle_pre_unwind_running,
                                  pre_unwind,
                                  SCM_F_WIND_EXPLICITLY);
      scm_dynwind_unwind_handler (toggle_pre_unwind_running,
                                  pre_unwind,
                                  SCM_F_WIND_EXPLICITLY);
      (pre_unwind->handler) (pre_unwind->handler_data, key, args);
      scm_dynwind_end (handler_token);
    }

    /* Now unwind and jump. */
    scm_dowinds (wind_goal, (scm_ilength (scm_i_dynwinds ())
                             - scm_ilength (wind_goal)));

    // jbr of catch (see scm_c_catch)
    jbr = (struct jmp_buf_and_retval *)JBJMPBUF (jmpbuf);

    jbr->throw_tag = key;
    jbr->retval = args;
    heap_reference (jbr, key);   // dereferenced by scm_c_catch
    heap_reference (jbr, args);  // dereferenced by scm_c_catch

    // scm_i_set_last_debug_frame (SCM_JBDFRAME (jmpbuf));

    dereference_parameter (key);
    dereference_parameter (args);
    dereference_parameter (jmpbuf);
    dereference_parameter (wind_goal);

    SCM_I_LONGJMP (*JBJMPBUF (jmpbuf), 1);
  }

  /* Otherwise, it's some random piece of junk.  */
  else {
    fprintf (stderr, " * VSI: broken catch * ");
    abort ();
  }
}

long *global_jmpbuffer_count   = NULL;
long *global_pre_unwind_data_count = NULL;

void
scm_init_throw ()
{
  tc16_jmpbuffer = scm_make_smob_type ("jmpbuffer", 0);
  scm_set_smob_print (tc16_jmpbuffer, jmpbuffer_print);

  tc16_pre_unwind_data = scm_make_smob_type ("pre-unwind-data", 0);
  scm_set_smob_print (tc16_pre_unwind_data, pre_unwind_data_print);

  global_jmpbuffer_count   = \
    &(scm_smobs[SCM_TC2SMOBNUM (tc16_jmpbuffer)].number_of_objects);
  global_pre_unwind_data_count = \
    &(scm_smobs[SCM_TC2SMOBNUM (tc16_pre_unwind_data)].number_of_objects);
}

void
x_inclusion_throw ()
{
#include "include/throw.x"
}

