// Strings & Brass (C0 intention)

/* regex-posix.c -- POSIX regular expression support.

   This code was written against Henry Spencer's famous regex package.
   The principal reference for POSIX behavior was the man page for this
   library, not the 1003.2 document itself.  Ergo, other `POSIX'
   libraries which do not agree with the Spencer implementation may
   produce varying behavior.  Sigh. */

#include "include/config.h"

#include <sys/types.h>

#include "include/_scm.h"

/* Supposedly, this file is never compiled unless we know we have
   POSIX regular expressions.  But we still put this in an #ifdef so
   the file is CPP'able (for dependency scanning) even on systems that
   don't have a <regex.h> header.  */
#ifdef HAVE_REGCOMP
#ifdef HAVE_REGEX_H
#include <regex.h>
#else
#ifdef HAVE_RXPOSIX_H
#include <rxposix.h>		/* GNU Rx library */
#else
#ifdef HAVE_RX_RXPOSIX_H
#include <rx/rxposix.h>		/* GNU Rx library on Linux */
#endif
#endif
#endif
#endif

#include "include/smob.h"
#include "include/symbols.h"
#include "include/vectors.h"
#include "include/strports.h"
#include "include/ports.h"
#include "include/feature.h"
#include "include/brass.h"
#include "include/byte-strings.h"

#include "include/validate.h"
#include "include/regex-posix.h"

/* This is defined by some regex libraries and omitted by others. */
#ifndef REG_BASIC
#define REG_BASIC 0
#endif

scm_t_tag scm_tc16_regex;


SCM_SYMBOL (scm_regexp_error_key, "regular-expression-syntax");

static SCM
scm_regexp_error_msg (int regerrno, regex_t *rx)
{
  char *errmsg;
  int l;

  errmsg = scm_malloc (80);
  l = regerror (regerrno, rx, errmsg, 80);
  if (l > 80)
    {
      free (errmsg);
      errmsg = scm_malloc (l);
      regerror (regerrno, rx, errmsg, l);
    }
  SCM ret = scm_from_sen (errmsg, strlen (errmsg));
  free (errmsg);
  return ret;
}

SCM_DEFINE (scm_regexp_p, "regexp?", 1, 0, 0,
            (SCM obj),
	    "Return @code{#t} if @var{obj} is a compiled regular expression,\n"
	    "or @code{#f} otherwise.")
#define FUNC_NAME s_scm_regexp_p
{
  reference_parameter (obj);
  SCM answer = scm_from_bool(SCM_RGXP (obj));
  dereference_parameter (obj);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_make_regexp, "make-regexp", 1, 0, 1, (SCM pat, SCM flags), "")
#define FUNC_NAME s_scm_make_regexp
{
  reference_parameter (pat);
  reference_parameter (flags);

  SCM flag;
  regex_t *rx;
  int status, cflags;
  SCM b_pat;

  SCM_VALIDATE_BRASS (1, pat);
  SCM_VALIDATE_REST_ARGUMENT (flags);

  /* Examine list of regexp flags.  If REG_BASIC is supplied, then
     turn off REG_EXTENDED flag (on by default). */
  cflags = REG_EXTENDED;
  flag = flags;
  while (!scm_is_null (flag))
    {
      if (scm_to_int (TMP_CAR (flag)) == REG_BASIC)
	cflags &= ~REG_EXTENDED;
      else
	cflags |= scm_to_int (TMP_CAR (flag));
      flag = FLY_CDR (flag);
    }

  rx = scm_gc_malloc (sizeof(regex_t), "regex");
  b_pat = scm_to_sen (pat);
  reference_parameter (b_pat);
  status = regcomp (rx, BYTE_STRING_DATA (b_pat),
		    /* Make sure they're not passing REG_NOSUB;
                       regexp-exec assumes we're getting match data.  */
		    cflags & ~REG_NOSUB);
  dereference_parameter (b_pat);
  if (status != 0)
    {
      SCM errmsg = scm_regexp_error_msg (status, rx);
      scm_gc_free (rx, sizeof(regex_t), "regex");
      scm_error_scm (scm_regexp_error_key,
		     scm_from_cen (FUNC_NAME),
		     errmsg,
		     SCM_BOOL_F,
		     scm_list_1 (pat));
      
      /* never returns */
    }
  dereference_parameter (pat);
  dereference_parameter (flags);
  SCM_RETURN_NEWSMOB (scm_tc16_regex, rx);
}
#undef FUNC_NAME

SCM_DEFINE (scm_regexp_exec, "regexp-exec", 2, 2, 0,
            (SCM rx, SCM str, SCM start, SCM flags), "")
#define FUNC_NAME s_scm_regexp_exec
{
  reference_parameter (rx);
  reference_parameter (str);
  reference_parameter (start);
  reference_parameter (flags);

  int status, nmatches, offset;
  regmatch_t *matches;
  SCM b_str;
  SCM mvec = SCM_BOOL_F;
  SCM substr;

  SCM_VALIDATE_RGXP (1, rx);
  SCM_VALIDATE_BRASS (2, str);

  if (SCM_UNBNDP (start)) {
    substr = str;
    offset = 0;

  } else {
    substr = scm_subrass (str, start, BR_LENGTH (str));
    offset = scm_to_int (start);
  }
  reference_parameter (substr);

  if (SCM_UNBNDP (flags))
    flags = SCM_INUM0;

  /* re_nsub doesn't account for the `subexpression' representing the
     whole regexp, so add 1 to nmatches. */

  nmatches = SCM_RGX(rx)->re_nsub + 1;
  matches = scm_malloc (sizeof (regmatch_t) * nmatches);
  b_str = scm_to_sen (substr);
  reference_parameter (b_str);
  status = regexec (SCM_RGX (rx), BYTE_STRING_DATA (b_str), nmatches, matches,
		    scm_to_int (flags));
  dereference_parameter (b_str);

  if (!status) {
    int i;
    /* The match vector must include a slot for the string that was matched,
       so add 1. */
    mvec = scm_c_make_vector (nmatches + 1, SCM_UNSPECIFIED);
    reference_parameter (mvec);
    SCM_SIMPLE_VECTOR_SET(mvec,0, str);
    for (i = 0; i < nmatches; ++i)

      if (matches[i].rm_so == -1) {
        SCM_SIMPLE_VECTOR_SET(mvec, i+1, \
          scm_cons (scm_from_int (-1), scm_from_int (-1)));

      } else {
        SCM_SIMPLE_VECTOR_SET(mvec, i+1, \
          scm_cons (scm_from_long (matches[i].rm_so + offset),
                    scm_from_long (matches[i].rm_eo + offset)));
      }
  }
  free (matches);

  if (status != 0 && status != REG_NOMATCH)
    scm_error_scm (scm_regexp_error_key,
		   scm_from_cen (FUNC_NAME),
		   scm_regexp_error_msg (status, SCM_RGX (rx)),
		   SCM_BOOL_F, SCM_BOOL_F);

  dereference_parameter (substr);
  dereference_parameter (rx);
  dereference_parameter (str);
  dereference_parameter (start);
  dereference_parameter (flags);
  return (rc_return_stack_reference (mvec));
}
#undef FUNC_NAME

static void
regex_release (SCM obj)
{
  regfree (SCM_RGX (obj));
  scm_gc_free (SCM_RGX (obj), sizeof(regex_t), "regex");
  return;
}

void
scm_init_regex_posix ()
{
  scm_tc16_regex = scm_make_smob_type ("regexp", sizeof (regex_t));
  scm_set_smob_release (scm_tc16_regex, 2, regex_release);
}

void
x_inclusion_regex ()
{
  /* Compilation flags.  */
  scm_c_define ("regexp/basic",    scm_from_int (REG_BASIC));
  scm_c_define ("regexp/extended", scm_from_int (REG_EXTENDED));
  scm_c_define ("regexp/icase",    scm_from_int (REG_ICASE));
  scm_c_define ("regexp/newline",  scm_from_int (REG_NEWLINE));

  /* Execution flags.  */
  scm_c_define ("regexp/notbol", scm_from_int (REG_NOTBOL));
  scm_c_define ("regexp/noteol", scm_from_int (REG_NOTEOL));

#include "include/regex-posix.x"

}
