// Strings & Brass

#include "include/config.h"

#include <fcntl.h>      /* for mingw */
#include <signal.h>
#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"

#include "include/eval.h"
#include "include/root.h"
#include "include/vectors.h"

#include "include/validate.h"
#include "include/signals.h"
#include "include/iselect.h"

#ifdef HAVE_IO_H
#include <io.h>  /* for mingw _pipe() */
#endif

#ifdef HAVE_PROCESS_H
#include <process.h>    /* for mingw */
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "include/byte-strings.h"
#include <sys/signalfd.h>

SCM_DEFINE (scm_signalfd, "signalfd", 1, 0, 0,
            (SCM signal_mask), "")
#define FUNC_NAME s_scm_signalfd
{
  reference_parameter (signal_mask);

  if (!BYTE_STRING_P (signal_mask)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (signal_mask) != sizeof (sigset_t)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  // If /signalfd/ (or alternative) is N/A you have to craft sth with a
  // pipe and let the signal handlers write into it, I guess.
  int answer = signalfd (-1, (sigset_t*)BYTE_STRING_DATA (signal_mask), 0);

  dereference_parameter (signal_mask);
  return scm_from_int (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_sigprocmask, "sigprocmask/unchecked", 1, 0, 0,
            (SCM how, SCM set, SCM oldset), "")
#define FUNC_NAME s_scm_sigprocmask
{
  reference_parameter (how);
  reference_parameter (set);
  reference_parameter (oldset);

  if (!BYTE_STRING_P (how)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (how) != sizeof (int)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  if (!BYTE_STRING_P (set)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 2, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (set) != sizeof (void*)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  if (!BYTE_STRING_P (oldset)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 3, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (oldset) != sizeof (void*)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  int answer = sigprocmask (*((int*)BYTE_STRING_DATA (how)), \
                            *((void**)BYTE_STRING_DATA (set)), \
                            *((void**)BYTE_STRING_DATA (oldset)));

  dereference_parameter (how);
  dereference_parameter (set);
  dereference_parameter (oldset);
  return scm_from_int (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_signal, "signal/unchecked", 1, 0, 0,
            (SCM signum, SCM handler), "")
#define FUNC_NAME s_scm_signal
{
  reference_parameter (signum);
  reference_parameter (handler);

  if (!BYTE_STRING_P (signum)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (signum) != sizeof (int)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }
  if (!BYTE_STRING_P (handler)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 2, SCM_BOOL_F, "byte-string");
  } else if (BYTE_STRING_LENGTH (handler) != sizeof (sighandler_t)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  sighandler_t res = signal (*((int*)BYTE_STRING_DATA (signum)), \
                             *((sighandler_t *)BYTE_STRING_DATA (handler)));

  SCM answer = scm_make_byte_string_with_copy ((void*)(&res), \
                                               sizeof (sighandler_t));
  dereference_parameter (signum);
  dereference_parameter (handler);
  return answer;
}
#undef FUNC_NAME


/* OLD CODE */
SCM_DEFINE (scm_alarm, "alarm", 1, 0, 0,
           (SCM i),
	    "Set a timer to raise a @code{SIGALRM} signal after the specified\n"
	    "number of seconds (an integer).  It's advisable to install a signal\n"
	    "handler for\n"
	    "@code{SIGALRM} beforehand, since the default action is to terminate\n"
	    "the process.\n\n"
	    "The return value indicates the time remaining for the previous alarm,\n"
	    "if any.  The new value replaces the previous alarm.  If there was\n"
	    "no previous alarm, the return value is zero.")
#define FUNC_NAME s_scm_alarm
{
  return scm_from_uint (alarm (scm_to_uint (i)));
}
#undef FUNC_NAME

#ifdef HAVE_SETITIMER
SCM_DEFINE (scm_setitimer, "setitimer", 5, 0, 0,
           (SCM which_timer,
            SCM interval_seconds, SCM interval_microseconds,
            SCM value_seconds, SCM value_microseconds),
            "Set the timer specified by @var{which_timer} according to the given\n"
            "@var{interval_seconds}, @var{interval_microseconds},\n"
            "@var{value_seconds}, and @var{value_microseconds} values.\n"
            "\n"
            "Return information about the timer's previous setting."
            "\n"
            "Errors are handled as described in the guile info pages under ``POSIX\n"
            "Interface Conventions''.\n"
            "\n"
            "The timers available are: @code{ITIMER_REAL}, @code{ITIMER_VIRTUAL},\n"
            "and @code{ITIMER_PROF}.\n"
            "\n"
            "The return value will be a list of two cons pairs representing the\n"
            "current state of the given timer.  The first pair is the seconds and\n"
            "microseconds of the timer @code{it_interval}, and the second pair is\n"
            "the seconds and microseconds of the timer @code{it_value}.")
#define FUNC_NAME s_scm_setitimer
{
  reference_parameter (which_timer);
  reference_parameter (interval_seconds);
  reference_parameter (interval_microseconds);
  reference_parameter (value_seconds);
  reference_parameter (value_microseconds);

  int rv;
  int c_which_timer;
  struct itimerval new_timer;
  struct itimerval old_timer;

  c_which_timer = SCM_NUM2INT(1, which_timer);
  new_timer.it_interval.tv_sec = SCM_NUM2LONG(2, interval_seconds);
  new_timer.it_interval.tv_usec = SCM_NUM2LONG(3, interval_microseconds);
  new_timer.it_value.tv_sec = SCM_NUM2LONG(4, value_seconds);
  new_timer.it_value.tv_usec = SCM_NUM2LONG(5, value_microseconds);

  SCM_SYSCALL(rv = setitimer(c_which_timer, &new_timer, &old_timer));
  
  if(rv != 0)
    SCM_SYSERROR;

  dereference_parameter (which_timer);
  dereference_parameter (interval_seconds);
  dereference_parameter (interval_microseconds);
  dereference_parameter (value_seconds);
  dereference_parameter (value_microseconds);
  return scm_list_2 (scm_cons (scm_from_long (old_timer.it_interval.tv_sec),
			       scm_from_long (old_timer.it_interval.tv_usec)),
		     scm_cons (scm_from_long (old_timer.it_value.tv_sec),
			       scm_from_long (old_timer.it_value.tv_usec)));
}
#undef FUNC_NAME
#endif /* HAVE_SETITIMER */

#ifdef HAVE_GETITIMER
SCM_DEFINE (scm_getitimer, "getitimer", 1, 0, 0,
  (SCM which_timer),
            "Return information about the timer specified by @var{which_timer}"
            "\n"
            "Errors are handled as described in the guile info pages under ``POSIX\n"
            "Interface Conventions''.\n"
            "\n"
            "The timers available are: @code{ITIMER_REAL}, @code{ITIMER_VIRTUAL},\n"
            "and @code{ITIMER_PROF}.\n"
            "\n"
            "The return value will be a list of two cons pairs representing the\n"
            "current state of the given timer.  The first pair is the seconds and\n"
            "microseconds of the timer @code{it_interval}, and the second pair is\n"
            "the seconds and microseconds of the timer @code{it_value}.")
#define FUNC_NAME s_scm_getitimer
{
  reference_parameter (which_timer);

  int rv;
  int c_which_timer;
  struct itimerval old_timer;

  c_which_timer = SCM_NUM2INT(1, which_timer);

  SCM_SYSCALL(rv = getitimer(c_which_timer, &old_timer));
  
  if(rv != 0)
    SCM_SYSERROR;
  
  dereference_parameter (which_timer);
  return scm_list_2 (scm_cons (scm_from_long (old_timer.it_interval.tv_sec),
			       scm_from_long (old_timer.it_interval.tv_usec)),
		     scm_cons (scm_from_long (old_timer.it_value.tv_sec),
			       scm_from_long (old_timer.it_value.tv_usec)));
}
#undef FUNC_NAME
#endif /* HAVE_GETITIMER */

#ifdef HAVE_PAUSE
SCM_DEFINE (scm_pause, "pause", 0, 0, 0,
           (),
	    "Pause the current process (thread?) until a signal arrives whose\n"
	    "action is to either terminate the current process or invoke a\n"
	    "handler procedure.  The return value is unspecified.")
#define FUNC_NAME s_scm_pause
{
  pause ();
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif

SCM_DEFINE (scm_sleep, "sleep", 1, 0, 0,
           (SCM i),
	    "Wait for the given number of seconds (an integer) or until a signal\n"
	    "arrives.  The return value is zero if the time elapses or the number\n"
	    "of seconds remaining otherwise.\n"
	    "\n"
	    "See also @code{usleep}.")
#define FUNC_NAME s_scm_sleep
{
  return scm_from_uint (scm_std_sleep (scm_to_uint (i)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_usleep, "usleep", 1, 0, 0,
           (SCM i),
	    "Wait the given period @var{usecs} microseconds (an integer).\n"
	    "If a signal arrives the wait stops and the return value is the\n"
	    "time remaining, in microseconds.  If the period elapses with no\n"
	    "signal the return is zero.\n"
	    "\n"
	    "On most systems the process scheduler is not microsecond accurate and\n"
	    "the actual period slept by @code{usleep} may be rounded to a system\n"
	    "clock tick boundary.  Traditionally such ticks were 10 milliseconds\n"
	    "apart, and that interval is often still used.\n"
	    "\n"
	    "See also @code{sleep}.")
#define FUNC_NAME s_scm_usleep
{
  return scm_from_ulong (scm_std_usleep (scm_to_ulong (i)));
}
#undef FUNC_NAME

SCM_DEFINE (scm_raise, "raise", 1, 0, 0,
           (SCM sig),
	    "Sends a specified signal @var{sig} to the current process, where\n"
	    "@var{sig} is as described for the kill procedure.")
#define FUNC_NAME s_scm_raise
{
  if (raise (scm_to_int (sig)) != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


unsigned long
scm_std_usleep (unsigned long usecs)
{
  struct timeval tv;
  tv.tv_usec = usecs % 1000000;
  tv.tv_sec = usecs / 1000000;
  scm_std_select (0, NULL, NULL, NULL, &tv);
  return tv.tv_sec * 1000000 + tv.tv_usec;
}

unsigned int
scm_std_sleep (unsigned int secs)
{
  struct timeval tv;
  tv.tv_usec = 0;
  tv.tv_sec = secs;
  scm_std_select (0, NULL, NULL, NULL, &tv);
  return tv.tv_sec;
}



void
x_inclusion_posix_signals ()
{
  scm_c_define ("NSIG", scm_from_long (NSIG));
  scm_c_define ("SIG_IGN", scm_from_long ((long) SIG_IGN));
  scm_c_define ("SIG_DFL", scm_from_long ((long) SIG_DFL));
#ifdef SA_NOCLDSTOP
  scm_c_define ("SA_NOCLDSTOP", scm_from_long (SA_NOCLDSTOP));
#endif
#ifdef SA_RESTART
  scm_c_define ("SA_RESTART", scm_from_long (SA_RESTART));
#endif

#if defined(HAVE_SETITIMER) || defined(HAVE_GETITIMER)
  /* Stuff needed by setitimer and getitimer. */
  scm_c_define ("ITIMER_REAL", scm_from_int (ITIMER_REAL));
  scm_c_define ("ITIMER_VIRTUAL", scm_from_int (ITIMER_VIRTUAL));
  scm_c_define ("ITIMER_PROF", scm_from_int (ITIMER_PROF));
#endif /* defined(HAVE_SETITIMER) || defined(HAVE_GETITIMER) */

#include "include/signals.x"
}
