#include "include/ihub.h"
#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>

/*
 * ioctl fun
 */

SCM_DEFINE (terminal_get_win_size, "%terminal-get-win-size", 1, 0, 0,
            (SCM file_descriptor), "Internal Terminal Support.")
{
  struct winsize answer = {0, 0, 0, 0};
  int ret = 0;
  int fd = -1;

  fd = scm_to_int (file_descriptor); /* ??? assert ? */
  ret = ioctl (fd, TIOCGWINSZ, &answer);
  if (ret == 0)
    {
      return (scm_list_4 (scm_from_ushort (answer.ws_col),
                          scm_from_ushort (answer.ws_row),
                          scm_from_ushort (answer.ws_xpixel),
                          scm_from_ushort (answer.ws_ypixel)));
    }
  return SCM_BOOL_F; /* we don't care about errno. */
}


/*
 * termios happiness
 */

SCM_DEFINE (terminal_raw_mode, "%terminal-raw-mode", 1, 0, 0,
            (SCM file_descriptor), "Internal Terminal Support.")
{
  struct termios mode;
  int fd = -1;

  if (!SCM_NUMBERP (file_descriptor)) return SCM_BOOL_F;
  fd = scm_to_int (file_descriptor); /* ??? assert ? */
  if (tcgetattr (fd, &mode))
    return SCM_BOOL_F;

  mode.c_iflag = 0;
  mode.c_oflag &= ~OPOST;
 /* BSD raw + VSI extensions */
  mode.c_lflag &= ~(ECHO | ECHOCTL | ECHONL | ICANON | ISIG | IEXTEN);

  mode.c_cc[VMIN] = 1;
  mode.c_cc[VTIME] = 0;

  if (tcsetattr (fd, TCSANOW, &mode) == -1)
    return SCM_BOOL_F;    /* errno is unimportant here */
  
  return SCM_BOOL_T;
}

SCM_DEFINE (terminal_cooked_mode, "%terminal-cooked-mode", 1, 0, 0,
            (SCM file_descriptor), "Internal Terminal Support.")
{
  struct termios mode;
  int fd = -1;

  if (!SCM_NUMBERP (file_descriptor)) return SCM_BOOL_F;
  fd = scm_to_int (file_descriptor); /* ??? assert ? */
  if (tcgetattr (fd, &mode))
    return SCM_BOOL_F;

  /* Cooked mode. */
  // mode.c_iflag |= BRKINT | IGNPAR | ISTRIP | ICRNL | IXON;
  // XXX patch (brass preliminaries): don't strip off the 8th bit
  //  [now it already looks as if but that is what utf-8 was meant for]
  // This really should be implemented with the unchecked SCA routines.
  // Perhaps we should use a complete prepared flag set to exclude
  // strange settings from elsewhere.
  mode.c_iflag |= BRKINT | IGNPAR | ICRNL | IXON;
  mode.c_oflag |= OPOST;
  mode.c_lflag |= ECHO | ISIG | ICANON;
  mode.c_lflag &= ~(ECHOCTL | IEXTEN);

  mode.c_cc[VEOF] = _POSIX_VDISABLE;
  mode.c_cc[VEOL] = _POSIX_VDISABLE;

  if (tcsetattr (fd, TCSANOW, &mode) == -1)
    return SCM_BOOL_F;    /* errno is unimportant here */

  return SCM_BOOL_T;
}

/* EXPERIMENTAL
 * System Call Abstraction Stage 1 PROTOTYPES
 * UNCHECKED argument abstractions
 */
#include "include/byte-strings.h"

/*
 * SCA experimental system interface with byte structure
 * argument passing.
 * (posix-misc.c is too large for this)
 */

SCM_DEFINE (unchecked_tcgetattr, "unchecked-tcgetattr", 1, 0, 0,
            (SCM file_descriptor), "Internal Terminal Support.")
{
  if (!SCM_NUMBERP (file_descriptor)) return SCM_BOOL_F;
  int fd = -1;
  fd = scm_to_int (file_descriptor); /* ??? assert ? */

  SCM answer = scm_c_make_byte_string (sizeof (struct termios));
  struct termios *mode = BYTE_STRING_DATA (answer);

  if (tcgetattr (fd, mode)) {
    dereference_return_value (answer);
    return SCM_BOOL_F;
  }

  return answer;
}

SCM_DEFINE (unchecked_tcsetattr, "unchecked-tcsetattr", 2, 0, 0,
            (SCM file_descriptor, SCM mode), "Internal Terminal Support.")
{
  if (!SCM_NUMBERP (file_descriptor)) return SCM_BOOL_F;
  if (!BYTE_STRING_P (mode) \
      || BYTE_STRING_LENGTH (mode) != sizeof (struct termios)) {
    return SCM_BOOL_F;
  }

  struct termios *modus = (struct termios *)BYTE_STRING_DATA (mode);
  int fd = scm_to_int (file_descriptor);

  if (tcsetattr (fd, TCSANOW, modus) == -1)
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}

// XXX system dependency
#define LINUX 1
#ifdef LINUX
#include <sys/inotify.h>
// Please note the kernel documentation for inotify describes a richer
// and more complicated interface but the scanned header files don't.
// No matter what the return value of /inotify_rm_watch/ remains in
// the realm of unspecified.
//
// XXX TODO CLR/3 no throw error handling prototype: please dereference
// on error.
SCM_DEFINE (unchecked_inotify_init, "unchecked-inotify-init", 0, 0, 0,
            (), "Experimental iNotify Support (SCA).")
{
  return scm_from_int (inotify_init ());
}

SCM_DEFINE (unchecked_inotify_add_watch, "unchecked-inotify-add-watch",
            3, 0, 0,
            (SCM file_descriptor, SCM bs_path, SCM bs_mask),
            "Experimental iNotify Support (SCA).")
{
  reference_parameter (file_descriptor);
  reference_parameter (bs_path);
  reference_parameter (bs_mask);

  if (!SCM_NUMBERP (file_descriptor)) return SCM_BOOL_F;
  if (!BYTE_STRING_P (bs_path))       return SCM_BOOL_F;
  int   fd   = scm_to_int (file_descriptor);
  char *path = (char*)BYTE_STRING_DATA (bs_path);

  if ('\0' != *(path + (BYTE_STRING_LENGTH(bs_path) - 1))) {
    return SCM_BOOL_F;
  }
  // We could do away with the triple checking and go unchecked for
  // real but SCA isn't even in its "preliminaries" yet.
  if (!BYTE_STRING_P (bs_mask) \
      || BYTE_STRING_LENGTH (bs_mask) != sizeof (uint32_t)) {
    return SCM_BOOL_F;
  }
  scm_t_uint32 mask = *(scm_t_uint32*)BYTE_STRING_DATA(bs_mask);

  int answer = inotify_add_watch (fd, path, mask);
  dereference_parameter (file_descriptor);
  dereference_parameter (bs_path);
  dereference_parameter (bs_mask);
  return scm_from_int (answer);
}

SCM_DEFINE (unchecked_inotify_rm_watch, "unchecked-inotify-rm-watch",
            2, 0, 0,
            (SCM file_descriptor, SCM watch_descriptor),
            "Experimental iNotify Support (SCA).")
{
  reference_parameter (file_descriptor);
  reference_parameter (watch_descriptor);

  if (!SCM_NUMBERP (file_descriptor))  return SCM_BOOL_F;
  if (!SCM_NUMBERP (watch_descriptor)) return SCM_BOOL_F;
  int   fd   = scm_to_int (file_descriptor);
  int   wd   = scm_to_int (watch_descriptor);

  int answer = inotify_rm_watch (fd, wd);
  dereference_parameter (file_descriptor);
  dereference_parameter (watch_descriptor);
  return scm_from_int (answer);
}

#endif // LINUX


/*
 * SCM interface
 */
void
scm_init_terminal_support ()
{
#include "include/terminal-support.x"
#ifdef LINUX
 #include "include/inotify.x"
#endif // LINUX
}
