/* Brass (as in Strings & Brass)
 *
 * A brass consists of characters.
 * Characters correspond to code points - unicode scalar values that is.
 * (Wha'? It's not me to decide what a code point is? Craft your own glyphs
 * then ...)
 *
 * What's the difference between a sequence (/list/) of characters and a
 * brass? The brass is still slightly more compact wrt memory layout.
 *
 * Beware of the fact that a character in a brass needs four bytes
 * (sizeof uint32_t) which isn't automatically respected by all
 * those traditional byte routines out there. (length * sizeof (uint32_t))
 */
#include "include/config.h"
#include "include/_scm.h"
#include "include/chars.h"
#include "include/trim.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/validate.h"

long global_number_of_brass = 0;

size_t inline
BR_NSLOTS (size_t c_len)
{
  size_t num_slots = (c_len >> 1); // Assert sizeof SCM = 2 * sizeof uint32
  if (ODD_P (c_len)) num_slots++;
  return num_slots+1;
}

SCM
scm_c_make_brass (size_t c_len)
{
  size_t num_slots = BR_NSLOTS (c_len);

  SCM br = scm_tagged_array (scm_tc7_brass, num_slots);
  SCM_SET_CELL_WORD_0 (br, c_len);

  global_number_of_brass++;
  return br;
}

SCM
scm_c_make_brass_with_copy (usv_t *usv, size_t len)
{
  SCM br = scm_c_make_brass (len);
  memcpy (BR_CHARS (br), usv, len * sizeof (usv_t));
  return br;
}

inline int
scm_is_brass (SCM br)
{
  return (SCM_BRASSP (br));
}

SCM_DEFINE (scm_brass_p, "brass?", 1, 0, 0, (SCM obj), "")
#define FUNC_NAME s_scm_brass_p
{
  reference_parameter (obj);
  SCM answer = scm_from_bool (SCM_BRASSP (obj));
  dereference_parameter (obj);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_length, "brass-length", 1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_length
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);

  size_t c_len = BR_LENGTH (br);

  dereference_parameter (br);
  return scm_from_size_t (c_len);
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_ref, "brass-ref", 2, 0, 0, (SCM br, SCM idx), "")
#define FUNC_NAME s_scm_brass_ref
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t c_idx = scm_to_size_t (idx);
  if (c_idx >= BR_LENGTH (br)) {
    scm_out_of_range (FUNC_NAME, br);
  }

  SCM ans = SCM_MAKE_UC_CHAR (BR_REF (br, c_idx));

  dereference_parameter (br);
  return ans;
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_set_x, "brass-set!", 3, 0, 0,
            (SCM br, SCM idx, SCM chr), "")
#define FUNC_NAME s_scm_brass_set_x
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t c_idx = scm_to_size_t (idx);
  if (c_idx >= BR_LENGTH (br)) {
    scm_out_of_range (FUNC_NAME, idx);
  }
  SCM_VALIDATE_CHAR (3, chr);

  BR_SET (br, c_idx, SCM_CHAR (chr));

  dereference_parameter (br);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM
scm_brass_equal_p (SCM a, SCM b)
#define FUNC_NAME "equal?(brass)"
{
  reference_parameter (a);
  reference_parameter (b);
  SCM answer = SCM_UNSPECIFIED;

  if (SCM_BRASSP (a) && SCM_BRASSP (b)) {
    size_t length = BR_LENGTH (a);
    if (length != BR_LENGTH (b)) {
      answer = SCM_BOOL_F;
    } else {
      void* d1 = (void*)BR_CHARS (a);
      void* d2 = (void*)BR_CHARS (b);
      answer = (bcmp (d1, d2, length * sizeof (uint32_t)) == 0) \
               ? SCM_BOOL_T : SCM_BOOL_F;
    }
  } else {
    scm_wrong_type_arg_msg (FUNC_NAME, 3, SCM_BOOL_F, "brass");
  }

  dereference_parameter (a);
  dereference_parameter (b);
  return (answer); 
}
#undef FUNC_NAME

SCM_DEFINE (scm_make_brass, "make-brass", 2, 0, 0,
            (SCM len, SCM chr), "")
#define FUNC_NAME s_scm_make_brass
{ // RC parameters are supposed to be immediates
  size_t c_len     = scm_to_size_t (len);
  SCM_VALIDATE_CHAR (2, chr);
  uint32_t usv     = SCM_CHAR (chr);

  SCM br = scm_c_make_brass (c_len);
  size_t i = 0;
  for ( i = 0; i < c_len; i++) {
    BR_SET (br, i, usv);
  }

  return br;
}
#undef FUNC_NAME

SCM
scm_i_copy_brass (SCM br)
{
  size_t c_len = BR_LENGTH (br);
  SCM ans = scm_c_make_brass (c_len);
  memcpy (BR_CHARS (ans), BR_CHARS (br), c_len * sizeof (uint32_t));

  return ans;
}

SCM_DEFINE (scm_brass_copy, "brass-copy", 1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_copy
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t c_len = BR_LENGTH (br);

  SCM ans = scm_c_make_brass (c_len);
  memcpy (BR_CHARS (ans), BR_CHARS (br), c_len * sizeof (uint32_t));

  dereference_parameter (br);
  return ans;
}
#undef FUNC_NAME

SCM_REGISTER_PROC (s_scm_list_to_brass, "list->brass", 1, 0, 0, scm_brass);

SCM_DEFINE (scm_brass, "brass", 0, 0, 1, (SCM chrs), "")
#define FUNC_NAME s_scm_brass
{
  reference_parameter (chrs);
  size_t c_len = 0;

  // Scan arguments and determine length.
  SCM rst = chrs;
  while (scm_is_pair (rst)) {
    SCM_VALIDATE_CHAR (SCM_ARGn, TMP_CAR (rst));
    c_len++;
    rst = TMP_CDR (rst);
  }
  if (!scm_is_null (rst)) SCM_VALIDATE_LIST (1, chrs);

  SCM br = scm_c_make_brass (c_len);
  size_t i = 0;
  rst = chrs;
  while (scm_is_pair (rst)) {
    BR_SET (br, i, SCM_CHAR (TMP_CAR (rst)));
    i++;
    rst = TMP_CDR (rst);
  }

  dereference_parameter (chrs);
  return br;
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_to_list, "brass->list", 1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_to_list
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t idx = BR_LENGTH (br);

  SCM ans = SCM_EOL;

  if (idx > 0) {

  do {
    idx--;
    ans = scm_i_cons (SCM_MAKE_UC_CHAR (BR_REF (br, idx)), ans);
  } while (idx != 0);

  }

  dereference_parameter (br);
  return (ans);
}
#undef FUNC_NAME

SCM_DEFINE (scm_byte_string_to_brass_utf8, "byte-string->brass/utf8",
            1, 0, 0, (SCM bs), "")
#define FUNC_NAME s_scm_byte_string_to_brass_utf8
{
  reference_parameter (bs);
  if (!BYTE_STRING_P (bs)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, bs, "byte-string");
  }
  size_t len      = BYTE_STRING_LENGTH (bs);
  const char* dat = BYTE_STRING_DATA (bs);

  SCM ans = utf8_to_brass (dat, len);

  dereference_parameter (bs);
  return (ans);
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_to_c0_byte_stringz, "brass->byte-string/utf8-c0z",
            1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_to_c0_byte_stringz
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t len = BR_LENGTH (br);
  size_t idx = 0;

  SCM ans = scm_c_make_byte_string (len+1); // zero padding
  char *adat = BYTE_STRING_DATA (ans);

  while (idx < len) {
    usv_t t = BR_REF (br, idx);
    if (t >= 128) {
      dereference_return_value (ans);
      dereference_parameter (br);
      return SCM_BOOL_F;
    }
    *adat = (char) t;
    adat++; idx++;
  }
  *adat = '\0'; // zero padding

  dereference_parameter (br);
  return (ans);
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_to_byte_string_utf8z, "brass->byte-string/utf8z",
            1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_to_byte_string_utf8z
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);

  size_t len = BR_LENGTH (br);
  usv_t *chars = BR_CHARS (br);
  size_t num_bytes = 1;  // zero padding

  while (len != 0) {
    len--;
    size_t nb = utf8_usv_num_bytes (*chars);
    if (!nb) {
      dereference_parameter (br);
      return SCM_BOOL_F;  // utf-8 out of range / broken brass
    }
    num_bytes += nb;
    chars++;
  }

  SCM ans = scm_c_make_byte_string (num_bytes);
  uint8_t *adat = BYTE_STRING_DATA (ans);
  len = BR_LENGTH (br);
  chars = BR_CHARS (br);
  while (len != 0) {
    len--;
    adat += utf8_usv_to_bytes (*chars, adat);
    chars++;
  }
  *adat = '\0';  // zero padding

  dereference_parameter (br);
  return (ans);
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_to_c0_byte_string, "brass->byte-string/utf8-c0",
            1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_to_c0_byte_string
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t len = BR_LENGTH (br);
  size_t idx = 0;

  SCM ans = scm_c_make_byte_string (len);
  char *adat = BYTE_STRING_DATA (ans);

  while (idx < len) {
    usv_t t = BR_REF (br, idx);
    if (t >= 128) {
      dereference_return_value (ans);
      dereference_parameter (br);
      return SCM_BOOL_F;
    }
    *adat = (char) t;
    adat++; idx++;
  }

  dereference_parameter (br);
  return (ans);
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_to_byte_string_utf8, "brass->byte-string/utf8",
            1, 0, 0, (SCM br), "")
#define FUNC_NAME s_scm_brass_to_byte_string_utf8
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);

  size_t len = BR_LENGTH (br);
  usv_t *chars = BR_CHARS (br);
  size_t num_bytes = 0;

  while (len != 0) {
    len--;
    size_t nb = utf8_usv_num_bytes (*chars);
    if (!nb) {
      dereference_parameter (br);
      return SCM_BOOL_F;  // utf-8 out of range / broken brass
    }
    num_bytes += nb;
    chars++;
  }

  SCM ans = scm_c_make_byte_string (num_bytes);
  uint8_t *adat = BYTE_STRING_DATA (ans);
  len = BR_LENGTH (br);
  chars = BR_CHARS (br);
  while (len != 0) {
    len--;
    adat += utf8_usv_to_bytes (*chars, adat);
    chars++;
  }

  dereference_parameter (br);
  return (ans);
}
#undef FUNC_NAME

// Internal C*
SCM
scm_from_cen (const char *c_str)
{
  return utf8_to_brass (c_str, strlen (c_str));
}

SCM 
scm_makfromstrs (int argc, char **argv)
{
  int i = argc;
  SCM lst = SCM_EOL;
  // Wow! A null terminated list mapped onto an array. TECHNIQUE!
  // NULL is FALSE and has become the pair tag. Sorry for that one.
  // (That wasn't even me then ... again. What was the name of that
  //  register again? Go count ones in /dev/zero or compute the
  //  checksum of the parity bit. ;-)
  if (0 > i) for (i = 0; argv[i]; i++);
  while (i--)
    lst = scm_cons (scm_from_sen (argv[i], strlen (argv[i])), lst);
  return lst;
}

// Editing Support

SCM_DEFINE (scm_subrass, "subrass", 3, 0, 0,
            (SCM br, SCM start, SCM exclusive_end), "")
#define FUNC_NAME s_scm_subrass
{
  reference_parameter (br);
  SCM_VALIDATE_BRASS (1, br);
  size_t c_len   = BR_LENGTH (br);
  size_t c_start = scm_to_size_t (start);          // >= 0
  size_t c_eend  = scm_to_size_t (exclusive_end);  // >= 0

  if (!(c_start <= c_eend && c_eend <= c_len)) {
    scm_out_of_range (FUNC_NAME, SCM_BOOL_F);
  }

  size_t nlen = c_eend - c_start;
  SCM ans = scm_c_make_brass (nlen);
  memcpy (BR_CHARS (ans), &((BR_CHARS (br))[c_start]), nlen * sizeof (uint32_t));

  dereference_parameter (br);
  return ans;
}
#undef FUNC_NAME

SCM_DEFINE (scm_brass_append, "brass-append", 0, 0, 1,
            (SCM rest_arg), "")
#define FUNC_NAME s_scm_brass_append
{
  reference_parameter (rest_arg);
  size_t c_len = 0;

  // Scan arguments and determine length.
  SCM rst = rest_arg;
  while (scm_is_pair (rst)) {
    SCM el = TMP_CAR (rst);
    SCM_VALIDATE_BRASS (SCM_ARGn, el);
    c_len = c_len + BR_LENGTH (el);
    rst = TMP_CDR (rst);
  }
  if (!scm_is_null (rst)) SCM_VALIDATE_LIST (1, rest_arg);

  SCM ans         = scm_c_make_brass (c_len);
  uint32_t *chars = BR_CHARS (ans);
  rst = rest_arg;
  size_t idx = 0;
  while (scm_is_pair (rst)) {
    SCM el = TMP_CAR (rst);
    memcpy (&(chars[idx]), BR_CHARS (el), BR_LENGTH (el) * sizeof (uint32_t));
    idx += BR_LENGTH (el);
    rst = TMP_CDR (rst);
  }

  dereference_parameter (rest_arg);
  return ans;
}
#undef FUNC_NAME

inline void
trim_release_brass (SCM br)
{

  scm_free_array (BR_NSLOTS(BR_LENGTH(br)), br);

  global_number_of_brass--;
  return;
}

void
x_inclusion_brass ()
{
  #include "include/brass.x"
}
