#include "include/config.h"

#include "include/_scm.h"
#include "include/root.h"
#include "include/smob.h"
#include "include/vectors.h"
#include "include/brass.h"
#include "include/byte-strings.h"

#include "include/validate.h"
#include "include/eq.h"


#ifdef HAVE_STRING_H
#include <string.h>
#endif


SCM_DEFINE1 (scm_eq_p, "eq?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	    "Return @code{#t} if @var{x} and @var{y} are the same object,\n")
#define FUNC_NAME s_scm_eq_p
{
  if (SCM_NIMP (x) || SCM_NIMP (y)) {
    reference_parameter (x);
    reference_parameter (y);
    SCM answer = scm_from_bool (scm_is_eq (x, y));
    dereference_parameter (x);
    dereference_parameter (y);
    return (answer);

  } else {
    return scm_from_bool (scm_is_eq (x, y));
  }
}
#undef FUNC_NAME

/* We compare doubles in a special way for 'eqv?' to be able to
   distinguish plus and minus zero and to identify NaNs.
*/

static int
real_eqv (double x, double y)
{
  return !memcmp (&x, &y, sizeof(double)) || (x != x && y != y);
}

#include <stdio.h>
SCM_PRIMITIVE_GENERIC_1 (scm_eqv_p, "eqv?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	    "Return @code{#t} if @var{x} and @var{y} are the same object, or\n"
	    "for characters and numbers the same value.")
#define FUNC_NAME s_scm_eqv_p
{
  reference_parameter (x);
  reference_parameter (y);

  if (scm_is_eq (x, y)) {
    dereference_parameter (x);
    dereference_parameter (y);
    return SCM_BOOL_T;

  } else if (SCM_IMP (x)) {
    dereference_parameter (y);
    return SCM_BOOL_F;

  } else if (SCM_IMP (y)) {
    dereference_parameter (x);
    return SCM_BOOL_F;
  }

  SCM answer = SCM_UNSPECIFIED;

  if (SCM_CELL_TYPE (x) != SCM_CELL_TYPE (y)) {
    /* treat mixes of real and complex types specially */
    if (SCM_INEXACTP (x)) {
        
      if (SCM_REALP (x)) {
        answer = scm_from_bool (SCM_COMPLEXP (y)
                                && real_eqv (SCM_REAL_VALUE (x),
                                             SCM_COMPLEX_REAL (y))
                                && SCM_COMPLEX_IMAG (y) == 0.0);
      } else {
        answer = scm_from_bool (SCM_REALP (y)
                                && real_eqv (SCM_COMPLEX_REAL (x),
                                             SCM_REAL_VALUE (y))
                                && SCM_COMPLEX_IMAG (x) == 0.0);
      }
      dereference_parameter (x);
      dereference_parameter (y);
      return answer;
    }
      
    if (SCM_FRACTIONP (x) && SCM_FRACTIONP (y))
      answer = scm_i_fraction_equalp (x, y);
    else
      answer = SCM_BOOL_F;

    dereference_parameter (x);
    dereference_parameter (y);
    return (answer);


  } else if (SCM_NUMP (x)) { // && same cell type

    if (SCM_BIGP (x)) {
      answer = scm_from_bool (scm_i_bigcmp (x, y) == 0);
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);

    } else if (SCM_REALP (x)) {
      answer = scm_from_bool (real_eqv (SCM_REAL_VALUE (x), \
                                        SCM_REAL_VALUE (y)));
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);

    } else if (SCM_FRACTIONP (x)) {
      answer = scm_i_fraction_equalp (x, y);
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);

    } else { /* complex */
      answer = \
       scm_from_bool ( \
           real_eqv (SCM_COMPLEX_REAL (x), SCM_COMPLEX_REAL (y)) \
        && real_eqv (SCM_COMPLEX_IMAG (x), SCM_COMPLEX_IMAG (y)));
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);
    }
  }

  dereference_parameter (x);
  dereference_parameter (y);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_PRIMITIVE_GENERIC_1 (scm_equal_p, "equal?", scm_tc7_rpsubr,
			 (SCM x, SCM y),
	    "Return @code{#t} if @var{x} and @var{y} are the same type, and\n"
	    "their contents or value are equal.")
#define FUNC_NAME s_scm_equal_p
{
  reference_parameter (x);
  reference_parameter (y);

 tailrecurse:  // tack
  if (scm_is_eq (x, y)) {
    dereference_parameter (x);
    dereference_parameter (y);
    return SCM_BOOL_T;

  } if (SCM_IMP (x)) {
    dereference_parameter (x);
    dereference_parameter (y);
    return SCM_BOOL_F;

  } if (SCM_IMP (y)) {
    dereference_parameter (x);
    dereference_parameter (y);
    return SCM_BOOL_F;
  }

 if (scm_is_pair (x) && scm_is_pair (y))
    {
      if (scm_is_false (scm_equal_p (SCM_CAR (x), SCM_CAR (y)))) {
        dereference_parameter (x);
        dereference_parameter (y);
	return SCM_BOOL_F;
      }
      x = FLY_CDR(x);
      y = FLY_CDR(y);
      goto tailrecurse;
    }

  if (SCM_TYP7 (x) == scm_tc7_brass && SCM_TYP7 (y) == scm_tc7_brass) {
    return scm_brass_equal_p (rc_Fly (x), rc_Fly (y));
  }
  if (SCM_TYP7 (x) == scm_tc7_byte_string \
      && SCM_TYP7 (y) == scm_tc7_byte_string) {
    return scm_byte_string_equal_p (rc_Fly (x), rc_Fly (y));
  }

  if (SCM_TYP7 (x) == scm_tc7_smob && SCM_TYP16 (x) == SCM_TYP16 (y)) {
    int i = SCM_SMOBNUM (x);
    if (!(i < scm_numsmob)) {
      dereference_parameter (x);
      dereference_parameter (y);
      return SCM_BOOL_F;

    } if (scm_smobs[i].equalp) {
      SCM answer = (scm_smobs[i].equalp) (x, y);
      reference_count_return (answer);
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);

    } else {
      goto generic_equal;
    }
  }

  if (SCM_CELL_TYPE (x) != SCM_CELL_TYPE (y)) {
    SCM answer = SCM_UNSPECIFIED;
    /* treat mixes of real and complex types specially */
    if (SCM_INEXACTP (x) && SCM_INEXACTP (y)) {
      if (SCM_REALP (x)) { // ??? real_eqv
        answer = scm_from_bool (SCM_COMPLEXP (y) \
                                && SCM_REAL_VALUE (x) == SCM_COMPLEX_REAL (y)
                                && SCM_COMPLEX_IMAG (y) == 0.0);
      } else {
        answer = scm_from_bool (SCM_REALP (y) \
                                && SCM_COMPLEX_REAL (x) == SCM_REAL_VALUE (y)
                                && SCM_COMPLEX_IMAG (x) == 0.0);
      }
      dereference_parameter (x);
      dereference_parameter (y);
      return (answer);

    } else {
      dereference_parameter (x);
      dereference_parameter (y);
      return SCM_BOOL_F;
    }
  }

  switch (SCM_TYP7 (x)) {
    SCM answer = SCM_UNSPECIFIED;
  default:
    break;
  case scm_tc7_number:
    switch SCM_TYP16 (x) {
      case scm_tc16_big:
        answer = scm_bigequal (x, y);
        dereference_parameter (x);
        dereference_parameter (y);
        return (answer);
        
      case scm_tc16_real:
        answer = scm_real_equalp (x, y);
        dereference_parameter (x);
        dereference_parameter (y);
        return (answer);

      case scm_tc16_complex:
        answer = scm_complex_equalp (x, y);
        dereference_parameter (x);
        dereference_parameter (y);
        return (answer);

      case scm_tc16_fraction:
        answer = scm_i_fraction_equalp (x, y);
        dereference_parameter (x);
        dereference_parameter (y);
        return (answer);
      }
  case scm_tc7_vector:
    answer = scm_i_vector_equal_p (x, y);
    dereference_parameter (x);
    dereference_parameter (y);
    return (answer);
  }

 generic_equal:
  dereference_parameter (x);
  dereference_parameter (y);
  return SCM_BOOL_F;
}
#undef FUNC_NAME



void
x_inclusion_equality ()
{
#include "include/eq.x"
}
