#include "include/config.h"

#include "include/_scm.h"
#include "include/alist.h" /* for SCM_EXTEND_ENV (well...) */
#include "include/eval.h"
#include "include/ports.h"
#include "include/print.h"
#include "include/root.h"
#include "include/smob.h"

#include "include/validate.h"
#include "include/macros.h"

scm_t_tag scm_tc16_macro;


static int
macro_print (SCM macro, SCM port, scm_print_state *pstate)
{
  SCM code = SCM_MACRO_CODE (macro);

  if (!SCM_CLOSUREP (code))
    scm_puts ("#<primitive-", port);
  else
    scm_puts ("#<", port);

  if (SCM_MACRO_TYPE (macro) == 0)
    scm_puts ("syntax", port);
  if (SCM_MACRO_TYPE (macro) == 1)
    scm_puts ("macro", port);
  if (SCM_MACRO_TYPE (macro) == 2)
    scm_puts ("macro!", port);
  if (SCM_MACRO_TYPE (macro) == 3)
    scm_puts ("builtin-macro!", port);

  scm_putc (' ', port);
  scm_iprin1 (scm_macro_name (macro), port, pstate);

  scm_putc ('>', port);
    

  return 1;
}

static SCM
makmac (SCM code, scm_t_bits flags)
{
  reference_parameter (code);

  SCM z;
  SCM_NEWSMOB (z, scm_tc16_macro, SCM_UNPACK (code));
  reference (z, code); dereference_parameter (code);
  SCM_SET_SMOB_FLAGS (z, flags);
  return z;
}

/* Return a mmacro that is known to be one of the built in macros. */
SCM
scm_i_makbimacro (SCM code)
#define FUNC_NAME "scm_i_makbimacro"
{
  SCM_VALIDATE_PROC (1, code);
  return makmac (code, 3);
}
#undef FUNC_NAME


SCM_DEFINE (scm_makmmacro, "procedure->memoizing-macro", 1, 0, 0, 
           (SCM code),
	    "Return a @dfn{macro} which, when a symbol defined to this value\n"
	    "appears as the first symbol in an expression, evaluates the\n"
	    "result of applying @var{code} to the expression and the\n"
	    "environment.\n\n"
	    "@code{procedure->memoizing-macro} is the same as\n"
	    "@code{procedure->macro}, except that the expression returned by\n"
	    "@var{code} replaces the original macro expression in the memoized\n"
	    "form of the containing code.")
#define FUNC_NAME s_scm_makmmacro
{
  SCM_VALIDATE_PROC (1, code);
  return makmac (code, 2);
}
#undef FUNC_NAME


SCM_DEFINE (scm_makacro, "procedure->syntax", 1, 0, 0,
            (SCM code),
	    "Return a @dfn{macro} which, when a symbol defined to this value\n"
	    "appears as the first symbol in an expression, returns the\n"
	    "result of applying @var{code} to the expression and the\n"
	    "environment.")
#define FUNC_NAME s_scm_makacro
{
  SCM_VALIDATE_PROC (1, code);
  return makmac (code, 0);
}
#undef FUNC_NAME


SCM_DEFINE (scm_makmacro, "procedure->macro", 1, 0, 0, 
           (SCM code),
	    "Return a @dfn{macro} which, when a symbol defined to this value\n"
	    "appears as the first symbol in an expression, evaluates the\n"
	    "result of applying @var{code} to the expression and the\n"
	    "environment.  For example:\n"
	    "\n"
	    "@lisp\n"
	    "(define trace\n"
	    "  (procedure->macro\n"
	    "   (lambda (x env) `(set! ,(cadr x) (tracef ,(cadr x) ',(cadr x))))))\n\n"
	    "(trace @i{foo}) @equiv{} (set! @i{foo} (tracef @i{foo} '@i{foo})).\n"
	    "@end lisp")
#define FUNC_NAME s_scm_makmacro
{
  SCM_VALIDATE_PROC (1, code);
  return makmac (code, 1);
}
#undef FUNC_NAME


SCM_DEFINE (scm_macro_p, "macro?", 1, 0, 0, 
            (SCM obj),
	    "Return @code{#t} if @var{obj} is a regular macro, a memoizing macro or a\n"
	    "syntax transformer.")
#define FUNC_NAME s_scm_macro_p
{
  reference_parameter (obj);
  SCM answer = scm_from_bool (SCM_SMOB_PREDICATE (scm_tc16_macro, obj));
  dereference_parameter (obj);
  return (answer);
}
#undef FUNC_NAME


SCM_SYMBOL (scm_sym_syntax, "syntax");
#if SCM_ENABLE_DEPRECATED == 1
SCM_SYMBOL (scm_sym_macro, "macro");
#endif
SCM_SYMBOL (scm_sym_mmacro, "macro!");
SCM_SYMBOL (scm_sym_bimacro, "builtin-macro!");

SCM_DEFINE (scm_macro_type, "macro-type", 1, 0, 0, 
            (SCM m),
	    "Return one of the symbols @code{syntax}, @code{macro} or\n"
	    "@code{macro!}, depending on whether @var{m} is a syntax\n"
	    "transformer, a regular macro, or a memoizing macro,\n"
	    "respectively.  If @var{m} is not a macro, @code{#f} is\n"
	    "returned.")
#define FUNC_NAME s_scm_macro_type
{
  reference_parameter (m);

  if (!SCM_SMOB_PREDICATE (scm_tc16_macro, m))
    return SCM_BOOL_F;

  SCM answer;
  switch (SCM_MACRO_TYPE (m))
    {
    case 0: answer = scm_sym_syntax; break;
#if SCM_ENABLE_DEPRECATED == 1
    case 1: answer = scm_sym_macro; break;
#endif
    case 2: answer = scm_sym_mmacro; break;
    case 3: answer = scm_sym_bimacro; break;
    default: scm_wrong_type_arg (FUNC_NAME, 1, m);
    }
  reference_count_return (answer);
  dereference_parameter (m);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_macro_name, "macro-name", 1, 0, 0, 
            (SCM m),
	    "Return the name of the macro @var{m}.")
#define FUNC_NAME s_scm_macro_name
{
  reference_parameter (m);
  SCM_VALIDATE_SMOB (1, m, macro);
  SCM answer = scm_procedure_name (SCM_PACK (SCM_SMOB_DATA (m)));
  dereference_parameter (m);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_macro_transformer, "macro-transformer", 1, 0, 0, 
            (SCM m),
	    "Return the transformer of the macro @var{m}.")
#define FUNC_NAME s_scm_macro_transformer
{
  reference_parameter (m);
  SCM_VALIDATE_SMOB (1, m, macro);
  SCM answer = ((SCM_CLOSUREP (SCM_PACK (SCM_SMOB_DATA (m)))) ?
                SCM_PACK(SCM_SMOB_DATA (m)) : SCM_BOOL_F);
  reference_count_return (answer);
  dereference_parameter (m);
  return (answer);
}
#undef FUNC_NAME

SCM
scm_make_synt (const char *name, SCM (*macroizer) (), SCM (*fcn)() )
{
  SCM var = scm_c_define ((char*)name, SCM_UNDEFINED);
  SCM transformer = scm_c_make_subr (name, scm_tc7_subr_2, fcn);
  SCM_VARIABLE_SET (var, macroizer (transformer));
  return SCM_UNSPECIFIED;
}

static void
macro_release (SCM macro)
{
  dereference (macro, SCM_CELL_OBJECT_1 (macro));
  return;
}

void
scm_init_macros ()
{
  scm_tc16_macro = scm_make_smob_type ("macro", 0);
  scm_set_smob_release (scm_tc16_macro, 2, macro_release);
  scm_set_smob_print (scm_tc16_macro, macro_print);
}

void
x_inclusion_macros ()
{
#include "include/macros.x"
}
