#include "include/config.h"

#include "include/_scm.h"
#include "include/validate.h"

#include "include/pairs.h"
#include "include/trim.h"



/* {Pairs}
 */

#if (SCM_DEBUG_PAIR_ACCESSES == 1)

#include "include/ports.h"
#include "include/brass.h"

void scm_error_pair_access (SCM non_pair)
{
  static unsigned int running = 0;
  SCM message = utf8_to_brass ("Non-pair accessed with SCM_C[AD]R: `~S'\n");

  if (!running)
    {
      running = 1;
      scm_simple_format (scm_current_error_port (),
			 message, scm_list_1 (non_pair));
      abort ();
    }
}

#endif

/*
 * Pair Statistics
 */
long global_number_of_pairs = 0;
long global_last_number_of_pairs    = 0;

SCM_DEFINE (scm_global_pair_count, "global-pair-count", 0, 0, 0,
            (),
            "Returns the total number of pairs.")
{
  return scm_from_long (global_number_of_pairs);
}

SCM_DEFINE (scm_global_pair_count_delta, "global-pair-count-delta",
            0, 0, 0, (), "")
{
  long delta;

  delta = global_number_of_pairs - global_last_number_of_pairs;
  global_last_number_of_pairs = global_number_of_pairs;

  return scm_from_long (delta);
}


/*
 * Pairs
 */


/* Counting Low Level Accessors corresponding to counting-lambda 2/a */

inline SCM
scm_i_cons (SCM car_value, SCM cdr_value)
{
  reference_parameter (car_value);
  reference_parameter (cdr_value);

  SCM pair = one_cell_please ();   // optimized counting
  SCM_SET_CELL_TYPE (pair, scm_tc3_cons);
  global_number_of_pairs++;

  reference (pair, car_value);
  reference (pair, cdr_value);

  SCM_SET_CELL_OBJECT_0 (pair, car_value);
  SCM_SET_CELL_OBJECT_1 (pair, cdr_value);

  dereference_parameter (car_value);
  dereference_parameter (cdr_value);

  RC_RETURN (pair);
}

inline SCM
scm_i_car (SCM object)
{ // simplified parameter count
  SCM answer = SCM_CELL_OBJECT_0 (object);
  reference_parameter (answer);

  dereference_return_value (object);
  return rc_return_stack_reference (answer);
}

inline SCM
scm_i_cdr (SCM object)
{ // simplified parameter count
  SCM answer = SCM_CELL_OBJECT_1 (object);
  reference_parameter (answer);

  dereference_return_value (object);
  return rc_return_stack_reference (answer);
}

inline void
scm_i_set_car_x (SCM pair, SCM value)
{
  // Counting for /pair/ has been optimized.
  SCM old = SCM_CELL_OBJECT_0 (pair);
  reference_parameter (value);

  reference (pair, value);
  SCM_SET_CELL_OBJECT_0 (pair, value);
  dereference (pair, old);

  dereference_return_value (pair);
  dereference_parameter (value);
}

inline void
scm_i_set_cdr_x (SCM pair, SCM value)
{
  // Counting for /pair/ has been optimized.
  SCM old = SCM_CELL_OBJECT_1 (pair);
  reference_parameter (value);

  reference (pair, value);
  SCM_SET_CELL_OBJECT_1 (pair, value);
  dereference (pair, old);

  dereference_return_value (pair);
  dereference_parameter (value);
}


inline SCM
fly_car (SCM object) // out of protocol / abbreviation (see reference)
{
  // Fly
  reference_count_return (object);
  dereference_parameter (object);

  // CAR
  reference_parameter (object);
  SCM answer = SCM_CELL_OBJECT_0 (object);
  reference_count_return (answer);
  dereference_parameter (object);

  // X
  reference_parameter (answer);

  return (answer);
}

inline SCM
fly_cdr (SCM object) // out of protocol / abbreviation (see reference)
{
  // Fly
  reference_count_return (object);
  dereference_parameter (object);

  // CDR
  reference_parameter (object);
  SCM answer = SCM_CELL_OBJECT_1 (object);
  reference_count_return (answer);
  dereference_parameter (object);

  // X
  reference_parameter (answer);

  return (answer);
}


/* inline SCM */
/* fly_cdr (SCM object) // out of protocol / abbreviation (see reference) */
/* { */
/*   SCM answer = SCM_CELL_OBJECT_1 (object); */

/*   reference_parameter (answer); */
/*   dereference_parameter (object); */

/*   return (answer); */
/* } */


/*
 * Pair Primitives
 */

SCM_DEFINE (scm_cons, "cons", 2, 0, 0,
	    (SCM x, SCM y),
	    "Return a newly allocated pair whose car is @var{x} and whose\n"
	    "cdr is @var{y}.  The pair is guaranteed to be different (in the\n"
	    "sense of @code{eq?}) from every previously existing object.")
#define FUNC_NAME s_scm_cons
{
  return scm_i_cons (x, y);
}
#undef FUNC_NAME


SCM 
scm_cons2 (SCM w, SCM x, SCM y)
{
  return scm_i_cons (w, scm_i_cons (x, y));
}


SCM_DEFINE (scm_pair_p, "pair?", 1, 0, 0, 
            (SCM x),
	    "Return @code{#t} if @var{x} is a pair; otherwise return\n"
	    "@code{#f}.")
#define FUNC_NAME s_scm_pair_p
{
  return scm_from_bool (scm_is_pair (x));
}
#undef FUNC_NAME

SCM
scm_car (SCM pair)
{
  if (!scm_is_pair (pair))
    scm_wrong_type_arg_msg (NULL, 0, pair, "pair");
  return SCM_CAR (pair);
}

SCM
scm_cdr (SCM pair)
{
  if (!scm_is_pair (pair))
    scm_wrong_type_arg_msg (NULL, 0, pair, "pair");
  return SCM_CDR (pair);
}

SCM
scm_i_chase_pairs (SCM tree, scm_t_uint32 pattern)
{
  reference_parameter (tree);

  do
    {
      if (!scm_is_pair (tree))
	scm_wrong_type_arg_msg (NULL, 0, tree, "pair");

      if (pattern & 1) {
        tree = fly_car (tree);
      } else {
        tree = fly_cdr (tree);
      }

      pattern >>= 2;
    }
  while (pattern);

  return rc_return_stack_reference (tree);
}

SCM_DEFINE (scm_set_car_x, "set-car!", 2, 0, 0,
            (SCM pair, SCM value),
            "Stores @var{value} in the car field of @var{pair}.  The value returned\n"
            "by @code{set-car!} is unspecified.")
#define FUNC_NAME s_scm_set_car_x
{
  SCM_VALIDATE_CONS (1, pair);
  SCM_SETCAR (pair, value);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (scm_set_cdr_x, "set-cdr!", 2, 0, 0,
            (SCM pair, SCM value),
            "Stores @var{value} in the cdr field of @var{pair}.  The value returned\n"
            "by @code{set-cdr!} is unspecified.")
#define FUNC_NAME s_scm_set_cdr_x
{
  SCM_VALIDATE_CONS (1, pair);
  SCM_SETCDR (pair, value);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME



/* Every cxr-pattern is made up of pairs of bits, starting with the two least
 * significant bits.  If in a pair of bits the least significant of the two
 * bits is 0, this means CDR, otherwise CAR.  The most significant bits of the
 * two bits is only needed to indicate when cxr-ing is ready.  This is the
 * case, when all remaining pairs of bits equal 00.  */

typedef struct {
  const char *name;
  unsigned char pattern;
} t_cxr;

static const t_cxr cxrs[] = 
{
  {"cdr",    0x02}, /* 00000010 */
  {"car",    0x03}, /* 00000011 */
  {"cddr",   0x0a}, /* 00001010 */
  {"cdar",   0x0b}, /* 00001011 */
  {"cadr",   0x0e}, /* 00001110 */
  {"caar",   0x0f}, /* 00001111 */
  {"cdddr",  0x2a}, /* 00101010 */
  {"cddar",  0x2b}, /* 00101011 */
  {"cdadr",  0x2e}, /* 00101110 */
  {"cdaar",  0x2f}, /* 00101111 */
  {"caddr",  0x3a}, /* 00111010 */
  {"cadar",  0x3b}, /* 00111011 */
  {"caadr",  0x3e}, /* 00111110 */
  {"caaar",  0x3f}, /* 00111111 */
  {"cddddr", 0xaa}, /* 10101010 */
  {"cdddar", 0xab}, /* 10101011 */
  {"cddadr", 0xae}, /* 10101110 */
  {"cddaar", 0xaf}, /* 10101111 */
  {"cdaddr", 0xba}, /* 10111010 */
  {"cdadar", 0xbb}, /* 10111011 */
  {"cdaadr", 0xbe}, /* 10111110 */
  {"cdaaar", 0xbf}, /* 10111111 */
  {"cadddr", 0xea}, /* 11101010 */
  {"caddar", 0xeb}, /* 11101011 */
  {"cadadr", 0xee}, /* 11101110 */
  {"cadaar", 0xef}, /* 11101111 */
  {"caaddr", 0xfa}, /* 11111010 */
  {"caadar", 0xfb}, /* 11111011 */
  {"caaadr", 0xfe}, /* 11111110 */
  {"caaaar", 0xff}, /* 11111111 */
  {0, 0}
};



void inline
trim_release_pair (SCM pair)
{
  global_number_of_pairs--;

  reference_count_minus1 ((SCM) TRIM_CELL_WORD_0 (pair));
  reference_count_minus1 ((SCM) TRIM_CELL_WORD_1 (pair));

  return;
}

void
scm_init_pairs ()
{
  unsigned int subnr = 0;

  for (subnr = 0; cxrs[subnr].name; subnr++)
    {
      SCM (*pattern) () = (SCM (*) ()) (scm_t_bits) cxrs[subnr].pattern;
      scm_c_define_subr (cxrs[subnr].name, scm_tc7_cxr, pattern);
    }

#include "include/pairs.x"
}
