#include "include/config.h"

#include "include/_scm.h"

#include "include/vectors.h"
#include "include/root.h"
#include "include/eval.h"  // scm_closure

#include "include/validate.h"
#include "include/procs.h"


/*
 * Procedure Statistics
 */


/* {Procedures}
 */

#define SCM_MAX_SUBR 2400
long scm_subr_table_entries = 0;
long scm_subr_table_room    = SCM_MAX_SUBR;
static scm_t_subr_entry scm_i_subrs[SCM_MAX_SUBR];
scm_t_subr_entry *scm_subr_table = scm_i_subrs;

SCM 
scm_c_make_subr (const char *name, long type, SCM (*fcn) ())
{
  register SCM z;
  long entry;

  if (scm_subr_table_entries >= scm_subr_table_room)
    {
      fprintf (stderr, " * VSI-C: scm-procs: scm_subr_table"
                       " is too small * ");
      abort ();
    }

  entry = scm_subr_table_entries;
  z = scm_cell (type, entry, (scm_t_bits) fcn);
  scm_subr_table[entry].handle  = z;
  scm_subr_table[entry].name    = scm_from_locale_symbol (name);
  scm_subr_table[entry].generic = 0;
  scm_subr_table[entry].properties = SCM_EOL;
  scm_subr_table[entry].gary    = 0;  // this has to be set by the caller
  scm_subr_table_entries++;

  heap_reference (scm_subr_table, scm_subr_table[entry].handle);
  heap_reference (scm_subr_table, scm_subr_table[entry].name);
  
  reference_count_return (z);
  return z;
}

SCM
scm_stand_in_scm_proc (SCM subr)
{
  reference_parameter (subr);

  long i;
  for (i = 0; i < scm_subr_table_entries; i++) {

    if (subr == scm_subr_table[i].handle) {

      // Initialize if necessary
      if (SCM_EOL == scm_subr_table[i].properties) {
        scm_subr_table[i].properties = \
          scm_closure (SCM_EOL, SCM_BOOL_F, SCM_EOL);
        heap_reference (scm_subr_table, scm_subr_table[i].properties);
      }

      dereference_parameter (subr);
      RC_RETURN (scm_subr_table[i].properties);
    }
  }
  fprintf (stderr, " * VSI-C missed a subroutine * ");
  abort ();
  // We could have returned NULL but what would that be good for?
}


SCM
scm_c_define_subr (const char *name, long type, SCM (*fcn) ())
{
  SCM subr = scm_c_make_subr (name, type, fcn);
  scm_define (SCM_SUBR_ENTRY(subr).name, subr);
  return subr;
}

/* This function isn't currently used since subrs are never freed. */
void
scm_free_subr_entry (SCM subr)
{
  /* long entry = SCM_SUBRNUM (subr); */
  /* /\* Move last entry in table to the free position *\/ */
  /* scm_subr_table[entry] = scm_subr_table[scm_subr_table_size - 1]; */
  /* SCM_SET_SUBRNUM (scm_subr_table[entry].handle, entry); */
  /* scm_subr_table_size--; */
}

SCM
scm_c_make_subr_with_generic (const char *name, 
			      long type, SCM (*fcn) (), SCM *gf)
{
  SCM subr = scm_c_make_subr (name, type, fcn);
  SCM_SUBR_ENTRY(subr).generic = gf;
  if (!(gf == NULL) && !(*gf == (SCM) NULL))  // XXX
    heap_reference (scm_subr_table, *gf);
  return subr;
}

SCM
scm_c_define_subr_with_generic (const char *name, 
				long type, SCM (*fcn) (), SCM *gf)
{
  SCM subr = scm_c_make_subr_with_generic (name, type, fcn, gf);
  scm_define (SCM_SUBR_ENTRY(subr).name, subr);
  return subr;
}

void
scm_mark_subr_table ()
{
  return;
  /* long i; */
  /* for (i = 0; i < scm_subr_table_size; ++i) */
  /*   { */
  /*     scm_gc_mark (scm_subr_table[i].name); */
  /*     if (scm_subr_table[i].generic && *scm_subr_table[i].generic) */
  /*       scm_gc_mark (*scm_subr_table[i].generic); */
  /*     if (SCM_NIMP (scm_subr_table[i].properties)) */
  /*       scm_gc_mark (scm_subr_table[i].properties); */
  /*   } */
}

/*
 * Counting Closures
 */

inline SCM
scm_i_get_environment (SCM closure)
{
  reference_parameter (closure);

  SCM answer = SCM_CELL_OBJECT_3 (closure);

  reference_count_return (answer);
  dereference_parameter (closure);
  return (answer);
}

inline void
scm_i_set_environment_x (SCM closure, SCM environment)
{
  reference_parameter (closure);
  reference_parameter (environment);

  SCM old = TMP_ENV (closure);

  reference   (closure, environment);
  SCM_SET_CELL_OBJECT_3  (closure, environment);
  dereference (closure, old);

  dereference_parameter (closure);
  dereference_parameter (environment);
  return;
}

inline void
scm_i_set_code_x (SCM closure, SCM formals, SCM body)
{
  reference_parameter (closure);
  reference_parameter (formals);
  reference_parameter (body);

  dereference (closure, TMP_CLOSURE_FORMALS (closure));
  dereference (closure, TMP_CLOSURE_BODY    (closure));
  reference (closure, formals);
  reference (closure, body);
  SCM_SET_CELL_OBJECT_1 (closure, formals);
  SCM_SET_CELL_OBJECT_2 (closure, body);

  dereference_parameter (closure);
  dereference_parameter (formals);
  dereference_parameter (body);
  return;
}

inline SCM
scm_i_closure_formals (SCM closure)
{
  reference_parameter (closure);

  SCM answer = TMP_CLOSURE_FORMALS (closure);
  reference_count_return (answer);
  dereference_parameter (closure);
  return (answer);
}

inline SCM
scm_i_closure_body (SCM closure)
{
  reference_parameter (closure);

  SCM answer = TMP_CLOSURE_BODY (closure);
  reference_count_return (answer);
  dereference_parameter (closure);
  return (answer);
}


inline SCM
scm_i_get_procprops (SCM closure)
{
  reference_parameter (closure);

  SCM answer = TMP_PROCPROPS (closure);
  if (!answer) answer = SCM_EOL;

  reference_count_return (answer);
  dereference_parameter (closure);
  return (answer);
}

inline void
scm_i_set_procprops_x (SCM closure, SCM properties)
{
  reference_parameter (closure);
  reference_parameter (properties);

  SCM old = TMP_PROCPROPS (closure);
  if (old) dereference (closure, TMP_PROCPROPS (closure));

  reference (closure, properties);
  TMP_SETPROCPROPS (closure, properties);

  dereference_parameter (closure);
  dereference_parameter (properties);
  return;
}


SCM_DEFINE (scm_procedure_p, "procedure?", 1, 0, 0, 
	    (SCM obj),
	    "Return @code{#t} if @var{obj} is a procedure.")
#define FUNC_NAME s_scm_procedure_p
{
  reference_parameter (obj);

  SCM answer;

  if (SCM_NIMP (obj))
    switch (SCM_TYP7 (obj))
      {
      case scm_tcs_closures:
      case scm_tcs_subrs:
        dereference_parameter (obj);
        return SCM_BOOL_T;
      default:
        dereference_parameter (obj);
	return SCM_BOOL_F;
      }

  dereference_parameter (obj);
  return SCM_BOOL_F;
}
#undef FUNC_NAME

SCM_DEFINE (scm_closure_p, "closure?", 1, 0, 0, 
           (SCM obj),
	    "Return @code{#t} if @var{obj} is a closure.")
#define FUNC_NAME s_scm_closure_p
{
  reference_parameter (obj);

  SCM answer = scm_from_bool (SCM_CLOSUREP (obj));

  reference_count_return (answer);
  dereference_parameter (obj);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_thunk_p, "thunk?", 1, 0, 0, 
	    (SCM obj),
	    "Return @code{#t} if @var{obj} is a thunk.")
#define FUNC_NAME s_scm_thunk_p
{
  reference_parameter (obj);

  SCM answer;

  if (SCM_NIMP (obj))
    {
      switch (SCM_TYP7 (obj))
	{
	case scm_tcs_closures:
          answer = scm_from_bool (!scm_is_pair (TMP_CLOSURE_FORMALS (obj)));
          reference_count_return (answer);
          dereference_parameter (obj);
          return answer;
	case scm_tc7_subr_0:
	case scm_tc7_subr_1o:
	case scm_tc7_lsubr:
	case scm_tc7_rpsubr:
	case scm_tc7_asubr:
          dereference_parameter (obj);
	  return SCM_BOOL_T;
	default:
	  ;
	}
    }

  dereference_parameter (obj);
  return SCM_BOOL_F;
}
#undef FUNC_NAME

/* Only used internally. */
int
scm_subr_p (SCM obj)
{
  if (SCM_NIMP (obj))
    switch (SCM_TYP7 (obj))
      {
      case scm_tcs_subrs:
	return 1;
      default:
	;
      }
  return 0;
}


void inline
trim_release_closure (SCM closure)
{
  if (TMP_PROCPROPS (closure)) // XXX
      reference_count_minus1 (TMP_PROCPROPS (closure));

  reference_count_minus1 (TMP_CLOSURE_FORMALS (closure));
  reference_count_minus1 (TMP_CLOSURE_BODY (closure));
  reference_count_minus1 (TMP_ENV (closure));

  global_number_of_closures--;
  return;
}

void
x_inclusion_procs ()
{
#include "include/procs.x"
}
