#include "include/config.h"

#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"

#include "include/ports.h"

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "include/smob.h"



/* scm_smobs scm_numsmob
 * implement a fixed sized array of smob records.
 * Indexes into this table are used when generating type
 * tags for smobjects (if you know a tag you can get an index and conversely).
 */

#define MAX_SMOB_COUNT 256
long scm_numsmob;
scm_smob_descriptor scm_smobs[MAX_SMOB_COUNT];

/* Lower 16 bit of data had to be zero. Could be anything now ...
*/
void
scm_i_set_smob_flags (SCM x, scm_t_bits data)
{
  SCM_SET_CELL_WORD_0 (x, (SCM_CELL_WORD_0 (x) & 0xFFFF) | data);
}

void
scm_assert_smob_type (scm_t_tag tag, SCM val)
{
  if (!SCM_SMOB_PREDICATE (tag, val))
    scm_wrong_type_arg_msg (NULL, 0, val, scm_smobs[SCM_TC2SMOBNUM(tag)].name);
}

/* {Mark}
 */

/* This function is vestigial.  It used to be the mark function's
   responsibility to set the mark bit on the smob or port, but now the
   generic marking routine in gc.c takes care of that, and a zero
   pointer for a mark function means "don't bother".  So you never
   need scm_mark0.

   However, we leave it here because it's harmless to call it, and
   people out there have smob code that uses it, and there's no reason
   to make their links fail.  */

SCM 
scm_mark0 (SCM ptr SCM_UNUSED)
{
  return SCM_BOOL_F;
}

SCM 
scm_markcdr (SCM ptr)
{
  return SCM_CELL_OBJECT_1 (ptr);
}

/* {Free}
 */

size_t 
scm_free0 (SCM ptr SCM_UNUSED)
{
  return 0;
}

size_t
scm_smob_free (SCM obj)
{
  long n = SCM_SMOBNUM (obj);
  if (scm_smobs[n].size > 0)
    scm_gc_free ((void *) SCM_CELL_WORD_1 (obj), 
		 scm_smobs[n].size, SCM_SMOBNAME (n));
  return 0;
}

/* {Print}
 */

int
scm_smob_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  long n = SCM_SMOBNUM (exp);
  scm_puts ("#<", port);
  scm_puts (SCM_SMOBNAME (n) ? SCM_SMOBNAME (n) : "smob", port);
  scm_putc (' ', port);
  if (scm_smobs[n].size)
    scm_uintprint (SCM_CELL_WORD_1 (exp), 16, port);
  else
    scm_uintprint (SCM_UNPACK (exp), 16, port);
  scm_putc ('>', port);
  return 1;
}

scm_t_tag 
scm_make_smob_type (char const *name, size_t size)
#define FUNC_NAME "scm_make_smob_type"
{
  long new_smob;

  new_smob = scm_numsmob;
  if (scm_numsmob != MAX_SMOB_COUNT)
    ++scm_numsmob;

  if (new_smob == MAX_SMOB_COUNT)
    scm_misc_error (FUNC_NAME, "maximum number of smobs exceeded", SCM_EOL);

  scm_smobs[new_smob].name = name;
  if (size != 0)
    {
      scm_smobs[new_smob].size = size;
    }

  return scm_tc7_smob + new_smob * 256;
}
#undef FUNC_NAME


void
scm_set_smob_release (scm_t_tag tc, int cell_size, void (*release) (SCM))
{
  scm_smobs[SCM_TC2SMOBNUM (tc)].release = release;
  scm_smobs[SCM_TC2SMOBNUM (tc)].cell_size = cell_size;
}

void
scm_set_smob_print (scm_t_tag tc, int (*print) (SCM, SCM, scm_print_state*))
{
  scm_smobs[SCM_TC2SMOBNUM (tc)].print = print;
}

void
scm_set_smob_equalp (scm_t_tag tc, SCM (*equalp) (SCM, SCM))
{
  scm_smobs[SCM_TC2SMOBNUM (tc)].equalp = equalp;
}

SCM
scm_make_smob (scm_t_tag tc) /* currently unused - see smobs.h */
{
  long n = SCM_TC2SMOBNUM (tc);
  size_t size = scm_smobs[n].size;
  scm_t_bits data = (size > 0
		     ? (scm_t_bits) scm_gc_malloc (size, SCM_SMOBNAME (n))
		     : 0);
  scm_smobs[n].number_of_objects += 1;
  return scm_cell (tc, (scm_t_bits) 0, data);
}


/* {Initialization for the type of free cells}
 */

static int
free_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  char buf[100];
  sprintf (buf, "#<free cell %p>",
	   (void *) SCM_UNPACK (exp));
  scm_puts (buf, port);
  
  return 1;
}

int
generic_smob_release (SCM smob)
{
  // !! Skeleton of a bug: /d/ was a copy of the smob descriptor on
  //    our stack frame. The counter was off then ... now it's a pointer.
  /* d.number_of_objects -= 1; */

  scm_smob_descriptor *d = &SCM_SMOB_DESCRIPTOR(smob);
  if (d->release) d->release (smob);
  d->number_of_objects -= 1;

  return d->cell_size;
}

void
scm_smob_init ()
{
  long i;
  scm_t_tag tc;

  if (global_debug_interpreter)
    fprintf (stderr, "VSI initialization trace:"
                     " scm_smob_init\n");

  scm_numsmob = 0;
  for (i = 0; i < MAX_SMOB_COUNT; ++i)
    {
      scm_smobs[i].name       = 0;
      scm_smobs[i].size       = 0;
      scm_smobs[i].release    = 0;
      scm_smobs[i].cell_size  = 0;
      scm_smobs[i].print      = scm_smob_print;
      scm_smobs[i].equalp     = 0;
      scm_smobs[i].number_of_objects = 0;
    }

  /* WARNING: This scm_make_smob_type call must be done first.  */
  tc = scm_make_smob_type ("free-cell-tag", 0);
  scm_set_smob_print (tc, free_print);
}
