#include "include/config.h"

#include "include/_scm.h"
#include "include/eq.h"
#include "include/root.h"
#include "include/tags.h"

#include "include/validate.h"
#include "include/vectors.h"

// That could be SCM_T_BITS
#define VECTOR_MAX_LENGTH (SCM_T_BITS_MAX >> 8)

/*
 * Vector Statistics
 */
long global_number_of_vectors = 0;
long global_last_number_of_vectors    = 0;

SCM_DEFINE (scm_global_vector_count, "global-vector-count", 0, 0, 0,
            (),
            "Returns the total number of vectors.")
{
  return scm_from_long (global_number_of_vectors);
}

SCM_DEFINE (scm_global_vector_count_delta,
            "global-vector-count-delta", 0, 0, 0, (), "")
{
  long delta;

  delta = global_number_of_vectors - global_last_number_of_vectors;
  global_last_number_of_vectors = global_number_of_vectors;

  return scm_from_long (delta);
}


/*
 * Vectors
 */

int
scm_is_vector (SCM obj) // XXX deprecated: use SCM_VECTOR_P instead
{
  return SCM_I_IS_VECTOR (obj);
}

int
scm_is_simple_vector (SCM obj) // XXX deprecated: use SCM_VECTOR_P instead
{
  return SCM_I_IS_VECTOR (obj);
}

const SCM *
scm_vector_elements (SCM vec, size_t *lenp, ssize_t *incp)
{
  if (lenp) /* XXX */
    {
      *lenp = SCM_I_VECTOR_LENGTH (vec);
      *incp = sizeof (SCM);
    }
  return (const SCM *) SCM_I_VECTOR_ELTS(vec);
}

SCM *
scm_vector_writable_elements (SCM vec, size_t *lenp, ssize_t *incp)
{
  if (lenp) /* XXX */
    {
      *lenp = SCM_I_VECTOR_LENGTH (vec);
      *incp = sizeof (SCM);
    }
  return SCM_I_VECTOR_WELTS(vec);
}

SCM_DEFINE (scm_vector_p, "vector?", 1, 0, 0, 
	    (SCM obj),
	    "Return @code{#t} if @var{obj} is a vector, otherwise return\n"
	    "@code{#f}.")
#define FUNC_NAME s_scm_vector_p
{
  return scm_from_bool (scm_is_vector (obj));
}
#undef FUNC_NAME

SCM_GPROC (s_vector_length, "vector-length", 1, 0, 0, scm_vector_length, g_vector_length);

SCM
scm_vector_length (SCM v)
{
  reference_parameter (v);

  if (SCM_I_IS_VECTOR (v)) {
    SCM answer = scm_from_size_t (SCM_I_VECTOR_LENGTH (v));
    reference_count_return (answer);
    dereference_parameter (v);
    return answer;

  }  else {
    SCM_WTA_DISPATCH_1 (g_vector_length, v, 1, NULL);
  }

  dereference_parameter (v);
  return SCM_BOOL_F;
}

size_t
scm_c_vector_length (SCM v)
{
  if (SCM_I_IS_VECTOR (v))
    return SCM_I_VECTOR_LENGTH (v);

  return 0;
}

SCM_DEFINE (scm_vector_ref, "vector-ref", 2, 0, 0,
            (SCM v, SCM k), "")
#define FUNC_NAME s_scm_vector_ref
{
  reference_parameter (v);
  reference_parameter (k);

  SCM answer;
  if (SCM_VECTOR_P (v)) {
    size_t i = scm_to_size_t (k);

    if (i >= SCM_I_VECTOR_LENGTH (v))
      scm_out_of_range (FUNC_NAME, k); 

    answer = (SCM_I_VECTOR_ELTS(v))[i];

  } else {
    scm_wrong_type_arg (FUNC_NAME, 1, v);
    // or SCM_WTA_DISPATCH_2 (g_vector_ref, v, scm_from_size_t (k), 2, NULL);
  }

  reference_count_return (answer);
  dereference_parameter (v);
  dereference_parameter (k);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_vector_set_x, "vector-set!", 3, 0, 0,
            (SCM v, SCM k, SCM obj), "")
#define FUNC_NAME s_scm_vector_set_x
{
  reference_parameter (v);
  reference_parameter (k);
  reference_parameter (obj);

  if (SCM_VECTOR_P (v)) {
    size_t i = scm_to_size_t (k);

    if (i >= SCM_I_VECTOR_LENGTH (v))
      scm_out_of_range (FUNC_NAME, k); 

    SCM_SIMPLE_VECTOR_SET (v, i, obj);

  } else {
    scm_wrong_type_arg (FUNC_NAME, 1, v);
  }

  dereference_parameter (v);
  dereference_parameter (k);
  dereference_parameter (obj);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

inline void
scm_i_vector_set_x (SCM vector, unsigned long index, SCM element)
{
  reference_parameter (vector);
  reference_parameter (element);

  reference (vector, element);
  dereference (vector, (SCM_I_VECTOR_ELTS(vector))[index]);
  (SCM_I_VECTOR_WELTS(vector))[index] = element;

  dereference_parameter (vector);
  dereference_parameter (element);
  return;
}

inline SCM
scm_i_vector_ref (SCM vector, unsigned long index)
{
  SCM answer = (SCM_I_VECTOR_WELTS(vector)[index]);
  reference_parameter (answer);

  dereference_return_value (vector);
  return rc_return_stack_reference (answer);
}

SCM_DEFINE (scm_make_vector, "make-vector", 1, 1, 0,
            (SCM k, SCM fill),
	    "Return a newly allocated vector of @var{k} elements.  If a\n"
	    "second argument is given, then each position is initialized to\n"
	    "@var{fill}.  Otherwise the initial contents of each position is\n"
	    "unspecified.")
#define FUNC_NAME s_scm_make_vector
{
  reference_parameter (k);
  reference_parameter (fill);

  size_t l = scm_to_unsigned_integer (k, 0, VECTOR_MAX_LENGTH);

  if (SCM_UNBNDP (fill))
    fill = SCM_UNSPECIFIED;
  
  SCM ret = scm_c_make_vector (l, fill);

  reference_count_return (ret);
  dereference_parameter (k);
  dereference_parameter (fill);
  return ret;
}
#undef FUNC_NAME

/*
 * Vector Construction Machines
 * scm_c_make_vector, scm_vector = scm_list_to_vector, scm_copy_vector
 *
 * Testing suggests the length in its own slot is faster than the former
 * combination of tag and length. (3% for v.scm) Have fun!
 *
 * Using a slot location (pointer) in vector->list resulted in reduced
 * performance: 0.3 seconds for a reported user run time of about
 * 14.7 seconds for Aftermath/v.scm.
 *
 * Please update VECTOR_MAX_LENGTH: it perhaps should remain an immediate
 * number though.
 */ 

SCM
scm_c_make_vector (size_t nslots, SCM fill)
#define FUNC_NAME s_scm_make_vector
{
  SCM array;
  SCM *base;
  unsigned long int j;

  SCM_ASSERT_RANGE (1, scm_from_ulong (nslots), nslots <= VECTOR_MAX_LENGTH);

  array = scm_tagged_array (scm_tc7_vector, 1+nslots);
  SCM_SET_CELL_WORD_0 (array, (scm_t_bits) nslots);

  base  = SCM_I_VECTOR_ELTS (array);
  for (j = nslots; j != 0; j--, base++)
    *base = fill;

  multiple_reference (array, nslots, fill);

  global_number_of_vectors++;
  RC_RETURN (array);
}
#undef FUNC_NAME

SCM_REGISTER_PROC (s_list_to_vector, "list->vector", 1, 0, 0, scm_vector);
SCM_DEFINE (scm_vector, "vector", 0, 0, 1, (SCM list), "")
#define FUNC_NAME s_scm_vector
{
  reference_parameter (list);

  SCM l = list;
  SCM res;
  long i, len;

  SCM_VALIDATE_LIST_COPYLEN (1, l, len);

  res = scm_tagged_array (scm_tc7_vector, 1+len);
  SCM_SET_CELL_WORD_0 (res, (scm_t_bits) len);

  SCM *slot_loc = SCM_I_VECTOR_ELTS (res);
  i = len;
  while (i--) { // && scm_is_pair (l) but see the list validation above
    SCM element = TMP_CAR (l);
    reference (res, element);
    *slot_loc = element;
    slot_loc++;
    l = TMP_CDR (l);
  }
  global_number_of_vectors++;

  dereference_parameter (list);
  RC_RETURN (res);
}
#undef FUNC_NAME


SCM_DEFINE (scm_vector_copy, "vector-copy", 1, 0, 0, (SCM vec), "")
#define FUNC_NAME s_scm_vector_copy
{
  reference_parameter (vec);

  size_t i, len;
  ssize_t inc;
  const SCM *src;
  SCM *dst;
  SCM ret = 0;

  if (SCM_VECTOR_P (vec)) {
    len = SCM_I_VECTOR_LENGTH (vec);

    ret = scm_tagged_array (scm_tc7_vector, 1+len);
    SCM_SET_CELL_WORD_0 (ret, (scm_t_bits) len);

    src = SCM_I_VECTOR_ELTS (vec);
    dst = SCM_I_VECTOR_ELTS (ret);
    for (i = 0; i < len; i++) {
      reference (ret, src[i]);
      dst[i] = src[i];
    }

    global_number_of_vectors++;

  } else {
    scm_wrong_type_arg (FUNC_NAME, 1, vec);
  }

  reference_count_return (ret);
  dereference_parameter (vec);
  return ret;
}
#undef FUNC_NAME

SCM_DEFINE (scm_vector_to_list, "vector->list", 1, 0, 0, (SCM v), "")
#define FUNC_NAME s_scm_vector_to_list
{
  reference_parameter (v);

  SCM res = SCM_EOL;
  size_t len;
  long i;

  if (SCM_VECTOR_P (v)) {
    len = SCM_SIMPLE_VECTOR_LENGTH (v);
    for (i = len - 1; i >= 0; i--) {
      res = scm_cons (SCM_SIMPLE_VECTOR_REF (v, i), res);
    }
  } else {
    scm_wrong_type_arg (FUNC_NAME, 1, v);
  }

  dereference_parameter (v);
  return res;
}
#undef FUNC_NAME


SCM_DEFINE (scm_vector_fill_x, "vector-fill!", 2, 0, 0, (SCM v, SCM fill), "")
#define FUNC_NAME s_scm_vector_fill_x
{
  reference_parameter (v);
  reference_parameter (fill);

  SCM *data;
  int len;
  int i;

  if (SCM_VECTOR_P (v)) {
    len = SCM_SIMPLE_VECTOR_LENGTH (v);
    for (i = 0; i < len; i++) {
      SCM_SIMPLE_VECTOR_SET (v, i, fill);
    }

  } else {
    scm_wrong_type_arg (FUNC_NAME, 1, v);
  }

  dereference_parameter (v);
  dereference_parameter (fill);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM
scm_i_vector_equal_p (SCM x, SCM y)
{
  long i;
  
  // The length check is implicitly done by the CELL_TYPE check in
  // /scm_eq/ but that may as well have been a coincidence.
  if (SCM_I_VECTOR_LENGTH (x) != SCM_I_VECTOR_LENGTH (y))
    return SCM_BOOL_F;

  for (i = SCM_I_VECTOR_LENGTH (x) - 1; i >= 0; i--)
    if (scm_is_false (scm_equal_p (SCM_I_VECTOR_ELTS (x)[i],
				   SCM_I_VECTOR_ELTS (y)[i])))
      return SCM_BOOL_F;

  return SCM_BOOL_T;
}


inline void
trim_release_vector (SCM vec)
{
  SCM *elements = SCM_I_VECTOR_ELTS (vec);
  size_t length = SCM_I_VECTOR_LENGTH (vec);
  unsigned long i;

  for (i = 0; i < length; i++) {
    dereference (vec, elements[i]);
  }

  scm_free_array (length+1, vec);

  global_number_of_vectors--;
  return;
}

void
x_inclusion_vectors ()
{
#include "include/vectors.x"
}

