#include "include/ihub.h"

/*
 * Arrays aren't necessarily faster than data graphs composed
 * of regular cells / pairs. They are more compact though and
 * easier to extend but you'll still need to adapt all the
 * interested accessors of your data type.
 *  There is no restriction to what data is found behind cell
 * word 0 but your data type implementation needs to ensure
 * it remains in bounds. [XXX tag is in M1A now; no restr. on word0]
 *
 * NOPHALANX to turn it off. PHALANX is a cheap copy&paste hack
 * as fast as malloc. It's much less dynamic than /malloc/ with
 * respect to the memory usage profile but you can adapt the steps
 * in /init_phalanx/ to match needs.
 * 
 */

unsigned long global_counter_non_phalanx = 0;

typedef struct scm_array_t_header {
     scm_t_tag cell_word_m1a;
scm_t_refcount cell_word_m1b;
           // SCM cell_word_0;
} scm_array_t_header;


/*
 * PHALANX
 */

typedef struct phalanx_descriptor_s {
  SCM free_list_head;
  unsigned long num_slots;
  unsigned long num_segments;
  unsigned long num_requests;
  unsigned long segment_size; // number of arrays in one segment
  void *segment_list;
} phalanx_descriptor;

#define PHALANX_MAX 2048
phalanx_descriptor scm_phalanx[PHALANX_MAX];

#define PD(n) (&(scm_phalanx[n]))
#define PSIZE(pd) (sizeof(scm_array_t_header) + (pd->num_slots * sizeof (SCM)))


void
init_phalanx ()
{
  unsigned long i;

  for (i = 0; i < PHALANX_MAX; i++) {
    scm_phalanx[i].num_slots      = i;
    scm_phalanx[i].free_list_head = TRIM_EOL;
    scm_phalanx[i].num_segments   = 0;
    scm_phalanx[i].num_requests   = 0;
    scm_phalanx[i].segment_list   = (void*) TRIM_EOL;
  }

  // Phalanx Steps

  // An array size of 0 conflicts with the current free list
  // routines which use slot 0 (cell word 0).
  scm_phalanx[0].segment_size = 12; // dummy; shouldn't happen

  // Last Step Adaption [2016-06-01]
  // - tests/init-9.number.scm: Level-V/a needed too much memory
  //   thus we needed to reduce segment sizes
  // - we expect objects, structs and vectors and presume most of
  //   them won't exceed 31 respectively 30 slots
  //
  // => Level-V/a passed with reduced memory usage when compared to
  //   malloc which sometimes exaggerated; no performance penalty for
  //   Aftermath/v.scm nor for tests/struct.scm. Reduced footprint
  //   for the corrected D-Stress (tests/rug.scm); no noticeable
  //   change of the memory footprint of the original D-Stress in
  //   tests/init-9.rug.scm (which can't be loaded twice or after
  //   tests/rug.scm; see rug.scm).
  //
  // Maybe worth noting special steps:
  // 3: 2D vector calculus
  // 4: 3D vector calculus
  // 5: D-Lists
  // PHALANX_MAX: counts as non-phalanx: /trim-phalanx-maps/
  //
  // Some days later: experimental reduction of the default segment
  // size to 6000 for cells, double cells and the first phalanx step.

  for (i = 1; i < 32; i++) {
    scm_phalanx[i].segment_size = 6000;
  }

  for (i = 32; i < 64; i++) {
    scm_phalanx[i].segment_size = 128;
  }

  for (i = 64; i < 256; i++) {
    scm_phalanx[i].segment_size = 64;
  }

  for (i = 256; i < PHALANX_MAX; i++) {
    scm_phalanx[i].segment_size = 16;
  }
}

// Phalanx Segment Lists
#define NEXT_SEGMENT(x)          (*(void**)x)
#define SEGMENT_CELLS(x)    (scm_array_t_header*)(((char*)x)+sizeof(void*))
#define SET_NEXT_SEGMENT_X(x, y) (*(void**)x = y)
#define SEGMENT_EOL_P(x)         ((void*)x == (void*)TRIM_EOL)

static void
drop_free_list (phalanx_descriptor *pd)
{
  pd->free_list_head = TRIM_EOL;
}

static void
add_free_list_segment_x (phalanx_descriptor *pd, SCM head, SCM tail)
{
  TRIM_SET_CELL_WORD_0 (tail, pd->free_list_head);
  pd->free_list_head = head;
}

static void
add_segment (phalanx_descriptor *pd, void *segment)
{
  SET_NEXT_SEGMENT_X (segment, pd->segment_list);
  pd->segment_list = segment;
}

static size_t
segment_list_length (phalanx_descriptor *pd)
{
  size_t answer = 0;
  void *segment = pd->segment_list;
  while (!SEGMENT_EOL_P (segment)) {
    segment = NEXT_SEGMENT (segment);
    answer++;
  }
  return answer;
}


// Returns 0 on zero saturation else non zero. This procedure is on
// the edge between SCM, cells and arbitray bits.
static int
add_freelist_segment (phalanx_descriptor *pd, void *segment)
{
  size_t array_size = PSIZE (pd);

  scm_array_t_header *arrays = SEGMENT_CELLS (segment);
  SCM free_list_segment_tail = TRIM_EOL;
  SCM free_list_segment_head = TRIM_EOL;
  int zero_saturation = 0;
  scm_array_t_header *arrayp = (scm_array_t_header*) \
    (((unsigned long) arrays) + (pd->segment_size - 1) * array_size);

  // Scan the segment to initialize the tail of the free list segment.
  while (arrayp >= arrays) {
    if (FREE_CELL_P (PTR2SCM(arrayp))) break;
    zero_saturation = 1;
    arrayp = (scm_array_t_header*)(((unsigned long) arrayp) - array_size);
  }
  if (arrayp >= arrays) {
    free_list_segment_tail = PTR2SCM (arrayp);
    free_list_segment_head = free_list_segment_tail;
    TRIM_SET_CELL_WORD_0 (free_list_segment_tail, TRIM_EOL);
    arrayp = (scm_array_t_header*)(((unsigned long) arrayp) - array_size);
  }

  // Scan the segment constructing the free list segment and keeping
  // track of the zero saturation state.
  while (arrayp >= arrays) {
    if (!FREE_CELL_P (PTR2SCM(arrayp))) {
      zero_saturation = 1;
      arrayp = (scm_array_t_header*)(((unsigned long) arrayp) - array_size);
      continue;
    }
    TRIM_SET_CELL_WORD_0 (PTR2SCM (arrayp), free_list_segment_head);
    free_list_segment_head = PTR2SCM (arrayp);
    arrayp = (scm_array_t_header*)(((unsigned long) arrayp) - array_size);
  }

  // Deliver the result depending on wether we found a segment with
  // zero saturation or not.
  if (zero_saturation != 0) {
    if (!TRIM_EOL_P (free_list_segment_head)) {
      add_free_list_segment_x (pd, free_list_segment_head, \
                                   free_list_segment_tail);
    }
    return 1;
  } else {
    return 0;
  }
}

static void
trim_phalanx_allocation (phalanx_descriptor *pd)
{
  drop_free_list (pd);

  void *segment = pd->segment_list;
  void *last    = (void*) TRIM_EOL;

  while (!SEGMENT_EOL_P (segment)) {

    int zero_saturation  = add_freelist_segment (pd, segment); 
    void *next = NEXT_SEGMENT (segment);

    if (zero_saturation == 0) {

      if (!SEGMENT_EOL_P (last))
        SET_NEXT_SEGMENT_X (last, next);

      if (segment == pd->segment_list)
        pd->segment_list = next;
      
      free (segment);
      pd->num_segments--;
      segment = next;

    } else {
      last = segment;
      segment = next;
    }
  }
  return;
}

void
trim_trim_phalanx ()
{
  size_t i;
  for (i = 0; i != PHALANX_MAX; i++) {
    trim_phalanx_allocation(PD(i));
  }
  return;
}

// Phalanx Free Lists

static void
s_alloc (phalanx_descriptor *pd)
{
  size_t size = PSIZE (pd);

  void *array_segment = \
    malloc ((size * pd->segment_size) + sizeof (void*));
  if (array_segment == NULL) {
    scm_i_memory_backup_ignition ();
    fprintf (stderr, " * VSI Fatal Error: memory allocation failed * \n");
    abort ();
  }
  scm_array_t_header *segment = SEGMENT_CELLS (array_segment);
  int n = 0;


  /* Initialize the new segment. */
  scm_array_t_header *p  = segment;
  p->cell_word_m1a = scm_tc_free_cell;
  p->cell_word_m1b = 0;
  TRIM_SET_CELL_WORD_0 (p, TRIM_EOL);
  scm_array_t_header *previous  = p;
  for (n = 1; n < pd->segment_size; n++) {
    p = (void*)(((unsigned long) p) + size);
    p->cell_word_m1a = scm_tc_free_cell;
    p->cell_word_m1b = 0;
    TRIM_SET_CELL_WORD_0 (p, (SCM) previous);
    previous = p;
  }

  /* Install the new free list. */
  if (!(TRIM_EOL_P (pd->free_list_head))) {
    fprintf (stderr, "s_alloc: assertion failed: free list not empty\n");
    exit (253); /* ??? */
  }
  pd->free_list_head = (SCM) p;

  /* Update the segment list */
  add_segment (pd, array_segment);
  return;
}


static inline void
assert_free_cell (SCM cell)
{
  if (TRIM_EOL_P (cell) || \
      !((SCM_CELL_TYPE (cell) == scm_tc_free_cell) \
        && (TRIM_REFCOUNT (cell) == 0))) {
    fprintf (stderr, " * VSI Fatal Error: assert_free_cell failed * ");
    abort ();
  }
}

SCM
one_array_please (unsigned long num_slots)
{
  phalanx_descriptor *pd = PD (num_slots);

  pd->num_requests++;

  if (TRIM_EOL_P (pd->free_list_head)) {
    s_alloc (pd);
    pd->num_segments++;
  }

  SCM array = pd->free_list_head;
  // if (global_debug_interpreter) assert_free_cell (array);
  assert_free_cell (array);
  pd->free_list_head = TRIM_CELL_WORD_0 (array);

  TRIM_SET_CELL_WORD_0 (array, SCM_UNSPECIFIED);

  RC_RETURN (array);
}

void
rc_array_release_reinsert (SCM array, unsigned long num_slots)
{
  phalanx_descriptor *pd = PD (num_slots);
  SCM_SET_CELL_TYPE (array, scm_tc_free_cell);
  TRIM_SET_CELL_WORD_0 (array, pd->free_list_head);
  pd->free_list_head = array;

  return;
}

// PHALANX saturation maps

static int
saturation (phalanx_descriptor *pd, void *segment)
{
  int c = 0;
  size_t array_size = PSIZE (pd);  // XXX precompute array_size into pd
  scm_array_t_header *arrays = SEGMENT_CELLS (segment);
  scm_array_t_header *arrayp = (scm_array_t_header*) \
    (((unsigned long) arrays) + (pd->segment_size - 1) * array_size);
  
  while (arrayp >= arrays) {
    if (!FREE_CELL_P (PTR2SCM(arrayp))) {
      c++;
    }
    arrayp = (scm_array_t_header*)(((unsigned long) arrayp) - array_size);
  }
  return c;
}

// Returns a saturation map as a list in allocation order.
static SCM
pd_saturation_map (phalanx_descriptor *pd)
{
  void * segment  = pd->segment_list;
  size_t segments = segment_list_length (pd);
  size_t index    = 0;

  int *counts     = malloc (segments * sizeof (int));
  if (!counts) return SCM_BOOL_F;

  while (index < segments) {
    if (!SEGMENT_EOL_P (segment)) {
      counts[index] = saturation (pd, segment);
    } else {
      counts[index] = -1;
    }
    segment = NEXT_SEGMENT (segment);
    index++;
  }

  SCM answer = SCM_EOL; // RC: consumed & pass through
  SCM segment_size_scm = scm_from_ulong (pd->segment_size);
  reference_parameter (segment_size_scm);

  for (index = 0; index < segments; index++) {
    answer = scm_cons (scm_divide (scm_from_int (counts[index]),
                                   segment_size_scm),
                       answer);
  }
  // RC: answer pass through

  free (counts);
  dereference_parameter (segment_size_scm);
  return answer;
}

SCM
phalanx_saturation_maps ()
{
  SCM ans = scm_c_make_vector (PHALANX_MAX, SCM_UNSPECIFIED);
  reference_parameter (ans);
  size_t i;
  for (i = 0; i < PHALANX_MAX; i++) {
    phalanx_descriptor *pd = PD(i);
    SCM ent = scm_list_5 (scm_from_ulong (pd->num_slots),
                          scm_from_ulong (pd->segment_size),
                          scm_from_ulong (pd->num_requests),
                          scm_from_ulong (pd->num_segments),
                          pd_saturation_map (PD(i)));
    SCM_SIMPLE_VECTOR_SET (ans, i, ent);
  }
  return rc_return_stack_reference (ans);
}

// PHALANX Interface

inline SCM
scm_tagged_array (scm_t_tag tag, size_t nslots)
{
  SCM array;

  if (nslots < PHALANX_MAX) {
    array = one_array_please (nslots);
  } else { // NOPHALANX
    size_t size;
    if (sizeof (SCM) == 8) {
      size = sizeof (scm_array_t_header) + (nslots << 3);
    } else {
      size = sizeof (scm_array_t_header) + (sizeof (SCM) * nslots);
    }
    array = PTR2SCM (scm_malloc (size));
    global_counter_non_phalanx++;
  }
  TRIM_SET_REFCOUNT (array, 0);
  SCM_SET_CELL_TYPE (array, tag);

  RC_RETURN (array);
}

inline void
scm_free_array (unsigned long num_slots, SCM array)
{
  if (num_slots < PHALANX_MAX) { // 10% perf check ???
    rc_array_release_reinsert (array, num_slots);
  } else { // NOPHALANX
    free (SCM2PTR (array));
  }
}
