#include <stdio.h>
#include "include/trim.h"
#include "include/debug.h"

/* Supported Data Type Header Inclusions */
#include "include/pairs.h"
#include "include/numbers.h"
#include "include/byte-strings.h"

#include "include/smob.h"
#include "include/dynwind.h"
#include "include/throw.h"

unsigned long global_counter_reinsertions = 0;
unsigned long global_counter_drops        = 0;


/* The caller needs to make sure /object/ denotes a valid cell.
 * (There is no check for immediates).
 */
void
rc_release (SCM object)
{
  if (0) return; // compile time debug switch

  int cell_size = 0;

  /* Dispatch on type specifics. */

  if (SCM_I_CONSP (object)) {
    trim_release_pair (object);
    cell_size = 2;

  } else {
    switch (SCM_TYP3 (object)) {
      case scm_tc3_closure:
        trim_release_closure (object);
        cell_size = 4; break;

      case scm_tc3_tc7_1:
      case scm_tc3_tc7_2:
     switch (SCM_TYP7 (object)) {
     case scm_tc7_vector:
       trim_release_vector (object);
       return; // SCM Array managed by PHALANX: nothing to do here.

     case scm_tc7_brass:
       trim_release_brass (object);
       return; // SCM Array managed by PHALANX: nothing to do here.

     case scm_tc7_byte_string:
       if (BYTE_STRING_MAPPING_P (object)) {
         global_number_of_byte_string_mappings--;
         cell_size = 2;
       } else {
         trim_release_byte_string (object);
         return; // SCM Array managed by PHALANX: nothing to do here.
       }

     case scm_tc7_number:
       switch (SCM_TYP16 (object)) {

       case scm_tc16_real:
         // inlined number data
         global_number_of_reals--;
         cell_size = 4; break;

       case scm_tc16_fraction:
         trim_release_num_fraction (object);
         cell_size = 4; break;

       case scm_tc16_complex:
         trim_release_num_complex (object);
         cell_size = 2; break;

       case scm_tc16_big:
         trim_release_num_big (object);
         cell_size = 4; break;
       } break; // closes SCM_TYP16 scm_tc7_number

     case scm_tc7_smob:
       if (FRAME_P (object)) {
         (*global_frame_count)--;
         cell_size = 2; break;

       } else if (WINDER_P (object)) {
         trim_release_winder (object);
         cell_size = 4; break;

       } else if (SCM_JMPBUFP (object)) {
         (*global_jmpbuffer_count)--;
         cell_size = 4; break;

       } else if (SCM_PRE_UNWIND_DATA_P (object)) {
         (*global_pre_unwind_data_count)--;
         cell_size = 2; break;

       } else if (FREE_CELL_P (object)) {
       // case scm_tc_free_cell:
         if (global_debug_interpreter) {
           fprintf (stderr, "VSI TRIM/RC: releasing a free cell\n");
           abort ();
         } break;

       } else {
         cell_size = generic_smob_release (object);
         break;
       }
       break;

     case scm_tc7_port:
       // trim_release_port (object); this is done when removing the table entry
       global_number_of_ports--;
       cell_size = 2; break;

     case scm_tc7_variable:  // This shouldn't happen yet.
       trim_release_variable (object);
       cell_size = 2; break;

     } // closes SCM_TYP7
    } // closes SCM_TYP3
  }

  global_counter_reinsertions++;

  /*
   * Clear the cell.
   */

   SCM_SET_CELL_TYPE   (object, scm_tc_free_cell);
  TRIM_SET_REFCOUNT    (object, 0);
  TRIM_SET_CELL_WORD_0 (object, 0);
  TRIM_SET_CELL_WORD_1 (object, TRIM_EOL);


  /*
   * Reinsert the cell into the free list.
   */

  if (cell_size == 2) {
    rc_release_reinsert (object);
  } else if (cell_size == 4) {
    rc_release_reinsert_double (object);
  } else {
    global_counter_drops++;
  }

  return;
}


/*
 * Debugging Routines
 */
SCM global_trim_abort_on_access = 0;

void trim_debug_abort_on_access (SCM object);

static int trim_debug_abort_tick = 10;

inline void
trim_debug_abort_on_access (SCM object)
{
  if (global_trim_abort_on_access != 0 \
      && object == global_trim_abort_on_access) {
    if (--trim_debug_abort_tick == 0) {
      fprintf (stderr, " * TRIM abort on access trap * ");
      debug_trim_print_rc (object);
      abort();
    }
  }
  return;
}

void
trim_i_debug_purple (SCM object, int ticks)
{
  trim_debug_abort_tick       = ticks;
  global_trim_abort_on_access = object;
  return;
}

inline SCM
trim_debug_cell_abort (SCM object, int n)
{
  if (n == 0) trim_debug_abort_on_access (object);
  return ((SCM *)SCM2PTR (object))[n];
}

inline void
trim_debug_set_cell_abort (SCM object, int n, long value)
{
  if (n == 0 && value != (long)(((SCM *)SCM2PTR (object))[n]))
    trim_debug_abort_on_access (object);
  (((SCM *)SCM2PTR (object))[n]) = (SCM) value;
return;
}

void
debug_trim_print_rc (SCM object)
{
  if (SCM_NIMP (object)) {
    fprintf (stderr, "TRIM/RC debug print reference count: %p: %.1f\n",
                     SCM2PTR (object),
                     (TRIM_REFCOUNT (object) / 2.0));
  } else {
    fprintf (stderr, "TRIM/RC debug print reference count: %p: N/A\n",
                     SCM2PTR (object));
  }
}
