;; SCA: System Call and Library Function Abstractions
;; PROTOTYPES                       Argument Handling
;;

(define bs:system-types
  '(long int short pointer byte))

(define (system-type-size type) ; public
 (case type
  ((long)    (size-of-long))
  ((int)     (size-of-int))
  ((short)   (size-of-short))
  ((pointer) (size-of-pointer))
  ((byte)  1)
  (else
   (error "system type size unknown" type))))


;;
;; Byte Structure Description
;;

(define bs:member-name   first)
(define bs:member-length second)
(define (bs:set-member-name! member name)
  (set-car! member name))
(define (bs:set-member-length! member length)
  (set-car! (cdr member) length))

(define (bs:parse-member-length! element)
   (or (and (symbol? (bs:member-length element))
            (begin (bs:set-member-length! element
                     (system-type-size (bs:member-length element)))
             #t))
       (and (number? (bs:member-length element))
            #t)
       (error "mischiefed structure description")))

(define (bs:keyword-filter member)
  (case (bs:member-name member)
    ((padding) #f)
    (else      #t)))
(define bs:padding-filter bs:keyword-filter)

(define (bs:unique-name? name rest-members)
 (null? (filter (lambda (member) (eq? name (bs:member-name member)))
                rest-members)))
(define (bs:check-name-space members)
  (if (null? members)
     #t
     (if (bs:unique-name? (bs:member-name (first members))
                          (cdr members))
         (bs:check-name-space (cdr members))
         (error "duplicate byte structure member name:"
                (bs:member-name (first members))) )))

(define (bs:parse-description . members)
 (bs:check-name-space (filter bs:keyword-filter members))
 (for-each bs:parse-member-length! members)
 members)

(define (bs-length description) ; public
  (apply + (map bs:member-length description)))

(define (bs:resolve-rest offset description name)
 (if (null? description)
   (error "byte structure member name unknown:" name)
   (let ((current (first description)))
     (if (eq? name (bs:member-name current))
       (values offset (bs:member-length current))
       (bs:resolve-rest (+ offset (bs:member-length current))
                               (cdr description)
                               name)) )))
(define (bs:resolve description name)
 (bs:resolve-rest 0 description name))

;;
;; Byte Structure Description Syntax Form
(define (bs-description:transform-member! member environment)
 (let ((size (bs:member-length member)))
  (cond ((symbol? size)
           (if (memq size bs:system-types)
             (bs:parse-member-length! member)
             (let ((value (eval! size environment)))
              (if (number? value)
                (bs:set-member-length! member value)
                (error "mischiefed byte structure member size"
                       (bs:member-name member))) )))
        ((and (not symbol?) (not (number? size)))
           (error "mischiefed byte structure member size"
                  (bs:member-name member))) )))

(define (bs-description:transform-member-case! member environment)
 (if (eq? (bs:member-name member) bs:description-literal)
  (error "TODO structure in structure support")
  (bs-description:transform-member! member environment)))

(define (bs-description:transformer expression environment)
 (let ((description (cdr expression)))
  (bs:check-name-space (filter bs:keyword-filter description))
  (for-each (lambda (member)
              (bs-description:transform-member-case! member environment))
            description)
  (set-car! expression 'quote)
  (set-cdr! expression (list description)))
 expression)

(define bs:description-literal 'byte-structure-description)
(define byte-structure-description
  (procedure->macro bs-description:transformer))


;;
;; Byte Structure Accessor Computation
;;

;;
;; Stage 2: byte string level accessors
(define (bs:make-getter OFFSET SIZE)
  (lambda (byte-string)
    (byte-string-select byte-string OFFSET SIZE)))

(define (bs:make-setter OFFSET SIZE)
  (lambda (byte-structure byte-string)
    (when (not (= SIZE (byte-string-length byte-string)))
      (error "assumption didn't hold"))
    (byte-string-set! byte-structure OFFSET byte-string)))

;;
;; Stage 3: we love symbols (byte structure level)
(define (byte-structure-getter structure-description member-name) : public
 (apply bs:make-getter (bs:resolve structure-description member-name)))

(define (byte-structure-setter structure-description member-name) ; public
 (apply bs:make-setter (bs:resolve structure-description member-name)))

;;
;; Stage 4: this works if /define/ is well defined

;; DEPRECATED namespace imcompatibilities
;; (define (bs-getters:define-getter structure-description prefix name)
;;   (define/global (bs-getters:coin-symbol prefix name)
;;                  (byte-structure-getter structure-description name)))

;; (define (bs:define-bs-getters structure-description prefix)
;;   (for-each (lambda (name)
;;               (bs-getters:define-getter structure-description
;;                                         prefix name))
;;             (map bs:member-name
;;                  (filter bs:padding-filter
;;                          structure-description))))

;; (define (bs-setters:define-setter structure-description prefix name)
;;   (define/global (bs-setters:coin-symbol prefix name)
;;                  (byte-structure-setter structure-description name)))

;; (define (bs:define-bs-setters structure-description prefix)
;;   (for-each (lambda (name)
;;               (bs-setters:define-setter structure-description
;;                                         prefix name))
;;             (map bs:member-name
;;                  (filter bs:padding-filter
;;                          structure-description))))

;; (define (define-bs-accessors prefix structure-description) ; public
;;   (bs:define-bs-getters structure-description prefix)
;;   (bs:define-bs-setters structure-description prefix))

(define (bs-getters:coin-symbol prefix name)
  (string->symbol
   (string-append (symbol->string prefix)
                  ":"
                  (symbol->string name))))

(define (bs-setters:coin-symbol prefix name)
  (string->symbol
   (string-append (symbol->string prefix)
                  ":" "set-"
                  (symbol->string name)
                  "!")))

(define (bs-accessors:bindings expression environment)
 (when (not (= (length expression) 3))
   (error "mischiefed bs-accessors statement" expression))
 (let ((prefix                (eval! (second expression) environment))
       (structure-description (eval! (third expression)  environment)))
  (when (not (and (symbol? prefix)
                  (list? structure-description)))
                  ;XXX(byte-structure-description? structure-description)))
    (error "mischiefed bs-accessors arguments" expression))
   (let ((names (map bs:member-name (filter bs:padding-filter
                                            structure-description)) ))
  (cons 'begin
   (append
    (map (lambda (name)
          `(define ,(bs-getters:coin-symbol prefix name)
                   ,(byte-structure-getter structure-description
                                           name)))
         names)
    (map (lambda (name)
          `(define ,(bs-setters:coin-symbol prefix name)
                   ,(byte-structure-setter structure-description
                                           name)))
         names))
        )) ))

(define define-bs-accessors (procedure->macro bs-accessors:bindings))

;;
;; Byte Structure Functional Interface
;;

(define (byte-structure->list structure-description byte-string) ; public
 (let loop ((structure-description structure-description)
            (offset 0) (answer '()))
  (if (null? structure-description)
   (reverse! answer)
   (let ((member-length (bs:member-length (first structure-description))))
    (begin ; debugging paragraph
     (display "offset: ")         (display offset)
     (display " member-length: ") (display member-length)
     (display " member: ")        (display (car structure-description))
    (newline))
    (loop (cdr structure-description)
     (+ offset member-length)
     (if (eq? 'padding (bs:member-name (first structure-description)))
      answer
      (cons (byte-string-select byte-string offset member-length)
            answer))) )) ))

(define (call-with-bytes structure-description byte-string procedure) ; public
  (apply procedure (byte-structure->list structure-description byte-string)))

(define (values->byte-structure values structure-description) ; public
 ;; XXX todo: convert scheme data to byte strings
 (apply byte-string-adjoin (bsf:padding values structure-description)))

(define (bsf:padding rest-values rest-description)
 (cond
  ((null? rest-description) '())
  ((eq? 'padding (bs:member-name (first rest-description)))
   (cons (byte-string-clear!
          (make-byte-string (bs:member-length (first rest-description))))
         (bsf:padding rest-values (cdr rest-description))))
  (else
   (cons (car rest-values) 
         (bsf:padding (cdr rest-values)
                      (cdr rest-description))) )))


;;
;; Byte Size Casts
;;

(cond (system-is-big-endian
  (define (cast:4-1 b) (byte-string-select b 3 1))
  (define (cast:4-2 b) (byte-string-select b 2 2)))
      (else
  (define (cast:4-1 b) (byte-string-select b 0 1))
  (define (cast:4-2 b) (byte-string-select b 0 2)))
)
