;;
;; DICTIONARY DATA TYPE
;;
;; You shouldn't call it dictionary. Let's call it:
;; DINOSAURS! or Acyclic Directed Graphs
;;
;; Keep the alfabet small and it'll be fast.
;; While this implementation is designed to work
;; with strings extension to work with symbols or
;; adaptions to any other system with tokens should
;; be easy.


;;
;; Entries
;;

(define (make-entry key definition successors)
 (cons key (cons definition successors)))

(define (definition? entry)
  (not (eq? (get-definition entry) 'undefined)))

(define (get-key entry)        (car entry))
(define (get-definition entry) (cadr entry))
(define (get-successors entry) (cddr entry))
(define (set-definition! entry definition)
  (set-car! (cdr entry) definition))

(define (add-successor! entry successor-entry)
  (set-cdr! (cdr entry) (cons successor-entry
                              (get-successors entry))))
(define (get-successor entry key)
  (assq key (get-successors entry)))

(define (get-entry branch keys)
 (cond
  ((not branch)     #f)
  ((null? keys) branch)
  (else
   (get-entry (assq (car keys) (get-successors branch))
              (cdr keys)) )))

(define (create-entry! parent keys)
 (let* ((key      (first keys))
        (entry    (get-successor parent key))
        (last-key (null? (cdr keys))))

   (when (not entry)
     (set! entry (make-entry key 'undefined '()))
     (add-successor! parent entry))

   (if last-key
     entry
     (create-entry! entry (cdr keys))) ))

(define (insert-definition* root keys definition)
 (let ((entry (create-entry! root keys)))
   (set-definition! entry definition)))


;;
;; Internal Extra

(define (number-of-entries-with-definition branch)
 (+ (length (filter definition? (get-successors branch)))
    (apply + (map number-of-entries-with-definition
                  (get-successors branch)) )))

(define (number-of-nodes branch)
  (+ 1 (apply + (map number-of-nodes (get-successors branch)))))

(define (extract-entries branch symbol-root tail)
 (set-cdr! tail (cons (get-key branch) '() ))
 (set! tail (cdr tail))
 (append! (if (not (definition? branch))
              '()
              (acons (list->string (cdr symbol-root))
                     (get-definition branch)
                     '() ))
          (apply append!
            (map (lambda (entry) (extract-entries entry symbol-root tail))
                 (get-successors branch))) ))


;;
;; Dictionary Data Type
;;

(define (dictionary? object)  ;; Heuristic Type Check
 (and (pair? object) (eq? (car object) 'dictionary)))

(define (make-dictionary)
 (cons 'dictionary (make-entry 'root 'undefined '())))

(define (get-root dictionary)
 (when (not (dictionary? dictionary))
   (error "get-root: wrong type argument"))
 (cdr dictionary))


(define (insert-definition! dictionary string definition)
 (when (not (and (dictionary? dictionary)
                 (string? string)))
   (error "insert-definition!: wrong type argument"))
 (let ((characters (string->list string))
       (root       (get-root dictionary)))
  (insert-definition* root characters definition) ))

(define (lookup-definition dictionary string)
 (when (not (and (dictionary? dictionary)
                 (string? string)))
   (error "insert-definition!: wrong type argument"))
 (let ((entry (get-entry (get-root dictionary)
                         (string->list string))))
  (and entry (get-definition entry)) ))

;;
;; Public Extra: Meta Information and Conversion

(define (number-of-definitions dictionary)
 (when (not (dictionary? dictionary))
   (error "number-of-definitions: wrong type argument"))
 (number-of-entries-with-definition
  (get-root dictionary)))

(define (describe-dictionary dictionary)
  (when (not (dictionary? dictionary))
    (error "describe-dictionary: wrong type argument"))
  (acons 'number-of-definitions (number-of-definitions dictionary)
    (acons 'number-of-nodes (number-of-nodes (get-root dictionary))
      '() )))

(define (dictionary->alist dictionary)
 (when (not (dictionary? dictionary))
   (error "dictionary->alist: wrong type argument"))
 (apply append!
   (map (lambda (start-entry)
         (let ((symbol-root (cons (get-key start-entry) '()) ))
          (extract-entries start-entry symbol-root symbol-root)))
     (get-successors (get-root dictionary)) )))

(define (lookup-prefix dictionary prefix)
 (when (not (and (dictionary? dictionary)
                 (string?     prefix)
                 (>= (string-length prefix) 1)))
   (error "lookup-prefix: wrong type argument"))
 (let* ((prefix-keys (string->list prefix))
        (branch (get-entry (get-root dictionary) prefix-keys)))
  (if (not branch)
   '()
   (let ((prefix-root
           (cons #t (string->list
                     (substring prefix 0 (- (string-length prefix) 1))) )))
     (extract-entries branch prefix-root (last-pair prefix-root))) )))

(define (scan-dictionary dictionary predicate?)
 (when (not (and (dictionary? dictionary)
                 (procedure?  predicate?)))
   (error "scan-dictionary: wrong type argument"))
 (let ((root (get-root dictionary))
       (answer-box (cons #f '())))

 ;; CLR4 inner: We cannot use /once/ here as this inner (inner) define
 ;; relies on bindings in the surrounding environment, i.e. /predicate?/.
   (define (dictionary-scan-recurse answer-box entry)
    (let ((definition (get-definition entry)))
     (when (and (not (eq? definition 'undefined))
                (predicate? definition))
       (set-cdr! answer-box (cons definition (cdr answer-box))))
     (for-each (lambda (entry) (dictionary-scan-recurse answer-box entry))
               (get-successors entry))))

  (for-each (lambda (entry) (dictionary-scan-recurse answer-box entry))
            (get-successors root))
  (set! dictionary-scan-recurse #f) ; Cut our ring (CLR4 inner)
  (cdr answer-box)))

;;
;; Tests
;;

;; (define D (make-dictionary))
;; (insert-definition! D "hallo" 500)

;; (define sys-define define)
;; (define (nature:transformer expression environment)
;;  (let* ((id               (first expression))
;;         (procedure-syntax
;;                    (pair? (second expression)))
;;         (second     (second expression))
;;         (symbol     (if procedure-syntax
;;                       (car second)
;;                       second))
;;         (definition (if procedure-syntax
;;                       (append (list 'lambda (cdr second))
;;                               (third expression))
;;                       (third expression))))
;;    (insert-definition! D
;;                        (symbol->string symbol)
;;                        (cons 'definition definition))
;;    (cons 'sys-define (cdr expression))))

;; (define nature-define (procedure->macro nature:transformer))
