(define (list-index l k)
  (let loop ((n 0)
	     (l l))
    (and (not (null? l))
	 (if (eq? (car l) k) ;; XXX don't use eq? here unconditionally
	     n
	     (loop (+ n 1) (cdr l))))))

;; VSI nextgen list processing
(define (interleave . lists )
  (append (apply map list lists)))


;; VSI improper list detection abstraction prototype
;;
;; make-list-clock returns null for an empty list or
;; a list-clock closure for the given perhaps improper
;; list /graph/.
;;
;; The list-clock closure calls /bang/ which receives /length/
;; when it detects a loop else returns the next /cdr/ or null.
;;
;; TODO we need a C version w/o too much lambda bells&whistles
;;      -> goes into metacircular.objection
;; ??? are static variables thread local for each stack?
;;
(define (make-list-clock graph bang)
 (or (and (null? graph) '())
  (let ((hour graph) (minute graph) (second (cdr graph))
        (length 1) (pendulum #f))
   (lambda ()
    (when pendulum (set! minute (cdr minute)))
    (cond ((null? second)      '())
          ((not (pair? second)) (bang length)) ; improper list
          ((eq? second minute)  (bang length))
          (else
             (set! length (+ 1 length))
             (set! hour second)
             (set! second (cdr second))
             (set! pendulum (not pendulum))
             hour))) )))
;; Improper List Clock Usage Example (XXX goes to reference)
#!
VSI> (define (bang n) (display "BANG!") (newline))
VSI> (define l '())
VSI> (define m (make-list-clock l bang ) )
VSI> m
()
VSI> (define l (circular-list 1 2 3))
VSI> (define m (make-list-clock l bang ) )
VSI> (m)
(2 3 1 . #-2#)
VSI> (m)
(3 1 2 . #-2#)
VSI> (m)
(1 2 3 . #-2#)
VSI> (m)
(2 3 1 . #-2#)
VSI> (m)
BANG!
VSI>
!#

;;; Constructors

(define (relp-copy-list l)
  (cond
   ((null? l) '())
   (else
    (cons (car l) (relp-copy-list (cdr l))) )))

(define (dotted-rest-list head-element . tail)
 (if (not (pair? tail)) head-element
 (begin ; else
  (cons head-element
        (cond
         ((not (pair? tail))       tail )
         ((not (pair? (cdr tail))) (car tail) )
         (else
           (apply dotted-rest-list (car tail) (cdr tail))) )))))
(define cons* dotted-rest-list) ; GOOPS compatibility
(define list* dotted-rest-list) ; interrim CLOS remembrence
              

(define (circular-list elt1 . elts)
  (set! elts (cons elt1 elts))
  (set-cdr! (last-pair elts) elts)
  elts)

;;; {IOTA functions: generating lists of numbers}
(define (iota- n)
  (let loop ((count (1- n)) (result '()))
    (if (< count 0) result
        (loop (1- count) (cons count result)))))

;;; Predicates

(define (proper-list? x)
 (or (null? x)
     (and (pair? x)
       (let ((next (make-list-clock x (lambda (.args) #f))))
         (let loop ((current (next)))
           (cond ((pair? current) (loop (next)))
                 ((null? current) #t)
                 (else            #f)) )) )))

(define (circular-list? x)
  (if (not-pair? x)
    #f
    (let lp ((hare (cdr x)) (tortoise x))
      (if (not-pair? hare)
	#f
	(let ((hare (cdr hare)))
	  (if (not-pair? hare)
	    #f
	    (if (eq? hare tortoise)
	      #t
	      (lp (cdr hare) (cdr tortoise)))))))))

(define (dotted-rest-list? list)
 (let ((x (last-pair l))) ; last-pair throws in circular lists
  (cond
    ((null? x)     #f)
    ((not-pair? x) #t)
    (else
     (error "unexplainable error condition met")) )))


;;; Selectors

(define first car)
(define second cadr)
(define third caddr)
(define fourth cadddr)

(define take list-head)
(define drop list-tail)

(define (list-pair n list)
 (cond
  ((and (> n 0) (not (null? list)))
    (list-pair (- n 1) (cdr list)))

  ((= n 0)       list)
  ((null? list)   '()) ; null for n>list when the list is a proper list
  (else
    (error "list-pair: error condition"))))

(define (take! n list)
 (cond
  ((>= n 1)
     (let ((section (list-pair (- n 1) list)))
       (when (not (null? section)) (set-cdr! section '()))
       list))
  ((= n 0) '())
  (else
    (error "take! argument out of range" n))))

(define (drop-right! n list)
 (let ((l (length list)))
  (cond
   ((= n 0) list)
   ((> l n)
      (set-cdr! (list-pair n list) '())
      list)
   ((= n l)
      '())
   (else
    (error "drop-right!: error condition (list-length tail-length)"
           l n)) )))

(define (drop-right n list)
 (let ((l (length list)))
  (cond
   ((= n 0) (list-copy list))
   ((> l n)
     (list-head (- l n) list))
   ((= n l)
     '())
   (else
    (error "drop-right: error condition (list-length tail-length)"
           l n)) )))
       

;;; Miscelleneous: length, append, concatenate, reverse, zip & count

;; Reference Implementation of an imaginary /Length/ clock work.
;; [You don't want to map these dimensions in your mind.]
;; (define (O/clock graph)
;;  (or (and (null? graph) 0)
;;   (let ((length 1) (infinity +inf.0))
;;    (let half-pass ((minute graph) (second (cdr graph)) (pendulum #f))
;;     (when pendulum (set! minute (cdr minute)))
;;      (cond ((null? second)
;;               length)
;;            ((eq? second minute)
;;               (make-rectangular infinity (/ (+ 1 length) 2)))
;;            (else
;;               (set! length (+ 1 length))
;;               (half-pass minute (cdr second) (not pendulum))) ))) ))


;; Reference Implementation of /Reverse/
;; (define (O/reverse list)
;;  (let ((last '()))
;;   (let loop ((rest list))
;;     (cond
;;      ((null? rest) last)
;;      (else
;;       (set! last (cons (car rest) last))
;;       (loop (cdr rest)) )) )))


;; (flatten! proper-list)
;; Use /copy-tree/ for a non destructive /flatten/ when needed.
;; '(1 2 3)       => (1 2 3)
;; '((1 2) 3)     => (1 2 3)
;; '((1) (2) (3)) => (1 2 3)
;; '()            => ()
(define (flatten! alist)
 (let ((answer-box (cons #f '())))
  (let loop ((rest alist) (last answer-box))
   (cond
   ((null? rest)       (cdr answer-box))
   ((list? (car rest)) (set-cdr! last (car rest))
                       (loop (cdr rest) (last-pair last)))
   (else               (set-cdr! last (list (car rest)))
                       (loop (cdr rest) (cdr last))) )) ))


;;; Fold & Reduce

(define (reduce operation operands)
 (let ((answer-box (cons 'reduce '())))
  (let loop ((operands operands) (last answer-box))
   (cond
    ((null? (cdr operands)) (cdr answer-box))
    (else (let ((a (first operands)) (b (second operands)))
     (set-cdr! last (cons (operation a b) '()))
     (loop (cdr operands) (cdr last)) ))) )))

;; The following should (roughly) correspond to rp_subr in VSI-C.
;; It breaks as soon as /predicate?/ returns falsity to avoid
;; unecessary iterations.
(define (reduce-predicate predicate? operands)
 (let loop ((operands operands))
  (or (null? (cdr operands))
      (and (predicate? (first operands) (second operands))
           (loop (cdr operands)) ))
))


;;; Association lists

(define alist-cons acons)

;; SCM->C: That iterative "monster" came straight out of C ;-)
(define (alist-copy alist)  ;; XXX to be tested
 (let ((answer-box (cons 'alist-copy '()) ))
  (let loop ((last answer-box) (rest alist))
   (cond
    ((null? rest)       (cdr answer-box))

    ((not (pair? rest))
       ;; This reflects the behavior of the former C version but we
       ;; could as well just return what we have copied up to here.
       (error "alist-copy: argument is not a proper association list"))

    (else
       (let ((key-value (car rest)))
        (cond ((not (pair? key-value))
               (error
                "alist-copy: argument contains an invalid key value pair"))
              (else
               (set-cdr! last (acons (car key-value) (cdr key-value)
                                     '()))
               (loop (cdr last) (cdr rest)) ))) ))) ))

(define (alist-delete key alist . rest)
  (let ((k= (if (pair? rest) (car rest) equal?)))
    (let lp ((a alist) (rl '()))
      (if (null? a)
	(reverse! rl)
	(if (k= key (caar a))
	  (lp (cdr a) rl)
	  (lp (cdr a) (cons (car a) rl)))))))

(define (alist-delete! key alist . rest)
  (let ((k= (if (pair? rest) (car rest) equal?)))
    (alist-delete key alist k=)))	; XXX:optimize


;;; Duplicates

;; (define (list-contains? equality? object list) ; VSI-C
;;   (cond ((not (pair? list))             #f)
;;         ((equality? object (car list)) list)
;;         (else
;;          (list-contains? object equality? (cdr list)))))

(define (delete-duplicates list . equality?)
  (apply delete-duplicates! (list-copy list) equality?))

(define (delete-duplicates! list . rest-equality)
 (let ((equality? (if (null? rest-equality) equal? (car rest-equality))))
  (let loop ((rest list))
   (if (null? rest)
     list
     (begin (set-cdr! rest (delete! (car rest) (cdr rest) equality?))
            (loop (cdr rest))) )) ))

(define (list-duplicates list . rest-equality)
 (let ((equality? (if (null? rest-equality) equal? (car rest-equality))))
  (let loop ((rest list) (res '()))
   (cond
   ((null? rest)                                     (reverse! res))
   ((list-contains? equality? (car rest) (cdr rest)) (loop (cdr rest) (cons (car rest) res)))
   (else                                             (loop (cdr rest) res)) )) ))


;;
;; Quantifiers
;;

(define (all? booleans) (reduce-predicate      (lambda (a b) (and a b))      booleans))
(define (any? booleans) (not (reduce-predicate (lambda (a b) (not (or a b))) booleans)))
