;;
;; Namespace Prototype
;;  missing syntax checks
;;
;; There are some shortcomings here:
;;
;; ... Another way we could try would be to redefine /define/ as
;; a "macro" that knows how to add bindings to our namespace. In
;; a certain sphere we would need to modify the memory object denoted
;; by /define/ but we could as well establish /define/ to be a lexical
;; keyword in a namespace. As namespaces are considered a fundamental
;; extension of the algorithm notation scheme we could get away with
;; this in the prototype.
;;
;; Right now it's about functionality or the possibility to create
;; and use namespaces but when it comes to performance and /ilocs/
;; the effective cost of value retrieval during evaluation rises
;; linear with the number of bindings before the one we want to have
;; resolved.
;;
;; REVISIONS of the namespace prototype
;;
;; * Limits
;;
;; BTW: "Frame numbers have 11 bits, relative offsets have 12 bits."
;;   - That's 23 + 8 bit tag + 1 cdr bit change lock
;;   - 2047 is the maximum depth of the environment of a closure
;;   * 4097 is the presumed maximum of bindings in a frame *
;;   - There don't seem to [have] be[en] any actual range checks nowhere.
;; Have Fun!
;;
;; * Inner Defines, Conditional Definitions and Redefinition of Define
;;
;;   Remove our definition of define completely. Else inner
;;   defines in namespaces will see our stale and voided definition
;;   and produce really nice and confusing error messages. (Maybe that
;;   was the right thing: just abandon inner defines?) The change
;;   has been tested with the revised /format/ routines which originally
;;   revealed the failure. Other tests confirmed the new functionality.
;;   This is specific to the prototype and its current coexistence with
;;   /m_expand_body/.
;;   It's presumed that because of the above change conditional
;;   definitions should still work as expected but redefinitions of
;;   /define/ or the use of its value, i.e. the syntax object, after the
;;   evaluation of the namespace, e.g. within closures in the namespace,
;;   may lead to unexpected behavior or failure or not depending on use
;;   and context. Just retaining the syntax object in its location in
;;   the environment frame wouldn't help as the actual symbol lookup is
;;   synchronized to the actual evaluation of individual code paths.
;;   Retaining the complete binding on the other hand would have bad
;;   effects on inner defines: they might appear top level within the
;;   name space without any further warning. Inner defines and redefinitions
;;   of /define/ within a namespace seem to be mutual exclusive with this
;;   prototype. An actual primary primitive implementing the /binder/ as
;;   well as /system environment/ definitions as well as custom environment
;;   bindings should not be affected by this shortcoming restricted to
;;   the prototype.

(define (namespace:canonicalize-define expression)
 (if (symbol? (second expression))
   expression
   (let ((symbol   (car (second expression)))
         (formals  (cdr (second expression)))
         (sequence (cddr expression)))
     `(define ,symbol ,(cons 'lambda (cons formals sequence))) )))

(define (namespace:resolve symbol frame)
 (let loop ((i 1) (symbols (cdar frame)))
  (if (null? symbols)
   #f
   (if (eq? (car symbols) symbol)
     (list-pair i (cdr frame))
     (loop (+ 1 i) (cdr symbols))) )))

(define (namespace-environment:transformer expression environment)
 (let* ((formals  (cons 'namespace:unhygienic-box-token '() ))
        (sequence (cdr expression))
        (values   (cons *unspecified* '() ))
        (frame    (cons formals values))
        (env      (cons frame environment))
        (last-value  (last-pair values))
        (last-formal (last-pair formals))
        (binder    (lambda (exp env)
                     (let* ((e (namespace:canonicalize-define exp))
                            (s (second e))
                            (v (third  e))
                            (p (namespace:resolve s frame)))
                      (or (and p (set-car! p (eval! v env)))
                       (begin
                        (set-cdr! last-formal (cons (second e) '()))
                        (set! last-formal (cdr last-formal))
                        ;; see also namespace visibility policy
                        (set-cdr! last-value (cons *unspecified* '()))
                        (set! last-value (cdr last-value))
                        (set-car! last-value (eval! (third e) env))
                        *unspecified*))) )))
   (set-car! formals 'define)
   (set-car! values (procedure->syntax binder))
   (for-each (lambda (e) (eval! e env)) sequence)
   (begin ; Unhygienic hygiene paragraph: like a bath in dust.
      ; Remove our definition of /define/ completely:
      ;  (we could even compute and install a cutting routine in an
      ;  unhygienic way here?)
     (set-car! formals 'vsi:ans:namespace-prototype-unygienic-void-symbol)
      ; Cut the syntax object or it leaks.
     (set-car! values *unspecified*) (set! binder *unspecified*)
   )
   env
))
;; The split into two syntax transformers allows to use the original
;; namespace environment as such by the means of the syntax transformer
;; /namespace-environment:transformer/.
;; This is needed by Viper Objects (see VOS/vo.scm) and reduces code duplication.
(define (namespace:transformer expression environment)
 (let ((frame (car (namespace-environment:transformer expression environment))))
   (cons 'namespace (map cons (cdr (car frame)) (cdr (cdr frame))) )))

(define (namespace? object)
  (and (pair? object)
       (symbol? (car object))
       (eq? (car object) 'namespace)))

;; With these checks you could go for a tc32 and use unchecked cdr-consing ;-)

(define (namespace->alist namespace)
 (when (not (namespace? namespace)) (error "domain error"))
 (cdr namespace))

(define (namespace:normalized-binding? b)
  (and (pair? b) (symbol? (car b))))
(define (alist->namespace alist)
 (let ((alength (length alist))
       (subnormals (filter (lambda (p) (not (namespace:normalized-binding? p)))
                            alist)))
  (if (and (> alength 0) (= 0 (length subnormals)))
    (cons 'namespace alist)
    (error "alist->namespace: mischiefed entries: " subnormals)) ))

;;
;; (namespace expressions)
;;
(define namespace (procedure->syntax namespace:transformer))

(define (call namespace name . arguments)
 (let ((proc (assq-ref (cdr namespace) name)))
   (when (not proc) (error (format "namespace call undefined: ~A" name)))
   (apply proc arguments)))
  

;; DEFUNCT experimental fast call adapting to argument changes
;; - arguments aren't evaluated
;; - it's too slow
;; - it isn't easy to implement
;;
;; What we would have liked to have here would have been:
;;
;; (fast-call namespace name argument-1 argument-2) =>
;; => (begin (when (not eq-namespace-and-name-last?)
;;             do-the-lookup-again-to-set-proc)
;;           (proc argument-1 argument-2))
;;
;; Perhaps someone would even try to XOR namespace and name to
;; reduce the number of comparisions but that wouldn't work in
;; every theory. ;-) In practice we need some place to store the
;; last signature, i.e. namespace and name, which results in a
;; surrounding closure and even as a closure without open arguments
;; this results in an additional application where /begin/ would
;; only have been a sequence ... we could somehow memoize the
;; data of the signature into
;;  (and (eq? name-to-be-evaluated memoized-variable-last-name) ...
;; but that is currently out of "my scope" and seems to be too
;; complicated.
;;
;; (define (fast-call:transformer expression environment)
;;  (let* ((memoized-namespace (second expression))
;;         (memoized-name      (third  expression))
;;         (arguments          (cdddr  expression))
;;         (memoized-proc      #f))
;;   (set-cdr! expression (append! (list memoized-namespace memoized-name)
;;                                 arguments))
;;   (set-car! expression
;;    (lambda (namespace name . arguments)
;;      (when (not (and (eq? memoized-namespace namespace)
;;                      (eq? memoized-name      name)))
;;        (set! memoized-proc (cdr (assq name (cdr namespace))))
;;        (set! memoized-namespace namespace)
;;        (set! memoized-name name))
;;      (unchecked-apply memoized-proc arguments)))
;;   expression))
;; (define fast-call (procedure->macro fast-call:transformer))

;; fixed call without further lookups
(define (fixed-call:transformer expression environment)
 (let* ((memoized-namespace (eval! (second expression) environment))
        (memoized-name      (eval! (third  expression) environment))
        (arguments          (cdddr  expression))
        (memoized-proc      (assq-ref (cdr memoized-namespace) memoized-name)))
  (when (not memoized-proc)
    (error (format "namespace fixed-call undefined: ~A in ~a" memoized-name expression)))
  (set-car! expression memoized-proc)
  (set-cdr! expression arguments)
  expression))
(define fixed-call (procedure->macro fixed-call:transformer))
    

(define (from-namespace name-space name)
 (let ((res (assq name (cdr name-space))))
  (if res
      (cdr res)
      (error (format "namespace from-namespace undefined: ~A" name))) ))

;;
;; (namespace-import name-space . possible future extensions)
;; ??? This imports only once as it memoizes the result. Is this
;;     what we want? Other transformers like /scm_m_define/ and
;;     /m_expand_body/ might do even more evil things
;;     with these expressions or they might not work at all.
;;
(define (namespace:import-transformer expression environment)
 (let ((name-space (eval! (second expression) environment))
       ; possible future extension (prefix (assq 'prefix (third expression)))
       (new-expression '()))
  (set! new-expression
        (map (lambda (binding)
               (let ((name  (car binding)) (value (cdr binding)))
                (cond ((pair? value)
                       `(define ,name (quote ,value)))
                      (else
                       `(define ,name ,value))) ))
          (cdr name-space)))
  (set-car! expression begin)
  (set-cdr! expression new-expression)
  expression))

(define namespace-import (procedure->macro namespace:import-transformer))

;; TODO cfr with the regular loader VSI-core/vsi.scm
(define (namespace:loader file-name env)
 (let ((expressions (cons 'box '())))
  ;; Read in all expressions.
  (with-input-from-file file-name (lambda ()
   (let loop ((expression (read)) (last expressions))
    (when (not (eof-object? expression))
     (set-cdr! last (cons expression '()))
     (loop (read) (cdr last)) )) ))

  ;; Create our namespace
  (set! expressions (cdr expressions))
  (if (null? expressions)
    #f
    (eval! (cons namespace expressions) env) )))

(define (namespace-load:transformer expression environment)
 (when (not (and (= 2 (length expression))))
   (error "mischiefed namespace-load statement" expression))
 (let ((path (eval! (second expression) environment)))
   (namespace:loader path environment)))

(define namespace-load (procedure->syntax namespace-load:transformer))
