;; Data Structure Scheme
;;  (Ever More) Data Types w/o Domain
;; 

(define (is-a? obj type)
 (and (vector? obj)
      (>= (vector-length obj) 1)
      (eq? (vector-ref obj 0) type)))

;;
;; Structures - Vectors with Named Slots
;;
;; NB This is part of the requested lock against the past:
;; /list-ref/, /vector-ref/, /brass-ref/ all work on indices
;; whereas /slot-ref/ traditionally should have been the
;; "low level" accessor of tagged arrays:
;; (slot-ref v 0) would than have returned the length of
;; a vector /v/ w/o any additional range checks?
;; (slot-ref p 1) would then have been the implementation
;; of /unchecked-cdr/? And (slot-ref o -2) would return the
;; bits of the type tag! LOL. See also SCM_CELL_WORD and
;; SCM_CELL_OBJECT as well as other accessors in VSI-C/.../trim.h
;; In practice it's better to point to the type tag.
;; In another theory there has been an "array processor" which
;; was able to load a complete register set at once. These six
;; registers were: instruction pointer, stack pointer,
;; computing register A, B, C and D. Thus CDR just referred
;; to some part of a machine instruction to use the contents
;; of that register then? As part of the machine instruction
;; list processing snippets? Or they turned it around and used
;; a register set pointer into the first class RAM ...
;;
;; NB There isn't any good reason to use a vector for name
;; resolutions if not for "coherency".
;;

(define (struct-layout? obj) (is-a? obj 'struct-layout))

(define (structure-layout labels)
 (let ((labels (delete-duplicates labels)))
  (list->vector (cons 'struct-layout labels))))

;; (define struct (macro (label . labels)
;;  `(structure-layout (quote (,label ,@labels)))))
(define struct (macro labels ; VO needs empty structures
  `(structure-layout (quote (,@labels)))))

(define (make-struct struct-layout)
 (let ((s (make-vector (vector-length struct-layout))))
  (vector-set! s 0 struct-layout)
  s))
(define allocate-struct make-struct)

(define (slot-idx struct label)
 (let ((struct-desc (vector-ref struct 0)))
  (let loop ((i (- (vector-length struct-desc) 1)))
   (cond ((eq? label (vector-ref struct-desc i))  i)
         ((> i 1)                                 (loop (- i 1)))
         (else (error (format "not a struct member: ~a in ~a"
                              label struct-desc))
               #f)) )))

(define (slot-ref struct slot)
  (vector-ref struct (slot-idx struct slot)))

(define (slot-set! struct slot value)
  (vector-set! struct (slot-idx struct slot) value))


;;; Memoizing Slot Accessors
;; Text Positional Slot Index Equality Assumption
;;   (use type checks before these macros)
;; Expression Duplication and Multiple Evaluation: struct
;;   (use these macros with symbols and variables only)

(define (at-slot-idx:transformer expression environment)
 (when (not (= 3 (length expression)))
   (error (format "@slot-idx: wrong number of arguments in ~a" expression)))
 (let* ((struct (second expression))
        (label  (third  expression))
        (res    (eval `(,slot-idx ,struct ,label) environment)))
   (set-car! expression quote)
   (set-cdr! expression (cons res '()))
   expression))
(define @slot-idx (procedure->macro at-slot-idx:transformer))

;; (define @slot-idx  (macro (struct label) `(slot-idx ,struct ,label) ))
(define @slot-ref  (macro (struct slot)
 `(,vector-ref  ,struct (@slot-idx ,struct ,slot)) ))
(define @slot-set! (macro (struct slot value)
 `(,vector-set! ,struct (@slot-idx ,struct ,slot) ,value) ))


(define (struct-from-values struct . values)
  (and (= (length values) (- (vector-length struct) 1))
       (list->vector (cons struct values))))

(define (values-from-struct struct) (cdr (vector->list struct)))


;;
;; DList - Doubly Linked List
;; (this looks like the "byte code" of some class LOL)
;;
(define dlist       (struct head current tail))
(define dlist-element (struct next value prev))  ; private to DList
(define dnull       '())
(define dnull?      null?)

(define (make-dlist) (struct-from-values dlist dnull dnull dnull))
(define (dlist? obj) (is-a? obj dlist))

(define (dlist-destroy! dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-destroy! wrong type argument: ~a" dlist)))
 (let loop ((el (@slot-ref dlist 'head)))
  (when (not (dnull? el))
   (@slot-set! el 'value dnull)
   (@slot-set! el 'prev  dnull)
   (loop (@slot-ref el 'next))))
 (@slot-set! dlist 'head    dnull)
 (@slot-set! dlist 'current dnull)
 (@slot-set! dlist 'tail    dnull)
 *unspecified*)

(define (dlist-start? dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-start? wrong type argument: ~a" dlist)))
 (dlist-start/unchecked? dlist))
(define (dlist-start/unchecked? dlist)
  (eq? (@slot-ref dlist 'current) (@slot-ref dlist 'head)))

(define (dlist-end? dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-end? wrong type argument: ~a" dlist)))
 (dlist-end/unchecked? dlist))
(define (dlist-end/unchecked? dlist)
  (eq? (@slot-ref dlist 'current) (@slot-ref dlist 'tail)))

(define (dlist-head dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-head wrong type argument: ~a" dlist)))
 (dlist-head/unchecked dlist))
(define (dlist-head/unchecked dlist)
 (let ((el (@slot-ref dlist 'head)))
  (cond ((dnull? el)  dnull)
        (else         (@slot-set! dlist 'current el)
                      (@slot-ref el 'value)) )))

(define (dlist-tail dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-tail wrong type argument: ~a" dlist)))
 (dlist-tail/unchecked dlist))
(define (dlist-tail/unchecked dlist)
 (let ((el (@slot-ref dlist 'tail)))
  (cond ((dnull? el)  dnull)
        (else         (@slot-set! dlist 'current el)
                      (@slot-ref el 'value)) )))

(define (dlist-next dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-next wrong type argument: ~a" dlist)))
 (dlist-next/unchecked dlist))
(define (dlist-next/unchecked dlist)
 (let ((el (@slot-ref dlist 'current)))
  (when (not (dnull? el)) (set! el (@slot-ref el 'next)))
  (cond ((dnull? el)  dnull)
        (else         (@slot-set! dlist 'current el)
                      (@slot-ref el 'value)) )))

(define (dlist-prev dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-prev wrong type argument: ~a" dlist)))
 (dlist-prev/unchecked dlist))
(define (dlist-prev/unchecked dlist)
 (let ((el (@slot-ref dlist 'current)))
  (when (not (dnull? el)) (set! el (@slot-ref el 'prev)))
  (cond ((dnull? el)  dnull)
        (else         (@slot-set! dlist 'current el)
                      (@slot-ref el 'value)) )))

(define (dlist-ref dlist n)
 (when (not (dlist? dlist))
   (error (format "dlist-ref wrong type argument: ~a" dlist)))
 (dlist-ref/unchecked dlist n))
(define (dlist-ref/unchecked dlist n)
 (let loop ((i 0) (el (@slot-ref dlist 'head)))
  (cond ((dnull? el)  #f) ;; out of range but length might become a struct member
        ((= i n)     (@slot-set! dlist 'current el) (@slot-ref el 'value))
        (else        (loop (+ i 1) (@slot-ref el 'next)) )) ))

(define (dlist-add! dlist value)
 (when (not (dlist? dlist))
   (error (format "dlist-add! wrong type argument: ~a" dlist)))
 (dlist-add/unchecked! dlist value))
(define (dlist-add/unchecked! dlist value)
 (let ((cur (@slot-ref dlist 'current))
       (el  (struct-from-values dlist-element dnull value dnull)))
  (cond
   ((dnull? cur) (@slot-set! dlist 'current el)
                 (@slot-set! dlist 'head    el)
                 (@slot-set! dlist 'tail    el)
                 #t)
   (else  (let ((next (@slot-ref cur 'next)))
                 (@slot-set! el    'next    next)
                 (@slot-set! el    'prev    cur)
                 (@slot-set! cur   'next    el)
                 (if (dnull? next)
                   (@slot-set! dlist 'tail el)
                   (@slot-set! next  'prev el))
                 (@slot-set! dlist 'current el)
                 #t)) )))

(define (dlist-insert! dlist value)
 (when (not (dlist? dlist))
   (error (format "dlist-insert! wrong type argument: ~a" dlist)))
 (dlist-insert/unchecked! dlist value))

(define (dlist-insert/unchecked! dlist value)
 (let ((cur (@slot-ref dlist 'current))
       (el  (struct-from-values dlist-element dnull value dnull)))
  (cond
   ((dnull? cur) (@slot-set! dlist 'current el)
                 (@slot-set! dlist 'head    el)
                 (@slot-set! dlist 'tail    el)
                 #t)
   (else  (let ((prev (@slot-ref cur 'prev)))
                 (@slot-set! el    'next    cur)
                 (@slot-set! el    'prev    prev)
                 (@slot-set! cur   'prev    el)
                 (if (dnull? prev)
                   (@slot-set! dlist 'head el)
                   (@slot-set! prev  'next el))
                 (@slot-set! dlist 'current el)
                 #t)) )))

(define (dlist-remove! dlist)
 (when (not (dlist? dlist))
   (error (format "dlist-remove! wrong type argument: ~a" dlist)))
 (dlist-remove/unchecked! dlist))

(define (dlist-remove/unchecked! dlist)
 (let ((cur (@slot-ref dlist 'current)))
  (and (not (dnull? cur))
       (let ((prev (@slot-ref cur 'prev)) (next (@slot-ref cur 'next)))
        (@slot-set! cur 'next dnull)
        (@slot-set! cur 'prev dnull)
        (cond
         ((and (dnull? prev)
               (dnull? next)) (@slot-set! dlist 'current dnull)
                              (@slot-set! dlist 'head    dnull)
                              (@slot-set! dlist 'tail    dnull)
                              #t)
         ((dnull? prev)       (@slot-set! dlist 'current next)
                              (@slot-set! dlist 'head    next)
                              (@slot-set! next  'prev    dnull)
                              #t)
         ((dnull? next)       (@slot-set! dlist 'current prev)
                              (@slot-set! dlist 'tail    prev)
                              (@slot-set! prev  'next    dnull)
                              #t)
         (else                (@slot-set! dlist 'current next)
                              (@slot-set! prev  'next    next)
                              (@slot-set! next  'prev    prev)
                              #t)
)) )))

(define (values-from-dlist dlist)
 (when (not (dlist? dlist))
   (error (format "values-from-dlist wrong type argument: ~a" dlist)))
 (let ((ans '()))
  (let loop ((el (@slot-ref dlist 'tail)))
   (cond ((dnull? el)   ans)
         (else          (set! ans (cons (@slot-ref el 'value) ans))
                        (loop (@slot-ref el 'prev)))))))

(define (dlist-from-values values)
 (let ((ans (make-dlist)))
  (for-each (lambda (el) (dlist-add/unchecked! ans el)) values)
  (@slot-set! ans 'current (@slot-ref ans 'head))
  ans))

;; Equality
;;
;; While equality is well defined for /structures/ in general it's
;; practically turned off by circular data structures like double
;; linked lists. In addition the transparent handle holds volatile
;; state in the form of the cursor known as the /current/ element
;; which would have lead to false results anyway. Therefore we
;; define our own equality predicate which is itself not able to
;; cope with circularity apart from the chain of rings of the double
;; linked list. That means if a D-list has an element holding a
;; reference to the same D-list our equality predicate will fail
;; just like equal failed. ??? The mark bit is still free: maybe
;; it means: circular redundant link? Or it means already seen?
;; A solution to the problem of determing equality of circular
;; data structures would be nice because in practice the actual
;; data graph of an application might really contain vast rings
;; for practical means?
;;
;; /Eqv?/ should be defined, too. But seems like it doesn't work
;; for vectors as expected anyway. (flat list compare)

;;; XXX the following is crap! plz open a maintenance issue on equality
(define (dlist-equal:case? obj1 obj2)
 (cond ((and (dlist? obj1)
             (dlist? obj2)) (dlist-equal? obj1 obj2))
       ((dlist? obj1)       #f)
       ((dlist? obj2)       #f)
       (else                (equal? obj1 obj2)) ))

(define (dlist-equal? dlist1 dlist2)
 (when (not (dlist? dlist1))
   (error (format "dlist-equal? wrong type argument 1: ~a" dlist1)))
 (when (not (dlist? dlist2))
   (error (format "dlist-equal? wrong type argument 2: ~a" dlist1)))
 (or (eq?  dlist1 dlist2)
     (eqv? dlist1 dlist2) ; two different handles on the same list?
     (let loop ((el1 (@slot-ref dlist1 'head)) (el2 (@slot-ref dlist2 'head)))
      (cond ((and (dnull? el1)
                  (dnull? el2)) #t)
            ((or  (dnull? el1)
                  (dnull? el2)) #f)
            (else (and (dlist-equal:case? (@slot-ref el1 'value)
                                          (@slot-ref el2 'value))
                       (loop (@slot-ref el1 'next)
                             (@slot-ref el2 'next)) )) )) ))
