;;
;; "Robust" Vector Calculus Implementation
;;
;; These expressions strive to be a robust reference implementation of
;; vector calculus with Scheme vectors containing cartesian
;; coordinates. Reference implementation means the chosen forms should
;; reflect mathematical formulas with little translation to list
;; processing and explicit formulation of "implicit maths" where
;; needed.
;;
;; Operations on vectors of different lengths are supported where
;; possible. The basic definition for this says operations will only
;; be computed on the pair wise minimum of the length of the operands
;; (i.e. the number of dimensions of a vector).
;;
;; Where it tries to be mathematically correct it also is intended to
;; be useful in practical or system programming tasks otherwise
;; unrelated to vector calculus.
;;

(define (numerical-vector? obj) (and (vector? obj) (all? (map scalar? (vector->list obj)) )))

(if (not (and (defined? 'symbolic-vector-calculus)
              symbolic-vector-calculus))
  (begin ; numerical operation
    (define %vc-plus     +)
    (define %vc-minus    -)
    (define %vc-multiply *)
    (define %vc-divide   /)
    (define %vc-square-root sqrt)
    (define %vc-acos  acos)
    (define %vc-pi     pi))
  (begin ; symbolic operations (debugging formulas)
    (define (%vc-plus   .   operands) (cons '+ operands)) 
    (define (%vc-minus   .  operands) (cons '- operands))
    (define (%vc-multiply . operands) (cons '* operands))
    (define (%vc-divide  .  operands) (cons '/ operands))
    (define (%vc-square-root operand) (list 'sqrt operand))
    (define (%vc-acos        operand) (list 'acos operand))
    (define %vc-pi                                    'pi))
)

;;
;; Vector Operation Abstraction
(define (%vc-vector-operator operation A B)
 (let ((dimensions-A (vector-length A))
       (dimensions-B (vector-length B)))
  (let (( C (vector-copy (if (> dimensions-A dimensions-B) A B)) ))
   (let parallel-evaluation-serialized
       ;; compensate for zero based indexing of vectors
       ((n (- (min (vector-length A) (vector-length B)) 1)))
     (vector-set! C n (operation (vector-ref A n)
                                 (vector-ref B n)))
     (cond
      ((= n 0) C)
      ((< n 0) (error "out of range error")) ; should not happen
      (else
       (parallel-evaluation-serialized (- n 1))) ))) ))

;;
;; Vector |+| Vector -> Vector

(define (%vc:pair-wise-fold-plus A B) (%vc-vector-operator %vc-plus A B))
(define (vector-plus  A B . rest-operands)
 (fold %vc:pair-wise-fold-plus        ; operation
       (%vc-vector-operator %vc-plus A B) ; initial value
       rest-operands))

(define (%vc:pair-wise-fold-minus A B) (%vc-vector-operator %vc-minus A B))
(define (vector-minus A B . rest-operands)
 (fold %vc:pair-wise-fold-minus        ; operation
       (%vc-vector-operator %vc-minus A B) ; initial value
       rest-operands))

;;
;; Vector |+| Vector -> Scalar
(define (vector-dot A B)
  (let (( C (%vc-vector-operator %vc-multiply A B)))
    (fold %vc-plus 0 (vector->list C))))

;;
;; Vector -> Scalar
(define (vector-magnitude A)
  (%vc-square-root (vector-dot A A)))

;;
;; Scalar |+| Vector -> Vector
(define (vector-scale s A)
  (%vc-vector-operator %vc-multiply (make-vector (vector-length A) s) A))


;;
;; Example Session Transcript: Numerical Mode
;;
;; VSI> (load "vector-calculus.scm")
;; VSI> (vector-dot #(1 2 3) #(3 2 1))    => 10
;; VSI> (vector-cross #(1 1 1) #(1 1 1))  => #(0 0 0)
;; ...
;; VSI> (define v1 #(1 2 3))
;; VSI> (vector-plus v1 v1 v1)            => #(3 6 9)
;; VSI> (vector-minus v1 v1 v1)           => #(-1 -2 -3)
;; VSI> (vector-cross #(1 0 0) #(0 1 0))  => #(0 0 1)
;; VSI> (vector-scale (- (expt 2 258) 1) v1)  => ...
;; ...
;; VSI> (define v1 #(1 2))
;; VSI> (define v2 #(1 2 3))
;; VSI> (vector-plus v1 v2)               => #(2 4 3)
;; VSI> (vector-plus v1 v2 v1 v2)         => #(4 8 6)
;; VSI> (vector-minus v1 v2) #(0 0 3)
;; VSI> (vector-minus v1 v2 v1 v2)        => #(-2 -4 0)
;; VSI> (vector-cross v1 v2)              = undefined 
;;
;;
;; Example Session Transcript: Symbolic Mode
;;
;; VSI> (define A '#(a1 a2 a3))     
;; VSI> (define B '#(b1 b2 b3))
;; VSI> (define symbolic-vector-calculus #t)
;; VSI> (load "vector-calculus.scm")
;; VSI> (vector-cross A B)
;; => #((- (* a2 b3) (* a3 b2)) (- (* a3 b1) (* a1 b3)) (- (* a1 b2) (* a2 b1)))
;; VSI> (vector-dot A B)
;; => (+ (* a3 b3) (+ (* a2 b2) (+ (* a1 b1) 0)))
;; VSI> (vector-plus A B)
;; => #((+ a1 b1) (+ a2 b2) (+ a3 b3))
;; VSI> (vector-minus A B)
;; => #((- a1 b1) (- a2 b2) (- a3 b3))
;; VSI> (vector-scale 's A)
;; => #((* s a1) (* s a2) (* s a3))
