;;
;; VSI Algorithm Notation Scheme Core
;;
(define vsi (version))
(define VSI vsi)

(define (eval expression environment)
 (eval! (copy-tree expression) environment))

(define (when:transformer expression environment)
 (if (< (length expression) 2) (error "mischiefed when" expression))
 (let* ((condition (cadr expression))
        (sequence  (cddr expression))
        (new-exp   (list cond (cons condition sequence))))
   (set-car! expression (car new-exp))
   (set-cdr! expression (cdr new-exp))
   expression))
(define when (procedure->macro when:transformer))

(define (repeat n closure)
  (if (not (and (exact? n)
                (integer? n)
                (>= n 0)))
      (error "repeat: the parameter n must be an exact natural number or zero.")
      (let loop ((i 0))
        (when (< i n)
          (closure)
          (loop (1+ i)))) ))

(define unchecked-apply apply)
(define (apply proc . args)
  (@apply proc (apply:nconc2last args)))

(define (quit . args) (apply throw 'quit args))
(define exit quit)

;; let format alias simple-format until the more complete version is loaded
(define format simple-format)

;; The identity function.
(define (identity idem) idem)

;; Macros by Template
;; (macro constituents template) ; canonical form
;;
;; TODO Documentation! As we're deviating from the tradition and
;; opting for working syntax objects which allow for the canonical
;; form we really should document this. Quasiquotation (iqq) is
;; still a cludge:
;;
;; Accent, commas and ats (@) for list processing instructions?
;;
;; ' = quote
;; ` = quasiquote
;;   ,  = unquote
;;   ,@ = unquote-splicing (traditional name)
;; While the last two are only valid within a quasiquote expression
;; their abbreviations (, and @,) are recognized by the reader w/o
;; respect to context. The use outside of a quasiquote expression
;; leads to unspecified behavior.
;;
;;   (define a (list 1 2))
;;   '(a a)     => (a a)
;;   `(a a)     => (a a)
;;   `(,a a)    => ((1 2) a)
;;   `(,a ,a)   => ((1 2) (1 2))
;;   `(,a ,@a)  => ((1 2) 1 2)
;;   `(,@a ,@a) => (1 2 1 2)
;;
;; All of this doesn't really fit the SCHEME and it violates its
;; regular way of reading in symbols: ,@a should be one symbol
;; but it isn't? If you want to know more about where this oddity
;; comes from you can try to find out about /read tables/ in
;; LISP which seem to have allowed for syntactical transformations
;; of character sequences (where Scheme prefers transformations on
;; the (tokenized) list processing level.
;; --
;;
;; Example: (define ifnonly-if (macro (a b) `(if ,a ,b)))
;;
;; The following expression: (ifnonly-if (predicate?) (display test))
;; would then lead to a call of the macro where its arguments
;; are bound to the constituents of the given expression:
;;
;;   /a/ becomes (predicate?)   which equals (list 'predicate?)
;;   /b/ becomes (display test) which equals (list 'display 'test)
;; (This is syntax transformation of list processors: only the
;; evaluation would lead to actual results of the expressions.)
;;
;; The result of the actual macro (which should be the
;; replacement expression as a list or a value) is then /memoized/
;; which means every occurrence of /ifnonlyif/ is replaced
;; with the template once. Future repetitions of the same
;; occurrence simply evaluate the replacement as defined by the
;; template, i.e. the result of the macro.
;;
(define (macro:transformer expression environment)
 (when (< (length expression) 3) (error "mischiefed macro" expression))
 (let (;(formals     (cadr expression))
       (applicable   (eval (cons lambda (cdr expression)) environment)))
  (procedure->macro
    (lambda (expression environment)
      ; XXX dotted rest list arguments n stuff
      ;(when (not (= (length expression) (+ num-formals 1)))
      ;  (error "mischiefed number of arguments" expression))
      (let* ((arguments   (cdr expression))
             (replacement (list begin (apply applicable arguments))))
       ;; Memoize the replacement.
       (set-car! expression (car replacement))
       (set-cdr! expression (cdr replacement))
       replacement)) )))

(define macro (procedure->syntax macro:transformer))


;;
;; Special Values of Implementation (sharp symbols & other oddities)
;;
(define *unspecified* (cdr sys-boxed:unspecified))
(define *end-of-file* (cdr sys-boxed:eof))
(define sys-undefine! (macro (symbol)
                        `(set! ,symbol ,(cdr sys-boxed:undefined)) ))

; ANS Definition: Top Level Environment
;
; The tradtional top level environment sometimes refers to the
; system environment and sometimes it refers to some relative top
; level environment like a namespace or any other scope in ANS.
;
;(define system-environment '())

;; Former Synthetizations
;(define *unspecified* (if #f #t))
;(define the-eof-object (call-with-input-string "" (lambda (p) (read-char p))))
;(define the-root-environment (the-environment))


;;
;; Revised VSI Nature load prototype
;; (load file-name)
(define (vsi-nature:loader file-name env)
 (when sys-debug-interpreter
   (display "vsi-nature-load: " (current-error-port))
   (display file-name           (current-error-port))
   (newline))
 (let ((expressions (cons 'box '())))
  ;; Read in all expressions.
  (with-input-from-file file-name (lambda ()
   (let loop ((expression (read)) (last expressions))
    (when (not (eof-object? expression))
     (set-cdr! last (cons expression '()))
     (loop (read) (cdr last)) )) ))
  (when (not (null? expressions))
    (eval! (cons begin expressions) env)) ))

(define (load:transformer expression environment)
 (when (not (and (= 2 (length expression))))
   (error "mischiefed load statement:" expression))
 (let ((path (eval (second expression) environment)))
   (when (not (and (string? path) (> (string-length path) 0)))
     (error "mischiefed path in load statement:" expression))
   (vsi-nature:loader path environment)))

(define (sys-load:transformer expression environment)
 (when (not (and (= 2 (length expression))))
   (error "mischiefed sys-load statement:" expression))
 (let ((path     (eval (second expression) environment))
       (sys-path #f))
   ;; Error Reporting is one fine XXX

   ; The loader could do this in one spot but it can't provide
   ; the failed expression to the error reporting routine. Therefore
   ; we repeat this test in every variant of /load/?
   (when (not (and (string? path) (> (string-length path) 0)))
     (error "mischiefed path in sys-load statement:" expression))

   ; The result of /%search-load-path/ may be false and if we pass that
   ; on we can see a really nice error message: "Wrong Type (expecting string)"
   ; where we would expect: "File not found".
   (set! sys-path (%search-load-path path))
   (when (not sys-path)
     ; key? misc-error? /open-file/ uses /system-error/
     (error "sys-load: File not found: " path))
   (vsi-nature:loader sys-path environment)))


;(define load vsi-nature-load)
;; (define (sys-load file-name)
;;   (vsi-nature-load (%search-load-path file-name)))
(define load     (procedure->syntax load:transformer))
(define sys-load (procedure->syntax sys-load:transformer))

;(define load primitive-load)
;(define sys-load primitive-load-path)


;;
;; Custom Environments DEFUNCT
;;
;; (define (null-environment . compatibility-argument) ; XXX DEFUNCT
;;   (list #f))

;
; ANS Definition: Top Level Environment
;
; The tradtional top level environment sometimes refers to the
; system environment and sometimes it refers to some relative top
; level environment like a namespace or any other scope in ANS.
; Sometimes it even refers to the text aligned to the left most
; border in a file or the outermost expressions of a file: any
; file can be loaded into a namespace (and in the future also 
; into any scope if things do work out).
;

(define the-environment  ; useful with REPL-HERE (goes there?)
  (procedure->syntax
   (lambda (x e)
     e)))


;;
;; Multiple Return Values
;;
(define values list)
(define (call-with-values producer consumer)
 (apply consumer (producer)))
(define receive (macro (values variables . expressions)
  `(call-with-values (lambda () ,values)
     (lambda ,variables ,@expressions))))


;;; {Error Handling} {Trivial Function}
;;;
(define false-if-exception (macro (expr)
  `(catch #t (lambda () ,expr)
	  (lambda args #f))))

(define (error . args)
  (if (null? args)
      (scm-error 'misc-error #f "?" #f #f)
      (let loop ((msg "~A")
		 (rest (cdr args)))
	(if (not (null? rest))
	    (loop (string-append msg " ~S")
		  (cdr rest))
	    (scm-error 'misc-error #f msg args #f)))))


;;; {Load Paths}
;;;

;;; Here for backward compatability
;;
(define scheme-file-suffix (lambda () ".scm"))

(define (in-vicinity vicinity file)
  (let ((tail (let ((len (string-length vicinity)))
		(if (zero? len)
		    #f
		    (string-ref vicinity (- len 1))))))
    (string-append vicinity
		   (if (or (not tail)
			   (eq? tail #\/))
		       ""
		       "/")
		   file)))



(define command-line program-arguments)
     

;;;
;;;
;;; Experimental Forms and Future Extensions
;;;
;;;

;;
;; (once expressions) - EXPERIMENTAL Evaluation Instruction 
;;
;; Evaluates the given expressions once and replaces it with
;; the result (i.e. it memoizes the result). This isn't an
;; evaluation hint it is an evaluation instructions which
;; does what it does even if the resulting behavior of the
;; expressions is unexpected, unspecified or unwanted.
;;
;; This is UNSAFE! Use with care!
;;
(define (memoize-result:transformer expression environment)
  ;; To prevent eval from treating the result as code we need
  ;; to /quote/ the result. [with /begin/]
  (set-car! expression begin)
  (set-cdr! expression
   (list (eval (cons begin (cdr expression)) environment) ))
  expression)
(define once (procedure->macro memoize-result:transformer))


;;
;; CLR4 cutting named let PROTOTYPE
(define pre-clr4-let let)
(define (named-let:seperator formals-initialization formals initializations)
  (cond
   ((null? formals-initialization)
      (list formals initializations))

   ((and (pair? (car formals-initialization))
         (symbol? (caar formals-initialization)))
      (set! formals (cons (caar formals-initialization) formals))
      (if (not (pair? (cdar formals-initialization)))
        (error 'mischiefed-named-let-inits formals-initialization)
        (set! initializations (cons (cadar formals-initialization)
                                    initializations)))
      (named-let:seperator (cdr formals-initialization)
                           formals initializations))
   (else
    (error 'mischiefed-inits-2))))

(define (clr4:named-let expression environment)
;  (newline)(display 'named-let:) (display expression) (newline)
;  (display " => ")
  (if (not (and (pair? expression)
                (pair? (cdr expression))
                (symbol? (cadr expression))))
    (begin (set-car! expression pre-clr4-let)
;           (display expression) (newline)
           expression)
    (begin
     ;; Do some more checks. ??? More explicit, please.
     (let* ((name (cadr expression))
            (formals-initialization (caddr expression))
            (body (cdddr expression))
            (formals '())
            (new-expressions '())
            (initializations '())
            (seperation (named-let:seperator formals-initialization '() '())))

      (set! formals         (reverse (car seperation)))
      (set! initializations (reverse (cadr seperation)))
      (set! new-expressions
       `(letrec ((,name (quote undefined))) ; XXX prototype undefined
         (set! ,name ,(append (list lambda formals) body))
         (set! ,name ,(cons name initializations))
         ,name
         ))
;           (display new-expression) (newline)
      ;; NATURE memoizing macro resolution
      (set-car! expression (car new-expressions))
      (set-cdr! expression (cdr new-expressions))
      ;new-expressions
      expression
))))

;; NATURE memoizing macro resolution: self memoizing macro instead
;; of /memoizing-macro/.
(define clr4-cutting-named-let (procedure->macro clr4:named-let))
(define let clr4-cutting-named-let)


;; EXPERIMENTAL
;;
;; Cut! variable . value
;;
;; Fundamental Cutting Tool
;; RC Cut Bang: cut the last reference w/o keeping a reference
;;  XXX syntax checks once and then memoize the unchecked variant
;;
;; Void the /variable/ and return /value/ if specified else unspecified.

(define (cutnpass:transformer expression environment)
 (let* ((location (cadr expression))
        (value    (if (null? (cddr expression))
                    *unspecified*
                    (eval (caddr expression) environment))))
   ;(display "Check check ") ;; DON'T MEMOIZE: This needs to run every time!
   `(begin
      (set! ,location *unspecified*)
      (quote ,value))))
(define unchecked-cut! (procedure->macro cutnpass:transformer))

(define (cutnpass:checker expression environment)
 (if (and (pair? expression)
          (pair? (cdr expression))
          (or (null? (cddr expression))
              (and (pair? (cddr expression))
                   (null? (cdddr expression)) )))
     (set-car! expression unchecked-cut!)
     (error "mischiefed cut"))
 expression)
(define cut! (procedure->macro cutnpass:checker))
