;;;
;;; Text Line
;;;

(sys-load "VOS/vo.scm")
(sys-load "Viper/vchars.scm")

(define text-line-class (class (extension viper-object-class)
  (slots display-sequence
         atom-sequence)
  ; NB The /display length/ may depend on the /display length/ of the
  ; atom sequence before a given position. The actual /display sequence/
  ; is considered volatile and it is often recreated completely. We
  ; do this per line during editing to liften the burden of the actual
  ; display routine which needs to pump some thousand characters to the
  ; terminal.

(method initialize (characters)
 (cond
 ((list? characters)
  (this (set! 'atom-sequence    (list->brass       characters)))
  (this (set! 'display-sequence (list->display-seq characters))))
 ((brass? characters)
  (this (set! 'atom-sequence    characters))
  (this (set! 'display-sequence (atom-seq->display-seq characters))))
 (else
  (error "text-line: initialize wrong type argument")))

 ;; A text line object has exactly one new line control as its last atom but
 ;; we currently don't check for additional controls because of laxity.
 (let ((as (this atom-sequence)))
  (when (not (eq? #\newline (brass-ref as (- (brass-length as) 1))))
   (error "text-line: initialize protocol violation (missing newline)")))
)

(method destroy () #t)

(method copy () (make-object text-line-class (brass-copy (this atom-sequence))))

;;  All positional arguments refer to the actual display length.
(method display-seq (from up-to) (subrass (this display-sequence) from up-to))
(method select (from up-to)      (subrass (this atom-sequence) from up-to))

(method append-atoms (atoms)
 (let ((as (this atom-sequence)))
  (this (set! 'atom-sequence
    (brass-append (subrass as 0 (- (brass-length as) 1))
                  atoms)))
  (this (set! 'display-sequence (atom-seq->display-seq (this atom-sequence))))
))

(method append (text-line) (this (append-atoms (text-line atom-sequence))))

(method display-length ( . rest)
 (cond
  ;; Display Length of the complete text line
  ((null? rest) (brass-length (this display-sequence)))

  ;; Display Length of a character at a position.
  (else (let ((idx (car rest)))
   (when (not (and (= (length rest) 1) (number? idx) (integer? idx) 
                      (<= 0 idx (- (brass-length (this display-sequence)) 1))))
     (error (format "text-line: display-length: argument out of bounds: ~A" rest)))
   (let ((as (this atom-sequence)))
    (display-length (brass-ref as (apos-from-dpos idx as))) ))) ))

(method left-length (dpos)
 (let ((as (this atom-sequence)))
  (display-length (brass-ref as (- (apos-from-dpos dpos as) 1)))))

(method adjust-dpos (dpos) (adjust-dpos dpos (this atom-sequence)))

(define (apos-from-dpos dpos atoms)
 (let loop ((idx 0) (dpos dpos))
  (cond ((= dpos 0) idx)
        ((> dpos 0) (loop (+ idx 1)
                          (- dpos (display-length (brass-ref atoms idx)))))
        (else       (error "text-line: apos-from-dpos: index mismatch")))))

;; Adjust the given display sequence index to the nearest atom boundary.
;; Nearest: drag to left  but the last approach created some game like diagonal
;;          movement.
(define (adjust-dpos dpos atoms)
 (let ((alen (brass-length atoms)))
  (let loop ((idx 0) (dpos dpos) (last 0))
    (cond ((<= dpos 0) last)
          ((> dpos 0)  (let ((dl (if (< idx alen) (display-length (brass-ref atoms idx))
                                                  1))) ; free space step
                         ;; TODO make this branch be idx < alen
                         ;; copy and simplify for the free space step
                         (loop (+ idx 1)
                               (- dpos dl)
                               (+ last dl)
                               ))) )) ))

(method remove (pos)
 (let* ((as  (this atom-sequence))
        (pos (apos-from-dpos pos as)))

   (when (not (< pos (- (brass-length as) 1)))
    (error "text-line: remove out of bounds"))

   (this (set! 'atom-sequence (brass-remove pos as)))
   (this (set! 'display-sequence (atom-seq->display-seq (this atom-sequence))))
))

(method split  (pos)
 (let* ((as  (this atom-sequence))
        (pos (apos-from-dpos pos as))
        (spl (brass-split pos as)))

  (when (not (< pos (- (brass-length as) 1)))
    ;; We don't allow splits after the final newline control.
    (error "text-line: split out of bounds"))

  (this (set! 'atom-sequence (brass-append (first spl) "\n")))
  (this (set! 'display-sequence (atom-seq->display-seq (this atom-sequence))))

  (values this (make-object text-line-class (second spl)))
))


(method insert (pos char)
 (let* ((as  (this atom-sequence))
        (pos (apos-from-dpos pos as)))

  (when (not (< pos (brass-length as)))
    (error "text-line: insert out of bounds"))

  (this (set! 'atom-sequence (brass-insert pos as char)))
  (this (set! 'display-sequence (atom-seq->display-seq (this atom-sequence))))
))

;)) ; close preliminary text-line class

;; Text Line Auxiliaries

(define (atom-seq->display-seq as) (list->display-seq (brass->list as)))
(define (list->display-seq lst)
 (let* ((ans  (cons #f '()))
        (last ans))
  (let loop ((rest lst)
             (sbuf '()))
   (cond
   ((null? rest) (when (not (null? sbuf))
                   (set-cdr! last (reverse! sbuf)))
                 (list->brass (cdr ans)))
                 ;(make-object text-line-class (cdr ans)))
   (else (let* ((char (car rest))
                (dl   (display-length char)))
           (cond
           ((= dl 1) (loop (cdr rest) (cons char sbuf)))

           (else (let ((df (display-form char)))
                     (set-cdr! last (append! (reverse! sbuf) df))
                     (set! last (last-pair last))
                     (loop (cdr rest) '()) )
))) ))) ))

)) ; closes text-line class

(define (list->text-line chars) (make-object text-line-class chars))

(when #f
(sys-load "VSI-core/repl.scm")
(repl-transcript

(define str "ABC def\t\ttest")
;(list->display-seq (string->list str))
(define t (make-object text-line-class (brass->list str)))
(t atom-sequence)
(t display-sequence)

)
(REPL)
)

