;;
;; Text Line Atoms
;;
;;  Every character has an associated /display form/ and /display length/.
;; All that double layer fuzz because of some outdated controls from back
;; then and now: fixed width isn't even fixed anymore, height might change
;; as well dynamically and even the actual display form is dynamic?
;;
;; C0 alone seems to be too restrictive.
;; Black lists and white lists aren't readily available.
;; Just letting it flow doesn't coincide well with the task of the Viper
;; Editor on the terminal. 
;;
;; DEPRECATION
;;  The current character name tables / lookup vectors will be replaced
;; and extended by the /bars/.
;;
;; The endless issue of UC broken concepts on the terminal: 
;;  - ALT-J is treated like a character but seems to combine with other
;;    characters on the same terminal line. 
;;  - Double Width characters tend to overlap on the terminal when treated
;;    as characters 
;; -> We might opt for a much more rigid approach of putting on the
;;    terminal what is known only. Everything else is then a senseless
;;    hexadecimal code.
;;
;; Problematic scripts in examples.desktop when it comes to the terminal:
;; Overflow for: ja, ko, zh_CN, zh_HK, zh_TW 
;;               These look like double width characters which actually
;;               use two terminal cells. We can adapt for that in the
;;               display length vector. The wingdings double width
;;               characters only use one terminal cell but overlap
;;               depending on cursor position.
;; See also the screenshot viper-vtb-screen-overflow.png where coloring
;; exhibits unattended overflows. For the regular buffer size this results
;; in some unattended scrolling action of the terminal where text flows into
;; the command line but model and view remain in sync:  editing operations
;; occur where the cursor is. [Except for the case where the overflow happens
;; for the last screen line.] 
;; 
;; Some scripts change with cursor position which might be part of their
;; input methods (km, kl, mr, ml, my, ne, bn, ta). This might or not
;; produce display errors but currently doesn't. 
;;
;; Change in terminal cell height: km
;;  This might interfere with the box drawing. 
;;
;; TODO Review the description of the unicode database fields if they
;; contain information about the double width and combining issues.
;; 
;; Current policy is to let it flow.
;;
;; Futures: 
;; The Viper Editor will be an editor for latin and compatible scripts
;; with their input methods based on the Visual Display Editor. 
;; 

(define CHARS (namespace

;; Auxiliary I
;; Unicode C0 + Traditional Character Code and Name Mapping
;; TODO cfr with core/characters
;; NB There is a non-C0 replacement character chunk near the
;;    box drawing code points.
(define c0-character-name-vector
(let ((vec                            ; TTY CTRL key combinations
#( ("NUL" "nul" "null")               ; space - 32
   ("SOH" "soh")                      ;     A - 64
   ("STX" "stx") ("ETX" "etx")
   ("EOT" "eot") ("ENQ" "enq")
   ("ACK" "ack") ("BEL" "bel" "bell") ;     G - 64

   ; 8                                ;     H - 64
   ("BS" "bs" "backspace")
   ("HT" "ht" "tab")
   ("LF" "lf" "newline" "nl")         ;     J - 64
   ("VT" "vt")
   ("FF" "np" "page")
   ("CR" "cr" "return")               ;     M - 64
   ("SO" "so")
   ("SI" "si")

   ; 16                               ;     P - 64
   ("DLE" "dle") ("DC1" "dc1") ("DC2" "dc2") ("DC3" "dc3")
   ("DC4" "dc4") ("NAK" "nak") ("SYN" "syn") ("ETB" "etb")

   ; 24                               ;     X - 64
   ("CAN" "can") ("EM" "em") ("SUB" "sb") ("ESC" "esc")
   ("FS" "fs")   ("GS" "gs") ("RS" "rs")  ("US" "us")

   ; 32
   ("SP" "space" "sp") ("!") ("\"") ("#")
   ("$") ("%") ("&") ("'")

   ; 40
   ("(") (")") ("*") ("+")  (",") ("-") (".") ("/")

   ; 48
   ("0") ("1") ("2") ("3")  ("4") ("5") ("6") ("7")

   ; 56
   ("8") ("9") (":") (";")  ("<") ("=") (">") ("?")

   ; 64
   ("@") ("A") ("B") ("C")  ("D") ("E") ("F") ("G")
   ("H") ("I") ("J") ("K")  ("L") ("M") ("N") ("O")
   ; 80
   ("P") ("Q") ("R") ("S")  ("T") ("U") ("V") ("W")
   ("X") ("Y") ("Z") ("[")  ("\\")
   ; 93
   ("]") ("^") ("_")

   ; 96
   ("`") ("a") ("b") ("c")  ("d") ("e") ("f") ("g")
   ("h") ("i") ("j") ("k")  ("l") ("m") ("n") ("o")
   ("p") ("q") ("r") ("s")  ("t") ("u") ("v") ("w")
   ("x") ("y") ("z")

   ; 123
   ("{") ("|") ("}") ("~")
   
   ; 127
   ("DEL" "del")
)))
vec))

(define (vchars:c1cnv-up)
 (let loop ((idx 161))
   (cond 
   ((< idx 256) (cons (brass (integer->char idx))
                      (loop (+ idx 1))))
   (else        '())) ))

(define (vchars:print-c1cnv)
 (define (printpqln br)
   (display #\( ) (display #\" )
   (display br)
   (display #\" ) (display #\) )
   (newline))
 (for-each printpqln (vchars:c1cnv-up))
 (set! printpqln #f)) ; XXX missing CLR4 inner define autocut

(define c1-character-name-vector #(
 ; 128
 ("XXX" "0080") ("XXX" "0081")
 ("BPH") ("NBH")                         ("IND") ("NEL") ("SSA") ("ESA")
 ("HTS") ("HTJ") ("VTS") ("PLD")         ("PLU") ("RI")  ("SS2") ("SS3")
 
 ; 144
 ("DCS") ("PU1") ("PU2") ("STS")         ("CCH") ("MW")  ("SPA") ("EPA")
 ("SOS") ("XXX" "0099") ("SCI") ("CSI")  ("ST")  ("OSC") ("PM")  ("APC")

 ; 160
 ("NBSP") ("¡") ("¢") ("£")  ("¤") ("¥")       ("¦") ("§")
    ("¨") ("©") ("ª") ("«")  ("¬") ("SHY" "­") ("®") ("¯")

 ; 176
    ("°") ("±") ("²") ("³")  ("´") ("µ") ("¶") ("·")
    ("¸") ("¹") ("º") ("»")  ("¼") ("½") ("¾") ("¿")

 ; 192
    ("À") ("Á") ("Â") ("Ã")  ("Ä") ("Å") ("Æ") ("Ç")
    ("È") ("É") ("Ê") ("Ë")  ("Ì") ("Í") ("Î") ("Ï")

 ; 208
    ("Ð") ("Ñ") ("Ò") ("Ó")  ("Ô") ("Õ") ("Ö") ("×")
    ("Ø") ("Ù") ("Ú") ("Û")  ("Ü") ("Ý") ("Þ") ("ß")

 ; 224
    ("à") ("á") ("â") ("ã")  ("ä") ("å") ("æ") ("ç")
    ("è") ("é") ("ê") ("ë")  ("ì") ("í") ("î") ("ï")

 ; 240
    ("ð") ("ñ") ("ò") ("ó")  ("ô") ("õ") ("ö") ("÷")
    ("ø") ("ù") ("ú") ("û")  ("ü") ("ý") ("þ") ("ÿ")
))

(when (not (= 128 (vector-length c0-character-name-vector)
                  (vector-length c1-character-name-vector)))
 (error "vchars: fatal character table inconsistency"))

(define char-seq-vec (make-vector 256 #f))

(let fill-selfs ((i 32)) ; sic! here <space> coincides
 (when (<= i 126)
  (vector-set! char-seq-vec i (list (integer->char i)))
  (fill-selfs (+ i 1)) ))

(let fill-controls ((i 0))
 (when (< i 32)
  (vector-set! char-seq-vec i
   (append (list #\< )
           (string->list (first (vector-ref c0-character-name-vector i)))
           (list #\>)))
  (fill-controls (+ i 1)) ))

(vector-set! char-seq-vec 127
   (append (list #\< )
           (string->list (first (vector-ref c0-character-name-vector 127)))
           (list #\>)))


(set! c1-character-name-vector
 (list->vector (append (vector->list c0-character-name-vector)
                       (vector->list c1-character-name-vector))))

;; There are three codepoints for which maybe the byte bit sequence
;; would look like a UTF tag or there is some other reason why they
;; used the explicit marker on them.
(let fill-c1-controls ((i 128))
 (when (<= i 160)
  (let* ((dfs (vector-ref c1-character-name-vector i))
         (df  (if (equal? (first dfs) "XXX") (second dfs) (first dfs)) ))
    (vector-set! char-seq-vec i (append (list #\< )
                                        (string->list df)
                                        (list #\>)) ))
  (fill-c1-controls (+ i 1)) ))

(let fill-selfs ((i 161))
 (when (< i 256)
  (vector-set! char-seq-vec i (list (integer->char i)))
  (fill-selfs (+ i 1)) ))

(let ((i #x00ad))  ; SHY
 (vector-set! char-seq-vec i
  (append (list #\< )
          (string->list (first (vector-ref c1-character-name-vector i)))
          (list #\>))))

;; The LF or newline control is our null element.
(vector-set! char-seq-vec (char->integer #\newline) '())

;; Beware to insert color codes after this. (Currently unsupported?)
(define char-dlen-vec (list->vector (map length (vector->list char-seq-vec))))

)) ; closes CHARS namespace

(define CHAR-seq-vec  (from-namespace CHARS 'char-seq-vec))
(define CHAR-dlen-vec (from-namespace CHARS 'char-dlen-vec))

(define (display-length chr)
 (cond ((< (char->integer chr) (vector-length CHAR-dlen-vec))
          (vector-ref CHAR-dlen-vec (char->integer chr)))
       (else 1))) ; xxx high code point hack: this doesn't work well
                  ; with double width characters
(define (display-form   chr)
 (cond ((< (char->integer chr) (vector-length CHAR-seq-vec))
         (list-copy (vector-ref CHAR-seq-vec (char->integer chr))))
       (else (list chr))))
