;; CEE
;; Extremely simple C Pre Processor Definition Parser and framework
;; prototype
;;
;; (I don't think this one conforms to any standard out there ;-)
;;
;; Simplified read syntax:
;; #<white space>define<white space>numerical-constant
;; -> dictionary entry for further processing
;;
;; The behavior for non numerical constants is unspecified.
;; Numerical constants will be converted from decimal and hexadecimal
;; notation automatically.
;;
;; Comments, strings and preprocessor conditionals aren't recognized
;; and not handled as such. This is why we can read in files with
;; non inclusion checks and other funny inventions but some files
;; might not yield any useful results.
;;
;; Check the results before further processing.
;;
;; TODO Invent some to resolve type definitions to basic types, i.e.
;; for a given type we want to get the C type by a simple dictionary
;; lookup after having processed some header file trees.
;; That goes a little something like this:
;;
;; That something to be invented is called "cpp" I thought and you
;; just do a "cpp header-file-with-definitions-to-resolve.h > defs"
;; and then all types would be resolved. Well, kind of:
;;
;; - includes have been replaced with numbers but the hash remained
;;   so we can filter these out
;; - guess what: all the cpp defines are gone but we already have them
;;   in our dictionaries, havn't we
;; - there are now a lot of /typedef/ declarations after the former
;;   include directives and someone supposes that should be enough
;;   to resolve all type declarations in that header file (system
;;   header files are well formed anyway, aren't they!)
;; - thus we feed a dictionary with the typedefs for basic types and
;; - then we can do some with the nice structure definitions describing
;;   protocols or else as by then we should have that byte structure
;;   system finished (see SCA)
;;
;;                The name of a structure becomes the symbol of the
;; binding, we decided. That's a little text transformation but a
;; C structure parser is more appropriate to do that. (Perhaps someone
;; will even update the tidbit parts to conform to the standards,
;; e.g. handling of control characters that will never appear in
;; system header files.)
;;
;; cpp /usr/include/X11/Xproto.h > Xproto.i  etc will then become
;; part of the compilation (stage 2 - Scheme level system bindings)?
;;
;; BTW clang's cpp is called "clang -E" and seems to do just the same.
;; But they claim their preprocessor already does the tokenization
;; which means for this parser to do the same when need arises.
;; [Apart from that the string constant - comment - backslash clearing
;;  sequence should really be done by our /read-char/ routine before
;;  anything else. Following include directives might enable us to
;;  resolve known constant names without relying on private header
;;  files - we can't use cpp as that one just replaces the defines
;;  as soon as it has been defined which might become necessary for
;;  this parser, too. But cpp macros like 
;;
;;                         #define paste(a1,a2,a3)   a1##a2##a3
;;  include/convert.i.c:59:#define F(pre,T,suf)   paste(pre,T,suf)
;;
;;  might be less daunting on a per character parsing level.]
;;

(sys-load "VSI-core/dictionary.scm")

(define read-char-system read-char)
(define (debugging-read-char . port)
 (let ((char (read-char-system)))
  (display char)
  char))

;(define read-char debugging-read-char)

(define (hash-finder)
 (let ((char (read-char)))
  (if (eq? char #\#)
    #t
    (if (eof-object? char)
      (throw 'eof)
      (hash-finder)) )))

(define cpp-definition-expectation (string->list "define"))
(define (cpp-definition-keyword?)
 (let loop ((expectation cpp-definition-expectation))
  (if (null? expectation)
    #t
    (let ((char (read-char)))
      (if (eof-object? char)
        (throw 'unexpected-eof)
        (if (eq? char (car expectation))
          (loop (cdr expectation))
          #f))) )))

(define (skip-whitespace)
 (let ((char (read-char)))
  (cond
   ((eof-object? char)   (throw 'eof))
   ((white-space? char)  (skip-whitespace))
   ((eq? char #\\)       (read-char) ;; ???
                         (skip-whitespace))
   (else
     (unread-char char)) )))

(define (cpp-definition-symbol+value! dictionary)
 (let loop ((char  (read-char))
            (current-token '()))
   (if (eof-object? char)
     (throw 'unexpected-eof)
     (cond  ; ANS top down sequence conditional
        ((is-unicode-c0-printable? (char->byte char))
         ;; in token
           (loop (read-char)
                 (cons char current-token) ))

        ((not (null? current-token))
         ;; end of token
          (catch 'eof skip-whitespace
                 (lambda () (throw 'unexpected-eof)))
          (let ((value (cpp-definition-value)))
           (set-definition!
             (create-entry! (get-root dictionary) (reverse! current-token))
             value))
           #t ); finish

        (else
           #f) )) ))

(define (cpp-definition-value)
 (let ((answer-box (cons #f '() )))
  (let loop ((char (read-char)))
  ;; Usually you would read in the value until the next newline.
  ;; Here we simplify for numerical constants.
   (cond
    ((is-unicode-c0-printable? (char->byte char))
       (set-cdr! answer-box (cons char (cdr answer-box)))
       (loop (read-char)) )

    ;; If we wanted to read in up to the next newline we would need
    ;; a /read-char/ which would have already cleared the comments and
    ;; handled line continuation marks correctly.
    ;; ((not (eq? char #\newline))
    ;;    (set-cdr! answer-box (cons char (cdr answer-box)))
    ;;    (loop (read-char)) )

    (else
     (conversion (reverse! (cdr answer-box)) ))) )))

(define (conversion-and argument-one . rest)
 (if (null? rest)
   argument-one
   (apply conversion-and (and argument-one (car rest)) (cdr rest))))
  
(define (conversion list)
 (or (and (pair? list) (pair? (cdr list))
          (eq? #\0 (first list))
          (eq? #\x (second list))
     ;; Hexadecimal Number Conversion
       (string->number (list->string (cons #\# (cons #\x (cddr list) )) )))

     (and (pair? list)
          (or (null? (cdr list)) (digit? (car list))
              (pair? (cdr list)) (apply conversion-and (map digit? list)))
     ;; Number Conversion
       (string->number (list->string list)))

     ;; Other Conversion
     (list->string list) ))

(define (throwing-cpp-definitions! dictionary file-name)
 (with-input-from-file file-name
  (lambda () ;; oops, we should have cleared comments first ...
   (let loop ()
     (hash-finder)
     (skip-whitespace)
     (when (cpp-definition-keyword?)
       (skip-whitespace)
       (cpp-definition-symbol+value! dictionary))
     (loop))) ))
(define (cpp-definitions! dictionary file-name)
 (catch 'eof (lambda ()
               (throwing-cpp-definitions! dictionary file-name))
        (lambda (key . args) #t)))

(define (resolve-types c-file-path i-file-path)
  (system (string-append "cpp \""
                         c-file-path
                         "\" \""
                         i-file-path
                         "\"")))
