;;; Generic MVS
;;;
;;; Experimental Generic Matrix Vector Scalar
;;;              Functions of Evaluation with Coersion
;;;
;;; NB The predicate /vector?/ isn't distinct from /matrix?/ where
;;;    the result of /matrix?/ is definitive.
;;;    => We rely on top down sequence for cond (guaranteed by the ANS).
;;; 
;;;    Since matrices and vectors are "open" collections and because of
;;;    their read syntax which allows for "coercables" someone decided
;;;    to apply coersion to each argument as well as each return value.
;;;
;;;    Combine the above two then ...
;;;    What mixturs of exact and inexact?
;;;     -> Dimensions need to be distinct to count as such. A vector
;;;        is based on "dimensionality" and matrices can often be seen
;;;        as a collection of vectors.
;;;
;;;   What about the complex numbers? Complex numbers aren't scalars
;;;   and you can't just coerce them to vectors. Currently they are
;;;   not even exact.
;;;                            

(sys-load "lib/matrices.scm")

;; Identity Token for Folds (see also /matrix-identity/ which might be implied by the token.)
(when (not (defined? 'generic:identity)) ;; double load protection
 (define generic:identity '(generic . identity)))
(define (generic-identity? obj) (eq? obj generic:identity))

(define (generic-mvs:coerce obj)
 (cond ((generic-identity? obj) generic:identity)
       ((number? obj) obj)
       ((and (numerical-matrix? obj) (equal? (matrix-dimensions obj) '(1 1)))  (matrix-ref    obj 0 0)) ; matrix -> scalar
       ((and (numerical-matrix? obj) (= 1 (matrix:number-of-rows    obj)))     (matrix-row    obj 0))   ; matrix -> vector
       ((and (numerical-matrix? obj) (= 1 (matrix:number-of-columns obj)))     (matrix-column obj 0))   ; matrix -> vector
       ((and (numerical-vector? obj) (= 1 (vector-length obj))                 (vector-ref    obj 0)))  ; vector -> scalar
       ((or (numerical-matrix? obj) (numerical-vector? obj) (scalar? obj))     obj)                     ; identity
       (else +nan.0))) ; not a number (is our wta here)

(define (generic-plus A B)
 (let ((A (generic-mvs:coerce A)) (B (generic-mvs:coerce B)))
 ; no coersion of the return value needed here
 (cond
 ((and (generic-identity? A) (generic-identity? B)) generic:identity) ; degenerate
 ((generic-identity? A)                             B)
 ((generic-identity? B)                             A)
 ((and (numerical-matrix? A) (numerical-matrix? B)) (matrix-plus A B))
 ((and (numerical-vector? A) (numerical-vector? B)) (vector-plus A B))
 ((and (scalar? A) (scalar? B)) (+ A B))
 (else +nan.0) )))

(define (generic-dot M N)
 (let ((M (generic-mvs:coerce M)) (N (generic-mvs:coerce N)))
 (generic-mvs:coerce
 (cond ((and (generic-identity? M) (generic-identity? N)) generic:identity) ; degenerate
       ((generic-identity? M)                             N)
       ((generic-identity? N)                             M)
       ((and (numerical-matrix? M) (numerical-matrix? N)) (matrix-dot M N))
       ((and (numerical-matrix? M) (numerical-vector? N)) (vector-map (lambda (row) (vector-dot row N)) M))
                                                                                           ;; ??? ((and (numerical-vector? M) mn) (vector-map (lambda (col) (vector-dot col M)) (matrix-transpose N)))
       ((and (numerical-matrix? M) (scalar? N))           (matrix-scale N M))
       ((and (scalar? M)           (numerical-matrix? N)) (matrix-scale M N))
       ((and (numerical-vector? M) (numerical-vector? N)) (vector-dot   M N))
       ((and (scalar? M)           (numerical-vector? N)) (vector-scale M N))
       ((and (numerical-vector? M) (scalar? N))           (vector-scale N M))
       ((and (scalar? M)           (scalar? N))           (* M N))
       (else +nan.0))) ))

(define (generic-minus A B)
 (let ((A (generic-mvs:coerce A)) (B (generic-mvs:coerce B)))
 ; no coersion of the return value needed here
 (cond
 ((and (generic-identity? A) (generic-identity? B)) generic:identity) ; degenerate
 ((generic-identity? A)                             (generic-dot -1 B))
 ((generic-identity? B)                             (generic-dot -1 A))
 ((and (numerical-matrix? A) (numerical-matrix? B)) (matrix-plus A (matrix-scale -1 B)))
 ((and (numerical-vector? A) (numerical-vector? B)) (vector-minus A B))
 ((and (scalar? A)           (scalar? B))           (- A B))
 (else +nan.0)) ))

(define generic-sum         (lambda arguments (fold generic-plus  generic:identity arguments)))
(define generic-product     (lambda arguments (fold generic-dot                  1 arguments)))
(define generic-subtraction (lambda arguments (fold generic-minus generic:identity arguments)))


(define g+ generic-sum)
(define g- generic-subtraction)
(define g* generic-product)

(when #t (sys-load "VSI-core/repl.scm")
(repl-transcript
 (define A (matrix #(1 2) #(3 4)))
 (define B (vector 5 6 7))
 (define C 8)

 (numerical-matrix? A) (numerical-matrix? B) (numerical-matrix? C)
 (numerical-vector? A) (numerical-vector? B) (numerical-vector? C)
 (scalar? A) (scalar? B) (scalar? C)

 (g+ A B C)
 (g* A B C)
 (g- A B C)

 (g+)
 (g*)
 (g-)

 (g+ A)
 (g* B)
 (g- C)

 (g* C C)
 (g* C B)
 (g* B C)

 (generic-dot C B)
 (generic-dot B C)

 (g- A)
 (g- B)
 (g- C)
)
(REPL)
)
