;;
;; Association List with Hashed Index
;; Hashtable with collision protection.
;;
;; (make-hashtable size)
;; (hashtable? object)
;; (hashtable-length hashtable)
;;
;; The /key/ needs to be a numerical integer value by default.
;;
;; (hashtable-insert! hashtable key value)
;; (hashtable-lookup  hashtable key)
;;
;; The lookup returns falsity #f if no entry could be found for /key/.
;;
;; (hashtable->alist hashtable)
;;
;; (describe-hashtable hashtable)
;;
;;
;; Private: (hashtable:index hashtable) => index-vector of the hashtable
;;          (hashtable:size  hashtable) => length of the index vector
(define hashtable:modulo integer-modulo)
(define hashtable:assoc  assoc)

(define (make-hashtable size)
 (cons 'hashtable (make-vector size '()) ))
(define (hashtable? object)
  (and (pair? object)
       (eq? (car object) 'hashtable)
       (vector? (cdr object))))
(define (hashtable:size hashtable) (vector-length (cdr hashtable)))
(define hashtable:index cdr)
(define (hashtable-insert! hashtable key value)
 (let* ((index (hashtable:index hashtable))
        (hash  (hashtable:modulo key (vector-length index)))
        (table (vector-ref index hash))
        (entry (hashtable:assoc key table)))
  (cond (entry (set-cdr! entry value))
        (else  (vector-set! index hash (acons key value table)))
)))
(define (hashtable-lookup hashtable key)
 (let* ((index (hashtable:index hashtable))
        (hash  (hashtable:modulo key (vector-length index)))
        (table (vector-ref index hash))
        (entry (hashtable:assoc key table)))
  (and entry (cdr entry)) ))

(define (hashtable-length hashtable)
 (let* ((index      (hashtable:index hashtable))
        (size       (hashtable:size  hashtable))
        (res        0 ))
  (let loop ((i 0)) (when (< i size)
    (set! res (+ res (length (vector-ref index i))))
    (loop (+ i 1))))
  res))

(define (hashtable->alist hashtable)
 (let* ((index      (hashtable:index hashtable))
        (size       (hashtable:size  hashtable))
        (res        '() ))
  (let loop ((i 0)) (when (< i size)
    (for-each (lambda (e) (set! res (acons (car e) (cdr e) res)))
              (vector-ref index i))
    (loop (+ i 1))))
  res))

(define (describe-hashtable hashtable)
 (let* ((idx  (hashtable:index hashtable))
        (size (hashtable:size hashtable))
        (il   (vector->list idx))
        (ill  (map length il))
        (nill (filter (lambda (e) (not (zero? e))) ill)))
  (list (cons 'size    size)
        (cons 'min     (apply min  ill))
        (cons 'max     (apply max  ill))
        (cons 'n-mean  (exact->inexact (/ (apply + nill) (length nill))))
        (cons 'empty   (length (filter zero? ill)))
)))
