;; 
;; Matrices are two dimensional collections with 
;; zero based indexed rows and columns.
;;
;; The following matrix with 3 rows and three columns 
;; --     -- 
;; | 1 2 3 | 
;; | 4 5 6 | 
;; | 7 8 9 | 
;; --     -- 
;;
;; is represented as a vector of vectors: 
;; 
;; #( #(1 2 3) #(4 5 6) #(7 8 9) )  
;; 
;; All non empty vectors with non empty vectors of the same length are
;; recognized as matrices. (See /matrix?/ and perhaps /Vector/ in the
;; reference manual under primitive data types).
;; 
;; There isn't any implicit coersion:
;; #(#(1)) is a matrix of size 1x1
;;   #(1)  is a vector of length 1 and
;;     1   is a scalar.
;;
;; [lib/generic-mvs.scm provides generic functions of evaluation with coersion]

(define (make-matrix number-of-rows number-of-columns . initial-value)
 (list->vector (map (lambda (num-cols) (apply make-vector num-cols initial-value))
                    (make-list number-of-rows number-of-columns)))) 

(define (matrix-copy M) (vector-map vector-copy M))

(define (matrix-eqv? M N) ;; XXX this is defunct <- depends on a working eqv? (see maintenance issue) 
 (and (matrix? M) (matrix? N)
      (equal? (matrix-dimensions M) (matrix-dimensions N)) 
      (all? (map eqv? (vector->list M) (vector->list N))) )) 
 
(define (matrix? object)
 (and (vector? object)
      (> (vector-length object) 0) 
      (let* ((rows (vector->list object))  
             (arow (first rows)) 
             (len  (and (vector? arow) (vector-length arow)) )) 
       (and len
            (> len 0) 
            (all? (map (lambda (row) (and (vector? row) (= (vector-length row) len))) 
                       rows)) )) ))           

(define (matrix-ref  M row col) (vector-ref (vector-ref M row) col)) 
(define (matrix-set! M row col value) (vector-set! (vector-ref M row) col value)) 

(define (matrix . rows)
 (let ((res (list->vector rows)))
  (cond ((not (matrix? res)) (error "matrix: bad matrix vectors"))
        (else                res) ))) 

(define (matrix-from-columns . cols) (matrix-transpose (apply matrix cols)))

(define (matrix:number-of-rows M) (vector-length M))
(define (matrix:number-of-columns M) (vector-length (vector-ref M 0))) 
(define (matrix-dimensions M) (values (vector-length M) (vector-length (vector-ref M 0)) )) 

; These selectors return a vector representing the specified row or column. 
(define (matrix-row    M row-idx) (vector-ref M row-idx))
(define (matrix-column M col-idx) (vector-map (lambda (row) (vector-ref row col-idx)) M)) 

(when #f (sys-load "VSI-core/repl.scm")
(repl-transcript 
 (define A (make-matrix 2 3))
 (define B (make-matrix 4 5 #f)) 
 (matrix-ref A 1 1)
 (matrix-ref B 2 2) 
 (matrix-set! A 1 1 11)
 (matrix-set! B 2 2 12)
 (matrix-ref A 1 1)
 (matrix-ref B 2 2) 
 (matrix? A)
 (matrix? B)
 (matrix? (vector 1 2 3))
 (matrix? (list 1 2 3)) 
 (matrix? (make-dlist)) 
 (define C (matrix #(1 0 0) #(0 1 0) #(0 0 1))) 
 (matrix-dimensions A)
 (matrix-dimensions B) 
 (matrix-dimensions C) 
 (matrix-row A 1)
 (matrix-column B 2)
 (matrix-row C 0)
 (matrix-column C 1) 
 (define Cb (matrix-copy C))
 (equal? Cb C) 
 (eq? Cb C) 
 (eqv? Cb C) ;; XXX TODO eqv
 (matrix-eqv? Cb C)
 (matrix-eqv? Cb 'symbol) 
) 
(REPL)
) 


;;;
;;; Matrix Processing
;;;

(define (matrix-map      f . lM) (apply vector-map      (lambda rows (apply vector-map      f rows)) lM))
(define (matrix-for-each f . lM) (apply vector-for-each (lambda rows (apply vector-for-each f rows)) lM))
(define (matrix->list M)
 (let* ((answer-box (cons #t '() ))
        (tail       answer-box))
 (matrix-for-each (lambda (el) (set-cdr! tail (cons el '())) (set! tail (cdr tail)))
                  M)
 (cdr answer-box)))

(define (matrix-transpose M) 
 (let* ((ncol (matrix:number-of-columns M))
        (nrow (matrix:number-of-rows    M))
        (R    (make-vector ncol)))
  (let loop ((idx (- ncol 1)))
  (cond ((>= idx 0) (vector-set! R idx (matrix-column M idx))
                    (loop (- idx 1)))
        (else       R) )) ))



;;;
;;; Matrix Calculus
;;;

(define (numerical-matrix? obj) (and (matrix? obj) (all? (map scalar? (matrix->list obj)) )))

;; Connect to the current mode of the vector calculus. (See VSI-core/vector-calculus.scm)
(define matrix:minus    %vc-minus)
(define matrix:multiply %vc-multiply)
(define matrix:divide   %vc-divide)
(define matrix:num-eqv? =)  ; ??? how should we treat this in symbolic mode?


(define (matrix-plus M N)  (vector-map vector-plus M N)) 
(define (matrix-scale s M) (vector-map (lambda (V) (vector-scale s V)) M)) 

;; Matrix Dot Product
;; The number of rows has to equal the number of columns.
;; In computer science literature (and elsewhere?) you often see shortcuts where
;; you have to understand that maths doesn't have any good concept of algorithm
;; and most of the time things will work out when we see the matrix as a sequence
;; of vectors: each of them should then be used to calculate the vector dot product
;; which will yield another vector then. When the second matrix has one column only
;; but the first matrix has two rows that implies some sort of algorithm which
;; isn't provided by the regular definition of the dot product for matrices.
;; [see also lib/generic-mvs.scm for the interpreting /generic-product/]
(define (matrix-dot M N)
 (let* ((nrow (matrix:number-of-rows    M)) 
        (ncol (matrix:number-of-columns N))
        (R    (make-matrix nrow ncol))) 
 (let loop ((ridx 0) (cidx 0)) 
  (cond ((>= ridx nrow) R)
        ((>= cidx ncol) (loop (+ ridx 1) 0))
        (else           (matrix-set! R ridx cidx (vector-dot (matrix-row    M ridx)
                                                             (matrix-column N cidx)))
                        (loop ridx (+ cidx 1)) )) ))) 


;;
;; Square Matrices
(define (matrix-square? M) (= (matrix:number-of-rows    M)
                              (matrix:number-of-columns M))) 
(define (matrix-identity n)
 (let ((M (make-matrix n n 0)))
  (let loop ((idx 0))
   (cond ((>= idx n) M)
         (else       (matrix-set! M idx idx 1)
                     (loop (+ idx 1)) )) )))

;; 2x2 matrices
(define (matrix-det-2 M)
 (let ((a (matrix-ref M 0 0))
       (b (matrix-ref M 0 1))
       (c (matrix-ref M 1 0))
       (d (matrix-ref M 1 1)))
 (matrix:minus (matrix:multiply a d) (matrix:multiply b c))))

(define (matrix-singular-2? M) (and (matrix-square? M)
                                    (matrix:num-eqv? (matrix-det-2 M) 0)))

;; 2x2 inverse: Return the inverse of the matrix /M/ if possible else falsity.
(define (matrix-inverse-2 M)
 (and (matrix-square? M) (not (matrix-singular-2? M))
      (let ((R (make-matrix 2 2)))
       (matrix-set! R 0 0 (matrix-ref M 1 1))
       (matrix-set! R 1 1 (matrix-ref M 0 0))
       (matrix-set! R 0 1 (matrix:minus (matrix-ref M 0 1)))
       (matrix-set! R 1 0 (matrix:minus (matrix-ref M 1 0)))
       (matrix-scale (matrix:divide (matrix-det-2 M)) R) )))

;; 3x3
;; determinant: rule of sarrus

;; nxn
;; determinant: leibniz formula for determinants, lu decomposition
;; inverse: several; see Invertable matrix in wikipedia (that includes 2x2, 3x3, 4x4)


(when #f (sys-load "VSI-core/repl.scm")
(repl-transcript
(define A (matrix #(1 3 1) #(1 0 0)))
(define B (matrix #(0 0 5) #(7 5 0)))
(matrix-plus A B) 
(matrix-scale 2 (matrix #(1 8 -3) #(4 -2 5))) 
(matrix-transpose (matrix #(1 2 3) #(0 -6 7))) 
(matrix-dot (matrix #(2 3 4) #(1 0 0)) (matrix #(0 1000) #(1 100) #(0 10)))  
(define A (matrix #(1 2) #(3 4)))
(define B (matrix #(0 1) #(0 0))) 
(matrix-dot A B)
(matrix-dot B A) 
(define iA (matrix-inverse-2 A))
(matrix-dot A iA)
(define iB (matrix-inverse-2 B))
iB
(and iB (matrix-dot B iB))
(matrix-copy A)
)
) 
