;; Freeze: Polar Vector Addition (to that vector calculus)
;;
;; Polar Vectors are described in terms of a vector
;; with the magnitude in its first slot and the angles
;; in the other slots (relative to a full circle of
;; 2*pi clock wise and the preceding slot dimension as 0°
;; (just after 2*pi+(0-0)) as expected.
;;
;; Void dimensions are recognized as such and there is
;; no angle between void dimensions:
;;
;;  (cartesian->polar #(1 0 0)) => #(1.0 0.0 +nan.0)
;;
;; For many algorithms based on polar angles it's correct
;; to replace NaNs with zeroes but some (in particular
;; the n-dimensional polar->cartesian) might fail.
;; (see Examples/polar.scm for the affected endimensional variant)
;; 
;; For a vector space with two dimensions there's only one
;; vector with void dimensions:
;;
;;   (cartesian->polar #(0 0)) => #(0.0 +nan.0)
;;
;; The /cartesian->polar/ conversion implies an /exact->inexact/
;; conversion due to non linear and trigonometric functions used.
;; See Additional Inexact Number Routines in VSI-core/numbers.scm
;; for (inexact= accuracy . operands)
;;
 
;; 
;; Clock Wise Normalization Two Pi Periodic 
;; Let v be a vector #(a1 a2 a3 ...)
;; and wn be the pair wise angles between an and an+1 
;; 
;;      /a3                       /a1 
;;     /                         / 
;;    /______ a1                /______________ a2 
;;    |\ w1 |                   |\____ <- w2 
;;    | \  /                    |     \____  
;;    |  \                      |          \___ V (a2 a3)
;;    |   \                    a3 
;;    |    \ 
;;    |     \
;;   a2      V(a1 a2) 
;;
;; 
;;                  V (0 0) ->  .  (no angle here; this is represented by w0 = +nan.0 Not A Number NaN) 
;; 
;;     
;; /cartesian->polar/ is known to produce positive angles only as
;; well as NaNs for void dimensions. 

(define (cartesian->polar cartesian)
 (define (quadrant-correction x y angle)
  (cond ((<= y 0)  angle)
        (else      (%vc-minus (%vc-multiply 2 %vc-pi) angle)) ))

 (define (dimensions->angle a b)
  (let ((c (%vc-square-root
            (%vc-plus (%vc-multiply a a) (%vc-multiply b b)))))
   (quadrant-correction a b (%vc-acos (%vc-divide a c)))
 ))

 (let ((res
 (list->vector
  (cons (vector-magnitude cartesian)
        (reduce dimensions->angle
                (vector->list cartesian)))) ))
   (set! quadrant-correction #f) ;; XXX missing CLR4 autocut
   (set! dimensions->angle   #f)
   res))

;; For some "simple" conversions this is already
;; down to 10^(-6).
(define polar-inexactness (/ 1 (expt 10 12)))

(define (two-pi-periodic angle)
 (inexact->exact
  (rationalize (inexact->exact (%vc-divide angle (%vc-multiply 2 %vc-pi)))
               polar-inexactness)))

(define (unit-vector n) (make-vector n 1))
(define (polar-unit-vector n)
  (cartesian->polar (unit-vector n)))

(define (polar magnitude angles)
  (list->vector (cons magnitude angles)))
(define (polar* magnitude . angles)
  (list->vector (cons magnitude angles)))
(define (polar-magnitude polar) (vector-ref polar 0))
(define (polar-angles    polar) (cdr (vector->list polar)))

(define (1D-polar->cartesian-vector polar)
 (let ((magnitude (polar-magnitude polar)))
  (vector magnitude)))

(define (2D-polar->cartesian-vector polar)
 (let ((magnitude (polar-magnitude polar))
       (angles    (cdr (vector->list polar))))
   (vector (* magnitude (cos (first angles)))
           (* magnitude (- (sin (first angles)) )) )))

(define polar->cartesian 2D-polar->cartesian-vector)

(define (polar-rotate polar-vector . angles)
 (polar (polar-magnitude polar-vector)
        (map %vc-plus (polar-angles polar-vector)
                      angles)))

(define (polar-scale s polar-vector)
 (polar (%vc-multiply s (polar-magnitude polar-vector))
        (polar-angles polar-vector)))

;;
;; Applications / Higher Order Operations
;;

(define (rotate-vector rotation vector)
 (polar->cartesian
  (apply polar-rotate (cartesian->polar vector)
                      rotation)))
