;; Transitional Setup (VOS/VOT hardening)
;;
;; This file is meant to be loaded into a namespace
;; as it may override system bindings.
;;

(define uc:replacement-character (integer->char #xFFFD)) ; <- Put this somehow into the core when applicable.

;;
;; Decoding UTF-8

(define utf-8:two-byte-mask   (byte->byte-string #b11100000))
(define utf-8:two-byte-tag    (byte->byte-string #b11000000))

(define utf-8:three-byte-mask (byte->byte-string #b11110000))
(define utf-8:three-byte-tag  (byte->byte-string #b11100000))

(define utf-8:four-byte-mask  (byte->byte-string #b11111000))
(define utf-8:four-byte-tag   (byte->byte-string #b11110000))

(define utf-8:sequence-byte-mask (byte->byte-string #b11000000))
(define utf-8:sequence-byte-tag  (byte->byte-string #b10000000))

(define (utf-8:sequence-byte? byte-string)
  (equal? (bit-and utf-8:sequence-byte-mask byte-string)
          utf-8:sequence-byte-tag))

(define (utf-8:two-byte? byte-string)
  (equal? (bit-and utf-8:two-byte-mask byte-string)
          utf-8:two-byte-tag))
(define (utf-8:three-byte? byte-string)
  (equal? (bit-and utf-8:three-byte-mask byte-string)
          utf-8:three-byte-tag))
(define (utf-8:four-byte? byte-string)
  (equal? (bit-and utf-8:four-byte-mask byte-string)
          utf-8:four-byte-tag))

(define (utf-8:sequence-bits byte-string)
 (cond
  ((utf-8:sequence-byte? byte-string)
        (bit-and (bit-not utf-8:sequence-byte-mask) byte-string))
  (else (encoding-error 'decoding-utf-8)) ))

(define from-msb! from-network-byte-order!)
(define to-msb    network-byte-order!)


;;; Stream Reader

(define sys-read-byte read-byte)
(define (utf-8:sequence-bits byte-string) ; w/o validation; we do that here
 (bit-and (bit-not utf-8:sequence-byte-mask) byte-string))

;; Prerequisites
(define (read-byte port) ;; XXX use (read-byte-string p 1) TODO rename read-byts
 (let ((b (sys-read-byte port)))
  (cond ((eof-object? b) b)
        (else            (byte->byte-string  b)))))
(define (unread-byte b) (unread-char (integer->char b))) ;; XXX
(define null-4          #"0 0 0 0")

;; Error Handling TODO add logging when appropriate
; (define uc:error-code #xFFFD) ; replacement character
(define uc:error-code #f) ; Falsity might better suit some situations
(define (bad-sequence   port sel)
  (unread-byte (byte-string->number sel))
  uc:error-code)
(define (bad-terminator port sel) uc:error-code)
(define (underflow  n   port)     uc:error-code)
    
;; Multibyte Decoding Stages (outsoured inlines)
(define (cp-from-bytes! usv-1 usv-2 usv-3 usv-4)
  ;; The following byte order conversion isn't needed when
  ;; you read /byte-string-set! usv n byte/ as a left shift
  ;; of the byte bits in a four byte word. Thus the final
  ;; right shifts below could be taken into account and
  ;; both left and right shifts could be replaced with the
  ;; resulting operations which would be independent of the
  ;; endianess. But we would need to check back with the
  ;; standards and test again.
  (map from-msb! (list usv-1 usv-2 usv-3 usv-4))
  (integer->char (byte-string->number
    (bit-or usv-1 (bit-shift-right usv-2 2)
                  (bit-shift-right usv-3 4)
                  (bit-shift-right usv-4 6)))))

(define (two-byte port sel usv-1 usv-2)
  (byte-string-set! usv-2 2 (bit-and (bit-not utf-8:two-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
       (cp-from-bytes! usv-1 usv-2 null-4 null-4))
  (else
       (bad-sequence port sel) )))

(define (three-byte port sel usv-1 usv-2 usv-3)
  (byte-string-set! usv-3 1 (bit-and (bit-not utf-8:three-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-2 2 (utf-8:sequence-bits sel))
       (set! sel (read-byte port))
       (cond
        ((eof-object? sel)
             (underflow 3 port))
        ((utf-8:sequence-byte? sel)
             (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
             (cp-from-bytes! usv-1 usv-2 usv-3 null-4))
        (else
             (bad-sequence port sel) )))
  (else
       (bad-sequence port sel) )))

(define (four-byte port sel usv-1 usv-2 usv-3 usv-4)
  (byte-string-set! usv-4 0 (bit-and (bit-not utf-8:four-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-3 1 (utf-8:sequence-bits sel))
       (set! sel (read-byte port))
       (cond
        ((eof-object? sel)
             (underflow 3 port))
        ((utf-8:sequence-byte? sel)
             (byte-string-set! usv-2 2 (utf-8:sequence-bits sel))
             (set! sel (read-byte port))
             (cond
              ((eof-object? sel)
                  (underflow 4 port))
              ((utf-8:sequence-byte? sel)
                  (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
                  (cp-from-bytes! usv-1 usv-2 usv-3 usv-4))
              (else
                  (bad-sequence port sel) )))
        (else
             (bad-sequence port sel) )))
  (else
       (bad-sequence port sel) )))

;; Stream Decoding Dispatch
;; ("forgiving" read-char for Viper/VOS/VOT: returns falsity for mischiefed encoding forms)
(define (read-char:utf-8/stream port)
 (let ((usv-1 (byte-string-clear! (make-byte-string 4)))
       (usv-2 (byte-string-clear! (make-byte-string 4)))
       (usv-3 (byte-string-clear! (make-byte-string 4)))
       (usv-4 (byte-string-clear! (make-byte-string 4)))
       (sel   (read-byte port)))

 (cond
  ((eof-object? sel) sel)
  ((utf-8:four-byte?  sel) (four-byte  port sel usv-1 usv-2 usv-3 usv-4))
  ((utf-8:three-byte? sel) (three-byte port sel usv-1 usv-2 usv-3))
  ((utf-8:two-byte?   sel) (two-byte   port sel usv-1 usv-2))
  ((< (byte-string->number sel) 128) (integer->char (byte-string->number sel)))
  (else                    (bad-terminator port sel)))
 ))

