;;; Vector Geometry: Trigonometry and else
;;; Private prefix: vg

;; Connects to the mode of the vector calculus.
(if (not (and (defined? 'symbolic-vector-calculus)
              symbolic-vector-calculus))
  (begin ; numerical operation
    (define %vc-sin sin))
  (begin
    (define (%vc-sin operand) (list 'vc-sin operand)) ))



;;
;; n dimensions
;;

(define (vector-middle v1 v2)      (vector-plus v1 (vector-scale 1/2   (vector-minus v2 v1))))
(define (vector-ratio v1 v2 ratio) (vector-plus v1 (vector-scale ratio (vector-minus v2 v1))))

(define (vector-norm A) (vector-scale (%vc-divide 1 (vector-magnitude A)) A))
(define (vector-distance v1 v2)    (vector-magnitude (vector-minus v1 v2)))
(define (vector-line-ratio v1 v2)  (%vc-divide (vector-magnitude (vector-minus v1 v2))
                                               (vector-magnitude v1)))


;; (vector-angle A B) => angle between the vectors A and B
;;
;; A and B should have the same number of dimensions.
;; Void dimensions aren't a problem but it defaults
;; to 90° for void angles: (vector-angle #(0 0) #(0 0)) => 1.5707963267949
(define (vector-angle A B) (%vc-acos (vector-dot (vector-norm A) (vector-norm B))))  ; http://www.euclideanspace.com/


;;; Trigonometry
;;
;; In the following paragraph of trigonometry all identifiers comply with
;; the diagram and relations below. The functions of evaluations defined
;; are meant to be used in concordance with the diagram. Concordance might
;; be permutation of the function arguments to match the task.
;;
;;         B beta
;;        /|\
;;       / | \
;;    c /  |  \ a
;;     /   |   \
;;    /    |h   \
;; A /_____|__s__\ C
;; alfa    H b    gamma
;;         
;; A B and C           are points or vectors.
;; AB BC AC BA CB CA   are the differences of the denoted vectors (in reversed order: AB is (vector-minus B A))
;; a b c               are scalars (the magnitudes of the differences).
;; alfa beta and gamma are the angles between the coinciding sides.
;;
;; s is a scalar which represents the ratio in which BH coincides onto CA. (scalar-triangle-ratio a beta), (vector-triangle-ratio A B C)
;; h is a scalar which represents the height of the triangle.              (scalar-triangle-height a beta), (vector-triangle-H A B C)
;; H is a point / vector which results when /s/ is applied to CA.
;;
;; The following relations hold:
;;
;; a^2 = h^2 + s^2
;; s^2 = a^2 - h^2
;;
;; alfa + beta + gamma = 180°
;; <(CA, HB) = 90°
;;
;; which translates to s-expressions like this:
;;
;; (inexact= accuracy pi (+ alfa beta gamma)) => #t
;; (inexact= accuracy (/ pi 4) (vector-angle (vector-minus A C) (vector-minus B H))) => #t
;;

;; The predicates should check for <(l1,l2) < 180° - 0.0 because
;; in their current state as tautologies they aren't useful.
(define (scalar-triangle? a b c)
  (or (> (+ a b) c)
      (> (+ a c) b)
      (> (+ b c) a)))
(define (vector-triangle? A B C)
 (let ((a (vector-magnitude (vector-minus B C)))
       (b (vector-magnitude (vector-minus A C)))
       (c (vector-magnitude (vector-minus A B))))
    (scalar-triangle? a b c)))

(define (scalar-triangle-height a beta) (%vc-multiply a (%vc-sin beta)))
(define (scalar-triangle-ratio  a beta) (let ((h (scalar-triangle-height a beta)))
                                          (sqrt (- (%vc-multiply a a) (%vc-multiply h h))) ))
(define (vector-triangle-ratio A B C)
 (let* ((BA   (vector-minus A B))
        (BC   (vector-minus C B))
        (CA   (vector-minus A C))
        (a    (vector-magnitude BC))
        (beta (vector-angle BA BC))
        (s    (scalar-triangle-ratio a beta)))
   s)) 

;; (vector-triangle-H A B C) => the vector H as described in the diagram
(define (vector-triangle-H A B C)
 (let* ((CA (vector-minus A C)))
   (vector-plus C (vector-scale (vector-triangle-ratio A B C) CA)) ))

;; End of the concordance with the diagram of the Trigonometry paragraph.
;;;



(define (vector-cross A B)
 (if (= 3 (vector-length A) (vector-length B))
   (vector-cross-3 A B)
   (n-ary-vector-cross-product A B ) ; undefined
))
;;
;; 3 dimensions
;;

(define (vector-cross-3 A B)
 (vector (%vc-minus (%vc-multiply (vector-ref A 1) (vector-ref B 2))
                    (%vc-multiply (vector-ref A 2) (vector-ref B 1)))
         (%vc-minus (%vc-multiply (vector-ref A 2) (vector-ref B 0))
                    (%vc-multiply (vector-ref A 0) (vector-ref B 2)))
         (%vc-minus (%vc-multiply (vector-ref A 0) (vector-ref B 1))
                    (%vc-multiply (vector-ref A 1) (vector-ref B 0)))
))

(define (vector-axis-3 A B) (vector-norm (vector-cross-3 A B)))

